/*=========================================================================

 Program:   ORFEO Toolbox
 Language:  C++
 Date:      $Date$
 Version:   $Revision$


 Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
 See OTBCopyright.txt for details.


 This software is distributed WITHOUT ANY WARRANTY; without even
 the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the above copyright notices for more information.

 =========================================================================*/
#include "otbApplicationCLIParser.h"
#include "itkMacro.h"

#include <assert.h>
#include <iostream>
#include "otbConfigure.h" // For OTB_VERSION_STRING generated by ccmake

namespace otb
{

ApplicationCLIParser::ApplicationCLIParser()
{

}

bool ApplicationCLIParser::ParseCommandLine(int argc, char *argv[], bool reportFailedMsg,
                                             bool failOnUnknownTrailingParameters,
                                             ApplicationOptionsResult::Pointer outResult)
{
  // Clear the result
  outResult->Clear();

  int i;
  m_Name = std::string(argv[0]);
  int index(0);

  // Go through the arguments
  for (i = 1; i < argc; ++i)
    {
    // Get the next argument
    std::string arg(argv[i]);

    // Check if the argument is known
    bool findOption = FindOption(arg, index);
    if (findOption == false)
      {
      if (failOnUnknownTrailingParameters)
        {
        // Unknown argument found
        if (reportFailedMsg == true)
          {
          std::cerr << "The following '" << arg << "' option is unknown" << std::endl;
          }
        return false;
        }
      else return true;
      }
    //Check the option
    m_OptionList[index].Found = true;
    // If the number of parameters is predefined
    if (m_OptionList[index].NumberOfParametersFixed == true)
      {
      // Check if the number of parameters is correct
      int nParameters = m_OptionList[index].NumberOfParameters;
      if (i + nParameters >= argc)
        {
        // Too few parameters
        if (reportFailedMsg == true)
          {
          std::cerr << "Missing one (or more) parameter(s) for the following '" << arg << "' option." << std::endl;
          }
        return false;
        }
      // Tell the result that the option has been encountered
      outResult->AddOption(m_OptionList[index].CommonName);

      // Pass in the parameters
      for (int j = 0; j < nParameters; ++j, ++i)
        {
        outResult->AddParameter(m_OptionList[index].CommonName, std::string(argv[i + 1]));
        }
      }
    // If the number of parameters is not defined, read until the next option or the end of argv
    else
      {
      // Tell the result that the option has been encountered
      outResult->AddOption(m_OptionList[index].CommonName);
      bool goOnFlag(true);
      while (goOnFlag == true)
        {
        if (argv[i + 1] != NULL)
          {
          std::string strArgv = std::string(argv[i + 1]);
          if (strArgv[0] == '-')
            {
            goOnFlag = false;
            }
          else
            {
            outResult->AddParameter(m_OptionList[index].CommonName, strArgv);
            ++i;
            }
          }
        else goOnFlag = false;
        //         ++i;
        }
      }

    }

  // Check that all required arguments are specified
  for (unsigned int cpt = 0; cpt < m_OptionList.size(); cpt++)
    {
    if ((m_OptionList[cpt].Mandatory == true) && (m_OptionList[cpt].Found == false))
      {
      // Too few parameters
      if (reportFailedMsg == true)
        {
        std::cerr << "Error: Missing required argument:'" << m_OptionList[cpt].CommonName << "'" << std::endl;
        }
      return false;
      }
    }

  // Everything is correct
  return true;
}

bool ApplicationCLIParser::FindOption(const std::string& option, int& index)
{
  //Look through the option list
  bool optionFoundFlag(false);
  bool goOnFlag(true);
  unsigned int cpt(0);

  while (goOnFlag == true)
    {
    std::string commonName = "--";
    commonName += m_OptionList[cpt].CommonName;
    std::string shortName = "-";
    shortName += m_OptionList[cpt].Synonym;

    if (option == commonName || option == shortName)
      {
      index = cpt;
      goOnFlag = false;
      optionFoundFlag = true;
      }
    cpt++;
    if (cpt >= m_OptionList.size())
      {
      goOnFlag = false;
      }
    }
  return (optionFoundFlag);
}

void ApplicationCLIParser::PrintUsage(std::ostream& os) const
{
  os << std::endl;

  if (!m_Description.empty())
    {
    os << m_Description << std::endl << std::endl;
    }

  os << " Usage : " << m_Name << std::endl;
  // Compute the max width for option display
  int largeurmax(-1);
  unsigned int i;

  for (i = 0; i < m_OptionList.size(); ++i)
    {
    int largeur = 2 + m_OptionList[i].CommonName.size() + 1 + m_OptionList[i].Synonym.size();
    if (largeur > largeurmax) largeurmax = largeur;
    }

  // Check that all required arguments are present on the command line
  for (i = 0; i < m_OptionList.size(); ++i)
    {
    if (m_OptionList[i].CommonName != "--OTBTesting")
      {
      int largeur = 2 + m_OptionList[i].CommonName.size() + 1 + m_OptionList[i].Synonym.size();
      os << "      ";
      os << ( m_OptionList[i].Mandatory ? " " : "[" );
      os << "--" << m_OptionList[i].CommonName;
      if (!m_OptionList[i].Synonym.empty())
        {
        os << "|-" << m_OptionList[i].Synonym;
        }
      os << ( m_OptionList[i].Mandatory ? " " : "]" );

      //Align the text
      for (int b = largeur; b < largeurmax; b++)
        os << " ";

      os << "  :  " << m_OptionList[i].Description;

      if (m_OptionList[i].NumberOfParametersFixed)
        {
        switch (m_OptionList[i].NumberOfParameters)
          {
          case 0:
            break;
          case 1:
            os << "  (" << m_OptionList[i].NumberOfParameters << " parameter)";
            break;
          default:
            os << "  (" << m_OptionList[i].NumberOfParameters << " parameters)";
            break;
          }
        }
      else
        {
        os << "  (N parameters)";
        }
      os << std::endl;
      }
    }
  os << std::endl;
}

void ApplicationCLIParser::PrintVersion(std::ostream& os) const
{
  os << " OTB Version : " << OTB_VERSION_STRING << std::endl;
}

}
