/*=========================================================================

 Program:   ORFEO Toolbox
 Language:  C++
 Date:      $Date$
 Version:   $Revision$


 Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
 See OTBCopyright.txt for details.

 This code comes from a David Dubois contribution.

 This software is distributed WITHOUT ANY WARRANTY; without even
 the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the above copyright notices for more information.

 =========================================================================*/
#include "otbApplicationFLTKGui.h"
#include "itkMacro.h"

#include <assert.h>
#include <iostream>
#include "otbConfigure.h" // For OTB_VERSION_STRING generated by ccmake
namespace otb
{

ApplicationFLTKGui::ApplicationFLTKGui()
{

}

bool ApplicationFLTKGui::ParseGui(bool reportFailedMsg, bool failOnUnknownTrailingParameters,
                                  ApplicationOptionsResult::Pointer outResult)
{
  bool parseStatus = false;
  m_StartParsing = false;

  // *DD* Define main window with program name as title (argv[0]) *DD*
  inputwindow = new Fl_Window(400, m_OptionList.size() * 50, m_Name.c_str());
  inputwindow->box(FL_PLASTIC_DOWN_BOX);
  inputwindow->color(FL_DARK1);
  inputwindow->user_data((void *) (this));
  unsigned int pos = 1;

  // *DD* For each option, dynamically create an input box *DD*
  for (ListOptionType::iterator it = m_OptionList.begin(); it != m_OptionList.end(); ++it)
    {
    if (it->Type != ApplicationDescriptor::Help
        && it->Type != ApplicationDescriptor::Version)
      {
      // Handle the boolean type : put a check box
      if(it->Type == ApplicationDescriptor::Boolean)
        {
        Fl_Check_Button * boolButton = new Fl_Check_Button(130, pos*50, 90, 25, it->CommonName.c_str());
        boolButton->down_box(FL_DOWN_BOX);
        boolButton->color(FL_BACKGROUND2_COLOR);
        boolButton->selection_color((Fl_Color)23);
        boolButton->labelcolor((Fl_Color)186);
        m_CheckButtonList.push_back(boolButton);
        }
      else
        {
        Fl_Input * tmpInput = new Fl_Input(130, pos * 50, 200, 30, it->CommonName.c_str());
        tmpInput->box(FL_PLASTIC_DOWN_BOX);
        tmpInput->color((Fl_Color)23);
        tmpInput->selection_color((Fl_Color)23);
        tmpInput->labelcolor((Fl_Color)187);
        tmpInput->textcolor(187);

        // If ValueType of option is "file", *DD* add BrowseButton next to input box *DD*
        if (it->Type == ApplicationDescriptor::InputImage
            || it->Type == ApplicationDescriptor::FileName
            || it->Type == ApplicationDescriptor::OutputImage)
          {
          Fl_Button * tmpBrowseButton = new Fl_Button(330, pos * 50, 30, 30, "...");
          tmpBrowseButton->box(FL_PLASTIC_DOWN_BOX);
          tmpBrowseButton->color((Fl_Color)23);
          tmpBrowseButton->selection_color((Fl_Color)23);
          tmpBrowseButton->labelcolor((Fl_Color)186);

          // *DD* Attach BrowseButton to callback function *DD*
          tmpBrowseButton->callback((Fl_Callback *) &ApplicationFLTKGui::cb_BrowseButton, (void *) pos);
          m_ButtonList.push_back(tmpBrowseButton);
          }

        // *DD* Default value *DD*
        //tmpInput->value(it->DefaultValue.c_str());
        // *DD* tooltip value for the inputbox is the description of the option *DD*
        tmpInput->tooltip(it->Description.c_str());
        m_InputList.push_back(tmpInput); // *DD* add inputbox to list of
        // options *DD*
        }
      pos++;
      }
    }
  // *DD* add return button to accept options *DD*
  Fl_Return_Button * cmdOkButton = new Fl_Return_Button(230, pos * 50, 130, 30, "Ok");
  cmdOkButton->box(FL_PLASTIC_UP_BOX);
  cmdOkButton->shortcut(0xff0d);
  cmdOkButton->color((Fl_Color)23);
  cmdOkButton->selection_color((Fl_Color)23);
  cmdOkButton->labelcolor((Fl_Color)186);
  cmdOkButton->callback((Fl_Callback *) &Self::cb_OkButton); // *DD* attach OkButton to callback function *DD*
  cmdOkButton->tooltip(m_Description.c_str());
  inputwindow->end();
  inputwindow->show();
  Fl::run();
  // *DD* If OKButton was pressed: start parsing input options *DD*
  if (m_StartParsing == true)
    {
    parseStatus = TryParseGUI(outResult, true, failOnUnknownTrailingParameters);
    if (!parseStatus)
      {
      PrintUsage(std::cerr);
      //itkExceptionMacro(<< "ParseGUI() argument error");
      }
    }
  else
    {
    PrintUsage(std::cerr);
    //itkExceptionMacro(<< "ParseGUI() arguments not set");
    }
  // *DD* Clean FLTK objects *DD*
  for (ListInputType::iterator inIt = m_InputList.begin(); inIt != m_InputList.end(); ++inIt)
    {
    delete (*inIt);
    }
  for (ListButtonType::iterator cmdIt = m_ButtonList.begin(); cmdIt != m_ButtonList.end(); ++cmdIt)
    {
    delete (*cmdIt);
    }
  delete cmdOkButton;
  delete inputwindow;
  return parseStatus;
}

// *DD* Take each options and feed them to ParseResult object *DD*
bool ApplicationFLTKGui::TryParseGUI(ApplicationOptionsResult * outResult, bool reportFailedMsg,
                                     bool failOnUnknownTrailingParameters)
{
  // Clear the result
  outResult->Clear();

  // Go through the arguments
  for (ListInputType::iterator flIt = m_InputList.begin(); flIt != m_InputList.end(); ++flIt)
    {
    // Verify that a value was entered in the input box
    if ((*flIt)->value() != std::string(""))
      {
      // const char * to std::string : Method need const std::string
      // as input
      const std::string  value_str((*flIt)->value());
      const std::string  label_str((*flIt)->label());

      // Get the index of the option
      int optionIndex;
      if(FindOption(label_str, optionIndex))
        {
        // Get the number of parameters set for this
        // option in the GUI
        unsigned int nParameters = this->Tokenize(value_str).size();

        // CASE 1 : the number of parameters is predefined
        if(m_OptionList.at(optionIndex).NumberOfParametersFixed == true)
          {
          // Check if the number of parameters is correct
          if(static_cast<unsigned int>(m_OptionList.at(optionIndex).NumberOfParameters) == nParameters)
            {
            // Tell the result that the option has been encountered
            outResult->AddOption((*flIt)->label());

            // Pass in the parameter
            unsigned int numberOfParameters = static_cast<unsigned int>(m_OptionList.at(optionIndex).NumberOfParameters);
            for (unsigned int j = 0; j < numberOfParameters; ++j)
              {
              outResult->AddParameter((*flIt)->label(), this->Tokenize((*flIt)->value())[j]);
              }
            }
          else
            {
            std::cerr << "Missing one (or more) parameter(s) for the following '"
                << (*flIt)->label() << "' option."
                << std::endl;
            }
          }
        else // CASE 2 : undefined Number Of parameters
          {
          // Tell the result that the option has been encountered
          outResult->AddOption((*flIt)->label());

          // Pass in the parameter
          for (unsigned int j = 0; j < nParameters; ++j)
            {
            outResult->AddParameter((*flIt)->label(), this->Tokenize((*flIt)->value())[j]);
            }
          } // else case 2
        }// else FindOption
      else
        {
        std::cerr<<"Find option did not found the option "<< (*flIt)->label() << std::endl;
        }
      } // else empty field
    else
      {
      // *DD* If no value: if obligatory and report fail: show error message and stop *DD*
      int index;
      if (FindOption((*flIt)->label(), index))
        {
        if (m_OptionList.at(index).Mandatory == true)
          {
          if (reportFailedMsg == true)
            {
            std::cerr << "'" << (*flIt)->label() << "' no value present !!!" << std::endl;
            }
          return false;
          }
        }
      }
    }

  // Go through the arguments
  // stored in the checkbuttonList : Boolean type
  for (ListCheckButtonType::iterator flIt = m_CheckButtonList.begin(); flIt != m_CheckButtonList.end(); ++flIt)
    {
    // const char * to std::string
    const std::string  label_str((*flIt)->label());

    // Get the index of the option
    int optionIndex;
    if(FindOption(label_str, optionIndex))
      {
      // Verify that a value was entered in the input box
      if ((*flIt)->value())
        {
        if( m_OptionList.at(optionIndex).Type == ApplicationDescriptor::Boolean)
          {
          // Tell the result that the option has been encountered
          outResult->AddOption((*flIt)->label());
          }
        }
      }
    }

  // Everything is correct
  return true;
}

std::vector<std::string>
ApplicationFLTKGui::Tokenize(std::string currentParameterLine)
{
  std::string delimiters =" ";

  // vector to fill
  std::vector<std::string >  tokens;

  // Skip delimiters at beginning.
  std::string::size_type lastPos = currentParameterLine.find_first_not_of(delimiters, 0);

  // Find first "non-delimiter" from lastPos position.
  std::string::size_type pos    = currentParameterLine.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos)
    {
    // Found a token, add it to the vector.
    std::string   currentString = currentParameterLine.substr(lastPos, pos - lastPos);
    tokens.push_back(currentString);

    // Skip delimiters.  Note the "not_of"
    lastPos = currentParameterLine.find_first_not_of(delimiters, pos);

    // Find next "non-delimiter"
    pos = currentParameterLine.find_first_of(delimiters, lastPos);
    }

  return tokens;
}

// Open file_chooser dialog in current directory DD
void ApplicationFLTKGui::BrowseButton(long currentOption)
{
  char *filename = fl_file_chooser("Select file", "*.*", "./");

  // If user selected a file name assign it to corresponding input box DD
  if (filename != NULL)
    {
    m_InputList.at(currentOption - 1)->value(filename);
    }
  // Give focus back to corresponding intput box DD
  m_InputList.at(currentOption - 1)->take_focus();
}

// Tell ParseGUI we are ready to start parsing inputs and hide window DD
void ApplicationFLTKGui::OkButton()
{
  m_StartParsing = true;
  inputwindow->hide();
}


bool ApplicationFLTKGui::FindOption(const std::string& option, int& index)
{
  //Look through the option list
  bool optionFoundFlag(false);
  bool goOnFlag(true);
  unsigned int cpt(0);
  std::string strOption(option);
  while (goOnFlag == true)
    {
    if ((m_OptionList[cpt].CommonName == strOption) || (m_OptionList[cpt].Synonym == strOption))
      {
      index = cpt;
      goOnFlag = false;
      optionFoundFlag = true;
      }
    cpt++;
    if (cpt >= m_OptionList.size())
      {
      goOnFlag = false;
      }
    }
  return (optionFoundFlag);
}

}
