/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbApplicationDescriptor_h
#define __otbApplicationDescriptor_h

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <cstdlib>

#include "itkProcessObject.h"
#include "otbMacro.h"

namespace otb
{
/**
 * \class ApplicationDescriptor
 * \brief Utility to Parse command line argument.
 *
 * Usage:
 *  - Initialize the parser:
 * \code
 *    ApplicationDescriptor parser;
 *    parser.SetProgramDescription("This program is an example");
 *    parser.AddInputImage();
 *    parser.AddOption("-f", 1);
 *    parser.AddSynonim("-f","--filename");
 *    parser.AddOption("-v", 0);
 *    parser.AddSynonim("-v","--verbose");
 *    parser.AddOption("--DetailledName","Help explanation","-d", 1, false);
 * \endcode
 *  - Use the parser:
 * \code
 *    ApplicationDescriptorResult result;
 *    if(parser.TryParseCommandLine(argc, argv, result))
 *    {
 *      if(result.IsOptionPresent("--DetailledName"))
 *      {
 *        cout << "Option " << result.GetParameterString("--DetailledName") << endl;
 *        ...
 *      }
 *    }
 * \endcode
 */
class ITK_EXPORT ApplicationDescriptor : public itk::Object
{
public:
  typedef ApplicationDescriptor  Self;
  typedef itk::LightObject              Superclass;
  typedef itk::SmartPointer<Self>       Pointer;
  typedef itk::SmartPointer<const Self> ConstPointer;

  itkNewMacro(Self);
  itkTypeMacro(ApplicationDescriptor, itk::ProcessObject);

  typedef enum
  {
    Help,
    Version,
    InputImage,
    OutputImage,
    FileName,
    DirectoryName,
    String,
    Boolean,
    Integer,
    Real,
    Enumerated,
    StringList
  } ApplicationOptionType;


  /** Add an input image option */
  void AddInputImage(bool mandatory = true);
  /** Add an output image option */
  void AddOutputImage(bool mandatory = true);

  /* Short name for the application */
  itkSetStringMacro(Name);
  itkGetStringMacro(Name);

  /* Description of the application */
  itkSetStringMacro(Description);
  itkGetStringMacro(Description);

  /** Add an option with fixed number of parameters (words that follow it) */
  void AddOption(std::string name,
                 std::string comment,
                 std::string synonim,
                 int nParameters,
                 bool mandatory,
                 ApplicationOptionType optionType);

  /** Add an option with arbitrary number of parameters (words that follow it) */
  void AddOptionNParams(std::string name,
                        std::string comment,
                        std::string synonym,
                        bool mandatory,
                        ApplicationOptionType optionType);

  /*
  void AddDoubleOption(std::string name,
                       std::string comment,
                       std::string synonim,
                       int nParameters = 1,
                       bool mandatory = true,
                       bool hasMin,
                       bool hasMax,
                       double min,
                       double max);
  */


  void PrintUsage(std::ostream& os) const;
  void PrintVersion(std::ostream& os) const;


  struct OptionType
  {
    OptionType()
    : NumberOfParametersFixed(false),
      NumberOfParameters(-1),
      Mandatory(false),
      Found(false)
    {
    }

    std::string CommonName;             // option name
    std::string Description;            // option description
    std::string Synonym;                // shortcut
    bool NumberOfParametersFixed;       // required number of values
    int NumberOfParameters;             // number of values
    bool Mandatory;                     // is the option mandatory ?
    bool Found;                         // check if the option is present
    ApplicationOptionType Type;         // option type (file, string, numerical or info (not showed)) DD
  };
  typedef std::vector<OptionType> ListOptionType;

  const ListOptionType& GetOptionList() const
  {
    return m_OptionList;
  }

protected:
  ApplicationDescriptor();
  virtual ~ApplicationDescriptor() {}

  std::string    m_Name;
  std::string    m_Description;
  ListOptionType m_OptionList;
};

}

#endif // __otbApplicationDescriptor_h_
