/*=========================================================================

 Program:   ORFEO Toolbox
 Language:  C++
 Date:      $Date$
 Version:   $Revision$


 Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
 See OTBCopyright.txt for details.


 This software is distributed WITHOUT ANY WARRANTY; without even
 the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the above copyright notices for more information.

 =========================================================================*/

#include "otbApplicationQgisWidget.h"
#ifdef WIN32
#   define CORE_EXPORT __declspec( dllimport )
#   define GUI_EXPORT __declspec( dllimport )
#else
#   define CORE_EXPORT
#   define GUI_EXPORT
#endif

#include "qgisinterface.h"
#include "qgsrasterlayer.h"
#include "qgsmapcanvas.h"
#include "qgsmaplayer.h"
#include "qgsmaplayerregistry.h"

#include <boost/algorithm/string.hpp>

namespace otb
{

ApplicationQgisWidget::ApplicationQgisWidget(otb::Application* appli, QgisInterface* iFace)
  : ApplicationQtWidget(appli),
    m_QgisIFace(iFace)
{
  m_ComboboxMapper = new QSignalMapper;

  connect( m_ComboboxMapper, SIGNAL(mapped(QWidget*)), this, SLOT(SetInputImageValue(QWidget*)) );
}

ApplicationQgisWidget::~ApplicationQgisWidget()
{
}

QWidget* ApplicationQgisWidget::CreateInputImageWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);

  m_Combobox = new QComboBox;
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(m_Combobox, optionID);
  m_ComboboxMapper->setMapping(m_Combobox, m_Combobox);
  connect( m_Combobox, SIGNAL(activated(int)), m_TextChangedMapper, SLOT(map()) );
  connect( m_Combobox, SIGNAL(activated(int)), m_ComboboxMapper, SLOT(map()) );

  //insert available raster layers
  QMap<QString, QgsMapLayer*> mapLayers = QgsMapLayerRegistry::instance()->mapLayers();
  QMap<QString, QgsMapLayer*>::iterator layer_it = mapLayers.begin();

  for (; layer_it != mapLayers.end(); ++layer_it)
    {
    QgsRasterLayer* rl = qobject_cast<QgsRasterLayer *> (layer_it.value());
    if (rl)
      {
      m_Combobox->addItem(rl->name(), QVariant(rl->getLayerID()));
      }
    }

  // initialize
  SetActiveOption(optionID);
  SetInputImageValue(m_Combobox);

  hLayout->addWidget(m_Combobox);
  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQgisWidget::CreateOutputImageWidget(const OptionType& option)
{
  return this->CreateFileNameWidget(option);
}

void ApplicationQgisWidget::SetInputImageValue(QWidget* widget)
{
  QComboBox* combobox = qobject_cast<QComboBox*>(widget);
  int index = combobox->currentIndex();
  QString filename;
  if ( index != -1 )
  {
    QString layerId = combobox->itemData( index ).toString();
    QgsMapLayer* inputLayer = QgsMapLayerRegistry::instance()->mapLayer(layerId);
    if (inputLayer)
      {
      filename = inputLayer->source();
      }
  }
  this->SetActiveOptionValue(filename);
}

void ApplicationQgisWidget::EndExecute()
{
  // call the Superclass implementation
  ApplicationQtWidget::EndExecute();

  // Add the output image to the Qgis layer
  if (m_Options->IsOptionOutputImagePresent())
    {
    std::string outputImage =  m_Options->GetOutputImage();

    // OTB wants .hdr for ENVI files, Qgis does not (as gdal)
    if (boost::iends_with(outputImage, ".hdr"))
      boost::ierase_last(outputImage, ".hdr");

    m_QgisIFace->addRasterLayer( QString(outputImage.c_str()) );
    }
}

}
