/*=========================================================================

 Program:   ORFEO Toolbox
 Language:  C++
 Date:      $Date$
 Version:   $Revision$


 Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
 See OTBCopyright.txt for details.


 This software is distributed WITHOUT ANY WARRANTY; without even
 the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the above copyright notices for more information.

 =========================================================================*/
#include "otbApplicationQtWidget.h"
#include "itkMacro.h"

#include <assert.h>
#include <iostream>
#include "otbConfigure.h" // For OTB_VERSION_STRING generated by ccmake

namespace otb
{

ApplicationQtWidget::ApplicationQtWidget(otb::Application* appli)
  : m_Application(appli),
    m_Descriptor(otb::ApplicationDescriptor::New()),
    m_Options(otb::ApplicationOptionsResult::New()),
    m_TextChangedMapper(new QSignalMapper(this)),
    m_CheckboxChangedMapper(new QSignalMapper(this)),
    m_SelectFileMapper(new QSignalMapper(this)),
    m_SelectDirMapper(new QSignalMapper(this)),
    m_ExecutionThread(new otb::ApplicationExecutionThread(this)),
    m_ErrorCaught(false)
{
  m_Application->Describe(m_Descriptor);

  connect( m_TextChangedMapper, SIGNAL(mapped(const QString&)), this, SLOT(SetActiveOption(const QString&)) );
  connect( m_CheckboxChangedMapper, SIGNAL(mapped(const QString&)), this, SLOT(SetActiveOption(const QString&)) );
  connect( m_SelectFileMapper, SIGNAL(mapped(QWidget*)), this, SLOT(SelectFile(QWidget*)) );
  connect( m_SelectDirMapper, SIGNAL(mapped(QWidget*)), this, SLOT(SelectDir(QWidget*)) );

  connect(m_ExecutionThread, SIGNAL(finished()), this, SLOT(EndExecute()));
  connect(m_ExecutionThread, SIGNAL(terminated()),  this, SLOT(EndExecute()));
  connect(m_ExecutionThread, SIGNAL(ReportError(const QString&)), this, SLOT(CatchError(const QString&)));

}

ApplicationQtWidget::~ApplicationQtWidget()
{
}

QString ApplicationQtWidget::GetName()
{
  return QString(m_Descriptor->GetName());
}

QString ApplicationQtWidget::GetDescription()
{
  return QString(m_Descriptor->GetDescription());
}


QWidget* ApplicationQtWidget::CreateHeader()
{
  // an HLayout with the description of the application, and two icons
  QHBoxLayout *headerLayout = new QHBoxLayout;

  QGroupBox *headerGroup = new QGroupBox;
  headerGroup->setStyleSheet("border: 1px solid gray");

  headerGroup->setFixedHeight(50);
  headerGroup->setContentsMargins(0, 0, 0, 0);
  headerLayout->setContentsMargins(5, 5, 5, 5);

  QLabel *iconOTBLabel = new QLabel;
  iconOTBLabel->setStyleSheet("border-style: none");
  iconOTBLabel->setPixmap(QIcon( ":/otb_big.png" ).pixmap(32, QIcon::Normal, QIcon::On));

  QLabel *descriptionLabel = new QLabel;
  descriptionLabel->setStyleSheet("border-style: none");
  QString descriptionLabelText(m_Descriptor->GetDescription());
  descriptionLabel->setText(descriptionLabelText);

  QLabel *iconCNESLabel = new QLabel;
  iconCNESLabel->setStyleSheet("border-style: none");
  iconCNESLabel->setPixmap(QIcon( ":/cnes.png" ).pixmap(32, QIcon::Normal, QIcon::On));

  headerLayout->addWidget(iconOTBLabel);
  headerLayout->addStretch();
  headerLayout->addWidget(descriptionLabel);
  headerLayout->addStretch();
  headerLayout->addWidget(iconCNESLabel);
  headerGroup->setLayout(headerLayout);

  return headerGroup;
}

QWidget* ApplicationQtWidget::CreateInputWidgets()
{
  // a GridLayout with two colums : option name / option inputs
  QGroupBox *gridGroup = new QGroupBox;
  gridGroup->setFlat(true);

  QGridLayout *gridLayout = new QGridLayout;
  gridLayout->setSpacing(1);
  gridLayout->setContentsMargins(0, 0, 0, 0);
  gridGroup->setLayout(gridLayout);

  ListOptionType::const_iterator it  = m_Descriptor->GetOptionList().begin();
  ListOptionType::const_iterator end = m_Descriptor->GetOptionList().end();

  unsigned int i = 0;
  for (; it != end; ++it, ++i)
    {
    if (it->Type != ApplicationDescriptor::Help
        && it->Type != ApplicationDescriptor::Version)
      {
      gridLayout->addWidget(CreateInputWidgetLabel(*it), i, 0);
      gridLayout->addWidget(CreateInputWidgetParam(*it), i, 1);
      }
    }

  return gridGroup;
}

QWidget* ApplicationQtWidget::CreateFooter()
{
  // an HLayout with two buttons : Execute and Quit
  QGroupBox *footerGroup = new QGroupBox;
  QHBoxLayout *footerLayout = new QHBoxLayout;
  
  footerGroup->setFixedHeight(40);
  footerGroup->setContentsMargins(0, 0, 0, 0);
  footerLayout->setContentsMargins(5, 5, 5, 5);

  m_ProgressLabel = new otb::ProgressLabel(footerGroup);
  connect(m_ExecutionThread, SIGNAL(started()),     this, SLOT(StartExecute()) );
  connect(m_ExecutionThread, SIGNAL(finished()),    this, SLOT(AbortExecute()) );
  connect(m_ExecutionThread, SIGNAL(terminated()),  this, SLOT(EndExecute()) );

  m_ExecButton = new QPushButton(footerGroup);
  //m_ExecButton->setStyleSheet("border-style: none");
  m_ExecButton->setDefault(true);
  m_ExecButton->setText(QObject::tr("Execute"));
  connect( m_ExecButton, SIGNAL(clicked()), this, SLOT(Execute()) );

  QPushButton *quitButton = new QPushButton(footerGroup);
  quitButton->setText(QObject::tr("Quit"));
  connect( quitButton, SIGNAL(clicked()), this, SLOT(close()) );

  // Put the buttons on the right
  footerLayout->addWidget(m_ProgressLabel);
  footerLayout->addStretch();
  footerLayout->addWidget(m_ExecButton);
  footerLayout->addWidget(quitButton);
  footerGroup->setLayout(footerLayout);

  return footerGroup;
}

void ApplicationQtWidget::CreateGui()
{
  // Create a VBoxLayout with the header, the input widgets, and the footer
  QVBoxLayout *mainLayout = new QVBoxLayout(this);

  mainLayout->addWidget(CreateHeader());
  mainLayout->addWidget(CreateInputWidgets());
  mainLayout->addWidget(CreateFooter());

  QGroupBox *mainGroup = new QGroupBox(this);
  mainGroup->setLayout(mainLayout);

  // Put the main group inside a scroll area
  QScrollArea *scrollArea = new QScrollArea;
  scrollArea->setWidget(mainGroup);
  scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
  scrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);

  QVBoxLayout *scrollLayout = new QVBoxLayout(this);
  scrollLayout->addWidget(scrollArea);

  // Make the scroll layout the main layout
  this->setLayout(scrollLayout);
  this->setWindowIcon(QIcon( ":/otb_small.png" ));
  this->setWindowTitle(QString(m_Descriptor->GetName()).append(" - version ").append(OTB_VERSION_STRING));

  // Tweak the window size so that it looks ok (TODO : find the way to make it perfectly symmetric)
  this->resize(mainGroup->sizeHint() + scrollArea->verticalScrollBar()->size());
  
}

QWidget* ApplicationQtWidget::CreateInputWidgetLabel(const OptionType& option)
{
  // Set up label
  QLabel *label = new QLabel;
  QString optionName(option.CommonName.c_str());
  QString optionLabel(option.CommonName.c_str());
  label->setText(optionLabel);
  label->setToolTip(option.Description.c_str());
  QGroupBox *labelHGroup = new QGroupBox;
  QHBoxLayout *labelLayout = new QHBoxLayout;
  labelLayout->setSpacing(0);
  labelLayout->setContentsMargins(0, 0, 0, 0);
  labelLayout->addWidget(label);
  labelHGroup->setLayout(labelLayout);
  return labelHGroup;
}

QWidget* ApplicationQtWidget::CreateInputWidgetParam(const OptionType& option)
{
  QWidget* widget = 0;
  switch (option.Type)
  {
    case ApplicationDescriptor::InputImage:
      widget = this->CreateInputImageWidget(option);
      break;
    case ApplicationDescriptor::OutputImage:
      widget = this->CreateOutputImageWidget(option);
      break;
    case ApplicationDescriptor::FileName:
      widget = this->CreateFileNameWidget(option);
      break;
    case ApplicationDescriptor::DirectoryName:
      widget = this->CreateDirNameWidget(option);
      break;
    case ApplicationDescriptor::String:
      widget = this->CreateStringWidget(option);
      break;
    case ApplicationDescriptor::Boolean:
      widget = this->CreateBoolWidget(option);
      break;
    case ApplicationDescriptor::Integer:
      widget = this->CreateIntWidget(option);
      break;
    case ApplicationDescriptor::Real:
      widget = this->CreateRealWidget(option);
      break;
   case ApplicationDescriptor::StringList:
      widget = this->CreateStringListWidget(option);
      break;
    default:
      widget = this->CreateStringWidget(option);
      break;
  }
  return widget;
}

QWidget* ApplicationQtWidget::CreateInputImageWidget(const OptionType& option)
{
  return this->CreateFileNameWidget(option);
}


QWidget* ApplicationQtWidget::CreateOutputImageWidget(const OptionType& option)
{
  return this->CreateFileNameWidget(option);
}

QWidget* ApplicationQtWidget::CreateFileNameWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QLineEdit* input = new QLineEdit;
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(textChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(textChanged(const QString&)), this, SLOT(SetActiveOptionValue(const QString&)));
  hLayout->addWidget(input);

  // Set up input text edit
  QPushButton *button = new QPushButton;
  button->setText("...");
  button->setToolTip("Select file...");
  button->setMaximumWidth(button->width());
  m_SelectFileMapper->setMapping(button, input);
  connect( button, SIGNAL(clicked()), m_SelectFileMapper, SLOT(map()) );
  hLayout->addWidget(button);

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQtWidget::CreateDirNameWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QLineEdit* input = new QLineEdit;
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(textChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(textChanged(const QString&)), this, SLOT(SetActiveOptionValue(const QString&)));
  hLayout->addWidget(input);

  QPushButton *button = new QPushButton;
  button->setText("...");
  button->setToolTip("Select directory...");
  button->setMaximumWidth(button->width());
  m_SelectDirMapper->setMapping(button, input);
  connect( button, SIGNAL(clicked()), m_SelectDirMapper, SLOT(map()) );
  hLayout->addWidget(button);

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQtWidget::CreateStringWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QLineEdit* input = new QLineEdit;
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(textChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(textChanged(const QString&)), this, SLOT(SetActiveOptionValue(const QString&)));
  hLayout->addWidget(input);

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQtWidget::CreateRealWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QDoubleSpinBox* input = new QDoubleSpinBox;
  input->setDecimals(5);
  input->setRange(-1.0E6, 1.0E6);
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(valueChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(valueChanged(const QString&)), this, SLOT(SetActiveOptionValue(const QString&)));
  hLayout->addWidget(input);
  hLayout->addStretch();

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQtWidget::CreateIntWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QSpinBox* input = new QSpinBox;
  input->setRange(0, 100000);
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(valueChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(valueChanged(const QString&)), this, SLOT(SetActiveOptionValue(const QString&)));
  hLayout->addWidget(input);
  hLayout->addStretch();

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}

QWidget* ApplicationQtWidget::CreateBoolWidget(const OptionType& option)
{
  // Set up input text edit
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QCheckBox* checkbox = new QCheckBox;
  checkbox->setToolTip(option.Description.c_str());

  QString optionID(option.CommonName.c_str());
  m_CheckboxChangedMapper->setMapping(checkbox, optionID);
  connect( checkbox, SIGNAL(stateChanged(int)), m_CheckboxChangedMapper, SLOT(map()) );
  connect( checkbox, SIGNAL(stateChanged(int)), this, SLOT(SetActiveOptionBoolValue(int)));
  hLayout->addWidget(checkbox);
  hLayout->addStretch();

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}


QWidget* ApplicationQtWidget::CreateStringListWidget(const OptionType& option)
{
  QHBoxLayout *hLayout = new QHBoxLayout;
  hLayout->setSpacing(0);
  hLayout->setContentsMargins(0, 0, 0, 0);
  QLineEdit* input = new QLineEdit;
  input->setToolTip(option.Description.c_str());
  QString optionID(option.CommonName.c_str());
  m_TextChangedMapper->setMapping(input, optionID);
  connect( input, SIGNAL(textChanged(const QString&)), m_TextChangedMapper, SLOT(map()) );
  connect( input, SIGNAL(textChanged(const QString&)), this, SLOT(SetActiveEnumOptionValue(const QString&)));
  hLayout->addWidget(input);

  QGroupBox *paramHGroup = new QGroupBox;
  paramHGroup->setLayout(hLayout);
  return paramHGroup;
}


bool ApplicationQtWidget::Execute()
{
  typedef otb::ApplicationDescriptor::ListOptionType OptionList;
  OptionList::const_iterator it = m_Descriptor->GetOptionList().begin();
  OptionList::const_iterator end = m_Descriptor->GetOptionList().end();
  for (; it != end; ++it)
    {
    std::string option = it->CommonName;
    if ( it->Mandatory && !m_Options->IsOptionPresent(option))
      {
      QMessageBox msgBox;
      QString text("Missing mandatory option :\n");
      text.append(option.c_str());
      msgBox.setText(text);
      msgBox.exec();
      return false;
      }
    }
  m_ErrorCaught = false;
  m_ExecutionThread->Execute(m_Application, m_Options);
  return true;
}

void ApplicationQtWidget::CatchError(const QString& msg)
{
  m_ErrorCaught = true;

  QMessageBox msgBox;
  msgBox.setText(msg);
  msgBox.exec();
}

void ApplicationQtWidget::StartExecute()
{
  m_ProgressLabel->Start();
  m_ExecButton->setEnabled(false);
}

void ApplicationQtWidget::AbortExecute()
{
  m_ProgressLabel->End();
  m_ExecButton->setEnabled(true);
}

void ApplicationQtWidget::EndExecute()
{
  m_ProgressLabel->End();
  m_ExecButton->setEnabled(true);
}

void ApplicationQtWidget::SetActiveOption( const QString& option )
{
  m_ActiveOption = option.toStdString();
}

void ApplicationQtWidget::SetActiveOptionValue( const QString& value )
{
  if (value.isEmpty())
    m_Options->RemoveOption(m_ActiveOption);
  else
    m_Options->SetParameter(m_ActiveOption, value.toStdString(), 0);
}

void ApplicationQtWidget::SetActiveOptionBoolValue( int qtValue )
{
  bool value = (qtValue == Qt::Checked);
  if (value)
    {
    m_Options->AddOption(m_ActiveOption);
    }
  else
    {
    m_Options->RemoveOption(m_ActiveOption);
    }
}


void ApplicationQtWidget::SetActiveEnumOptionValue( const QString& value )
{
  QStringList l_val = value.split(" ");
  unsigned int size = l_val.size();
  if (size == 0 )
    m_Options->RemoveOption(m_ActiveOption);
  else
    {
      for(unsigned int i=0; i<size; i++)
        {
          m_Options->SetParameter(m_ActiveOption, l_val[i].toStdString(), i);
        }
    }
}


void ApplicationQtWidget::SelectFile( QWidget* inputWidget )
{
  QFileDialog fileDialog;
  fileDialog.setConfirmOverwrite(true);
  fileDialog.setFileMode(QFileDialog::AnyFile);
  fileDialog.setNameFilter("Any files (*)");

  if (fileDialog.exec())
    {
    QStringList fileNames = fileDialog.selectedFiles();
    QLineEdit* edit = (QLineEdit*)inputWidget;
    edit->setText(fileNames.at(0));
    }
}

void ApplicationQtWidget::SelectDir( QWidget* inputWidget )
{
  QFileDialog fileDialog;
  fileDialog.setFileMode(QFileDialog::Directory);
  //fileDialog.setOption(QFileDialog::ShowDirsOnly, true);

  if (fileDialog.exec())
    {
    QStringList fileNames = fileDialog.selectedFiles();
    QLineEdit* edit = (QLineEdit*)inputWidget;
    edit->setText(fileNames.at(0));
    }
}

}
