/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbFocusingModel_h
#define __otbFocusingModel_h

#include "itkObject.h"
#include "itkRGBPixel.h"
#include "otbImage.h"
#include "otbMVCModel.h"
#include "otbImageLayerRenderingModel.h"
#include "otbListenerBase.h"
#include "otbMultiChannelExtractROI.h"
#include "otbImageLayerGenerator.h"
#include "otbImageLayer.h"
#include "otbVectorData.h"
#include "otbVectorDataExtractROI.h"
#include "otbVectorDataGlComponent.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataFileReader.h"
#include "otbHistogramCurve.h"
#include "otbObjectList.h"


namespace otb
{
/** \class FocusingModel
 *
 *
 */
template < class TInputImage, class TOutputImage = otb::Image<itk::RGBAPixel<unsigned char>, 2 > >
class FocusingModel
: public MVCModel<ListenerBase>, public itk::Object
{
  public:
  /** Standard class typedefs */
  typedef FocusingModel                         Self;
  typedef MVCModel<ListenerBase>  Superclass;
  typedef itk::SmartPointer<Self>               Pointer;
  typedef itk::SmartPointer<const Self>         ConstPointer;

  /** Standards macros */
  itkNewMacro(Self);

      /** Runtime information */
  itkTypeMacro(FocusingModel, MVCModel);

  /** View typedef */
  typedef ImageLayerRenderingModel<TOutputImage>         ImageLayerRenderingModelType;
  typedef typename ImageLayerRenderingModelType::Pointer ImageLayerRenderingModelPointerType;

  /** Output image typedef */
  typedef TOutputImage                               OutputImageType;
  typedef typename OutputImageType::Pointer          OutputImagePointerType;
  typedef typename OutputImageType::RegionType       RegionType;
  typedef typename OutputImageType::PixelType        OutputPixelType;
  typedef typename RegionType::SizeType              SizeType;
  typedef typename RegionType::IndexType             IndexType;

  /** Input image typedef */
  typedef TInputImage                                   InputImageType;
  typedef typename InputImageType::Pointer              InputImagePointerType;
  typedef typename InputImageType::PixelType            PixelType;
  typedef typename itk::NumericTraits<PixelType>::ValueType ScalarPixelType;
  typedef typename InputImageType::PointType            PointType;
  typedef otb::ImageLayer<InputImageType, OutputImageType>               LayerType;
  typedef typename LayerType::Pointer                   LayerPointerType;
  typedef typename otb::ImageLayerGenerator<LayerType>  LayerGeneratorType;
/* Test */
  typedef typename LayerType::RenderingFunctionType                 RenderingFunctionType;
  typedef typename LayerType::RenderingFunctionPointerType          RenderingFunctionPointerType;
//   typedef typename LayerType::DefaultRenderingFunctionType          StandardRenderingFunctionType;
  typedef Function::StandardRenderingFunction<PixelType, OutputPixelType>
                                                                    StandardRenderingFunctionType;
  typedef typename StandardRenderingFunctionType::Pointer           StandardRenderingFunctionPointerType;


  /** Extract ROI typedef */
  typedef MultiChannelExtractROI<ScalarPixelType, ScalarPixelType>  ExtractROIType;
  typedef typename ExtractROIType::Pointer                        ExtractROIPointerType;

  /** Vector Data Typedefs */
  typedef VectorData<double>                                        VectorDataType;
  typedef typename VectorDataType::Pointer                          VectorDataPointerType;
  typedef VectorDataProjectionFilter<VectorDataType, VectorDataType> VectorDataProjectionFilterType;
  typedef VectorDataExtractROI<VectorDataType>                      VectorDataExtractROIType;
  typedef typename VectorDataExtractROIType::RegionType             RemoteSensingRegionType;
  typedef VectorDataGlComponent<VectorDataType>                     VectorDataGlComponentType;
  typedef typename VectorDataGlComponentType::Pointer               VectorDataGlComponentPointerType;
  typedef VectorDataFileReader<VectorDataType>                      VectorDataReaderType;
  typedef typename VectorDataReaderType::Pointer                    VectorDataReaderPointerType;
  typedef typename LayerType::HistogramType                         HistogramType;
  typedef HistogramCurve<HistogramType>                             HistogramCurveType;
  typedef typename HistogramCurveType::Pointer                      HistogramCurvePointerType;
  typedef typename HistogramCurveType::ColorType                    ColorType;
  // list declaration
  typedef ObjectList<VectorDataType>                                VectorDataListType;
  typedef ObjectList<VectorDataGlComponentType>                     VectorDataGlComponentListType;


  /** Get/Set pointers the models */
  itkGetObjectMacro(MasterModel, ImageLayerRenderingModelType);
  itkGetObjectMacro(WorkingModel, ImageLayerRenderingModelType);

  itkSetObjectMacro(MasterModel, ImageLayerRenderingModelType);
  itkSetObjectMacro(WorkingModel, ImageLayerRenderingModelType);

  itkGetObjectMacro(VectorDataList, VectorDataListType);
  itkGetObjectMacro(MasterVectorDataGlList, VectorDataGlComponentListType);
  itkGetObjectMacro(WorkingVectorDataGlList, VectorDataGlComponentListType);

  /** Get/Set the rendering Function */
  itkGetObjectMacro(RenderingFunction, RenderingFunctionType);
  itkSetObjectMacro(RenderingFunction, RenderingFunctionType);

  /** Get/Set the input image */
  InputImagePointerType GetExtractedImage()
  {
    return m_Extractor->GetOutput(); ;
  }

  void SetInputImage(InputImagePointerType pInputImage);
  itkGetObjectMacro( InputImage, InputImageType );

  /** Load a vector data file */
  void LoadVectorData();
  void LoadVectorData( std::string name )
  {
    this->SetVectorDataFileName(name);
    this->LoadVectorData();
    this->Modified();
  }
  void AddVectorData( VectorDataPointerType vectorData );
  void UpdateMasterVectorData(unsigned int vectorId);
  void UpdateWorkingVectorData(unsigned int vectorId);
  void UpdateVectorData();

  VectorDataGlComponentPointerType UpdateVectorData( InputImagePointerType image/*RegionType region*/, unsigned int vectorId );
  void ClearVectorData( unsigned int id );
  void ClearAllVectorData();
  void ChangeVectorDataColor( unsigned int id, ColorType color );

  void UpdateRGBChannelOrder(unsigned int red, unsigned int green, unsigned int blue);
  void UpdateGrayscaleChannel(unsigned int cannel);

  void AddMasterLayer(LayerType * layer)
  {
    m_MasterModel->AddLayer(layer);
  }

  void AddWorkingLayer(LayerType * layer)
  {
    m_WorkingModel->AddLayer(layer);
  }


  /** Change the extract region by giving the 2 points of the
   * region */
  void SetExtractRegionByIndex(const IndexType & startIndex, const IndexType & stopIndex );

  itkGetConstReferenceMacro(ExtractRegion, RegionType);

  void Update();
  void UpdateModels()
  {
    m_MasterModel->Update();
    m_WorkingModel->Update();
  }
  void UpdateQuicklook(unsigned int ratio);
  void UpdateWorkingView();
  void ClearLayers();
  void Init();

  /** Change the transparency */
  virtual void ChangeOpacity( double opacity );
  /** Change from boundaries to full polygon*/
  virtual void ChangeGlRepresentation();

  virtual void NotifyListener(ListenerBase * listener);

  itkGetMacro(WorkingHasNewVectorData, bool);
  itkGetMacro(MasterHasNewVectorData, bool);
  itkGetMacro(WorkingHasNewExtract, bool);
  itkGetMacro(HasSupressedVectorData, bool);
  itkGetMacro(HasToUpdate, bool);
  itkGetMacro(NewExtractVDDisplayed, bool);
  itkSetMacro(NewExtractVDDisplayed, bool);
  itkGetMacro(LastErasedVectorData, unsigned int);
  itkSetMacro(VectorDataFileName, std::string);
  itkGetMacro(VectorDataFileName, std::string);
  itkSetMacro(UseDEM, bool);
  itkGetMacro(UseDEM, bool);
  itkSetMacro(DEMDirectory, std::string);
  itkGetMacro(DEMDirectory, std::string);
  itkGetMacro(HasToInit, bool);

  itkSetMacro(MasterScreenRatio, double);
  itkGetMacro(MasterScreenRatio, double);

  itkSetMacro(Color, ColorType);
  itkGetMacro(Color, ColorType);

  std::vector<ColorType> GetUsedColorList()
  {
    return m_UsedColorList;
  }

  protected:
  /** Constructor */
  FocusingModel();
  /** Destructor */
  ~FocusingModel();
  /** Printself method */
  void PrintSelf(std::ostream& os, itk::Indent indent) const;


  private:

  ImageLayerRenderingModelPointerType m_MasterModel;
  ImageLayerRenderingModelPointerType m_WorkingModel;

  /** Extract to obtain the ROI to display in the working view. */
  ExtractROIPointerType            m_Extractor;
  /** Extracted ROI. */
  RegionType                        m_ExtractRegion;
  InputImagePointerType             m_InputImage;
  VectorDataListType::Pointer       m_VectorDataList;
  /** GL component containing the vector data of the master view. */
  VectorDataGlComponentListType::Pointer m_MasterVectorDataGlList;
  /** GL component containing the vector data of the working view. */
  VectorDataGlComponentListType::Pointer m_WorkingVectorDataGlList;
  /** Load new vector data -> have to update master display. */
  bool                          m_MasterHasNewVectorData;
  /** Update the working vector data display when ROI change. */
  bool                          m_WorkingHasNewVectorData;
  /** User has selected a new roi in master scroll. */
  bool                          m_WorkingHasNewExtract;
  /** End of the extractor generation (with VD refresh) */
  bool                          m_NewExtractVDDisplayed;
  /** User has supressed one or more vector data */
  bool                          m_HasSupressedVectorData;
  /** Just update the view */
  bool                          m_HasToUpdate;
  /** Last vector data erased */
  unsigned int                  m_LastErasedVectorData;
  /** Vector data file name. */
  std::string                   m_VectorDataFileName;
  /** Use vector DEM to correct the vector data display. */
  bool                          m_UseDEM;
  /** Path to the directory that contains the DEM for vector data display correction. */
  std::string                   m_DEMDirectory;
  /** Screen ratio value for Master View. */
  double                        m_MasterScreenRatio;
  /** bool if init*/
  bool                          m_HasToInit;
  /** Current Display Color */
  ColorType                     m_Color;
  /** Store the used color */
  std::vector<ColorType>        m_UsedColorList;
  /** GlComponent display type */
  bool                          m_PolygonBoundariesOnly;
  /** Save the rendering Function */
  RenderingFunctionPointerType  m_RenderingFunction;

}; // end class
} // end namespace otb

#ifndef  OTB_MANUAL_INSTANTIATION
#include "otbFocusingModel.txx"
#endif

#endif


