/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbFocusingModel_txx
#define __otbFocusingModel_txx

#include "otbFocusingView.h"
#include "otbFltkFilterWatcher.h"
#include "otbVectorDataFileWriter.h"

namespace otb
{
template < class TInputImage, class TOutputImage >
FocusingModel<TInputImage, TOutputImage>
::FocusingModel() : m_MasterModel(), m_WorkingModel(), m_Extractor(), m_InputImage(), m_VectorDataList(), m_MasterVectorDataGlList(), m_WorkingVectorDataGlList()
{
  m_MasterModel             = ImageLayerRenderingModelType::New();
  m_WorkingModel            = ImageLayerRenderingModelType::New();
  m_Extractor               = ExtractROIType::New();
  m_InputImage              = InputImageType::New();
  m_VectorDataList          = VectorDataListType::New();
  m_MasterVectorDataGlList  = VectorDataGlComponentListType::New();
  m_WorkingVectorDataGlList = VectorDataGlComponentListType::New();
  m_RenderingFunction = StandardRenderingFunctionType::New();
  m_WorkingHasNewExtract    = false;
  m_MasterHasNewVectorData  = false;
  m_WorkingHasNewVectorData = false;
  m_HasSupressedVectorData  = false;
  m_HasToUpdate             = false;
  m_LastErasedVectorData    = 0;
  m_UseDEM                  = false;
  m_HasToInit               = false;
  m_NewExtractVDDisplayed   = false;
  m_DEMDirectory            = "";
  m_VectorDataFileName      = "";
  m_MasterScreenRatio       = 0.5;
  // Init with blue
  m_Color.Fill(0);
  m_Color[2]=1.;
  m_Color[3]=1.;
  m_UsedColorList.clear();
  m_PolygonBoundariesOnly = false;
}

template < class TInputImage, class TOutputImage >
FocusingModel<TInputImage, TOutputImage>
::~FocusingModel()
{
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::PrintSelf(std::ostream& os, itk::Indent indent) const
{
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::SetExtractRegionByIndex( const IndexType & startIndex, const IndexType & stopIndex )
{
  // if input image exists
  if( this->GetMasterModel()->GetNumberOfLayers() != 0 )
    {
      m_MasterModel->SetExtractRegionByIndex(startIndex, stopIndex);
      m_ExtractRegion = m_MasterModel->GetExtractRegion();
    }

}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::Init()
{
  this->ClearLayers();
  RegionType region;
  IndexType id;
  id.Fill(0);
  SizeType size;
  size.Fill(0);
  region.SetSize(size);
  region.SetIndex(id);
  m_ExtractRegion = region;
  m_WorkingModel->SetExtractRegion( m_ExtractRegion );
  m_MasterModel->SetExtractRegion(m_ExtractRegion);
  m_HasToInit=true;
  this->NotifyAll();
  m_HasToInit=false;

  this->GetWorkingModel()->Update();
  this->GetMasterModel()->Update();
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::Update()
{
  // if input image exists
  if( this->GetMasterModel()->GetNumberOfLayers() != 0 )
    {
      this->UpdateWorkingView();
    }
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::SetInputImage(InputImagePointerType pInputImage)
{
  m_InputImage = pInputImage;
  typename  LayerGeneratorType::Pointer lGenerator = LayerGeneratorType::New();

  // Watcher
  FltkFilterWatcher qlwatcher(lGenerator->GetProgressSource(), 0, 0, 250, 20,"Reading Input Image...");

  lGenerator->SetImage(m_InputImage);
  lGenerator->SetScreenRatio(m_MasterScreenRatio);
  lGenerator->GenerateLayer();

  // save the standard rendering function
  m_RenderingFunction  = lGenerator->GetRenderingFunction();

  // Add the layer to the two models
  this->ClearLayers();
  this->AddMasterLayer(lGenerator->GetLayer());
}

/*************************************************************************
 ********************* WORKING VIEW UPDATE *******************************
 *************************************************************************/
template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateWorkingView()
{
  this->GetWorkingModel()->ClearLayers();
  m_WorkingVectorDataGlList->Clear();
  m_WorkingModel->Init();

  m_Extractor = ExtractROIType::New();
  m_Extractor->SetInput( m_InputImage );
  RegionType lSelectedRegion = m_MasterModel->GetExtractRegion();

  // Make an extract of the input image
  m_Extractor->SetStartX(lSelectedRegion.GetIndex()[0]);
  m_Extractor->SetStartY(lSelectedRegion.GetIndex()[1]);
  m_Extractor->SetSizeX(lSelectedRegion.GetSize()[0]);
  m_Extractor->SetSizeY(lSelectedRegion.GetSize()[1]);

  // Generate the layer
  typename LayerGeneratorType::Pointer lGenerator = LayerGeneratorType::New();

  // Watcher
  FltkFilterWatcher qlwatcher(lGenerator->GetProgressSource(), 0, 0, 250, 20,"Generating Working View ...");

  lGenerator->SetImage(m_Extractor->GetOutput());

  unsigned int ratio = lGenerator->GetOptimalSubSamplingRate();

  lGenerator->SetScreenRatio(std::max(1/static_cast<double>(ratio), 0.25));
  lGenerator->GenerateLayer();
  // use the master view rendering function to have the same image display dynamic
  LayerType * lLayer = static_cast<LayerType*>(this->GetMasterModel()->GetLayer(0));
  lGenerator->GetLayer()->SetRenderingFunction( lLayer->GetRenderingFunction() );
  //FIXME define a generic method SetParameter with a map for rendering function for example
  // to avoid these ugly cast...
  static_cast<StandardRenderingFunctionType *>(lGenerator->GetLayer()->GetRenderingFunction())->SetAutoMinMax( false );
  // Add the layer to the working model

  this->AddWorkingLayer(lGenerator->GetLayer());

  this->GetWorkingModel()->SetExtractRegion( m_ExtractRegion );
  this->GetWorkingModel()->Update();

   m_WorkingHasNewExtract = true;
  this->NotifyAll();
  m_WorkingHasNewExtract = false;

  ColorType tempColor = m_Color;
 // update vector data in working view ONLY if there's some
  for(unsigned int id = 0; id< m_VectorDataList->Size(); id++)
    {
      this->SetColor( m_UsedColorList[id] );
      this->UpdateWorkingVectorData( id );
    }
  m_Color = tempColor;

  m_NewExtractVDDisplayed = true;
  this->NotifyAll();
  m_NewExtractVDDisplayed = false;
}

/*************************************************************************
 ***************** VECTOR DATA MANIPULATION *****************************
 *************************************************************************/
template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::LoadVectorData()
{
  VectorDataReaderType::Pointer vectorDataReader = VectorDataReaderType::New();
  vectorDataReader->SetFileName(m_VectorDataFileName);
  vectorDataReader->Update();
  m_VectorDataList->PushBack( vectorDataReader->GetOutput() );
  m_UsedColorList.push_back(m_Color);

  unsigned int vectorId = m_VectorDataList->Size()-1;
  this->UpdateMasterVectorData(vectorId);
  this->UpdateWorkingVectorData(vectorId);
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::AddVectorData( VectorDataPointerType vectorData )
{
  m_VectorDataList->PushBack( vectorData );
  unsigned int vectorId = m_VectorDataList->Size()-1;
  m_UsedColorList.push_back(m_Color);
  this->UpdateMasterVectorData(vectorId);
  this->UpdateWorkingVectorData(vectorId);
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
:: ClearVectorData( unsigned int id )
{
  m_VectorDataList->Erase( id );
  m_MasterVectorDataGlList->Erase( id );
  m_WorkingVectorDataGlList->Erase( id );
  m_UsedColorList.erase( m_UsedColorList.begin()+id );
  m_LastErasedVectorData = id;

  m_HasSupressedVectorData = true;
  this->NotifyAll();
  m_HasSupressedVectorData = false;
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::ClearAllVectorData()
{
  m_VectorDataList->Clear();
  m_MasterVectorDataGlList->Clear();
  m_WorkingVectorDataGlList->Clear();
  m_UsedColorList.clear();

  m_HasSupressedVectorData = true;
  this->NotifyAll();
  m_HasSupressedVectorData = false;
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateVectorData()
{
  m_MasterVectorDataGlList->Clear();
  m_WorkingVectorDataGlList->Clear();

  m_HasSupressedVectorData = true;
  this->NotifyAll();
  m_HasSupressedVectorData = false;

  for(unsigned int i=0; i<m_VectorDataList->Size(); i++)
    {
      m_Color = m_UsedColorList[i];
      this->UpdateMasterVectorData(i);
      this->UpdateWorkingVectorData(i);
    }
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateMasterVectorData( unsigned int vectorId )
{
  VectorDataGlComponentPointerType vect;
  vect = this->UpdateVectorData( m_InputImage, vectorId );
  m_MasterVectorDataGlList->PushBack( vect );

  m_MasterHasNewVectorData = true;
  this->NotifyAll();
  m_MasterHasNewVectorData = false;

}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateWorkingVectorData( unsigned int vectorId )
{
  VectorDataGlComponentPointerType vect;
  vect = this->UpdateVectorData( m_Extractor->GetOutput(), vectorId );
  m_WorkingVectorDataGlList->PushBack( vect );

  m_WorkingHasNewVectorData = true;
  this->NotifyAll();
  m_WorkingHasNewVectorData = false;
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::ChangeVectorDataColor( unsigned int id, ColorType color )
{
  if( id>=m_WorkingVectorDataGlList->Size() || id>=m_MasterVectorDataGlList->Size() || id>=m_UsedColorList.size() )
    {
      itkExceptionMacro(<<"Can't change color of the vector data "<<id<<", invalid index");
    }
  m_WorkingVectorDataGlList->GetNthElement( id )->SetColor(color);
  m_MasterVectorDataGlList->GetNthElement( id )->SetColor(color);
  m_UsedColorList[id] = color;

  m_HasToUpdate = true;
  this->NotifyAll();
  m_HasToUpdate = false;

}


template < class TInputImage, class TOutputImage >
typename FocusingModel<TInputImage, TOutputImage>
::VectorDataGlComponentPointerType
FocusingModel<TInputImage, TOutputImage>
::UpdateVectorData( InputImagePointerType image, unsigned int vectorId )
{
  if( vectorId>=m_VectorDataList->Size() )
    {
      itkExceptionMacro(<<"Can't display the vector data "<<vectorId<<", invalid index");
    }

  // Extract The part of the VectorData that actually overlaps with
  // the image extent
  VectorDataExtractROIType::Pointer vdextract = VectorDataExtractROIType::New();
  vdextract->SetInput( m_VectorDataList->GetNthElement( vectorId ) );
  RegionType region = image->GetLargestPossibleRegion();
  // Find the geographic region of interest

  // Ge the index of the corner of the image
  IndexType ul, ur, ll, lr;
  PointType pul, pur, pll, plr;
  ul = region.GetIndex();
  ur = ul;
  ll = ul;
  lr = ul;
  ur[0]+=m_InputImage->GetLargestPossibleRegion().GetSize()[0];
  lr[0]+=m_InputImage->GetLargestPossibleRegion().GetSize()[0];
  lr[1]+=m_InputImage->GetLargestPossibleRegion().GetSize()[1];
  ll[1]+=m_InputImage->GetLargestPossibleRegion().GetSize()[1];

  // Transform to physical point
  m_InputImage->TransformIndexToPhysicalPoint(ul, pul);
  m_InputImage->TransformIndexToPhysicalPoint(ur, pur);
  m_InputImage->TransformIndexToPhysicalPoint(ll, pll);
  m_InputImage->TransformIndexToPhysicalPoint(lr, plr);

  // Build the cartographic region
  RemoteSensingRegionType rsRegion;
  typename RemoteSensingRegionType::IndexType rsOrigin;
  typename RemoteSensingRegionType::SizeType  rsSize;
  rsOrigin[0]= std::min(pul[0], plr[0]);
  rsOrigin[1]= std::min(pul[1], plr[1]);
  rsSize[0]=vcl_abs(pul[0]-plr[0]);
  rsSize[1]=vcl_abs(pul[1]-plr[1]);

  rsRegion.SetOrigin(rsOrigin);
  rsRegion.SetSize(rsSize);
  rsRegion.SetRegionProjection(m_InputImage->GetProjectionRef());
  rsRegion.SetKeywordList(m_InputImage->GetImageKeywordlist());

  // Set the cartographic region to the extract roi filter
  vdextract->SetRegion(rsRegion);
  if( m_UseDEM==true )
    {
    otb::DEMHandler::Instance()->OpenDEMDirectory(m_DEMDirectory);
    }

  // Reproject VectorData in image projection
  VectorDataProjectionFilterType::Pointer vproj = VectorDataProjectionFilterType::New();

  // Watcher on the Vector data projection
  FltkFilterWatcher qlwatcher(vproj, 0, 0, 250, 20,"Reading Vector Data ...");

  vproj->SetInput(vdextract->GetOutput());
  vproj->SetInputProjectionRef(m_VectorDataList->GetNthElement( vectorId )->GetProjectionRef());
  vproj->SetOutputProjectionRef(m_InputImage->GetProjectionRef());
  vproj->SetOutputKeywordList(m_InputImage->GetImageKeywordlist());
  vproj->SetOutputOrigin(image->GetOrigin());
  vproj->SetOutputSpacing(m_InputImage->GetSpacing());
  if( m_UseDEM==true )
    {
    otb::DEMHandler::Instance()->OpenDEMDirectory(m_DEMDirectory);
    }
  vproj->Update();

  // Create a VectorData gl component
  VectorDataGlComponentPointerType compo  = VectorDataGlComponentType::New();
  compo->SetVectorData(vproj->GetOutput());
  compo->SetVisible(true);
  compo->SetRenderPolygonBoundariesOnly( m_PolygonBoundariesOnly );
  // if m_PolygonBoundariesOnly is set to true later
  compo->SetLineWidth(1);
  compo->SetColor(m_Color);

  return compo;
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::ChangeOpacity( double opacity )
{
  if(opacity > 1)
    opacity = 1;
  if(opacity < 0)
    opacity = 0;

  for(unsigned int i=0; i<m_UsedColorList.size(); i++)
    {
      ColorType color = m_UsedColorList[i];
      color[3] = opacity;
      this->ChangeVectorDataColor(i, color);
    }
  m_Color[3] = opacity;
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::ChangeGlRepresentation()
{
  m_PolygonBoundariesOnly = !m_PolygonBoundariesOnly;
  for(unsigned int i=0; i<m_MasterVectorDataGlList->Size(); i++)
    {
      m_MasterVectorDataGlList->GetNthElement(i)->SetRenderPolygonBoundariesOnly( m_PolygonBoundariesOnly ); //!val );
    }
  for(unsigned int j=0; j<m_WorkingVectorDataGlList->Size(); j++)
    {
      m_WorkingVectorDataGlList->GetNthElement(j)->SetRenderPolygonBoundariesOnly( m_PolygonBoundariesOnly ); //!val );
    }

  m_HasToUpdate = true;
  this->NotifyAll();
  m_HasToUpdate = false;
}


template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::ClearLayers()
{
  m_MasterModel->ClearLayers();
  m_WorkingModel->ClearLayers();
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::NotifyListener(ListenerBase * listener)
{
    listener->Notify();
}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateRGBChannelOrder(unsigned int red, unsigned int green, unsigned int blue)
{
  StandardRenderingFunctionPointerType renderFunction;
  renderFunction = StandardRenderingFunctionType::New();

  renderFunction->GetPixelRepresentationFunction()->SetChannelIndex(0, red);
  renderFunction->GetPixelRepresentationFunction()->SetChannelIndex(1, green);
  renderFunction->GetPixelRepresentationFunction()->SetChannelIndex(2, blue);

  renderFunction->SetAutoMinMax( false );

  m_RenderingFunction = renderFunction; //FIXME evaluate the cost of evaluating the rendering function from scratch
//   m_RenderingFunction->SetRedChannelIndex(red);
//   m_RenderingFunction->SetGreenChannelIndex(green);
//   m_RenderingFunction->SetBlueChannelIndex(blue);

 /** Update Master View */
  m_MasterModel->ClearLayers();
  // Generate the layer
  typename LayerGeneratorType::Pointer lGenerator = LayerGeneratorType::New();

  // Watcher
  FltkFilterWatcher qlwatcher(lGenerator->GetProgressSource(), 0, 0, 250, 20,"Changing channels order...");

  lGenerator->SetImage(m_InputImage);
  lGenerator->SetScreenRatio(m_MasterScreenRatio);
  lGenerator->GenerateLayer();
  lGenerator->GetLayer()->SetRenderingFunction(m_RenderingFunction);
//   lGenerator->GetLayer()->SetAutoMinMax( false );
  // Add the layer to the master model
  this->AddMasterLayer(lGenerator->GetLayer());

  m_MasterModel->Update();

  /** Update Working View */
  this->Update();

}

template < class TInputImage, class TOutputImage >
void
FocusingModel<TInputImage, TOutputImage>
::UpdateGrayscaleChannel(unsigned int channel)
{

  StandardRenderingFunctionPointerType renderFunction;
  renderFunction = StandardRenderingFunctionType::New();
  renderFunction->GetPixelRepresentationFunction()->SetAllChannels(channel);

  renderFunction->SetAutoMinMax( false );

  m_RenderingFunction = renderFunction; //FIXME evaluate the cost of evaluating the rendering function from scratch

  //   m_RenderingFunction->SetAllChannels(channel);

  /** Update Master View */
  m_MasterModel->ClearLayers();
  // Generate the layer
  typename LayerGeneratorType::Pointer lGenerator = LayerGeneratorType::New();

  // Watcher
  FltkFilterWatcher qlwatcher(lGenerator->GetProgressSource(), 0, 0, 250, 20,"Changing channels order...");

  lGenerator->SetImage(m_InputImage);
  lGenerator->SetScreenRatio(m_MasterScreenRatio);
  lGenerator->GenerateLayer();
  lGenerator->GetLayer()->SetRenderingFunction(m_RenderingFunction);
//   lGenerator->GetLayer()->SetAutoMinMax( false );
  // Add the layer to the master model
  this->AddMasterLayer(lGenerator->GetLayer());

  m_MasterModel->Update();

  /** Update Working View */
  this->Update();
}

}

#endif
