/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbFocusingView_h
#define __otbFocusingView_h


#include "otbImageWidgetActionHandler.h"
#include "itkObject.h"
#include "otbImageView.h"
#include "otbMacro.h"
#include "otbExtractROI.h"
#include "otbMultiChannelExtractROI.h"
#include "otbFocusingModel.h"
#include "otbFocusingController.h"

namespace otb
{
/** \class FocusingView
 *  We use the scrollview of the "MasterViewer", where the user will select a particular area.
 *  Then the selected area will appear in the second scroll ("workingViewer").
 */
template <class TFocusingModel >
class FocusingView :
  public ListenerBase, public itk::Object
{
 public:
  /** Standard class typedefs */
  typedef FocusingView                     Self;
  typedef itk::Object                      Superclass;
  typedef itk::SmartPointer<Self>          Pointer;
  typedef itk::SmartPointer<const Self>    ConstPointer;

  /** Standards macros */
  itkNewMacro(Self);

  /** Runtime information */
  itkTypeMacro(FocusingView, itk::Object/*ListenerBase*/);

  /** Model typedef */
  typedef TFocusingModel                                           FocusingModelType;
  typedef typename FocusingModelType::Pointer                      FocusingModelPointerType;
  typedef typename FocusingModelType::ImageLayerRenderingModelType ModelType;
  typedef typename ModelType::Pointer                              ModelPointerType;
  typedef typename ModelType::OutputImageType                      ImageType;

  /** View typedef */
  typedef ImageView<ModelType>                         ImageViewType;
  typedef typename ImageViewType::Pointer              ImageViewPointerType;
  typedef typename ImageType::RegionType               RegionType;
  typedef typename RegionType::SizeType                SizeType;
  typedef typename ImageType::PixelType                PixelType;
  typedef typename RegionType::IndexType               IndexType;
  typedef typename ImageType::Pointer                  ImagePointerType;
  typedef typename ImageViewType::ImageWidgetType      ImageWidgetType;
  typedef typename ImageWidgetType::Pointer            ImageWidgetPointerType;

  /** Controller typedef */
  typedef FocusingController<FocusingModelType, Self>  FocusingControllerType;
  typedef typename FocusingControllerType::Pointer     FocusingControllerPointerType;
  typedef ImageWidgetController                        ControllerType;
  typedef typename ControllerType::Pointer             ControllerPointerType;

  typedef MultiChannelExtractROI<PixelType, PixelType> ExtractROIType;
  typedef typename ExtractROIType::Pointer             ExtractROIPointerType;

  /** Get pointers the widgets */
  itkGetObjectMacro(MasterViewer, ImageViewType);
  itkGetObjectMacro(WorkingViewer, ImageViewType);

  /** Set Master view model. */
  void SetMasterModel(ModelPointerType model){
    m_MasterViewer->SetModel(model);
  }
  /** Set Working view model. */
  void SetWorkingModel(ModelPointerType model){
    m_WorkingViewer->SetModel(model);
  }
  /** Set Master view controller. */
  void SetMasterController(ControllerPointerType controller){
    m_MasterViewer->SetController(controller);
  }
  /** Set Working view controller. */
  void SetWorkingController(ControllerPointerType controller){
    m_WorkingViewer->SetController(controller);
  }

  /** Get Master view scroll widget. */
  ImageWidgetPointerType GetMasterScrollWidget(){
    return m_MasterViewer->GetScrollWidget();
  }
  /** Get Working view scroll widget. */
  ImageWidgetPointerType GetWorkingScrollWidget(){
    return m_WorkingViewer->GetScrollWidget();
  }
  /** Get Working view full widget. */
  ImageWidgetPointerType GetWorkingFullWidget(){
    return m_WorkingViewer->GetFullWidget();
  }

  /**Set the focusing Model */
  void SetFocusingModel(FocusingModelType * model);

  /**Set the focusing Controller */
  void SetFocusingController(FocusingControllerType * controller);

  /** Update the red square on the masterScroll */
  void UpdateSelectedArea(IndexType startIndex, IndexType stopIndex);

  /** Clear GL Components of the views. */
  void ClearGLComponents();

  /** Clear the Gl Components of the working view :
   * first GlComponent mustn't be erase (scaled and full selection Glcomponent).
   */
  void ClearWorkingGLComponents();
  void ClearMasterGLComponents();
  /** Hide/Show the added GlComponent */
  // all gl from model
  void SetGLComponentsStatus( bool val );
  // a list
  void SetGLComponentsStatus(bool val, std::vector<unsigned int> vect);

  virtual void Notify();

  /** Printself method */
  void PrintSelf(std::ostream& os, itk::Indent indent) const;

 protected:
  /** Constructor */
  FocusingView();
  /** Destructor */
  ~FocusingView();

 private:
  /** Master view (allows roi selection). */
  ImageViewPointerType m_MasterViewer;
  /** Working view (displays selected roi).*/
  ImageViewPointerType m_WorkingViewer;
  /** Area selection indexes. */
  IndexType m_StartIndex;
  IndexType m_StopIndex;
  /** Pointer to the model. */
  FocusingModelPointerType m_FocusingModel;
  /** Pointer to the model. */
  FocusingControllerPointerType m_FocusingController;

}; // end class
} // end namespace otb

#ifndef  OTB_MANUAL_INSTANTIATION
#include "otbFocusingView.txx"
#endif

#endif


