/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

     =========================================================================*/
#ifndef __otbFocusingView_txx
#define __otbFocusingView_txx

#include "otbFocusingView.h"

namespace otb
{
template < class TFocusingModel >
FocusingView<TFocusingModel>
::FocusingView() : m_MasterViewer(), m_WorkingViewer(), m_FocusingModel()
{
  // Initializing the two viewers
  m_MasterViewer = ImageViewType::New();
  m_WorkingViewer = ImageViewType::New();
  
  m_MasterViewer->GetExtractRegionGlComponent()->SetVisible(false);
  m_MasterViewer->GetScrollWidget()->SetIdentifier("Master View");
  m_WorkingViewer->GetScrollWidget()->SetIdentifier("Working View");
}
  
  
template < class TFocusingModel >
FocusingView<TFocusingModel>
::~FocusingView()
{
}


template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::SetFocusingModel(FocusingModelType * model)
{
  // Unregister from previous model if nay
  if(m_FocusingModel.IsNotNull())
    {
      m_FocusingModel->UnRegisterListener(this);
    }
  // Set and register with new model
  m_FocusingModel = model;
  m_FocusingModel->RegisterListener(this);
  this->SetMasterModel(m_FocusingModel->GetMasterModel());
  this->SetWorkingModel(m_FocusingModel->GetWorkingModel());
}

template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::SetFocusingController(FocusingControllerType * controller)
{
  // Set and register with new model
  m_FocusingController = controller;
  this->SetMasterController(m_FocusingController->GetMasterController());
  this->SetWorkingController(m_FocusingController->GetWorkingController());
}


template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::UpdateSelectedArea(IndexType startIndex, IndexType stopIndex)
{
  RegionType lImageRegion;
  lImageRegion = m_MasterViewer->GetModel()->GetLayer(0)->GetExtent();
  
  SizeType lSize;
  lSize[0] = vcl_abs(stopIndex[0]-startIndex[0]);
  lSize[1] = vcl_abs(stopIndex[1]-startIndex[1]);
  
  IndexType lIndex;
  lIndex[0] = std::min(startIndex[0], stopIndex[0]);
  lIndex[1] = std::min(startIndex[1], stopIndex[1]);
  
  RegionType lRegion;
  lRegion.SetIndex(lIndex);
  lRegion.SetSize(lSize);
  
  if(lRegion.Crop(lImageRegion))
    {
      m_MasterViewer->GetExtractRegionGlComponent()->SetVisible(true);
      m_MasterViewer->GetExtractRegionGlComponent()->SetRegion(lRegion);
      m_MasterViewer->GetScrollWidget()->redraw();
    }
}

template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::ClearGLComponents()
{
  this->ClearMasterGLComponents();
  this->ClearWorkingGLComponents();
}

// separate method because of the selected area display for full window
template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::ClearWorkingGLComponents()
{
  // supress everything except the first one
  unsigned int sizeW = this->GetWorkingScrollWidget()->GetNumberOfGlComponents();
  for(unsigned int i=1; i<sizeW; i++ )
    {
      this->GetWorkingScrollWidget()->RemoveGlComponent(1);
    }
  // supress everything except the first one
  unsigned int sizeF = this->GetWorkingFullWidget()->GetNumberOfGlComponents();
  for(unsigned int i=1; i<sizeF; i++ )
    {
      this->GetWorkingFullWidget()->RemoveGlComponent(1);
    }
}


// separate method because of the selected area display for full window
template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::ClearMasterGLComponents()
{
  // supress everything except the first one
  unsigned int sizeM = this->GetMasterScrollWidget()->GetNumberOfGlComponents();
  for(unsigned int i=1; i<sizeM; i++ )
    {
      this->GetMasterScrollWidget()->RemoveGlComponent(1);
    }
}


template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::SetGLComponentsStatus(bool val)
{
  for(unsigned int i=0; i< m_FocusingModel->GetMasterVectorDataGlList()->Size(); i++ )
    {
      m_FocusingModel->GetMasterVectorDataGlList()->GetNthElement( i )->SetVisible(val);
    }
  for(unsigned int j=0; j< m_FocusingModel->GetWorkingVectorDataGlList()->Size(); j++ )
    {
      m_FocusingModel->GetWorkingVectorDataGlList()->GetNthElement( j )->SetVisible(val);
    }

  this->GetWorkingScrollWidget()->redraw();
  this->GetWorkingFullWidget()->redraw();
  this->GetMasterScrollWidget()->redraw();
}

template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::SetGLComponentsStatus(bool val, std::vector<unsigned int> vect)
{
  if( vect.size() > m_FocusingModel->GetMasterVectorDataGlList()->Size() ||  vect.size() > m_FocusingModel->GetWorkingVectorDataGlList()->Size() )
    itkExceptionMacro(<<"Can't change GlComponentsStatus, Dimensions mismatch");

  for(unsigned int i=0; i< vect.size(); i++ )
    {
      m_FocusingModel->GetMasterVectorDataGlList()->GetNthElement( vect[i] )->SetVisible(val);
      m_FocusingModel->GetWorkingVectorDataGlList()->GetNthElement( vect[i] )->SetVisible(val);
    }

  this->GetWorkingScrollWidget()->redraw();
  this->GetWorkingFullWidget()->redraw();
  this->GetMasterScrollWidget()->redraw();
}

template < class TFocusingModel >
void
FocusingView<TFocusingModel>
::Notify()
{
  if( m_FocusingModel->GetWorkingHasNewExtract() == true )
    {
      // Supress printed vector datas
      this->ClearWorkingGLComponents();
      this->GetWorkingScrollWidget()->ClearBuffer();
      this->GetWorkingFullWidget()->ClearBuffer();
      this->GetWorkingScrollWidget()->redraw();
      this->GetWorkingFullWidget()->redraw();
 
      // Update the full window : if select a little area (full image extract smaller than window)
      // and then select a bigger area, the full window keep the smaller one.
      int x = m_WorkingViewer->GetFullWidget()->x();
      int y = m_WorkingViewer->GetFullWidget()->y();
      int w = m_WorkingViewer->GetFullWidget()->w();
      int h = m_WorkingViewer->GetFullWidget()->h();
      // +1 and -1 to back to the original size
      m_WorkingViewer->GetFullWidget()->resize(x, y, w, h+1);
      m_WorkingViewer->GetFullWidget()->resize(x, y, w, h-1);
    }

  // Update vector data display
  if( m_FocusingModel->GetMasterHasNewVectorData() == true )
    {
      unsigned int id = m_FocusingModel->GetMasterVectorDataGlList()->Size()-1;
      this->GetMasterScrollWidget()->AddGlComponent( m_FocusingModel->GetMasterVectorDataGlList()->GetNthElement( id ) );
      this->GetMasterScrollWidget()->redraw();
    }

  if( m_FocusingModel->GetWorkingHasNewVectorData() == true )
    {
      unsigned int id = m_FocusingModel->GetWorkingVectorDataGlList()->Size()-1;
      this->GetWorkingScrollWidget()->AddGlComponent( m_FocusingModel->GetWorkingVectorDataGlList()->GetNthElement( id ) );
      this->GetWorkingFullWidget()->AddGlComponent( m_FocusingModel->GetWorkingVectorDataGlList()->GetNthElement( id ) );
      this->GetWorkingScrollWidget()->redraw();
      this->GetWorkingFullWidget()->redraw();
    }
  if( m_FocusingModel->GetHasSupressedVectorData() == true )
    {
      // model->ClearAllVectorData
      if( m_FocusingModel->GetWorkingVectorDataGlList()->Size()==0 &&  m_FocusingModel->GetMasterVectorDataGlList()->Size()==0)
       {
         this->ClearGLComponents();
       }
      // model->ClearVectorData(unsigned int)
      else
       {
         // +1 to shift the selection rectangles
         unsigned int id = m_FocusingModel->GetLastErasedVectorData()+1;
         this->GetWorkingScrollWidget()->RemoveGlComponent(id);
         this->GetWorkingFullWidget()->RemoveGlComponent(id);
         this->GetMasterScrollWidget()->RemoveGlComponent(id);
       }
      this->GetWorkingScrollWidget()->redraw();
      this->GetWorkingFullWidget()->redraw();
      this->GetMasterScrollWidget()->redraw();
    }
  if( m_FocusingModel->GetHasToInit() == true )
    {
      this->ClearGLComponents();
      this->GetWorkingScrollWidget()->ClearBuffer();
      this->GetWorkingScrollWidget()->redraw();
      this->GetWorkingFullWidget()->ClearBuffer();
      this->GetWorkingFullWidget()->redraw();
      this->GetMasterScrollWidget()->ClearBuffer();
      this->GetMasterScrollWidget()->redraw();
    }
  if( m_FocusingModel->GetHasToUpdate() == true )
    {
      this->GetWorkingScrollWidget()->redraw();
      this->GetWorkingFullWidget()->redraw();
      this->GetMasterScrollWidget()->redraw();
    }
}
  


template < class TFocusingModel>
void
FocusingView<TFocusingModel>
::PrintSelf(std::ostream& os, itk::Indent indent) const
{
}

}

#endif


