/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbFineRegistrationApplicationController.h"
#include "otbMsgReporter.h"

#include "otbFltkWriterWatcher.h"
#include "otbFltkFilterWatcher.h"
#include <FL/fl_ask.H>

namespace otb
{

FineRegistrationApplicationController
::FineRegistrationApplicationController()
{
  m_IsExtractRunning = false;
}

FineRegistrationApplicationController
::~FineRegistrationApplicationController()
{
  MsgReporter::GetInstance()->Hide();
}

void
FineRegistrationApplicationController
::OpenImagePair(const char * moving, const char * fixed)
{
  // Close already opened pair if any
  if (m_Model->GetState() != NO_IMAGES)
  {
    m_Model->ClosePair();
  }

  // Watch the shrinkers
  FltkFilterWatcher watcher1(m_Model->GetFixedShrinkFilter(), 0, 0, 200, 20, "Loading left image ...");
  FltkFilterWatcher watcher2(m_Model->GetMovingShrinkFilter(), 0, 0, 200, 20, "Loading right image ...");

  try
  {
    m_Model->OpenPair(fixed, moving);
    ViewType::RegionType region;
    IndexType newIndex;
    newIndex.Fill(0);
    ViewType::SizeType newSize;
    newSize[0]=m_View->gImageViewer->w();
    newSize[1]=m_View->gImageViewer->h();
    region.SetIndex(newIndex);
    region.SetSize(newSize);
    m_Model->SetCurrentRegion(region);
    m_Model->RunImageExtraction();
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void
FineRegistrationApplicationController
::IndexClicked(const IndexType & index)
{
  if ( !m_IsExtractRunning )
  {
    m_IsExtractRunning = true;
    ViewType::RegionType region = m_View->GetFullImageWidget()->GetBufferedRegion();
    IndexType newIndex;
    newIndex[0]=index[0]*m_Model->GetShrinkFactor()-region.GetSize()[0]/2;
    newIndex[1]=index[1]*m_Model->GetShrinkFactor()-region.GetSize()[1]/2;
    region.SetIndex(newIndex);
    m_Model->SetCurrentRegion(region);
    m_Model->RunImageExtraction();
    m_IsExtractRunning = false;
  }
}


void
FineRegistrationApplicationController
::Run()
{
  FltkFilterWatcher watcher2(m_Model->GetRegistrationFilter(), 0, 0, 300, 20, "Performing registration ...");
  m_Model->RunDeformationFieldEstimation();
}

void
FineRegistrationApplicationController
::ChangeNormalizationFactor(double factor)
{
  m_Model->SetClampThreshold(factor/100.);
}

void
FineRegistrationApplicationController
::SaveDisparityMap(const char * filename)
{
  FltkWriterWatcher watcher(m_Model->GetDeformationWriter(), m_Model->GetRegistrationFilter(), 0, 0, 250, 50,"Writing deformation field ...");

  try
  {
    m_Model->SaveDisparityMap(filename);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void
FineRegistrationApplicationController
::SaveRegisteredImage(const char * filename)
{
  FltkWriterWatcher watcher(m_Model->GetRegistrationWriter(), m_Model->GetRegistrationFilter(), 0, 0, 250, 50,"Writing registered image ...");
  try
  {
    m_Model->SaveRegisteredImage(filename);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void
FineRegistrationApplicationController
::ChangeMaxDeformationValue(double minVal, double maxVal)
{
  FineRegistrationApplicationModel::DeformationValueType val;
  val[0] = minVal;
  val[1] = maxVal;

  m_Model->SetMaxDeformationValue(val);
}


void
FineRegistrationApplicationController
::SetNumberOfIterations(int nbIter)
{
  m_Model->SetNumberOfIterations( nbIter );
}

void
FineRegistrationApplicationController
::SetRadius(unsigned int rad, int id)
{
  RadiusType radius = m_Model->GetRadius();
  radius[id] = rad;
  this->SetRadius( radius );
}

void
FineRegistrationApplicationController
::SetRadius(RadiusType rad)
{
  m_Model->SetRadius( rad );
}


} // end namespace otb

