/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbFineRegistrationApplicationModel.h"

#include "otbImageFileWriter.h"

namespace otb
{

/** Initialize the singleton */
FineRegistrationApplicationModel::Pointer FineRegistrationApplicationModel::Instance = NULL;

/** Manage the singleton */
FineRegistrationApplicationModel::Pointer
FineRegistrationApplicationModel
::GetInstance()
{
  if (!Instance)
  {
    Instance = FineRegistrationApplicationModel::New();
  }
  return Instance;
}

void
FineRegistrationApplicationModel
::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

FineRegistrationApplicationModel
::FineRegistrationApplicationModel()
{
  m_QuicklookSize.Fill(256);
  m_ClampThreshold = 0.01;
  m_MaxDeformationValue.Fill(1);
  m_State = NO_IMAGES;
  this->InitFilters();
}

FineRegistrationApplicationModel

::~FineRegistrationApplicationModel()
{}

const char *
FineRegistrationApplicationModel
::GetMovingImageFilename()
{
  return m_MovingReader->GetFileName();
}

const char *
FineRegistrationApplicationModel
::GetFixedImageFilename()
{
  return m_FixedReader->GetFileName();
}

FineRegistrationApplicationModel::VectorImageType *
FineRegistrationApplicationModel
::GetImageColorComposition()
{
  return m_Image;
}

FineRegistrationApplicationModel::VectorImageType *
FineRegistrationApplicationModel
::GetDeformationFieldColorComposition()
{
  return m_Deformation;
}


FineRegistrationApplicationModel::VectorImageType *
FineRegistrationApplicationModel
::GetQuicklookColorComposition()
{
  return m_Quicklook;
}


FineRegistrationApplicationModel::ShrinkFilterType *
FineRegistrationApplicationModel
::GetMovingShrinkFilter()
{
  return m_MovingShrinker;
}


FineRegistrationApplicationModel::ShrinkFilterType *
FineRegistrationApplicationModel
::GetFixedShrinkFilter()
{
  return m_FixedShrinker;
}

unsigned int
FineRegistrationApplicationModel
::GetShrinkFactor()
{
  return m_ShrinkFactor;
}

FineRegistrationApplicationModel::RadiusType
FineRegistrationApplicationModel
::GetRadius() const
{
  return m_RegistrationFilter->GetNCCRadius();
}

void
FineRegistrationApplicationModel
::SetRadius(const RadiusType & radius)
{
  m_RegistrationFilter->SetNCCRadius(radius);
}

unsigned int
FineRegistrationApplicationModel
::GetNumberOfIterations() const
{
  return m_RegistrationFilter->GetNumberOfIterations();
}

void
FineRegistrationApplicationModel
::SetNumberOfIterations(unsigned int nbIterations)
{
  m_RegistrationFilter->SetNumberOfIterations(nbIterations);
}

FineRegistrationApplicationModel::PixelType
FineRegistrationApplicationModel
::GetMin() const
{

  PixelType min=m_Histogram->Quantile(0, m_ClampThreshold);
  return min;
}

FineRegistrationApplicationModel::PixelType
FineRegistrationApplicationModel
::GetMax() const
{
  PixelType max;
  max=m_Histogram->Quantile(0, 1-m_ClampThreshold);
  return max;
}

FineRegistrationApplicationModel::DeformationValueType
FineRegistrationApplicationModel
::GetMaxDeformationValue()
{
  return m_MaxDeformationValue;
}

void
FineRegistrationApplicationModel
::SetMaxDeformationValue( DeformationValueType maxVal )
{
  m_MaxDeformationValue = maxVal;
}


void
FineRegistrationApplicationModel
::InitFilters()
{
  m_MovingReader = ReaderType::New();
  m_FixedReader = ReaderType::New();
  m_MovingShrinker = ShrinkFilterType::New();
  m_FixedShrinker = ShrinkFilterType::New();
  m_RegistrationFilter = RegistrationFilterType::New();
  m_WarpFilter = WarpFilterType::New();
  m_Interpolator = InterpolatorType::New();
  m_XDeformationFieldExtractor   = ChannelExtractionFilterType::New();
  m_YDeformationFieldExtractor   = ChannelExtractionFilterType::New();
  m_DeformationFieldIntensityFilter = IntensityFilterType::New();
  m_Deformation = VectorImageType::New();
  m_RegistrationWriter = RegistrationWriterType::New();
  m_DeformationWriter = DeformationWriterType::New();
}

void FineRegistrationApplicationModel
::WirePipeline()
{
  m_MovingShrinker->SetInput(m_MovingReader->GetOutput());
  m_FixedShrinker->SetInput(m_FixedReader->GetOutput());

  // Setup the registration filter
  m_RegistrationFilter->SetMovingImage(m_MovingReader->GetOutput());
  m_RegistrationFilter->SetFixedImage(m_FixedReader->GetOutput());

  m_WarpFilter->SetInput(m_MovingReader->GetOutput());
  m_WarpFilter->SetDeformationField(m_RegistrationFilter->GetOutput());
  m_WarpFilter->SetInterpolator(m_Interpolator);

  m_XDeformationFieldExtractor->SetInput(m_RegistrationFilter->GetOutput());
  m_XDeformationFieldExtractor->SetChannel(1);
  m_YDeformationFieldExtractor->SetInput(m_RegistrationFilter->GetOutput());
  m_YDeformationFieldExtractor->SetChannel(2);
  m_DeformationFieldIntensityFilter->SetInput(m_RegistrationFilter->GetOutput());
}

void
FineRegistrationApplicationModel
::ClosePair()
{
  this->InitFilters();
  m_State = NO_IMAGES;
  this->NotifyAll();
}

void
FineRegistrationApplicationModel
::OpenPair(const char * movingName, const char * fixedName)
{
  if (m_State != NO_IMAGES)
  {
    itkExceptionMacro(<<"A pair of images is already loaded. Please call ClosePair() first.");
  }

  this->WirePipeline();

  // Trying to read left image
  m_MovingReader->SetFileName(movingName);
  m_MovingReader->GenerateOutputInformation();

  // Trying to read right image
  m_FixedReader->SetFileName(fixedName);
  m_FixedReader->GenerateOutputInformation();

  if (m_MovingReader->GetOutput()->GetLargestPossibleRegion()
      != m_FixedReader->GetOutput()->GetLargestPossibleRegion())
  {
    itkExceptionMacro(<<"Moving and fixed images do not have the same size !");
  }

  // Generate quicklook
  this->GenerateQuicklook();

  // Generate the joint histogram
  this->GenerateHistogram();

  m_State = NO_DEFORMATION;
}

void
FineRegistrationApplicationModel
::GenerateQuicklook()
{
  // Shrink factor
  SizeType largestRegionSize = m_MovingReader->GetOutput()->GetLargestPossibleRegion().GetSize();
  unsigned int maxQuicklook = std::max(m_QuicklookSize[0], m_QuicklookSize[1]);
  unsigned int maxLargest   = std::max(largestRegionSize[0], largestRegionSize[1]);

  m_ShrinkFactor =  maxLargest/maxQuicklook;

  if (m_ShrinkFactor == 0)
  {
    m_ShrinkFactor = 1;
  }

  // Moving shrinker
  m_MovingShrinker->SetShrinkFactor(m_ShrinkFactor);
  m_MovingShrinker->Update();

  // Fixed shrinker
  m_FixedShrinker->SetShrinkFactor(m_ShrinkFactor);
  m_FixedShrinker->Update();

  // Color composition shrinker
  ImageListPointerType il1 = ImageListType::New();
  il1->PushBack(m_FixedShrinker->GetOutput());
  il1->PushBack(m_MovingShrinker->GetOutput());
  il1->PushBack(m_MovingShrinker->GetOutput());

  ImageListToVectorImageFilterPointerType il2vi1 = ImageListToVectorImageFilterType::New();
  il2vi1->SetInput(il1);

  il2vi1->Update();
  m_Quicklook =il2vi1->GetOutput();
}

void
FineRegistrationApplicationModel
::SetupDeformationEstimation()
{
  // Setup the warp filter
  m_WarpFilter->SetOutputOrigin(m_FixedReader->GetOutput()->GetOrigin());
  m_WarpFilter->SetOutputSpacing(m_FixedReader->GetOutput()->GetSpacing());
  m_WarpFilter->SetMaximumDeformation( m_MaxDeformationValue );
}


void
FineRegistrationApplicationModel
::RunImageExtraction()
{
  if (m_State == NO_IMAGES)
  {
    itkExceptionMacro(<<"Can not run image extraction: there are no input images.");
  }

  if (!this->ConstrainsCurrentRegion())
  {
    m_CurrentRegion.Crop(m_FixedReader->GetOutput()->GetLargestPossibleRegion());
  }

  ExtractROIFilterType::Pointer extract = ExtractROIFilterType::New();

  // Extract images
  ImageListPointerType il = ImageListType::New();
  il->PushBack(m_FixedReader->GetOutput());
  il->PushBack(m_MovingReader->GetOutput());
  il->PushBack(m_MovingReader->GetOutput());

  ImageListToVectorImageFilterPointerType il2vi = ImageListToVectorImageFilterType::New();
  il2vi->SetInput(il);

  extract->SetInput(il2vi->GetOutput());
  extract->SetExtractionRegion(m_CurrentRegion);
  extract->Update();

  m_Image = extract->GetOutput();

  m_State = NO_DEFORMATION;

  this->NotifyAll();
}

bool FineRegistrationApplicationModel
::ConstrainsCurrentRegion()
{
  RegionType largest = m_FixedReader->GetOutput()->GetLargestPossibleRegion();

  if (  m_CurrentRegion.GetSize()[0]>largest.GetSize()[0]
        ||m_CurrentRegion.GetSize()[1]>largest.GetSize()[1])
  {
    return false;
  }
  RegionType::SizeType size = m_CurrentRegion.GetSize();
  RegionType::IndexType index = m_CurrentRegion.GetIndex();

  if ( index[0] < largest.GetIndex()[0])
  {
    index[0]=largest.GetIndex()[0];
  }

  if ( index[1] < largest.GetIndex()[1])
  {
    index[1]=largest.GetIndex()[1];
  }

  if (index[0]+size[0] > largest.GetIndex()[0]+largest.GetSize()[0])
  {
    index[0]=largest.GetIndex()[0]+largest.GetSize()[0]-size[0];
  }

  if (index[1]+size[1] > largest.GetIndex()[1]+largest.GetSize()[1])
  {
    index[1]=largest.GetIndex()[1]+largest.GetSize()[1]-size[1];
  }

  m_CurrentRegion.SetIndex(index);
  m_CurrentRegion.SetSize(size);

  return true;
}

void
FineRegistrationApplicationModel
::RunDeformationFieldEstimation()
{
  if (m_State == NO_IMAGES)
  {
    itkExceptionMacro(<<"Can not run registration: there are no input images.");
  }

  if (!this->ConstrainsCurrentRegion())
  {
    m_CurrentRegion.Crop(m_FixedReader->GetOutput()->GetLargestPossibleRegion());
  }

  this->SetupDeformationEstimation();

  // ROI Extraction filter
  ExtractROIFilterType::Pointer extract1 = ExtractROIFilterType::New();
  ExtractROIFilterType::Pointer extract2 = ExtractROIFilterType::New();

  // Extract deformation
  ImageListPointerType il1 = ImageListType::New();
  il1->PushBack(m_XDeformationFieldExtractor->GetOutput());
  il1->PushBack(m_YDeformationFieldExtractor->GetOutput());
  il1->PushBack(m_DeformationFieldIntensityFilter->GetOutput());

  ImageListToVectorImageFilterPointerType il2vi1 = ImageListToVectorImageFilterType::New();
  il2vi1->SetInput(il1);

  extract1->SetInput(il2vi1->GetOutput());
  extract1->SetExtractionRegion(m_CurrentRegion);
  extract1->Update();

  m_Deformation = extract1->GetOutput();

  // Extract images
  ImageListPointerType il2 = ImageListType::New();

  il2->PushBack(m_FixedReader->GetOutput());
  il2->PushBack(m_MovingReader->GetOutput());
  il2->PushBack(m_WarpFilter->GetOutput());

  ImageListToVectorImageFilterPointerType il2vi2 = ImageListToVectorImageFilterType::New();
  il2vi2->SetInput(il2);

  extract2->SetInput(il2vi2->GetOutput());
  extract2->SetExtractionRegion(m_CurrentRegion);
  extract2->Update();

  m_Image = extract2->GetOutput();

  // Change application stae
  m_State = PARTIAL_DEFORMATION;
  this->NotifyAll();
}

void
FineRegistrationApplicationModel
::GenerateHistogram()
{
  // Compute the normalization factors
  ImageType::Pointer leftQuicklook = m_MovingShrinker->GetOutput();
  ImageType::Pointer rightQuicklook = m_FixedShrinker->GetOutput();

  ListSampleType::Pointer sl = ListSampleType::New();

  IteratorType leftIt(leftQuicklook, leftQuicklook->GetLargestPossibleRegion());
  leftIt.GoToBegin();
  while ( !leftIt.IsAtEnd() )
  {
    PixelType pixel = leftIt.Get();
    sl->PushBack(pixel);
    ++leftIt;
  }

  IteratorType rightIt(rightQuicklook, rightQuicklook->GetLargestPossibleRegion());
  rightIt.GoToBegin();
  while ( !rightIt.IsAtEnd() )
  {
    PixelType pixel = rightIt.Get();
    sl->PushBack(pixel);
    ++rightIt;
  }

  HistogramGeneratorType::Pointer generator = HistogramGeneratorType::New();
  generator->SetListSample(sl);
  HistogramGeneratorType::HistogramType::SizeType size;
  size.Fill(256);
  generator->SetNumberOfBins(size);
  generator->Update();
  m_Histogram = const_cast<HistogramType *>(generator->GetOutput());
}


void
FineRegistrationApplicationModel
::SaveRegisteredImage(const char * filename)
{
  m_RegistrationWriter->SetFileName( filename );
  m_RegistrationWriter->SetInput( m_WarpFilter->GetOutput() );
  m_RegistrationWriter->Update();
  m_RegistrationWriter = RegistrationWriterType::New();
}

void
FineRegistrationApplicationModel
::SaveDisparityMap( const char * filename)
{
  m_DeformationWriter->SetFileName( filename );
  m_DeformationWriter->SetInput( m_RegistrationFilter->GetOutput() );
  m_DeformationWriter->Update();
  m_DeformationWriter = DeformationWriterType::New();
}

}
