/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbFineRegistrationApplicationModel_h
#define __otbFineRegistrationApplicationModel_h

#include "otbMVCModel.h"

#include "otbVectorImage.h"
#include "otbImageFileReader.h"
#include "otbNCCRegistrationFilter.h"
#include "itkWarpImageFilter.h"
#include "otbStreamingWarpImageFilter.h"
#include "itkLinearInterpolateImageFunction.h"
#include "otbVectorImageToColorAnaglyphVectorImageFilter.h"
#include "otbStreamingShrinkImageFilter.h"
#include "otbListenerBase.h"
#include "itkImageRegionIterator.h"
#include "itkListSample.h"
#include "itkListSampleToHistogramGenerator.h"
#include "otbImageList.h"
#include "otbImageOfVectorsToMonoChannelExtractROI.h"
#include "otbMultiChannelExtractROI.h"
#include "otbImageListToVectorImageFilter.h"
#include "otbVectorImageToIntensityImageFilter.h"
#include "otbStreamingImageFileWriter.h"

namespace otb
{

typedef enum
{
  NO_IMAGES,
  NO_DEFORMATION,
  PARTIAL_DEFORMATION
}
FineRegistrationApplicationStates;


/** \class FineRegistrationApplicationModel
 *
 *
 *
 */
class ITK_EXPORT FineRegistrationApplicationModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef FineRegistrationApplicationModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                              Pointer;
  typedef itk::SmartPointer<const Self>                        ConstPointer;

  /** Standard type macro */
  itkTypeMacro(FineRegistrationApplicationModel, Object);

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  typedef float                                                                           PixelType;
  typedef otb::Image<PixelType, 2>                                                         ImageType;
  typedef ImageType::SizeType                                                             SizeType;
  typedef ImageType::IndexType                                                            IndexType;
  typedef ImageType::RegionType                                                           RegionType;
  typedef otb::VectorImage<PixelType, 2>                                                   VectorImageType;
  typedef itk::Vector<PixelType, 2>                                                        DeformationPixelType;
  typedef otb::Image<DeformationPixelType, 2>                                              DeformationFieldType;
  typedef otb::NCCRegistrationFilter<ImageType, ImageType, DeformationFieldType>            RegistrationFilterType;
  typedef RegistrationFilterType::Pointer                                                 RegistrationFilterPointerType;
  typedef RegistrationFilterType::RadiusType                                              RadiusType;

  typedef StreamingWarpImageFilter<ImageType, ImageType, DeformationFieldType>              WarpFilterType;
  typedef WarpFilterType::DeformationValueType                                            DeformationValueType;
  typedef WarpFilterType::Pointer                                                         WarpFilterPointerType;

  typedef itk::LinearInterpolateImageFunction<ImageType, double>                           InterpolatorType;
  typedef InterpolatorType::Pointer                                                       InterpolatorPointerType;
  typedef otb::ImageFileReader<ImageType>                                                 ReaderType;
  typedef ReaderType::Pointer                                                             ReaderPointerType;
  typedef otb::StreamingShrinkImageFilter<ImageType, ImageType>                            ShrinkFilterType;
  typedef ShrinkFilterType::Pointer                                                       ShrinkFilterPointerType;
  typedef itk::ImageRegionConstIterator< ImageType >                                      IteratorType;
  typedef itk::Vector<itk::NumericTraits<PixelType>::RealType, 1>                          MeasurementVectorType;
  typedef itk::Statistics::ListSample<MeasurementVectorType>                              ListSampleType;
  typedef float                                                                           HistogramMeasurementType;
  typedef itk::Statistics::ListSampleToHistogramGenerator<
  ListSampleType,
  HistogramMeasurementType,
  itk::Statistics::DenseFrequencyContainer, 1>                                           HistogramGeneratorType;
  typedef HistogramGeneratorType::HistogramType                                           HistogramType;
  typedef HistogramType::Pointer                                                          HistogramPointerType;

  // Color composition typedefs
  typedef otb::ImageList<ImageType>                                                       ImageListType;
  typedef ImageListType::Pointer                                                          ImageListPointerType;
  typedef otb::ImageListToVectorImageFilter<ImageListType, VectorImageType>                ImageListToVectorImageFilterType;
  typedef ImageListToVectorImageFilterType::Pointer                                       ImageListToVectorImageFilterPointerType;
  typedef otb::ImageOfVectorsToMonoChannelExtractROI<DeformationFieldType, ImageType>      ChannelExtractionFilterType;
  typedef ChannelExtractionFilterType::Pointer                                            ChannelExtractionFilterPointerType;
  typedef otb::VectorImageToIntensityImageFilter<DeformationFieldType, ImageType>          IntensityFilterType;
  typedef IntensityFilterType::Pointer                                                    IntensityFilterPointerType;
  typedef FineRegistrationApplicationStates                                               StatesType;
  typedef MultiChannelExtractROI<PixelType, PixelType>                                     ExtractROIFilterType;
  typedef StreamingImageFileWriter<RegistrationFilterType::OutputImageType>               DeformationWriterType;
  typedef StreamingImageFileWriter<ImageType>                                             RegistrationWriterType;

  itkSetMacro(CurrentRegion, RegionType);
  itkGetMacro(CurrentRegion, RegionType);
  itkGetMacro(State, StatesType);
  itkGetObjectMacro(RegistrationFilter, RegistrationFilterType);
  itkGetObjectMacro(WarpFilter, WarpFilterType);
  itkGetObjectMacro(DeformationWriter, DeformationWriterType);
  itkGetObjectMacro(RegistrationWriter, RegistrationWriterType);

  /** Open a stereoscopic couple to the application */
  void OpenPair(const char * movingName, const char * fixedName);

  /** Close the current stereoscopic couple */
  void ClosePair();

  /** Estimate deformation on current region */
  void RunDeformationFieldEstimation();

  /** Run image extraction on current region */
  void RunImageExtraction();

  /** Get the nth left image filename */
  const char * GetMovingImageFilename ();

  /** Get the nth right image filename */
  const char * GetFixedImageFilename ();

  /** Get the image color composition */
  VectorImageType * GetImageColorComposition();

  /** Get the deformation field color composition */
  VectorImageType * GetDeformationFieldColorComposition();

  /** Get the shrink color composition */
  VectorImageType * GetQuicklookColorComposition();

  /** Get the nth Shrink filter for progress reporting */
  ShrinkFilterType * GetMovingShrinkFilter();

  /** Get the nth Shrink filter for progress reporting */
  ShrinkFilterType * GetFixedShrinkFilter();

  /** Get the nth shrink factor */
  unsigned int GetShrinkFactor();

  /** Save Registerd Image */
  void SaveRegisteredImage(const char * filename);

  /** Save Disparity Map */
  void SaveDisparityMap( const char * fiename);

  /** Get the mins and the maxs */
  PixelType GetMin() const;
  PixelType GetMax() const;

  /** Set the size of the quicklooks to generate */
  itkSetMacro(QuicklookSize, SizeType);

  /** Set the percentage used for normalization */
  itkSetMacro(ClampThreshold, double);
  itkGetMacro(ClampThreshold, double);

  /** Set the max deformation value */
  void SetMaxDeformationValue( DeformationValueType maxVal );
  DeformationValueType GetMaxDeformationValue();

  /** Set/Get the number of iterations */
  unsigned int GetNumberOfIterations() const;
  void SetNumberOfIterations(unsigned int nbIterations);

  /** Set/Get the radius */
  RadiusType GetRadius() const;
  void SetRadius(const RadiusType & radius);

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);
  /** Constructor */
  FineRegistrationApplicationModel();
  /** Destructor */
  ~FineRegistrationApplicationModel();

  /** Init every used filter */
  void InitFilters();
  /** Wire the pipeline */
  void WirePipeline();

  /** Generate the nth quicklook and normalization factor */
  void GenerateHistogram();

  /** Generate quicklook */
  void GenerateQuicklook();

  /** Setup deformation field estimation */
  void SetupDeformationEstimation();

  /** Constrains current region to largest possible region */
  bool ConstrainsCurrentRegion();

private:
  FineRegistrationApplicationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** The instance singleton */
  static Pointer Instance;

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  // Moving readers list
  ReaderPointerType m_MovingReader;
  // Fixed readers list
  ReaderPointerType m_FixedReader;
  /** Moving shrink filter */
  ShrinkFilterPointerType m_MovingShrinker;
  /** Fixed shrink filter */
  ShrinkFilterPointerType m_FixedShrinker;
  /** Size of the quicklook to generate */
  SizeType m_QuicklookSize;
  /** Shrink factors */
  unsigned int m_ShrinkFactor;
  /** Histogram list */
  HistogramPointerType m_Histogram;
  /** percentage used for normalization */
  double m_ClampThreshold;

  /** Application state */
  StatesType m_State;

  /** Images to display */
  VectorImageType::Pointer m_Quicklook;
  VectorImageType::Pointer m_Image;
  VectorImageType::Pointer m_Deformation;

  /** Region to work on */
  RegionType m_CurrentRegion;

  /** Registration filter */
  RegistrationFilterPointerType m_RegistrationFilter;

  /** Image warper */
  WarpFilterPointerType         m_WarpFilter;

  /** Interpolator */
  InterpolatorPointerType       m_Interpolator;

  /** Input for deformation field color composition */
  ChannelExtractionFilterPointerType m_XDeformationFieldExtractor;
  ChannelExtractionFilterPointerType m_YDeformationFieldExtractor;
  IntensityFilterPointerType         m_DeformationFieldIntensityFilter;

  /** Color composer */
  DeformationValueType                    m_MaxDeformationValue;

  /** Writers */
  RegistrationWriterType::Pointer m_RegistrationWriter;
  DeformationWriterType::Pointer  m_DeformationWriter;

};
}
#endif
