/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbImageToDBRegistrationModel.h"
#include "otbFltkFilterWatcher.h"
#include "otbFltkWriterWatcher.h"
#include <FL/fl_ask.H>
#include "itkExceptionObject.h"
#include "otbMacro.h"

#include "itkPreOrderTreeIterator.h"
#include "otbVectorDataFileWriter.h"

// for test only
#include "itkConnectedComponentImageFilter.h"
#include "itkRelabelComponentImageFilter.h"
#include "otbPersistentVectorizationImageFilter.h"
#include "otbPolygon.h"
#include "otbObjectList.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataExtractROI.h"
#include "otbStreamingImageFileWriter.h"
#include "itkLinearInterpolateImageFunction.h"

namespace otb
{

/** Initialize the singleton */
ImageToDBRegistrationModel::Pointer ImageToDBRegistrationModel::Instance = NULL;

ImageToDBRegistrationModel::ImageToDBRegistrationModel()
{
  m_InputImage = InputImageType::New();
  m_OutputImage = OutputImageType::New();
  m_FocusingVisuModel = FocusingVisuModelType::New();
  m_Reader = ReaderType::New();
  m_HasNewInput = false;
  m_ExtractHasNewDB = false;
  m_FullHasNewDB = false;
  m_HasNewExtract = false;
  m_HasExtract = false;
  m_HasInput = false;
  m_DBListId.clear();
  m_VectorData = VectorDataType::New();
  m_Alpha = 0.5;
  m_HasOutputImage = false;

  m_ValidSegmentThreshold = 1.;
  m_ColinearityThreshold = 10.;
  m_TripletMaxDistance = 50.;

  // FOR TESTS ONLY
  m_IntensityFilter = IntensityFilterType::New();
  m_Thresholder =  ThresholderType::New();
  m_UpperThresh = static_cast<PixelType>(0);
  m_LowerThresh = static_cast<PixelType>(1);

  // the appli model registered to the focusing model to receive its notifycation
  m_FocusingVisuModel->RegisterListener(this);
}


ImageToDBRegistrationModel
::~ImageToDBRegistrationModel() {}

/** Manage the singleton */
ImageToDBRegistrationModel::Pointer
ImageToDBRegistrationModel::GetInstance()
{
  if (!Instance)
    {
    Instance = ImageToDBRegistrationModel::New();
    }
  return Instance;
}

void
ImageToDBRegistrationModel
::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}


void
ImageToDBRegistrationModel
::Notify()
{
  if ( m_FocusingVisuModel->GetNewExtractVDDisplayed() == true )
    {
    // remove from GUI list before clearing DBListId
    m_HasNewExtract = true;
    this->NotifyAll();
    m_HasNewExtract = false;
    // clear vector data "Results"
    unsigned int sizeList = m_DBListId.size();
    unsigned int sizeList2 = m_FocusingVisuModel->GetVectorDataList()->Size();
    if (sizeList2>0 && sizeList>0)
      {
      // Sort in decrease values the vector to erase it from the end (insertion sort)
      unsigned int j;
      for (unsigned int i=0; i<m_DBListId.size(); ++i)
        {
        unsigned int val = m_DBListId[i];
        for( j=i; (j>0 && m_DBListId[j-1]<val); --j )
          {
          m_DBListId[j] = m_DBListId[j-1];
          }
        m_DBListId[j] = val;
        }
          
      for(unsigned int k=0; k<sizeList; k++)
        {
        this->ClearVectorData(m_DBListId[0]);
        }
      }
    m_DBListId.clear();
    m_HasOutputImage = false;
    m_HasExtract = true;
    }
}


void
ImageToDBRegistrationModel
::InitFilters()
{
  // FOR TESTS ONLY
  m_RegistrationFilter = RegistrationFilterType::New();
  m_RegistrationFilter->SetColinearityThreshold(m_ColinearityThreshold);
  m_RegistrationFilter->SetValidSegmentThreshold(m_ValidSegmentThreshold);
  m_RegistrationFilter->SetTripletMaxDistance(m_TripletMaxDistance);

  m_IntensityFilter = IntensityFilterType::New();
  m_Thresholder =  ThresholderType::New();
  m_Thresholder->SetInsideValue(1);
  m_Thresholder->SetOutsideValue(0);
  m_Thresholder->SetUpperThreshold( m_UpperThresh );
  m_Thresholder->SetLowerThreshold( m_LowerThresh );
  m_Thresholder->SetInput(m_IntensityFilter->GetOutput());
}

void
ImageToDBRegistrationModel
::OpenImage()
{
  m_Reader = ReaderType::New();
  m_Reader->SetFileName(m_InputFileName);
  m_Reader->GenerateOutputInformation();
  this->SetInputImage( m_Reader->GetOutput() );

  m_FocusingVisuModel->Init();
  m_FocusingVisuModel->SetInputImage( m_InputImage );
  m_FocusingVisuModel->UpdateModels();

  // Notify the observers
  m_HasNewInput = true;
  this->NotifyAll();
  m_HasNewInput = false;

  m_HasInput = true;
}


void
ImageToDBRegistrationModel
::LoadVectorData( std::string name )
{
  m_FocusingVisuModel->LoadVectorData( name );
}


void
ImageToDBRegistrationModel
::SetDEMDirectory( std::string name )
{
  m_FocusingVisuModel->SetDEMDirectory( name );
}

void
ImageToDBRegistrationModel
::UseDEM( bool b )
{
  if ( m_FocusingVisuModel->GetUseDEM() != b && m_FocusingVisuModel->GetVectorDataList()->Size() != 0 )
    {
    m_FocusingVisuModel->SetUseDEM( b );
    m_FocusingVisuModel->UpdateVectorData();
    }
}

void
ImageToDBRegistrationModel
::ChangeVectorDataColor( unsigned int id, ColorType color)
{
  m_FocusingVisuModel->ChangeVectorDataColor( id, color );
}

void
ImageToDBRegistrationModel
::ChangeColor( ColorType color)
{
  m_FocusingVisuModel->SetColor( color );
}

void
ImageToDBRegistrationModel
::ClearAllVectorData()
{
  m_FocusingVisuModel->ClearAllVectorData();
  m_DBListId.clear();
}

void
ImageToDBRegistrationModel
::ClearVectorData( unsigned int id )
{
  // To avoid inifite loop : ChangeVectorDataColor calls Notify...
  bool oldStatus = m_FocusingVisuModel->GetNewExtractVDDisplayed();
  if (oldStatus==true)
    m_FocusingVisuModel->SetNewExtractVDDisplayed( false );

  m_FocusingVisuModel->ClearVectorData(id);

  // CHECK IF THE DELETED VD IS A LOADED ONE OR NOT
  bool wasLoaded = true;
  unsigned int k=0;
  while ( k<m_DBListId.size() && wasLoaded == true )
    {
    if (m_DBListId[k]==id)
      wasLoaded = false;
    k++;
    }


  // if loaded vd
  if ( wasLoaded==true )
    {
    for (unsigned int i=0; i<m_DBListId.size(); i++)
      {
      m_DBListId[i]--;
      }
    }
  // if result vd
  else
    {
    unsigned int j = 0;
    bool ok = false;
    while (j<m_DBListId.size() && ok == false)
      {
      if (m_DBListId[j]==id)
        {
        m_DBListId.erase(m_DBListId.begin()+j);
        ok = true;
        }
      j++;
      }
    //  Shift the m_ResultInListId higher values of 1
    if (ok == true)
      {
      for (unsigned int i=0; i<m_DBListId.size(); i++)
        {
        if (m_DBListId[i]>id)
          m_DBListId[i]--;
        }
      }
    // the delete VD is a loaded VB
    }

  if (oldStatus==true)
    m_FocusingVisuModel->SetNewExtractVDDisplayed( true );
}

ImageToDBRegistrationModel::VectorDataPointerType
ImageToDBRegistrationModel
::ExtractVD(VectorDataPointerType vect, InputImageEnumType type)
{
  InputImagePointerType image;
  if (type==otb::EXTRACT)
    {
    image =  m_FocusingVisuModel->GetExtractedImage();
    }
  else
    {
    image = m_InputImage;
    }

  //Extract the VextorData included in the Largest Possible Remote Region of the image
  typedef VectorDataExtractROI<VectorDataType >          VExtractROIType;
  typedef VExtractROIType::RegionType RemoteSensingRegionType;
  VExtractROIType::Pointer     vdextract = VExtractROIType::New();
  RemoteSensingRegionType      rsRegion;

  //Fill the rsRegion
  // Ge the index of the corner of the image
  InputImageType::IndexType ul, ur, ll, lr;
  InputImageType::PointType pul, pur, pll, plr;
  ul = m_FocusingVisuModel->GetExtractedImage()->GetLargestPossibleRegion().GetIndex();
  ur = ul;
  ll = ul;
  lr = ul;
  ur[0] += image->GetLargestPossibleRegion().GetSize()[0];
  lr[0] += image->GetLargestPossibleRegion().GetSize()[0];
  lr[1] += image->GetLargestPossibleRegion().GetSize()[1];
  ll[1] += image->GetLargestPossibleRegion().GetSize()[1];

  // Transform to physical point
  image->TransformIndexToPhysicalPoint(ul, pul);
  image->TransformIndexToPhysicalPoint(ur, pur);
  image->TransformIndexToPhysicalPoint(ll, pll);
  image->TransformIndexToPhysicalPoint(lr, plr);

  // Build the cartographic region
  RemoteSensingRegionType::IndexType rsOrigin;
  RemoteSensingRegionType::SizeType  rsSize;
  rsOrigin[0] = min(pul[0], plr[0]);
  rsOrigin[1] = min(pul[1], plr[1]);
  rsSize[0] = vcl_abs(pul[0]-plr[0]);
  rsSize[1] =vcl_abs(pul[1]-plr[1]);

  rsRegion.SetOrigin(rsOrigin);
  rsRegion.SetSize(rsSize);
  rsRegion.SetRegionProjection(image->GetProjectionRef());
  rsRegion.SetKeywordList(image->GetImageKeywordlist());

  //set the cartographix region & the input to the
  vdextract->SetInput( vect );
  vdextract->SetRegion(rsRegion);
  vdextract->Update();

  return vdextract->GetOutput();
}

void
ImageToDBRegistrationModel
::Run(OutputEnumType type, int refVDId)
{
  if ( m_HasInput == true )
    {
    if ( type==otb::IMAGE )
      {
      this->ComputeImage( otb::FULL );
      //m_HasOutputImage = true;
      //this->DisplayOutputImage();
      m_HasOutputImage = false;
      }
    else if ( type==otb::DATABASE )
      {
      m_VectorData = m_FocusingVisuModel->GetVectorDataList()->GetNthElement( refVDId );
      this->SaveVectorData( otb::FULL );
      m_FullHasNewDB = true;
      this->NotifyAll();
      m_FullHasNewDB = false;
      }
    }
  else
    {
    itkExceptionMacro(<<"No input image selected...");
    }
}

void
ImageToDBRegistrationModel
::RunOverExtract(OutputEnumType type, int refVDId)
{
  if ( m_HasExtract == true )
    {
    // FOR TESTS ONLY
    this->InitFilters();

    m_IntensityFilter->SetInput( m_FocusingVisuModel->GetExtractedImage() );
    m_RegistrationFilter->SetInputImage(m_IntensityFilter->GetOutput());

    m_VectorData = this->ExtractVD(m_FocusingVisuModel->GetVectorDataList()->GetNthElement(refVDId), otb::EXTRACT);

    //Set the founded vectorData
    m_RegistrationFilter->SetInputVectorData(m_VectorData);

    m_RegistrationFilter->Update();
    m_Transform = m_RegistrationFilter->GetOutput();

    if ( type==otb::IMAGE )
      {
      this->ComputeImage( otb::EXTRACT );
      m_HasOutputImage = true;
      this->DisplayOutputImage();
      }
    else if ( type==otb::DATABASE )
      {
      this->ImageToVDProjection( m_RegistrationFilter->GetVectorDataProjectedTransformed(), otb::EXTRACT );
      m_ExtractHasNewDB = true;
      this->NotifyAll();
      m_ExtractHasNewDB = false;
      m_DBListId.push_back( m_FocusingVisuModel->GetVectorDataList()->Size()-1 );
      }
    }
  else
    {
    itkExceptionMacro(<<"No working area selected...");
    }
}


void
ImageToDBRegistrationModel
::ImageToVDProjection(VectorDataPointerType vect, InputImageEnumType type)
{
  InputImagePointerType image;
  if (type==otb::EXTRACT)
    {
    image =  m_FocusingVisuModel->GetExtractedImage();
    }
  else
    {
    image = m_InputImage;
    }
  ProjectionFilterType::Pointer vproj =  ProjectionFilterType::New();

  vproj->SetInput(vect);
  vproj->SetInputProjectionRef(image->GetProjectionRef());
  vproj->SetInputSpacing(image->GetSpacing());
  vproj->SetInputOrigin(image->GetOrigin());
  vproj->SetInputKeywordList(image->GetImageKeywordlist());

  vproj->SetOutputProjectionRef(m_VectorData->GetProjectionRef());
  vproj->SetOutputSpacing(m_VectorData->GetSpacing());
  vproj->SetOutputOrigin(m_VectorData->GetOrigin());

  vproj->Update();

  m_VectorData = vproj->GetOutput();
  m_FocusingVisuModel->AddVectorData(m_VectorData);
}


void
ImageToDBRegistrationModel
::ComputeImage(InputImageEnumType type)
{
  InputImagePointerType image;
  if (type==otb::EXTRACT)
    {
    image =  m_FocusingVisuModel->GetExtractedImage();
    }
  else
    {
    image = m_InputImage;
    }

  TransformPointerType invTrans = TransformType::New();
  m_Transform->GetInverse(invTrans);

  PerBandImageResamplerType::Pointer resampler = PerBandImageResamplerType::New();
  ImageResamplerType::Pointer        internRes = ImageResamplerType::New();
  InterpolatorType::Pointer          interp    = InterpolatorType::New();

  internRes->SetOutputSpacing(    invTrans->TransformVector(image->GetSpacing()) );
  internRes->SetOutputSize(             image->GetLargestPossibleRegion().GetSize() );
  internRes->SetOutputOrigin(     m_Transform->TransformPoint(image->GetOrigin())  );
  internRes->SetOutputStartIndex( image->GetLargestPossibleRegion().GetIndex() );
  internRes->SetTransform(        invTrans );
  internRes->SetInterpolator(     interp );

  resampler->SetInput( image );
  resampler->SetFilter( internRes );

  if (type==otb::EXTRACT)
    {
    resampler->Update();
    }
  else
    {
    WriterType::Pointer writer = WriterType::New();
    itk::OStringStream oss;
    oss.str("");
    oss<<"Save ";
    oss<<m_OutputFileName;
    oss<<"...";
    FltkFilterWatcher qlwatcher(writer, 0, 0, 250, 20, oss.str().c_str());
    writer->SetInput(resampler->GetOutput());
    writer->SetFileName(m_OutputFileName);
    writer->Update();
    }


  m_OutputImage = resampler->GetOutput();
}

void
ImageToDBRegistrationModel
::DisplayOutputImage()
{
  if ( m_HasOutputImage == true )
    {
    LayerGeneratorType::Pointer  lGenerator = LayerGeneratorType::New();
    lGenerator->SetImage(m_OutputImage);
    lGenerator->GenerateLayer();

    /** Display the image result */
    unsigned int size = m_FocusingVisuModel->GetWorkingModel()->GetNumberOfLayers();
    if (size>1)
      {
      for(unsigned int j=1; j<size; j++)
        {
        m_FocusingVisuModel->GetWorkingModel()->DeleteLayer(j);
        }
      }
    LayerType * lLayer = static_cast<LayerType*>(m_FocusingVisuModel->GetWorkingModel()->GetLayer(0));
    //FIXME define a generic method SetParameter with a map for rendering function for example
    // to avoid these ugly cast...
    static_cast<Function::StandardRenderingFunction<InputImagePixelType, RGBPixelType>* >
      (lGenerator->GetLayer()->GetRenderingFunction())->SetAutoMinMax( false );
    lGenerator->GetLayer()->SetRenderingFunction( lLayer->GetRenderingFunction() );
    // lGenerator->GetLayer()->SetAutoMinMax( false );
    UniformAlphaBlendingFunctionType::Pointer blendfunc = UniformAlphaBlendingFunctionType::New();
    blendfunc->SetAlpha(m_Alpha);
    lGenerator->GetLayer()->SetBlendingFunction(blendfunc);
    m_FocusingVisuModel->GetWorkingModel()->AddLayer(lGenerator->GetLayer());
    m_FocusingVisuModel->GetWorkingModel()->Update();
    }
}

void
ImageToDBRegistrationModel
::SaveVectorData(InputImageEnumType type)
{
  // only full case, otherwise, the transformation is done in the ProcessFilter
  // transform the VD from VD proj to image proj and resample it with m_transform
  if ( type == otb::FULL )
    {
    // transform the VD from VD proj to image proj and resample it with m_transform
    ProjectionFilterType::Pointer vproj =  ProjectionFilterType::New();

    vproj->SetInput(this->ExtractVD(m_VectorData, otb::FULL)); //vdextract->GetOutput());
    vproj->SetInputProjectionRef(m_VectorData->GetProjectionRef());
    vproj->SetInputSpacing(m_VectorData->GetSpacing());
    vproj->SetInputOrigin(m_VectorData->GetOrigin());

    vproj->SetOutputKeywordList(m_InputImage->GetImageKeywordlist());
    vproj->SetOutputProjectionRef(m_InputImage->GetProjectionRef());
    vproj->SetOutputOrigin(m_InputImage->GetOrigin());
    vproj->SetOutputSpacing(m_InputImage->GetSpacing());
    vproj->Update();

    VectorDataResampleType::Pointer resampler = VectorDataResampleType::New();
    resampler->SetInput( vproj->GetOutput() );

    TransformType::InputPointType center;
    center[0] = m_FocusingVisuModel->GetExtractedImage()->GetOrigin()[0];
    center[1] = m_FocusingVisuModel->GetExtractedImage()->GetOrigin()[1];
    m_Transform->SetCenter(center);
    resampler->SetTransform(m_Transform);
    resampler->Update();

    this->ImageToVDProjection(resampler->GetOutput(), otb::FULL);
    }

  VectorDataWriterType::Pointer writerVD = VectorDataWriterType::New();
  itk::OStringStream oss;
  oss.str("");
  oss<<"Save ";
  oss<<m_OutputFileName;
  oss<<"...";
  //FltkFilterWatcher qlwatcher(writerVD, 0, 0, 250, 20, oss.str().c_str());
  writerVD->SetFileName(m_OutputFileName);
  writerVD->SetInput(m_VectorData);
  writerVD->Update();
}

// call to save the extract
void
ImageToDBRegistrationModel
::SaveImage(std::string name)
{
  WriterType::Pointer writer = WriterType::New();
  itk::OStringStream oss;
  oss.str("");
  oss<<"Save ";
  oss<<name;
  oss<<"...";
  FltkFilterWatcher qlwatcher(writer, 0, 0, 250, 20, oss.str().c_str());
  writer->SetFileName(name);
  writer->SetInput( m_OutputImage );
  writer->Update();
}

void
ImageToDBRegistrationModel
::ChangeAlpha(double alpha)
{
  m_Alpha = alpha;
  this->Modified();
  this->DisplayOutputImage();
}

}// end of cxx
