/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbImageToDBRegistrationModel_h
#define __otbImageToDBRegistrationModel_h


#include "otbMVCModel.h"
#include "otbImageToDBRegistrationEnum.h"
#include "otbListenerBase.h"
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageFileReader.h"
#include "otbObjectList.h"
#include "otbStreamingImageFileWriter.h"
#include "otbVectorDataFileWriter.h"

#include "itkRGBAPixel.h"
#include "otbImageLayer.h"
#include "otbImageLayerGenerator.h"
#include "otbImageLayerRenderingModel.h"
#include "otbUniformAlphaBlendingFunction.h"
#include "otbBlendingFunction.h"

#include "otbFocusingView.h"
#include "otbFocusingModel.h"
#include "otbListenerBase.h"
#include "otbImageToDBRegistrationProcessFilter.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbApplyTransformToVectorDataFilter.h"
#include "otbStreamingResampleImageFilter.h"
#include "otbPerBandVectorImageFilter.h"
#include "itkLinearInterpolateImageFunction.h"

// FOR TEST
#include "otbVectorImageToIntensityImageFilter.h"
#include "itkBinaryThresholdImageFilter.h"


namespace otb
{
/** \class ImageToDBRegistrationModel
 *
 *
 *
 */


class ITK_EXPORT ImageToDBRegistrationModel
  : public MVCModel<ListenerBase>, public ListenerBase, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef ImageToDBRegistrationModel                     Self;
  typedef MVCModel<ListenerBase>  Superclass;
  typedef itk::SmartPointer<Self>                        Pointer;
  typedef itk::SmartPointer<const Self>                  ConstPointer;

  /** Standard type macro */
  itkTypeMacro(ImageToDBRegistrationModel, MVCModel);

  /** Images typedefs */
  typedef double                                               PixelType;
  /** Input Image Type*/
  typedef VectorImage<PixelType, 2>                             InputImageType;
  typedef InputImageType::Pointer                              InputImagePointerType;
  typedef InputImageType::PixelType                            InputImagePixelType;
  typedef InputImageType::IndexType                            IndexType;
  typedef InputImageType::SizeType                             SizeType;
  typedef InputImageType::OffsetType                           OffsetType;
  /** Vector Image Output Type*/
  typedef VectorImage<PixelType, 2>                             OutputImageType;
  typedef OutputImageType::PixelType                           OutputImagePixelType;
  typedef OutputImageType::Pointer                             OutputImagePointerType;


  typedef itk::RGBAPixel<unsigned char>                        RGBPixelType;
  typedef Image<RGBPixelType, 2>                                ViewerImageType;
  typedef ImageLayer<InputImageType, ViewerImageType>          LayerType;
  typedef ImageLayerGenerator<LayerType>                       LayerGeneratorType;
  typedef LayerGeneratorType::Pointer                          LayerGeneratorPointerType;
  typedef Function::BlendingFunction<RGBPixelType>             BlendingFunctionType;
  typedef Function::UniformAlphaBlendingFunction<RGBPixelType> UniformAlphaBlendingFunctionType;


  typedef FocusingModel<InputImageType, ViewerImageType>       FocusingVisuModelType;
  typedef FocusingVisuModelType::Pointer                      FocusingVisuModelPointerType;
  typedef FocusingView<FocusingVisuModelType>                 FocusingVisuViewType;
  typedef FocusingVisuViewType::Pointer                       FocusingVisuViewPointerType;
  typedef FocusingVisuModelType::ImageLayerRenderingModelType VisuModelType;
  typedef VisuModelType::Pointer                              VisuModelPointerType;
  typedef FocusingVisuViewType::ImageViewType                 VisuViewType;
  typedef VisuViewType::Pointer                               VisuViewPointerType;
  typedef FocusingVisuModelType::ColorType                    ColorType;


  /** Vector Data and its filters*/
  typedef FocusingVisuModelType::VectorDataType                     VectorDataType;
  typedef VectorDataType::Pointer                                   VectorDataPointerType;
  typedef VectorDataType::DataNodeType                              DataNodeType;
  typedef VectorDataType::DataTreeType                              DataTreeType;
  typedef VectorDataFileWriter<VectorDataType>                      VectorDataWriterType;
  typedef VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  typedef ApplyTransformToVectorDataFilter<VectorDataType, VectorDataType>   VectorDataResampleType;


  /** Get the unique instanc1e of the model */
  static Pointer GetInstance();

  /** TYPEDEFS FOR TESTS */
  typedef Image<double, 2>                                                    SingleImageType;
  typedef VectorImageToIntensityImageFilter<InputImageType, SingleImageType> IntensityFilterType;
  typedef IntensityFilterType::Pointer                                       IntensityFilterPointerType;
  typedef itk::BinaryThresholdImageFilter<SingleImageType, SingleImageType>  ThresholderType;
  typedef ThresholderType::Pointer                                           ThresholderPointerType;

  /** REGISTRATION TYPEDEF*/
  typedef ImageToDBRegistrationProcessFilter<SingleImageType , VectorDataType>          RegistrationFilterType;
  typedef RegistrationFilterType::TransformType                                        TransformType;
  typedef RegistrationFilterType::TransformPointerType                                 TransformPointerType;
  typedef itk::LinearInterpolateImageFunction<SingleImageType, double>                  InterpolatorType;
  typedef StreamingResampleImageFilter<SingleImageType, SingleImageType>               ImageResamplerType;
  typedef PerBandVectorImageFilter<InputImageType, InputImageType, ImageResamplerType> PerBandImageResamplerType;

  /** Reader/Writer */
  typedef ImageFileReader<InputImageType>          ReaderType;
  typedef ReaderType::Pointer                      ReaderPointerType;
  typedef StreamingImageFileWriter<InputImageType> WriterType;


  /** Input Image Pointer */
  itkSetObjectMacro(InputImage, InputImageType);
  itkGetObjectMacro(InputImage, InputImageType);

  /** Input Image Filename */
  itkSetMacro(InputFileName, std::string);
  itkGetMacro(InputFileName, std::string);

  /** Get the output image */
  itkGetObjectMacro(OutputImage, OutputImageType);

  /** Transformation Accessors */
  itkGetObjectMacro(Transform, TransformType);
  itkSetObjectMacro(Transform, TransformType);

  /** Open Input Image */
  void OpenImage();
  void LoadVectorData( std::string name );
  void SetDEMDirectory( std::string name );
  void UseDEM( bool b );
  void RunOverExtract(OutputEnumType type, int refVDId);
  void Run(OutputEnumType type, int refVDId);
  void ChangeVectorDataColor( unsigned int id, ColorType color);
  void ChangeColor( ColorType color);
  void ClearAllVectorData();
  void ClearVectorData( unsigned int id );
  void InitFilters();
  void DisplayOutputImage();
  void SaveVectorData(InputImageEnumType type);
  void SaveImage(std::string name);
  // change the composition between 2 layers
  void ChangeAlpha(double alpha);
  void ImageToVDProjection(VectorDataPointerType vect, InputImageEnumType type);
  void ComputeImage(InputImageEnumType type);
  VectorDataPointerType ExtractVD(VectorDataPointerType vect, InputImageEnumType type);

  itkSetObjectMacro(FocusingVisuModel, FocusingVisuModelType);
  FocusingVisuModelPointerType GetFocusingVisuModel()
    {
      return m_FocusingVisuModel;
    }

  VisuModelPointerType GetMasterVisuModel(){
    return m_FocusingVisuModel->GetMasterModel();
  }
  VisuModelPointerType GetWorkingVisuModel(){
    return m_FocusingVisuModel->GetWorkingModel();
  }
  std::vector<ColorType> GetColorList(){
    return m_FocusingVisuModel->GetUsedColorList();
  }

  itkGetMacro(HasNewInput, bool);
  itkSetMacro(ColinearityThreshold, double);
  itkSetMacro(ValidSegmentThreshold, double);
  itkSetMacro(TripletMaxDistance, double);
  itkGetMacro(ExtractHasNewDB, bool);
  itkGetMacro(FullHasNewDB, bool);
  itkSetMacro(OutputFileName, std::string);
  itkGetMacro(OutputFileName, std::string);

  std::vector<unsigned int> GetDBListId()
    {
      return m_DBListId;
    };
  itkGetMacro(HasNewExtract, bool);
  itkGetMacro(Alpha, double);

  /** Receive events from the focusing model */
  virtual void Notify();

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  ImageToDBRegistrationModel();
  /** Destructor */
  virtual ~ImageToDBRegistrationModel();

private:
  ImageToDBRegistrationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);
  //virtual void NotifyListener(ListenerBase * listener);

  /** The instance singleton */
  static Pointer               Instance;
  /** Input Image Filename */
  std::string                  m_InputFileName;
  /** Imput Image Pointer */
  InputImagePointerType        m_InputImage;
  /** Output Image */
  OutputImagePointerType       m_OutputImage;
  /** Output filename, same variable used for image or vector datas */
  std::string                  m_OutputFileName;
  /** Reader */
  ReaderPointerType            m_Reader;
  /** Focusing model */
  FocusingVisuModelPointerType m_FocusingVisuModel;
  /** The Input has changed */
  bool                         m_HasNewInput;
  /** New vector data ouput */
  bool                         m_ExtractHasNewDB;
  /** New vector data full ouput */
  bool                         m_FullHasNewDB;
  /** List of theVD  output in the focsung model*/
  std::vector<unsigned int>    m_DBListId;
  /** New extract */
  bool                         m_HasNewExtract;
  /** Result vector data over full image */
  VectorDataType::Pointer      m_VectorData;
  /** alpha */
  double                       m_Alpha;
  /** Image output computed */
  bool                         m_HasOutputImage;
  /** Has an input image */
  bool                         m_HasInput;
  /** Has an extract (ie. working has an input */
  bool                         m_HasExtract;

  /** Registration instanciation filter*/
  RegistrationFilterType::Pointer      m_RegistrationFilter;

  /** Threshold for the length of the valid detected segment */
  double m_ValidSegmentThreshold;
  /** Threshold for the orientation between two segment when a Tuplet is build*/
  double m_ColinearityThreshold;
  /** Maximum distance authorized between 2 analyzed triplets*/
  double m_TripletMaxDistance;
  /** Stores computed transformation */
  TransformPointerType         m_Transform;

  // FOR TESTS ONLY
  IntensityFilterPointerType m_IntensityFilter; ;
  ThresholderPointerType m_Thresholder;
  PixelType m_UpperThresh;
  PixelType m_LowerThresh;
};

}
#endif
