/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITYor FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#ifndef __otbImageToDBRegistrationProcessFilter_h
#define __otbImageToDBRegistrationProcessFilter_h

#include  "itkProcessObject.h"
#include  "itkHistogram.h"


//#include "otbImageToLineSegmentVectorData.h"
#include "otbLineSegmentDetector.h"
#include "otbVectorDataToHistogramFilter.h"

#include "otbVectorDataToValidSegmentsFilter.h"
#include "otbSignificantLineVectorDataToTriplet.h"
#include "otbTripletToTransformFilter.h"

#include "otbVectorDataProjectionFilter.h"
#include "otbApplyTransformToVectorDataFilter.h"
#include "itkAffineTransform.h"


namespace otb
{
  /** \class otbImageToDBRegistrationProcessFilter
   *  \brief Full Class to determine the affine transformation
   *         to register a BD on a image
   *
   *
   */

template <class TImage, class TVectorData>
class ITK_EXPORT ImageToDBRegistrationProcessFilter :
  public itk::ProcessObject
  {
  public:
    /** Standard class typedefs. */
    typedef ImageToDBRegistrationProcessFilter                      Self;
    typedef itk::ProcessObject                                      Superclass;
    typedef itk::SmartPointer<Self>                                 Pointer;
    typedef itk::SmartPointer<const Self>                           ConstPointer;

    /** Run-time type information (and related methods). */
    itkTypeMacro(ImageToDBRegistrationProcessFilter, itk::ProcessObject);
    
    /** Method for creation through the object factory. */
    itkNewMacro(Self);

    /** typedef Image Type **/
    typedef TImage                                              ImageType;
    typedef typename ImageType::Pointer                         ImagePointerType;
    
    /** VectorData TypeDef*/
    typedef TVectorData                                         VectorDataType;
    typedef typename VectorDataType::Pointer                    VectorDataPointerType;
    typedef typename VectorDataType::DataNodeType               DataNodeType;
    typedef typename DataNodeType::Pointer                      DataNodePointerType;
    typedef typename DataNodeType::LineType                     LineType;
    typedef typename LineType::VertexType                       VertexType;
    
    typedef typename VectorDataType::DataTreeType::TreeNodeType InternalTreeNodeType;
    typedef typename InternalTreeNodeType::ChildrenListType     ChildrenListType;

    typedef std::vector<VertexType>                              LineVectorType;
    typedef std::vector<LineVectorType>                          VertexVectorType;

    //used Filters typedefs
    typedef LineSegmentDetector<ImageType , typename ImageType::InternalPixelType>   LSDFilterType;
    //typedef ImageToLineSegmentVectorData<ImageType>                                   LSDFilterType;
    typedef typename LSDFilterType::LineSpatialObjectListType                            LSDOutputType;
    
    typedef itk::Statistics::Histogram<typename VectorDataType::PrecisionType>           HistogramType;
    typedef typename HistogramType::Pointer                                              HistogramPointerType;
    typedef otb::VectorDataToHistogramFilter<VectorDataType, HistogramType>              VectorDataToHistogramType;
    typedef typename VectorDataToHistogramType::Pointer                                  VectorDataToHistogramPointerType;
 
    
    typedef otb::VectorDataToValidSegmentsFilter<VectorDataType>                         VectorDataToValidSegmentsType;
    typedef otb::SignificantLineVectorDataToTriplet<VectorDataType, ImageType>            SiginficantLineFilter;
    typedef otb::TripletToTransformFilter                                                TripletToTransformType;
    typedef typename TripletToTransformType::TransformType                               TransformType;
    typedef typename TripletToTransformType::TransformType::Pointer                      TransformPointerType;

    typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType>               ProjectionFilterType;
    typedef otb::ApplyTransformToVectorDataFilter<VectorDataType, VectorDataType>                 VectorDataResampleType;

    
    //Set/Get the threshold for the length of the valid detected segment
    itkSetMacro(ValidSegmentThreshold, double);

    //Set the threshold for the orientation between two segment when a Tuplet is build
    itkSetMacro(ColinearityThreshold, double);

    // Set the maximum distance authorized between 2 analyzed triplets
    itkSetMacro(TripletMaxDistance, double);
    
    //Get/Set Inputs : vectorData & inputImage
    const ImageType * GetInputImage();
    const VectorDataType* GetInputVectorData();
    void SetInputImage(const ImageType* inputImage);
    void SetInputVectorData(const VectorDataType * dvector);
    
    //Output
    TransformType* GetOutput()
      {
       return m_BestTransform;
      }

   
    //Get The best Transform estimated
    itkGetObjectMacro(VectorDataProjectedTransformed, VectorDataType);
    
  protected:
    void SetOutput(TransformType * vectord)
      {
       m_BestTransform = vectord;
       this->Modified();
      }
    virtual void GenerateData(void);
    virtual void BeforeGenerateData(void);
    
    
    VectorDataPointerType LineSpatialObjectListToVectorData(typename LSDOutputType::Pointer    lines);
    double HistogramCorrelation(VectorDataToHistogramPointerType  imageHistogram, VectorDataToHistogramPointerType  dbHistogram );
 
    ImageToDBRegistrationProcessFilter();
    virtual ~ImageToDBRegistrationProcessFilter() {};
    
  private:
    ImageToDBRegistrationProcessFilter(const Self&); //purposely not implemented
    void operator=(const Self&); //purposely not implemented
    
    double                       m_ValidSegmentThreshold;
    double                       m_ColinearityThreshold;
    double                       m_TripletMaxDistance;

    TransformPointerType         m_BestTransform;

    VectorDataPointerType        m_VectorDataProjected;
    VectorDataPointerType        m_VectorDataProjectedTransformed;
    
  };


}
#ifndef ITK_MANUAL_INSTANTIATION
#include "otbImageToDBRegistrationProcessFilter.txx"
#endif

#endif
