/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbImageToDBRegistrationProcessFilter_txx
#define __otbImageToDBRegistrationProcessFilter_txx

#include "otbImageToDBRegistrationProcessFilter.h"
#include "otbStandardImageViewer.h"
#include "otbVectorDataFileWriter.h"
#include "itkCannyEdgeDetectionImageFilter.h"

namespace otb
{

/**
 *
 */
template <class TImage, class TVectorData>
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::ImageToDBRegistrationProcessFilter()
{
  this->SetNumberOfRequiredInputs(2);
  this->SetNumberOfOutputs(1);
  this->SetNthOutput(0, VectorDataType::New());
  
  m_BestTransform = TransformType::New();
  m_VectorDataProjectedTransformed = VectorDataType::New();
  m_ValidSegmentThreshold = 1.;
  m_ColinearityThreshold = 10.;
  m_TripletMaxDistance = 50;
}


/**
 * Get Input Image
 */
template <class TImage, class TVectorData>
const typename  ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::ImageType *
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::GetInputImage()
{
  return static_cast<const ImageType *>(this->itk::ProcessObject::GetInput(0));
}

/**
 * Set Input Image
 */
template <class TImage, class TVectorData>
void
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::SetInputImage(const ImageType* inputImage)
{
  this->itk::ProcessObject::SetNthInput(0, const_cast<ImageType *>(inputImage));
}

/**
 * Get VectorData
 */
template <class TImage, class TVectorData>
const typename  ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::VectorDataType   *
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::GetInputVectorData()
{
  return static_cast<const VectorDataType *>(this->itk::ProcessObject::GetInput(1));
}

/**
 * Set Input Vector Data
 */
template <class TImage, class TVectorData>
void
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::SetInputVectorData(const VectorDataType * dvector )
{
  this->itk::ProcessObject::SetNthInput(1, const_cast<VectorDataType *>(dvector));
}


/**
 * BeforeGenerateData : Project the vectorData in the projection of the image
 */
template <class TImage, class TVectorData>
void
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::BeforeGenerateData(void)
{
  typename ProjectionFilterType::Pointer      vproj =  ProjectionFilterType::New();
  
  vproj->SetInput(this->GetInputVectorData());
  vproj->SetInputProjectionRef(this->GetInputVectorData()->GetProjectionRef());
  vproj->SetInputSpacing(this->GetInputVectorData()->GetSpacing());
  vproj->SetInputOrigin(this->GetInputVectorData()->GetOrigin());
  
  vproj->SetOutputKeywordList(this->GetInputImage()->GetImageKeywordlist());
  vproj->SetOutputProjectionRef(this->GetInputImage()->GetProjectionRef());
  vproj->SetOutputOrigin(this->GetInputImage()->GetOrigin());
  vproj->SetOutputSpacing(this->GetInputImage()->GetSpacing());
  vproj->Update();
  
  m_VectorDataProjected = vproj->GetOutput();
}


/**
 * Method Generate Data
 */
template <class TImage, class TVectorData>
void
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::GenerateData(void)
{
  //Call before Generate Data
  this->BeforeGenerateData();
  otbMsgDevMacro("ImageToDBRegistationFilter : Generate data ....");
  
  ImagePointerType       inputImage      = const_cast<ImageType*>(this->GetInputImage());
  VectorDataPointerType  inputVectorData = m_VectorDataProjected;
  
  //Detect Lines using LSD : for the image
  typename LSDFilterType::Pointer  lsdFilter = LSDFilterType::New();
  lsdFilter->SetInput(inputImage);
  //lsdFilter->SetThreadDistanceThreshold(5);
  lsdFilter->Update();

  //Convert LSD Filter Output (SpatialObjectListLine)  into a VectorData
  VectorDataPointerType  vimageLines = /*lsdFilter->GetOutputVectorData(); */this->LineSpatialObjectListToVectorData(lsdFilter->GetOutput());
  
  //Compute histogram of the dataBase and the image
  VectorDataToHistogramPointerType   imageHistogram = VectorDataToHistogramType::New();
  VectorDataToHistogramPointerType   dbHistogram    = VectorDataToHistogramType::New();
  
  imageHistogram->SetVectorData(vimageLines);
  imageHistogram->Update();
  
  dbHistogram->SetVectorData(m_VectorDataProjected);
  dbHistogram->Update();
  double angleOffset = this->HistogramCorrelation(imageHistogram, dbHistogram);
  double angleOffsetInRadian = 0.;
  if(vcl_abs(angleOffset) < 5.)
    angleOffsetInRadian = angleOffset *CONST_PI_180;
  else
    angleOffsetInRadian = 0.;
  otbMsgDevMacro(" Angle Offset Computed through Histogram Correlation" << angleOffset);
    
  //Apply the rotation to the DataBase
  typename VectorDataResampleType::Pointer   dbresampler = VectorDataResampleType::New();
  TransformPointerType                       affine     = TransformType::New();
  
  typename   TransformType::ParametersType     params(6);
  params.Fill(0.);
  params[0] = vcl_cos(angleOffsetInRadian);
  params[1] = -vcl_sin(angleOffsetInRadian);
  params[2] = vcl_sin(angleOffsetInRadian);
  params[3] = vcl_cos(angleOffsetInRadian);
  affine->SetParameters(params);
  otbMsgDevMacro(" Histogram Transform Parameters" << affine->GetParameters());
  
  dbresampler->SetInput(m_VectorDataProjected);
  dbresampler->SetTransform(affine);
  dbresampler->Update();
  
  //Line to Valid Segments
  typename VectorDataToValidSegmentsType::Pointer  imagehough  = VectorDataToValidSegmentsType::New();
  typename VectorDataToValidSegmentsType::Pointer  dbhough  = VectorDataToValidSegmentsType::New();
  
  imagehough->SetInput(vimageLines);
  imagehough->SetFrequencyThreshold(m_ValidSegmentThreshold);
  imagehough->Update();

  dbhough->SetInput(dbresampler->GetOutput());
  dbhough->SetFrequencyThreshold(m_ValidSegmentThreshold);
  dbhough->Update();
  
  //Valid Segment To Tuplets
  typename SiginficantLineFilter::Pointer  isignificantFilter  = SiginficantLineFilter::New();
  typename SiginficantLineFilter::Pointer  dbsignificantFilter = SiginficantLineFilter::New();
  
  isignificantFilter->SetVectorData(imagehough->GetOutput());
  isignificantFilter->SetColinearityThreshold(m_ColinearityThreshold);
  isignificantFilter->Update();
  
  dbsignificantFilter->SetVectorData(dbhough->GetOutput());
  dbsignificantFilter->SetColinearityThreshold(m_ColinearityThreshold);
  dbsignificantFilter->Update();
      
  //Tuplets To Transformation
  typename TripletToTransformType::Pointer   transform = TripletToTransformType::New();
  transform->SetFirstTripletVector(isignificantFilter->GetTupletOutput());
  transform->SetSecondTripletVector(dbsignificantFilter->GetTupletOutput());
  transform->SetFirstLineVector(imagehough->GetLineVector());
  transform->SetSecondLineVector(dbhough->GetLineVector());
  transform->SetTripletMaxDistance(m_TripletMaxDistance);
  transform->Update();
  
  //Copy the BestTransform from the class TripletToTransform
  m_BestTransform->SetParameters(transform->GetBestTransform()->GetParameters());
  otbMsgDevMacro(" Best  Transform Parameters" <<  m_BestTransform->GetParameters());

  //Apply the best affine transform found
  
  typename VectorDataResampleType::Pointer   finaldbresampler = VectorDataResampleType::New();
  finaldbresampler->SetInput(dbresampler->GetOutput());
  finaldbresampler->SetTransform(m_BestTransform);
  finaldbresampler->Update();
  m_VectorDataProjectedTransformed =  finaldbresampler->GetOutput();
  
  m_BestTransform->Compose(affine, true);

  this->SetOutput(m_BestTransform);
}


/**
*  Correlate the two histogram and find the orienation offset betwenn the input image histogram
*  and the dataBase histigral
*/
template <class TImage, class TVectorData>
double
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::HistogramCorrelation(VectorDataToHistogramPointerType  imageHistogram, VectorDataToHistogramPointerType  dbHistogram )
{
  double maxCorr   = 0.;
  double decalage  = 0.;
  double imaxCorr  = 0.;
  double idecalage = 0.;
  unsigned int vhistSize  = dbHistogram->GetHistogram()->Size();

  //Compute the correlation
  for(unsigned int i = 0; i<vhistSize; i++)
    {
      //Circular histogram cause the orientation are computed modulo PI
      typename HistogramType::Iterator    it = imageHistogram->GetHistogram()->Begin();
      double correlation = 0.;
      while(it != imageHistogram->GetHistogram()->End())
       {
         unsigned int posMove = ( (it.GetInstanceIdentifier()+i)%vhistSize );
         correlation  += it.GetFrequency()*dbHistogram->GetHistogram()->GetFrequency(posMove) - (imageHistogram->GetMean()* dbHistogram->GetMean());
         ++it;
       }

      correlation /=(imageHistogram->GetSigma()*dbHistogram->GetSigma());
      if(correlation > maxCorr)
       {
         maxCorr = correlation;
         decalage = i;
       }
    }
  
  /* Correlation with the inverse histogram of the image :
   *      - The orientation in the histogram class computation is done modulo 2PI,
   *        we have to check if the correlation (with inverse Image Histogram) is greater than
   *        the previsous one
   */
  HistogramPointerType inviHistogram = imageHistogram->GetInverseHistogram();
    
  //inverse correlation
  for(unsigned int i = 0; i<vhistSize; i++)
    {
      //Circular histogram cause the orientation are computed modulo PI
      typename HistogramType::Iterator    it = inviHistogram->Begin();
      double correlation = 0.;
      while(it != inviHistogram->End())
       {
         unsigned int posMove = ((it.GetInstanceIdentifier()+i)%vhistSize);
         correlation  += it.GetFrequency()*dbHistogram->GetHistogram()->GetFrequency(posMove) - (imageHistogram->GetMean()* dbHistogram->GetMean());
         ++it;
       }
      correlation /=(imageHistogram->GetSigma()*dbHistogram->GetSigma());

      if(correlation > imaxCorr)
       {
         imaxCorr = correlation;
         idecalage = i;
       }
    }

  double angleoffset = 0.;
  if(maxCorr>imaxCorr)
    angleoffset = decalage;
  else
    angleoffset = idecalage;
    
  if(angleoffset > 90 )
    angleoffset = -(180-angleoffset);

  return angleoffset;
}


/**
* Convert the LineSpatialObjectList to a Vector Data
*/
template <class TImage, class TVectorData>
typename ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::VectorDataPointerType
ImageToDBRegistrationProcessFilter<TImage, TVectorData>
::LineSpatialObjectListToVectorData(typename LSDOutputType::Pointer    lines)
{
  
  VectorDataPointerType              vlines    = VectorDataType::New();
  typename DataNodeType::Pointer     document  = DataNodeType::New();
  typename DataNodeType::Pointer     folder    = DataNodeType::New();
  typename DataNodeType::Pointer     root      = vlines->GetDataTree()->GetRoot()->Get();
  
  typename  LSDOutputType::const_iterator  it     = lines->begin();
  typename  LSDOutputType::const_iterator  itEnd = lines->end();
  
  document->SetNodeType(otb::DOCUMENT);
  folder->SetNodeType(otb::FOLDER);
  vlines->GetDataTree()->Add(document, root);
  vlines->GetDataTree()->Add(folder, document);
  vlines->SetProjectionRef(this->GetInputImage()->GetProjectionRef());

  while(it != itEnd)
    {
      VertexType p1, p2;
      typename LineType::Pointer l = LineType::New();
      typename LSDOutputType::LineType::PointListType & pointsList = (*it)->GetPoints();
      typename LSDOutputType::LineType::PointListType::const_iterator itPoints = pointsList.begin();
      p1[0] =(*itPoints).GetPosition()[0];     //First Vertex
      p1[1] =(*itPoints).GetPosition()[1];
      itPoints++;
      p2[0] =(*itPoints).GetPosition()[0];     //Second Vertex
      p2[1] =(*itPoints).GetPosition()[1];
      l->AddVertex(p1);
      l->AddVertex(p2);
      DataNodePointerType node = DataNodeType::New();
      node->SetNodeType(otb::FEATURE_LINE);
      node->SetLine(l);
      vlines->GetDataTree()->Add(node, folder);
      ++it;
    }  /** End lsd To VectorData */
  return vlines;
}


}

#endif
