/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbSignificantLineVectorDataToTriplet_txx
#define __otbSignificantLineVectorDataToTriplet_txx

#include "itkTimeProbe.h"
#include "otbSignificantLineVectorDataToTriplet.h"

namespace otb
{

template <class TVectorData, class TImage>
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::SignificantLineVectorDataToTriplet()
{
  this->SetNumberOfRequiredInputs(1);
  this->SetNumberOfOutputs(1);
  this->SetNthOutput(0, VectorDataType::New());
  
  m_OutputVectorData = VectorDataType::New();
  m_Image = ImageType::New();
  m_ColinearityThreshold = 1.;
  m_Origin.Fill(0);
}
  
/**
 * Set InputVectorData
 */
template  <class TVectorData, class TImage>
void
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::SetVectorData(VectorDataType * vector)
{
  this->itk::ProcessObject::SetNthInput(0, const_cast<VectorDataType *>(vector));
}
/**
 * GetInput
 */
template  <class TVectorData, class TImage>
typename SignificantLineVectorDataToTriplet<TVectorData, TImage>
::VectorDataType *
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::GetInputVectorData()
{
  return static_cast<VectorDataType *>(this->itk::ProcessObject::GetInput(0));
}


template <class TVectorData, class TImage>
typename SignificantLineVectorDataToTriplet<TVectorData, TImage>
::VectorDataType*
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::GetOutput()
{
  return static_cast<VectorDataType * >
    (this->itk::ProcessObject::GetOutput(0) );
}


template <class TVectorData, class TImage>
void
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::GenerateData()
{
  typename VectorDataType::Pointer   vectorData = this->GetInputVectorData();
  
  // Get the input tree root
  InternalTreeNodeType * inputRoot = const_cast<InternalTreeNodeType *>(vectorData->GetDataTree()->GetRoot());
  
  // Start recursive processing to read the vectorData
  itk::TimeProbe chrono;
  chrono.Start();
  ProcessNode(inputRoot);
  chrono.Stop();
  otbMsgDevMacro("Features processed in "<<chrono.GetMeanTime()<<" seconds.");
  otbMsgDevMacro("   Size of the processed list "<< m_LineVector.size() );
  
  //Edit the Tuplets list
  chrono.Start();
  this->ProcessTuplets();
  chrono.Stop();
  otbMsgDevMacro("Tuplet processed in "<<chrono.GetMeanTime()<<" seconds.");
  otbMsgDevMacro("   Size of the tuplets list "<< m_TupletVector.size());
}

  /**
   */
template <class TVectorData, class TImage>
void
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::ProcessNode(InternalTreeNodeType * source)
{
  // Get the children list from the input node
  typename InternalTreeNodeType::ChildrenListType children = source->GetChildrenList();
  // For each child
  for(typename InternalTreeNodeType::ChildrenListType::iterator it = children.begin(); it!=children.end(); ++it)
    {
      DataNodePointerType dataNode = (*it)->Get();
      
      switch(dataNode->GetNodeType())
       {
       case ROOT:
         {
           ProcessNode((*it));
           break;
         }
       case DOCUMENT:
         {
           ProcessNode((*it));
           break;
         }
       case FOLDER:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_POINT:
         {
           break;
         }
       case FEATURE_LINE:
         {
           LineVectorType                   currentList;
           typename VertexListType::ConstIterator it = dataNode->GetLine()->GetVertexList()->Begin();
           typename VertexListType::ConstIterator itEnd = dataNode->GetLine()->GetVertexList()->End();
           while(it != itEnd)
             {
              currentList.push_back(it.Value());
              ++it;
             }
           m_LineVector.push_back(currentList);
           break;
         }
       case FEATURE_POLYGON:
         {
           break;
         }
       case FEATURE_MULTIPOINT:
         {
           break;
         }
       case FEATURE_MULTILINE:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_MULTIPOLYGON:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_COLLECTION:
         {
           ProcessNode((*it));
           break;
         }
       }
      }
}


/**
 *
 */
template <class TVectorData, class TImage>
void
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::ProcessTuplets()
{
  //Preparing the Output vectordata
  typename VectorDataType::DataTreePointerType otree      = m_OutputVectorData->GetDataTree();
  InternalTreeNodeType * outputRoot = const_cast<InternalTreeNodeType *>(otree->GetRoot());
    
  //add the root
  typename DataNodeType::Pointer document = DataNodeType::New();
  typename DataNodeType::Pointer folder   = DataNodeType::New();
  typename DataNodeType::Pointer root     = outputRoot->Get();
 
  //add a document
  document->SetNodeType(otb::DOCUMENT);
  document->SetNodeId("DOCUMENT");
 
  //add a  folder
  folder->SetNodeType(otb::FOLDER);
  folder->SetNodeId("FOLDER");

  //Build the output tree
  otree->Add(document, root);
  otree->Add(folder, document);


  // Iterate through the all possible line intersections.
  // if lines not colinear with an ColinearityThreshold keep them
  typename VertexVectorType::iterator it1 = m_LineVector.begin();
  
  while(it1 != m_LineVector.end())
    {
      double orientation1 = this->ComputeSegmentOrientation((*it1)[0] , (*it1)[1] );
      typename VertexVectorType::iterator it2 = it1;
      ++it2;
      while(it2 != m_LineVector.end())
       {
         double orientation2 = this->ComputeSegmentOrientation((*it2)[0] , (*it2)[1] );
         double dx1 = vcl_abs(orientation1 - orientation2);
         if( dx1 >m_ColinearityThreshold && dx1 < 180. - m_ColinearityThreshold)
           {
             LineVectorType   tuplets;
             VertexType   p1 = this->ComputeLineIntersection(it1, it2);
             tuplets.push_back(p1);
             if(!this->IsPresent(tuplets))
              m_TupletVector.push_back(tuplets);
           }

         ++it2;
       }

      ++it1;
    }


}
template <class TVectorData, class TImage>
bool
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::IsPresent(LineVectorType tuplet)
{
 
  for(unsigned int i = 0; i<  m_TupletVector.size(); i++ )
    {
      if( this->ComputeSegmenLength(tuplet[0], m_TupletVector[i][0]) < 2.)
       return true;
    }
  return false;

}

/**
 *
 */

template <class TVectorData, class TImage>
double
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::ComputeSegmenLength(VertexType currentVertex, VertexType nextVertex)
{
  return vcl_sqrt((currentVertex[0]-nextVertex[0])*(currentVertex[0]-nextVertex[0])+(currentVertex[1]-nextVertex[1])*(currentVertex[1]-nextVertex[1]));
}


/**
 *  return the orientation in degrees
 */
template <class TVectorData, class TImage>
double
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::ComputeSegmentOrientation(VertexType currentVertex, VertexType nextVertex)
{
  //Compute the orientation
  double dx = currentVertex[0]-nextVertex[0];
  double dy = currentVertex[1]-nextVertex[1];
  double orientation = vcl_atan2(dy, dx);
  if(orientation < 0)
    orientation += CONST_PI;

  return (orientation*180./CONST_PI);
}

  
  /**
   */
template <class TVectorData, class TImage>
typename SignificantLineVectorDataToTriplet<TVectorData, TImage>
::VertexType
SignificantLineVectorDataToTriplet<TVectorData, TImage>
::ComputeLineIntersection(typename VertexVectorType::iterator it1, typename VertexVectorType::iterator it2 )
{
  VertexType    P;
    
  //Vertex of the first Line
  double Xp1 = (*it1)[0][0]+m_Origin[0];
  double Yp1 = (*it1)[0][1]+m_Origin[1];
  double Xp2 = (*it1)[1][0]+m_Origin[0];
  double Yp2 = (*it1)[1][1]+m_Origin[1];
 
  //Vertex of the second line
  double Xq1 =  (*it2)[0][0]+m_Origin[0];
  double Yq1 =  (*it2)[0][1]+m_Origin[1];
  double Xq2 =  (*it2)[1][0]+m_Origin[0];
  double Yq2 =  (*it2)[1][1]+m_Origin[1];
  
  
  /** Compute the equation of the lines A and B  which are support of the segments Src & Dst*/
  // - Line 1 : slope and origin
  double originA = 0. , slopeA = 0.;
  double originB = 0. , slopeB = 0.;
  double LengthSegmentAy = 0., lengthSegmentBy = 0.;

  /** Equation of the first Line*/
  if(vcl_abs(Xp2 - Xp1) <1e-10)
    Xp2 = 0.0001;

  if(Xp1 < Xp2 )
    LengthSegmentAy = Yp2 - Yp1;
  else
    LengthSegmentAy = Yp1 - Yp2;
  
  slopeA = LengthSegmentAy/(Xp2-Xp1);
  originA = Yp1 - (slopeA * Xp1);

  /** Equation of the second Line*/
  if(vcl_abs(Xq2 - Xq1) <1e-10)
    Xq2 = 0.0001;

  if(Xq1 < Xq2 )
    lengthSegmentBy = Yq2 - Yq1;
  else
     lengthSegmentBy = Yq1 - Yq2;
  
  slopeB = lengthSegmentBy/(Xq2-Xq1);
  originB = Yq1 - (slopeB * Xq1);

  /** Avoid the case of parallel lines*/
  double denum = 0.;
  if(vcl_abs(slopeA - slopeB ) < 1e-5)
    denum = 0.001;
  else
    denum = slopeA - slopeB;
 
  /** Compute the coordinate of the intersection point Y =AX+B*/
  P[0] =  (originB - originA)/denum;
  P[1] =  slopeA * static_cast<double>(P[0]) + originA;
  
  return P;
}


} // end namespace otb

#endif
