/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbTripletToTransformFilter.h"
#include "itkMatrix.h"

namespace otb
{
 
TripletToTransformFilter
::TripletToTransformFilter()
{
  this->SetNumberOfRequiredInputs(0);
  this->SetNumberOfOutputs(1);
  this->SetNthOutput(0, VectorDataType::New());
  
  m_Transform = TransformType::New();
  m_BestTransform = TransformType::New();
  m_ThresholdDistance = 20.;
  m_InliersRatio = 0.05;
  m_BestInlier = 0;
  m_BestCumulatedDistance = 1000.;
  m_TripletMaxDistance = 200.;
  
}

/**
 * Get Second Line Vector
 */
TripletToTransformFilter
::VectorDataType*
TripletToTransformFilter
::GetOutput()
{
  return static_cast<VectorDataType * >
    (this->itk::ProcessObject::GetOutput(0) );
}


/**
 *
 */
void
TripletToTransformFilter
::GenerateData()
{
  unsigned int sizeFirst  = m_FirstTupletVector.size();
  unsigned int sizeSecond = m_SecondTupletVector.size();

  if( sizeFirst==0 || sizeSecond == 0 )
    {
      itkExceptionMacro(<<"Lists must have size greater than 0 ...");
    }

  double inliersMin = m_InliersRatio*( std::min( static_cast<double>(sizeFirst) , static_cast<double>(sizeSecond) ) );
  if(inliersMin == 0)
    {
      itkExceptionMacro(<< " Lists must have size greater than 0 ...");
    }

  //find the best transform
  VertexVectorType::iterator itSecond    = m_SecondTupletVector.begin();
  VertexVectorType::iterator itEndSecond = m_SecondTupletVector.end();
  unsigned int count1 =0;
  unsigned int count2 =0;
  VertexType centroidSecond;
  VertexType centroidFirst;
  while(itSecond != itEndSecond)
    {
      centroidSecond[0] = static_cast<double>(  (*itSecond)[0][0]);
      centroidSecond[1] = static_cast<double>(  (*itSecond)[0][1]);

      VertexVectorType::iterator itFirst    = m_FirstTupletVector.begin();
      VertexVectorType::iterator itEndFirst = m_FirstTupletVector.end();
      count1 = 0;
      while(itFirst != itEndFirst )
       {
         centroidFirst[0] = static_cast<double>(  (*itFirst)[0][0]);
         if( vcl_abs(centroidFirst[0]-centroidSecond[0]) <= m_TripletMaxDistance)
           {
             centroidFirst[1] = static_cast<double>(  (*itFirst)[0][1]);
             if( vcl_abs(centroidFirst[1]-centroidSecond[1]) <= m_TripletMaxDistance)
              {
                this->InitTransformation(count1, count2);
              }
           }
         count1++;
         ++itFirst;
       }
      count2++;
      ++itSecond;
    }
}

/**
 *  InitTransform : pick randomly 3 a Tuplet in each set, and compute a transform
 */
void
TripletToTransformFilter
::InitTransformation(unsigned int count, unsigned int count2)
{
  //Generate Possible Permutation
  m_PossiblePermutation.clear();
  m_PossiblePermutation.push_back(m_FirstTupletVector[count]);

  //Compute the possible transformation
  this->UpToDateTransform(count2);
}

/**
*
*/
void
TripletToTransformFilter
::UpToDateTransform(unsigned int relement)
{
  itk::Matrix<double, 6, 6>       R, Rinv, V, H;
  V.Fill(0.);
  
  TransformType::MatrixType              rotationMatrix;
  TransformType::OutputVectorType        translationVector;
  //Fill the rotation matrix
  rotationMatrix(0, 0) = 1;
  rotationMatrix(0, 1) = 0;
  rotationMatrix(1, 0) = 0;
  rotationMatrix(1, 1) = 1;
  
  //Fill the translation vector
  translationVector[0] = -m_PossiblePermutation[0][0][0] + m_SecondTupletVector[relement][0][0];
  translationVector[1] = -m_PossiblePermutation[0][0][1] + m_SecondTupletVector[relement][0][1];
  
  //Fill the transform
  m_Transform->SetMatrix(rotationMatrix);
  m_Transform->SetTranslation(translationVector);
  
  m_CurrentCumulatedDistance = 0.;
  
  //Match the segments
  double nbInliers = this->SegmentMatching();
  
  if(nbInliers >= static_cast<double> (m_InliersRatio*std::min(m_FirstLineVector.size(), m_SecondLineVector.size())) && nbInliers > m_BestInlier)
    {
      m_BestInlier = nbInliers;
      m_BestCumulatedDistance = m_CurrentCumulatedDistance;
      m_BestTransform->SetParameters( m_Transform->GetParameters());
      
      otbMsgDevMacro(" -------------------------------------- ");
      otbMsgDevMacro(" Transform Parameters " << m_Transform->GetParameters() << " Inliers found "  <<  nbInliers );
      otbMsgDevMacro("        m_BestCumulatedDistance " << m_BestCumulatedDistance << " m_CurrentCumulatedDistance  " << m_CurrentCumulatedDistance  );
      otbMsgDevMacro("            Best  Transform"   << m_BestTransform->GetParameters() );
      otbMsgDevMacro(" -------------------------------------- ");
    }
  if(nbInliers == m_BestInlier && nbInliers > 0 &&  m_CurrentCumulatedDistance < m_BestCumulatedDistance)
    {
      m_BestCumulatedDistance = m_CurrentCumulatedDistance;
      m_BestTransform->SetParameters( m_Transform->GetParameters());

      otbMsgDevMacro(" -------------------------------------- ");
      otbMsgDevMacro(" Transform Parameters " << m_Transform->GetParameters() << " Inliers found "  <<  nbInliers );
      otbMsgDevMacro("        m_BestCumulatedDistance " << m_BestCumulatedDistance << " m_CurrentCumulatedDistance  " << m_CurrentCumulatedDistance  );
      otbMsgDevMacro("            Best  Transform"   << m_BestTransform->GetParameters() );
      otbMsgDevMacro(" -------------------------------------- ");
    }
}


/**
 * Using the transform , we compute the number of inliers
 *
 */
double
TripletToTransformFilter
::SegmentMatching()
{
  TransformType::OutputPointType                      nullP;
  nullP.Fill(0);
  double    inliers    = 0;
  
  VertexVectorType::iterator itSecond    = m_SecondLineVector.begin();
  VertexVectorType::iterator itEndSecond = m_SecondLineVector.end();

  while(itSecond != itEndSecond)
    {
      //project the points of the second set
      std::vector<TransformType::OutputPointType >        pDst(2, nullP);

      for(unsigned int i = 0; i< (*itSecond).size(); i++)
       {
         pDst[i] = m_Transform->TransformPoint( (*itSecond)[i] );
       }
            
      VertexVectorType::iterator itFirst    = m_FirstLineVector.begin();
      VertexVectorType::iterator itEndFirst = m_FirstLineVector.end();
      
      while(itFirst != itEndFirst )
       {
         //Check if the current Tuplets are the same
         if(this->IsMatchedSegment(itFirst, pDst))
           {
             ++inliers;
             break;
           }
         ++itFirst;
       }
      ++itSecond;
    }
 
 return inliers;

}

/**
 *
 */
bool
TripletToTransformFilter
::IsMatchedSegment(VertexVectorType::iterator  itFirst, std::vector<TransformType::OutputPointType > pDst)
{

  double dist1  = this->ComputeEuclideanDistance((*itFirst)[0], pDst[0]);
  double dist2  = this->ComputeEuclideanDistance((*itFirst)[1], pDst[1]);
  double dist3  = this->ComputeEuclideanDistance((*itFirst)[0], pDst[1]);
  double dist4  = this->ComputeEuclideanDistance((*itFirst)[1], pDst[0]);

  double finalDistance = std::min(dist1+dist2, dist3+dist4);
      
  if(finalDistance<m_ThresholdDistance )
    {
      m_CurrentCumulatedDistance += finalDistance;
      return true;
    }
  return false;
}

/**
 *
 */
double
TripletToTransformFilter
::ComputeEuclideanDistance(VertexType pSrc, TransformType::OutputPointType pDst )
{
  return vcl_sqrt( (pSrc[0]-pDst[0])*(pSrc[0]-pDst[0]) + (pSrc[1]-pDst[1])*(pSrc[1]-pDst[1]) );
}

} // end namespace otb
