/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITYor FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#ifndef __otbTripletToTransformFilter_h
#define __otbTripletToTransformFilter_h

#include  "itkProcessObject.h"
#include "otbVectorData.h"
#include  "itkAffineTransform.h"
#include  "itkSize.h"
#include  "itkVector.h"
#include  "itkPoint.h"


namespace otb
{
  /** \class TripletToTransformFilter
   *  \brief this class computes the best rigid affine transformation
   *         between two sets of triplets :
   *         p' = A*p + T
   *         This aims to approximate A and T.
   *         Two ways to choose the affineTransform : compute score using triplets
   *                                                  compute score using valid segment
   *
   */
  class /*ITK_EXPORT*/ TripletToTransformFilter :
  public itk::ProcessObject
  {
  public:
    /** Standard class typedefs. */
    typedef TripletToTransformFilter                               Self;
    typedef itk::ProcessObject                                     Superclass;
    typedef itk::SmartPointer<Self>                                Pointer;
    typedef itk::SmartPointer<const Self>                          ConstPointer;

    /** Method for creation through the object factory. */
    itkNewMacro(Self);

    /** Run-time type information (and related methods). */
    itkTypeMacro(TripletToTransformFilter, itk::ProcessObject);
    
    
    typedef VectorData<double, 2>                                VectorDataType;
    typedef VectorDataType::DataNodeType::LineType::VertexType  VertexType;
    
    typedef std::vector<VertexType>                             LineVectorType;
    typedef std::vector<LineVectorType>                         VertexVectorType;

    typedef itk::Vector<double, 2>                              SpacingType;
    typedef itk::Point<double, 2>                               OriginType;
    typedef itk::Size<2>                                        SizeType;

    //SimilarTransform
    typedef itk::AffineTransform<double, 2>                      TransformType;
    typedef TransformType::Pointer                              TransformPointerType;
    
    /* mandatory call to begin the process **/
    /*     void Update() */
    /*       { */
    /*        this->Run(); */
    /*       }; */
    
    //Set the vector of Tuplets
    void SetFirstTripletVector(VertexVectorType vfirst)
      {
       m_FirstTupletVector = vfirst;
       this->Modified();
      }
    void SetSecondTripletVector(VertexVectorType vsecond)
      {
       m_SecondTupletVector = vsecond;
       this->Modified();
      }

    //Get the transform
    itkGetObjectMacro(Transform, TransformType);
    itkGetObjectMacro(BestTransform, TransformType);
    

    //Get/Set ThresholdDistance
    itkGetMacro(ThresholdDistance, double);
    itkSetMacro(ThresholdDistance, double);

    //Get/Set Inliers Ratio
    itkGetMacro(InliersRatio, double);
    itkSetMacro(InliersRatio, double);

    //Get/Set Maximum Iteration to find a valid transform
    itkGetMacro(IterationMaximum, int);
    itkSetMacro(IterationMaximum, int);

    //Get/Set the maximum distance between Triplet to allow Transform Computation
    itkGetMacro(TripletMaxDistance, double);
    itkSetMacro(TripletMaxDistance, double);
    
    
    //Set Line Vector
    void SetFirstLineVector(VertexVectorType fvector)
      {
       m_FirstLineVector = fvector;
       this->Modified();
      }

    void SetSecondLineVector(VertexVectorType svector)
      {
       m_SecondLineVector = svector;
       this->Modified();
      }

    virtual VectorDataType* GetOutput();
      

    virtual void InitTransformation(unsigned int count, unsigned int count2);
    virtual void UpToDateTransform(unsigned int relement);
    virtual double SegmentMatching();
    
  protected:
    virtual void GenerateData();
    virtual double ComputeEuclideanDistance(VertexType pSrc, TransformType::OutputPointType pDst );
    virtual bool IsMatchedSegment(VertexVectorType::iterator  itFirst, std::vector<TransformType::OutputPointType > pDst);
    
    TripletToTransformFilter();
    ~TripletToTransformFilter(){};

  private:
    TripletToTransformFilter(const Self&); //purposely not implemented
    void operator=(const Self&); //purposely not implemented
    
    double                        m_ThresholdDistance;
    double                        m_InliersRatio;
    double                        m_BestInlier;
    int                           m_IterationMaximum;
    double                        m_CurrentCumulatedDistance;
    double                        m_BestCumulatedDistance;
    
    // Maximum distance authorized between 2 analyzed triplets
    double                        m_TripletMaxDistance;
    
    TransformPointerType          m_Transform;
    TransformPointerType          m_BestTransform;
    
    VertexVectorType              m_FirstTupletVector;
    VertexVectorType              m_SecondTupletVector;
    VertexVectorType              m_PossiblePermutation;
    VertexVectorType              m_FirstLineVector;
    VertexVectorType              m_SecondLineVector;
  };

}


#endif
