/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbVectorDataToHistogramFilter_txx
#define __otbVectorDataToHistogramFilter_txx

namespace otb
{

  template <class TVectorData, class THistogram>
  VectorDataToHistogramFilter<TVectorData, THistogram>
  ::VectorDataToHistogramFilter()
  {
    this->SetNumberOfRequiredInputs(1);
    this->SetNumberOfOutputs(1);
    this->SetNthOutput(0, VectorDataType::New());

    m_Histogram  = HistogramType::New();
    m_InverseHistogram = HistogramType::New();
    m_Size.Fill(180);
    m_Mean = 0;
    m_Sigma = 0.;
    m_Origin.Fill(0);
    m_Histogram->Initialize(m_Size);
    m_InverseHistogram->Initialize(m_Size);


  }

template <class TVectorData, class THistogram>
typename VectorDataToHistogramFilter<TVectorData, THistogram>
::VectorDataType *
VectorDataToHistogramFilter<TVectorData, THistogram>
::GetOutput()
{
  return static_cast< VectorDataType * >
    (this->itk::ProcessObject::GetOutput(0) );
}
  

  
/**
 * Set Input VectorData
 */
template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::SetVectorData(VectorDataType * vector)
{
  this->itk::ProcessObject::SetNthInput(0, const_cast<VectorDataType *>(vector));
}

/**
 * GetVectorData
 */
template <class TVectorData, class THistogram>
const typename VectorDataToHistogramFilter<TVectorData, THistogram>
::VectorDataType *
VectorDataToHistogramFilter<TVectorData, THistogram>
::GetVectorData()
{
  return static_cast<const VectorDataType *>(this->itk::ProcessObject::GetInput(0));
}


template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::GenerateData()
{
  VectorDataPointerType vectorData  = const_cast<VectorDataType *>(this->GetVectorData());

  // Get the input tree root
  InternalTreeNodeType * inputRoot = const_cast<InternalTreeNodeType *>(vectorData->GetDataTree()->GetRoot());
  
  // Start recursive processing
  ProcessNode(inputRoot);
  
  //Compute the mean and the variance of the histogram
  this->ComputeHistogramVariance();
}

  /**
   */
template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::ProcessNode(InternalTreeNodeType * source)
{
  // Get the children list from the input node
  typename InternalTreeNodeType::ChildrenListType children = source->GetChildrenList();
  // For each child
  for(typename InternalTreeNodeType::ChildrenListType::iterator it = children.begin(); it!=children.end(); ++it)
    {
      DataNodePointerType dataNode = (*it)->Get();
      
      switch(dataNode->GetNodeType())
       {
       case ROOT:
         {
           ProcessNode((*it));
           break;
         }
       case DOCUMENT:
         {
           ProcessNode((*it));
           break;
         }
       case FOLDER:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_POINT:
         {
           break;
         }
       case FEATURE_LINE:
         {
           this->VertexToHistogram(dataNode->GetLine()->GetVertexList());
           break;
         }
       case FEATURE_POLYGON:
         {
           this->VertexToHistogram(dataNode->GetPolygonExteriorRing()->GetVertexList());
           break;
         }
       case FEATURE_MULTIPOINT:
         {
           break;
         }
       case FEATURE_MULTILINE:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_MULTIPOLYGON:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_COLLECTION:
         {
           ProcessNode((*it));
           break;
         }
       }
      }
}

  /**
   */
template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::VertexToHistogram(typename VertexListType::ConstPointer vertexlist)
{
  VertexType currentVertex, nextVertex;
  typename VertexListType::ConstIterator it = vertexlist->Begin();

  if (vertexlist->Size() > 0 )
    {
      currentVertex = it.Value();
      ++it;
      while (it != vertexlist->End())
       {
         nextVertex = it.Value();
         this->UpdateHistogram(currentVertex, nextVertex);
         currentVertex = nextVertex;
         ++it;
       }
    }
}

  /**
   */
template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::UpdateHistogram(VertexType currentVertex, VertexType nextVertex)
{
  currentVertex[0] += m_Origin[0];
  currentVertex[1] += m_Origin[1];
  
  nextVertex[0]    += m_Origin[1];
  nextVertex[1]    += m_Origin[1];
  
  //Compute the length
  double length = vcl_sqrt((currentVertex[0]-nextVertex[0])*(currentVertex[0]-nextVertex[0])
                        +(currentVertex[1]-nextVertex[1])*(currentVertex[1]-nextVertex[1]));
  
  //Compute the orientation
  double dx = currentVertex[0]-nextVertex[0];
  double dy = currentVertex[1]-nextVertex[1];
  double orientation = vcl_atan2(dy, dx);
  if(orientation < 0)
    orientation += CONST_PI;
    
  //Update the histogram
  unsigned int instanceIdentifier = static_cast<unsigned int >(orientation*180./CONST_PI);
  m_Histogram->IncreaseFrequency(instanceIdentifier, length);
  
  //Update the mean of the histogram
  m_Mean += length;
}

/**
 */
template <class TVectorData, class THistogram>
void
VectorDataToHistogramFilter<TVectorData, THistogram>
::ComputeHistogramVariance()
{
  //Normalize the mean
  m_Mean /= m_Histogram->Size();
  
  //Compute the variance
  typename HistogramType::Iterator it = m_Histogram->Begin();
  while(it != m_Histogram->End())
    {
      m_Sigma += (it.GetFrequency() - m_Mean)*(it.GetFrequency() - m_Mean);
      ++it;
    }

  m_Sigma  = vcl_sqrt(m_Sigma)/(m_Histogram->Size());
}
  
/**
 */
template <class TVectorData, class THistogram>
typename VectorDataToHistogramFilter<TVectorData, THistogram>
::HistogramPointerType
VectorDataToHistogramFilter<TVectorData, THistogram>
::GetInverseHistogram()
{
  typename HistogramType::Iterator it = m_Histogram->Begin();
  unsigned int size = m_Histogram->Size();
  
  while(it != m_Histogram->End())
    {
      m_InverseHistogram->SetFrequency(size - it.GetInstanceIdentifier()-1, it.GetFrequency());
      ++it;
    }
  return m_InverseHistogram;
}


} // end namespace otb

#endif
