/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITYor FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/

#ifndef __otbVectorDataToValidSegmentsFilter_h
#define __otbVectorDataToValidSegmentsFilter_h

#include  "otbVectorDataToVectorDataFilter.h"
#include  "itkSize.h"
#include  "itkPoint.h"
#include  "itkHistogram.h"

namespace otb
{
  /** \class otbVectorDataToValidSegmentsFilter
   *  \brief This class detect the reliable segments in a VectorData by
   *         applying the hough transform:
   *         we build an accumulator (rho, tetha ) where rho is the distance from the center
   *         and  theta the oritentation of the segment.
   *         The valid kept segment are the segments (rho, theta) with an accumulated length
   *         larger than a threshold set by the user.(SetFrequencyThreshold).
   */

template <class TVectorData>
class ITK_EXPORT VectorDataToValidSegmentsFilter :
  public VectorDataToVectorDataFilter<TVectorData, TVectorData>
  {
  public:
    /** Standard class typedefs. */
    typedef VectorDataToValidSegmentsFilter                            Self;
    typedef VectorDataToVectorDataFilter<TVectorData, TVectorData>      Superclass;
    typedef itk::SmartPointer<Self>                                Pointer;
    typedef itk::SmartPointer<const Self>                          ConstPointer;

    /** Run-time type information (and related methods). */
    itkTypeMacro(otbVectorDataToValidSegmentsFilter, VectorDataToVectorDataFilter);
    
    /** Method for creation through the object factory. */
    itkNewMacro(Self);

    /** VectorData TypeDef*/
    typedef TVectorData                                         VectorDataType;
    typedef typename VectorDataType::Pointer                    VectorDataPointerType;
    typedef typename VectorDataType::DataNodeType               DataNodeType;
    typedef typename DataNodeType::Pointer                      DataNodePointerType;
    typedef typename DataNodeType::PolygonType::VertexType      VertexType;
    typedef typename DataNodeType::PolygonType::VertexListType  VertexListType;
    
    typedef typename VectorDataType::DataTreeType::TreeNodeType InternalTreeNodeType;
    typedef typename InternalTreeNodeType::ChildrenListType     ChildrenListType;

    typedef std::vector<VertexType>                              LineVectorType;
    typedef std::vector<LineVectorType>                          VertexVectorType;

    //Accumulator
    typedef itk::Statistics::Histogram<typename VectorDataType::PrecisionType, 2>     AccumulatorType;
    typedef typename AccumulatorType::Pointer                                        AccumulatorPointerType;
    typedef itk::Index<2>                                                            IndexType;
    typedef std::pair<VertexType, VertexType>                                         PairType;
    typedef itk::Point<double, 2>                                                     PointType;
    
    //Get/Set the Frequency  threshold
    itkSetMacro(FrequencyThreshold, double);
    itkGetMacro(FrequencyThreshold, double);

    //Get/Set the Number of segment kept
    itkGetMacro(NumberOfKeptLines, unsigned int);
    itkSetMacro(NumberOfKeptLines, unsigned int);

    //Get/Set Image origin
    itkSetMacro(Origin, PointType);
    
    //Get the LineVector
    VertexVectorType GetLineVector()
      {
       return m_LineVector;
      }
    
  protected:
    virtual void   GenerateData(void);
    virtual void   ProcessNode(InternalTreeNodeType * source);
    virtual double ComputeEuclideanDistanceToCenter(VertexType q1, VertexType q2);
    virtual void   ComputeHoughParameters(typename VertexListType::ConstPointer  vertexList);
    virtual double ComputeSegmentOrientation(VertexType currentVertex, VertexType nextVertex);
    virtual double ComputeSegmentLength(VertexType currentVertex, VertexType nextVertex);
    virtual void   UpdateStoredSegment(IndexType index, VertexType currentVertex, VertexType nextVertex);
    virtual void   WriteOutputVectorData();
 
    VectorDataToValidSegmentsFilter();
    virtual ~VectorDataToValidSegmentsFilter() {};
    
  private:
    VectorDataToValidSegmentsFilter(const Self&); //purposely not implemented
    void operator=(const Self&); //purposely not implemented
    
    double                                              m_FrequencyThreshold;
    unsigned int                                        m_NumberOfKeptLines;
    
    itk::Size<2>                                        m_Size;
    AccumulatorPointerType                              m_Accumulator;
    std::vector< PairType >                             m_StoreVector;
    std::vector<bool>                                   m_HasUpToDate;
    PointType                                           m_Origin;

    VertexVectorType                                    m_LineVector;
  };


}
#ifndef ITK_MANUAL_INSTANTIATION
#include "otbVectorDataToValidSegmentsFilter.txx"
#endif

#endif
