/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbVectorDataToValidSegmentsFilter_txx
#define __otbVectorDataToValidSegmentsFilter_txx

#include "itkTimeProbe.h"
#include "itkIndex.h"

namespace otb
{

template <class TVectorData>
VectorDataToValidSegmentsFilter<TVectorData>
::VectorDataToValidSegmentsFilter()
{
  m_Accumulator = AccumulatorType::New();
  m_Size[0] = 180;
  m_Size[1] = 13000;
  //m_StoreVector.resize(m_Size[0]*m_Size[1]);
  m_HasUpToDate.resize(m_Size[0]*m_Size[1]);
  for(unsigned int i = 0; i<m_HasUpToDate.size(); i++)
    m_HasUpToDate[i] = false;

  m_FrequencyThreshold = 5.;
  m_NumberOfKeptLines  = 40;
  m_Origin.Fill(0.);

}


template <class TVectorData>
void
VectorDataToValidSegmentsFilter<TVectorData>
::GenerateData()
{
    
  // Get the input tree root
  InternalTreeNodeType * inputRoot = const_cast<InternalTreeNodeType *>(this->GetInput()->GetDataTree()->GetRoot());
  
  //Initialize the accumulator
  m_Accumulator->Initialize(m_Size);
    
  // Start recursive processing
  itk::TimeProbe chrono;
  chrono.Start();
  ProcessNode(inputRoot);
  chrono.Stop();
  otbMsgDevMacro("Features processed in "<<chrono.GetMeanTime()<<" seconds.");
  
  //
  this->WriteOutputVectorData();
  otbMsgDevMacro("VectorDataToValidSegment: Size of the Lines Kept  is " << m_LineVector.size());
}

  /**
   */
template <class TVectorData>
void
VectorDataToValidSegmentsFilter<TVectorData>
::ProcessNode(InternalTreeNodeType * source)
{
  // Get the children list from the input node
  typename InternalTreeNodeType::ChildrenListType children = source->GetChildrenList();
  // For each child
  for(typename InternalTreeNodeType::ChildrenListType::iterator it = children.begin(); it!=children.end(); ++it)
    {
      DataNodePointerType dataNode = (*it)->Get();
      
      switch(dataNode->GetNodeType())
       {
       case ROOT:
         {
           ProcessNode((*it));
           break;
         }
       case DOCUMENT:
         {
           ProcessNode((*it));
           break;
         }
       case FOLDER:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_POINT:
         {
           break;
         }
       case FEATURE_LINE:
         {
           this->ComputeHoughParameters(dataNode->GetLine()->GetVertexList());
           break;
         }
       case FEATURE_POLYGON:
         {
           this->ComputeHoughParameters(dataNode->GetPolygonExteriorRing()->GetVertexList());
           break;
         }
       case FEATURE_MULTIPOINT:
         {
           break;
         }
       case FEATURE_MULTILINE:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_MULTIPOLYGON:
         {
           ProcessNode((*it));
           break;
         }
       case FEATURE_COLLECTION:
         {
           ProcessNode((*it));
           break;
         }
       }
      }
}

  /**
   */
template <class TVectorData>
double
VectorDataToValidSegmentsFilter<TVectorData>
::ComputeEuclideanDistanceToCenter(VertexType q1, VertexType q2)
{
  double Xq1 = q1[0]+m_Origin[0];
  double Yq1 = q1[1]+m_Origin[1];
  double Xq2 = q2[0]+m_Origin[0];
  double Yq2 = q2[1]+m_Origin[1];
  
  double SegmentLength = this->ComputeSegmentLength(q1, q2);
  double CrossProduct  =   Xq1*Yq2 - Xq2*Yq1;
  double Num   = vcl_abs(CrossProduct);
  
  /** distance from Point P to Segment Q1Q2*/
  return  (Num/SegmentLength);
}

/**
 *
 */
template <class TVectorData>
double
VectorDataToValidSegmentsFilter<TVectorData>
::ComputeSegmentLength(VertexType currentVertex, VertexType nextVertex)
{
  //Compute the length
  double length = vcl_sqrt((currentVertex[0]-nextVertex[0])*(currentVertex[0]-nextVertex[0])
                        +(currentVertex[1]-nextVertex[1])*(currentVertex[1]-nextVertex[1]));
  return length;
}

/**
 *
 */
template <class TVectorData>
double
VectorDataToValidSegmentsFilter<TVectorData>
::ComputeSegmentOrientation(VertexType currentVertex, VertexType nextVertex)
{
  //Compute the orientation
  double dx          = currentVertex[0]-nextVertex[0];
  double dy          = currentVertex[1]-nextVertex[1];
  double orientation = vcl_atan2(dy, dx);
  if(orientation <0)
    orientation += CONST_PI;
  return orientation;
}
/**
* \brief : compute the theta and rho parameters
*/
template <class TVectorData>
void
VectorDataToValidSegmentsFilter<TVectorData>
::ComputeHoughParameters(typename VertexListType::ConstPointer  vertexList)
{
  VertexType currentVertex, nextVertex;
  typename VertexListType::ConstIterator it = vertexList->Begin();

  if (vertexList->Size() > 0 )
    {
      currentVertex = it.Value();
      ++it;
      while (it != vertexList->End())
       {
         nextVertex = it.Value();
         IndexType        index;
         index.Fill(0);
         this->UpdateStoredSegment(index, currentVertex , nextVertex);
         currentVertex = nextVertex;
         ++it;
       }
    }
}

/**
 * Udpate with the min and the max of the vertex
 **/
template <class TVectorData>
void
VectorDataToValidSegmentsFilter<TVectorData>
::UpdateStoredSegment(IndexType index, VertexType currentVertex, VertexType nextVertex)
{
  PairType   pair;
  pair.first = currentVertex;
  pair.second = nextVertex;
  m_StoreVector.push_back(pair);
}

/**
 *
 */
template <class TVectorData>
void
VectorDataToValidSegmentsFilter<TVectorData>
::WriteOutputVectorData()
{
  //Get The output
  typename VectorDataType::Pointer             outputPtr = this->GetOutput();
  typename VectorDataType::DataTreePointerType tree      = outputPtr->GetDataTree();

  InternalTreeNodeType * outputRoot = const_cast<InternalTreeNodeType *>(this->GetOutput()->GetDataTree()->GetRoot());
  
  //add the root
  typename DataNodeType::Pointer document = DataNodeType::New();
  typename DataNodeType::Pointer folder   = DataNodeType::New();
  typename DataNodeType::Pointer root     =   outputRoot->Get();
 
  //add a document
  document->SetNodeType(otb::DOCUMENT);
  document->SetNodeId("DOCUMENT");
 
  //add a  folder
  folder->SetNodeType(otb::FOLDER);
  folder->SetNodeId("FOLDER");

  //Build the output tree
  tree->Add(document, root);
  tree->Add(folder, document);

  typename AccumulatorType::Iterator it = m_Accumulator->Begin();
  typedef std::pair<double, unsigned int >     pairType2;
  std::vector<pairType2 >    instanceIdVector;
 
  for(unsigned int i = 0; i< m_StoreVector.size(); i++)
    {
      pairType2  pair;
      pair.first = this->ComputeSegmentLength(m_StoreVector[i].first, m_StoreVector[i].second);
      pair.second = i;
      instanceIdVector.push_back(pair);
      
    }
  
  //sort the instanceIdentifier
  std::sort(instanceIdVector.begin(), instanceIdVector.end());

  unsigned int count = 0;
  for(unsigned int s = instanceIdVector.size()-1; s > instanceIdVector.size() - m_NumberOfKeptLines; s--)
    {

      if(count < instanceIdVector.size())
       {
         double length = this->ComputeSegmentLength(m_StoreVector[instanceIdVector[s].second].first, m_StoreVector[instanceIdVector[s].second].second);
         if(length > m_FrequencyThreshold )
         {
           DataNodePointerType line = DataNodeType::New();
           line->SetNodeType(otb::FEATURE_LINE);
           line->SetNodeId("FEATURE_LINE");
           typename  DataNodeType::LineType::Pointer l = DataNodeType::LineType::New();
           l->AddVertex(m_StoreVector[instanceIdVector[s].second].first);
           l->AddVertex(m_StoreVector[instanceIdVector[s].second].second);
           line->SetLine(l);
           tree->Add(line, folder);
                      
           //Fill the line Vector
           LineVectorType         significantLine;
           significantLine.push_back(m_StoreVector[instanceIdVector[s].second].first);
           significantLine.push_back(m_StoreVector[instanceIdVector[s].second].second);
           m_LineVector.push_back(significantLine);
         }
       }
      count++;
    }
}


} // end namespace otb

#endif
