/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbLandCoverMapModel.h"
#include "itkExceptionObject.h"
#include "otbMacro.h"

#include "itkImageRegionConstIterator.h"

namespace otb
{
/** Initialize the singleton */
LandCoverMapModel::Pointer LandCoverMapModel::Instance = NULL;


LandCoverMapModel
::LandCoverMapModel()
{
  /** Initial parameters */
  m_InputImageList = VectorToImageListType::New();
  m_NumberOfComponent = 0;
  std::vector<bool> sf(3, true);
  m_SelectedFilters = sf;
  m_Reader               =  ReaderType::New();
  m_Classifier = ClassificationFilterType::New();
  m_Model = ModelType::New();
  m_OutputImage = LabeledVectorImageType::New();
  m_HasInput    = false;
  m_HasOutput   = false;

  // Filter List
  //m_MeanShift = MeanShiftFilterType::New();
  m_NDVI = NDVIFilterType::New();
  m_ARVI = ARVIFilterType::New();
  m_NDVIGR = NDVIFilterType::New();

  m_OutputList =    ImageListType::New();
  m_InputImageList =  VectorToImageListType::New();
  m_IL2VI =   ImageListToVectorImageFilterType::New();
  m_Rescaler = RescaleFilterType::New();
  m_VectorRescaler = VectorRescaleFilterType::New();
  m_InputRescaler = RescaleFilterType::New();
  m_InputVectorRescaler = VectorRescaleFilterType::New();
  m_BlackGenerator = BlackImageType::New();

  m_Index.clear();

  m_ChangeLabelFilter = ChangeLabelFilterType::New();
  m_Writer = WriterType::New();

  this->InitColor();
}


LandCoverMapModel
::~LandCoverMapModel() {}


void
LandCoverMapModel
::InitColor()
{
  m_ChangeLabelFilter->SetNumberOfComponentsPerPixel(3);

  LabelVectorType vect;
  vect.SetSize( 3 );
  vect.Fill(1);

  vect[0] = 0;
  vect[1] = 255;
  vect[2] = 0;
  m_ChangeLabelFilter->SetChange(1, vect);
  vect[0] = 0;
  vect[1] = 0;
  vect[2] = 255;
  m_ChangeLabelFilter->SetChange(2, vect);
  vect[0] = 255;
  vect[1] = 0;
  vect[2] = 0;
  m_ChangeLabelFilter->SetChange(3, vect);
  vect[0] = 125;
  vect[1] = 125;
  vect[2] = 125;
  m_ChangeLabelFilter->SetChange(4, vect);
  vect[0] = 200;
  vect[1] = 125;
  vect[2] = 0;
  m_ChangeLabelFilter->SetChange(5, vect);
  vect[0] = 0;
  vect[1] = 0;
  vect[2] = 0;
  m_ChangeLabelFilter->SetChange(6 , vect);
}


/** Manage the singleton */
LandCoverMapModel::Pointer
LandCoverMapModel::GetInstance()
{
  if (!Instance)
  {
    Instance = LandCoverMapModel::New();
  }
  return Instance;
}


void
LandCoverMapModel
::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}


void
LandCoverMapModel
::LoadModel(const char * modelFilename)
{
  m_Model = ModelType::New();
  m_Model->LoadModel(modelFilename);
  std::string m_ModelFileName( modelFilename );
}


void
LandCoverMapModel
::LoadImage(InputImagePointerType image)
{
  // Set the input image
  m_InputImage = InputImageType::New();
  m_InputImage = image;
  m_InputImage->UpdateOutputInformation();
  m_NumberOfComponent = m_InputImage->GetNumberOfComponentsPerPixel();
  m_BlackGenerator = BlackImageType::New();
  m_BlackGenerator->SetRegion( m_InputImage->GetLargestPossibleRegion());
  m_BlackGenerator->UpdateOutputInformation();

  this->WirePipeline();
}

void
LandCoverMapModel
::SetInputImage(std::string filename)
{

  m_Reader = ReaderType::New();
  m_InputFileName = filename;
  m_Reader->SetFileName(filename);
  m_Reader->UpdateOutputInformation();
  if ( m_Reader->GetOutput()->GetNumberOfComponentsPerPixel() != 4)
  {
    itkExceptionMacro(<<"Invalid Input Image : 4 channels are required instead of "<<m_Reader->GetOutput()->GetNumberOfComponentsPerPixel() );
  }
  this->LoadImage( m_Reader->GetOutput() );

  m_HasInput = true;
  this->NotifyAll();
  m_HasInput = false;

}

void
LandCoverMapModel
::WireNDVI()
{
  m_NDVI = NDVIFilterType::New();
  m_NDVI->SetInput( m_Reader->GetOutput() );
  m_NDVI->SetRedIndex(3);
  m_NDVI->SetNIRIndex(4);

  m_NDVI->UpdateOutputInformation();
}


void
LandCoverMapModel
::WireNDVIGR()
{
  // Use Green Channel instead of NIR channel
  m_NDVIGR = NDVIFilterType::New();
  m_NDVIGR->SetInput( m_Reader->GetOutput() );
  m_NDVIGR->SetRedIndex(3);
  m_NDVIGR->SetNIRIndex(2);

  m_NDVIGR->UpdateOutputInformation();
}


void
LandCoverMapModel
::WireARVI()
{
  m_ARVI = ARVIFilterType::New();
  m_ARVI->SetInput( m_Reader->GetOutput() );
  m_ARVI->SetRedIndex(3);
  m_ARVI->SetBlueIndex(1);
  m_ARVI->SetNIRIndex(4);
  m_ARVI->GetFunctor().SetGamma(0.5);

  m_ARVI->UpdateOutputInformation();
}


void
LandCoverMapModel
::WirePipeline()
{
  //m_InputImageList       = VectorToImageListType::New();
//  m_MeanShift       = MeanShiftFilterType::New();
//  m_MeanShiftImageList  = VectorToImageListType::New();
  m_OutputList      = ImageListType::New();
  m_IL2VI           = ImageListToVectorImageFilterType::New();

  m_Index.clear();
  unsigned int id = 0;

  m_InputRescaler = RescaleFilterType::New();
  m_InputVectorRescaler = VectorRescaleFilterType::New();

  m_InputVectorRescaler->SetFilter( m_InputRescaler );

  m_InputVectorRescaler->SetInput( m_Reader->GetOutput() );
  PixelType pix = static_cast<PixelType>(0);
  m_InputRescaler->SetOutputMinimum(pix);
  pix = static_cast<PixelType>(1);
  m_InputRescaler->SetOutputMaximum(pix);

  pix = static_cast<PixelType>(0);
  m_InputRescaler->SetWindowMinimum(pix);
  pix = static_cast<PixelType>(vcl_pow(2.0, 10)); //Max
  //for
  //a
  //QB pixel

  m_InputRescaler->SetWindowMaximum(pix);


  m_InputImageList->SetInput(m_InputVectorRescaler->GetOutput() );
  m_InputImageList->UpdateOutputInformation();


//   m_MeanShift->SetInput(m_Reader->GetOutput() );
//   m_MeanShift->SetSpatialRadius(8);
//   m_MeanShift->SetRangeRadius(20);
//   m_MeanShift->UpdateOutputInformation();

//   m_MeanShiftImageList->SetInput( m_MeanShift->GetOutput() );
//   m_MeanShiftImageList->UpdateOutputInformation();

  // Add Input Image
  m_Index.push_back(id);
  for ( unsigned int w=0; w<m_InputImageList->GetOutput()->Size(); w++)
  {
    m_OutputList->PushBack(m_InputImageList->GetOutput()->GetNthElement(w));
    id++;
  }

  // Add Mean Shift Image
//   m_Index.push_back(id);
//   for( unsigned int w=0; w<m_MeanShiftImageList->GetOutput()->Size(); w++)
//     {
//       m_OutputList->PushBack(m_MeanShiftImageList->GetOutput()->GetNthElement(w));
//       id++;
//     }


  // NDVI
  m_Index.push_back(id);
  this->WireNDVI();
  m_OutputList->PushBack( m_NDVI->GetOutput() );
  id++;

  // NDVI RG
  m_Index.push_back(id);
  this->WireNDVIGR();
  m_OutputList->PushBack( m_NDVIGR->GetOutput() );
  id++;

  // ARVI
  m_Index.push_back(id);
  this->WireARVI();
  m_OutputList->PushBack( m_ARVI->GetOutput() );
  id++;


  m_OutputList->UpdateOutputInformation();

  m_IL2VI->SetInput( m_OutputList );
  m_IL2VI->UpdateOutputInformation();

}


// If not selected, set the image to black
void
LandCoverMapModel
::CheckSelected()
{
//   // Mean Shift
//   if(m_SelectedFilters[0] == false)
//     {
//       for(unsigned int i=m_Index[1]; i<m_Index[2]; i++)
//   {
//     m_OutputList->SetNthElement(  i , m_BlackGenerator->GetOutput() );
//   }
//     }


  // NDVI
  if (m_SelectedFilters[0] == false)
  {
    m_OutputList->SetNthElement(  m_Index[1] , m_BlackGenerator->GetOutput() );
  }
  // NDVI RG
  if (m_SelectedFilters[1] == false)
  {
    m_OutputList->SetNthElement(  m_Index[2] , m_BlackGenerator->GetOutput() );
  }
  // ARVI
  if (m_SelectedFilters[2] == false)
  {
    m_OutputList->SetNthElement(  m_Index[3] , m_BlackGenerator->GetOutput() );
  }

  m_OutputList->UpdateOutputInformation();
  m_IL2VI->UpdateOutputInformation();
}

void
LandCoverMapModel
::SetupProcessing()
{
  m_Rescaler          = RescaleFilterType::New();
  m_VectorRescaler          = VectorRescaleFilterType::New();

  m_VectorRescaler->SetFilter( m_Rescaler );

  m_Classifier        = ClassificationFilterType::New();
  m_OutputImage       = LabeledVectorImageType::New();

  this->WirePipeline();
  this->InitColor();
  this->CheckSelected();


  m_VectorRescaler->SetInput( m_IL2VI->GetOutput() );
  PixelType pix = static_cast<PixelType>(-1);
  m_Rescaler->SetOutputMinimum(pix);
  pix = static_cast<PixelType>(1);
  m_Rescaler->SetOutputMaximum(pix);

  pix = static_cast<PixelType>(0);
  m_Rescaler->SetWindowMinimum(pix);
  pix = static_cast<PixelType>(1);
  m_Rescaler->SetWindowMaximum(pix);

  m_VectorRescaler->UpdateOutputInformation();


}


void
LandCoverMapModel
::SaveFeatureImage()
{
  this->SetupProcessing();


  FeatureWriterType::Pointer writer = FeatureWriterType::New();
  writer->SetFileName( m_FeatureFileName );
  writer->SetInput( m_VectorRescaler->GetOutput() );
  writer->SetAutomaticTiledStreaming();

  try
  {
    writer->Update();
  }
  catch ( itk::ExceptionObject & err )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Following Exception caught : ";
    oss<<err;
    return;
  }
  catch ( std::bad_alloc & err )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Bad alloc, following exception caught : ";
    oss<<(char*)err.what();
    return;
  }
  catch ( ... )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Unknown Exception found !";
    return;
  }
}

void
LandCoverMapModel
::SaveResult()
{
  this->SetupProcessing();

  m_Classifier->SetInput( m_VectorRescaler->GetOutput() );
  m_Classifier->SetModel( m_Model );

  m_ChangeLabelFilter->SetInput( m_Classifier->GetOutput() );
  /*  m_ChangeLabelFilter->GetOutput()->UpdateOutputInformation(); */

  m_Writer->SetFileName( m_OutputFileName );
  m_Writer->SetInput( m_ChangeLabelFilter->GetOutput() );
  m_Writer->SetAutomaticTiledStreaming();

  // StandardFilterWatcher watcher1(m_NDVI,"NDVI");
  // StandardFilterWatcher watcher2(m_NDVIGR,"NDVIGR");
  // StandardFilterWatcher watcher3(m_ARVI,"ARVI");
  // StandardFilterWatcher watcher4(m_Rescaler,"Rescaler");
  // StandardFilterWatcher watcher5(m_InputImageList,"VectorImageToImageList");
  // StandardFilterWatcher watcher6(m_ChangeLabelFilter,"ChangeLabelFilter");
  // StandardFilterWatcher watcher7(m_Classifier,"Classifier");

  m_Writer->Update();
  m_OutputImage = m_ChangeLabelFilter->GetOutput();
  m_Writer = WriterType::New();

  m_HasOutput = true;
  this->NotifyAll();
  m_HasOutput = false;
}

}
