/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbCorinLandCoveModel_h
#define __otbCorinLandCoveModel_h

#include "otbMVCModel.h"
#include "otbListenerBase.h"
#include "otbSVMImageClassificationFilter.h"

#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageFileReader.h"
#include "otbObjectList.h"
#include "itkImageSource.h"
#include "otbImageList.h"
#include "otbImageListToVectorImageFilter.h"
#include "otbStreamingImageFileWriter.h"
#include "otbVectorImageToImageListFilter.h"
#include "otbImageToVectorImageCastFilter.h"
#include "otbPerBandVectorImageFilter.h"
#include "itkIntensityWindowingImageFilter.h"
#include "otbBlackImage.h"
#include "otbMinimumMaximumVectorImageFilter.h"
#include "otbChangeLabelImageFilter.h"

//Filter include
#include "otbMeanShiftVectorImageFilter.h"
#include "otbVegetationIndicesFunctor.h"
#include "otbMultiChannelRAndNIRIndexImageFilter.h"
#include "otbMultiChannelRAndBAndNIRIndexImageFilter.h"
#include "otbStreamingShrinkImageFilter.h"

#include "itkListSample.h"
#include "itkListSampleToHistogramGenerator.h"

namespace otb
{
/** \class LandCoverMapModel
 *
 * This class modelize a process to compute a descriptor image and compute a SVM classification.
 *
 */


class ITK_EXPORT LandCoverMapModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef LandCoverMapModel                         Self;
  typedef MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                        Pointer;
  typedef itk::SmartPointer<const Self>                  ConstPointer;

  /** Standard type macro */
  itkTypeMacro(LandCoverMapModel, MVCModel);

  /** Images typedefs */
  typedef double                                 PixelType;
  /** Inpout Image Type*/
  typedef VectorImage<PixelType, 2>               InputImageType;
  typedef Image<PixelType, 2>                     ScalarInputImageType;
  typedef InputImageType::Pointer                InputImagePointerType;
  typedef InputImageType::IndexType              IndexType;
  typedef InputImageType::SizeType               SizeType;
  typedef InputImageType::PixelType              InputPixelType;
  /** Vector Image Output Filter */
  typedef VectorImage<PixelType, 2>               OutputImageType;
  typedef Image<PixelType, 2>                     ScalarOutputImageType;
  typedef OutputImageType::PixelType             OutputImagePixelType;
  typedef OutputImageType::InternalPixelType     OutputImageInternalPixelType;
  typedef OutputImageType::Pointer               OutputImagePointerType;
  /** Image Type */
  typedef Image<PixelType, 2>                     SingleImageType;
  typedef SingleImageType::Pointer               SingleImagePointerType;
  typedef SingleImageType::PixelType             SinglePixelType;
  typedef SingleImageType::InternalPixelType     SingleInternalPixelType;
  typedef SingleImageType::RegionType            SingleRegionType;
  typedef SingleImageType::IndexType             SingleIndexType;
  typedef SingleImageType::RegionType            SingleSpacingType;
  typedef SingleImageType::SizeType              SingleSizeType;
  typedef StreamingShrinkImageFilter<InputImageType, InputImageType>                         ShrinkFilterType;
  typedef ShrinkFilterType::Pointer                                                       ShrinkFilterPointerType;
  typedef itk::Vector<itk::NumericTraits<PixelType>::RealType, 1>                          MeasurementVectorType;
  typedef itk::Statistics::ListSample<MeasurementVectorType>                              ListSampleType;
  typedef float                                                                           HistogramMeasurementType;
  typedef itk::Statistics::ListSampleToHistogramGenerator<
  ListSampleType,
  HistogramMeasurementType,
  itk::Statistics::DenseFrequencyContainer, 1>                                           HistogramGeneratorType;
  typedef HistogramGeneratorType::HistogramType                                           HistogramType;
  typedef HistogramType::Pointer                                                          HistogramPointerType;

  /** Classifier */
  typedef unsigned char                                                        LabeledPixelType;
  typedef Image<LabeledPixelType, 2>                                            LabeledImageType;
  typedef VectorImage<LabeledPixelType, 2>                                      LabeledVectorImageType;
  typedef SVMImageClassificationFilter<OutputImageType, LabeledImageType>       ClassificationFilterType;
  typedef ClassificationFilterType::ModelType                                  ModelType;
  typedef ChangeLabelImageFilter<LabeledImageType, LabeledVectorImageType>      ChangeLabelFilterType;
  typedef LabeledVectorImageType::PixelType                                    LabelVectorType;
  typedef std::map<LabeledPixelType, LabelVectorType>                          ColorMapType;

  typedef std::vector<int>                                                     FilterTypeListType;
  typedef std::vector<std::string>                                             OutputFilterInformationType;

  /** Concatenation typedef */
  typedef ImageList< SingleImageType >                                         ImageListType;
  typedef ImageListToVectorImageFilter< ImageListType, OutputImageType >       ImageListToVectorImageFilterType;
  typedef StreamingImageFileWriter<LabeledVectorImageType/*LabeledImageType*//*OutputImageType*/> WriterType;
  typedef StreamingImageFileWriter<OutputImageType> FeatureWriterType;

  typedef ImageFileReader<InputImageType>                                      ReaderType;
  typedef ReaderType::Pointer                                                  ReaderPointerType;
  typedef VectorImageToImageListFilter<OutputImageType, ImageListType>         VectorToImageListType;
  typedef ImageToVectorImageCastFilter<SingleImageType, OutputImageType>       CasterType;

  typedef BlackImage<SingleImageType>                                          BlackImageType;

  typedef itk::IntensityWindowingImageFilter<ScalarOutputImageType, ScalarOutputImageType>
  RescaleFilterType;

  typedef PerBandVectorImageFilter<OutputImageType,
  OutputImageType, RescaleFilterType>                                 VectorRescaleFilterType;


  /***************************/
  /** Filter type declaration*/
  /***************************/
  // Pre treatment
  typedef MeanShiftVectorImageFilter<InputImageType, InputImageType>                                        MeanShiftFilterType;

  // NDVI and NDVI RG
  typedef Functor::NDVI< PixelType, PixelType, PixelType >                                                  NDVIFunctorType;
  typedef MultiChannelRAndNIRIndexImageFilter<InputImageType, SingleImageType, NDVIFunctorType>               NDVIFilterType;
  // ARVI
  typedef Functor::ARVI< PixelType, PixelType, PixelType, PixelType >                                       ARVIFunctorType;
  typedef MultiChannelRAndBAndNIRIndexImageFilter<InputImageType, SingleImageType, ARVIFunctorType>           ARVIFilterType;

  /** Get the unique instanc1e of the model */
  static Pointer GetInstance();

  /** Input Image Pointer */
  void LoadImage(InputImagePointerType image);
  InputImagePointerType GetInputImage()
  {
    return m_InputImage;
  };

  /** Get the output image */
  itkGetObjectMacro(OutputImage, LabeledVectorImageType);

  /** Get the output writer */
  itkGetObjectMacro(ChangeLabelFilter, ChangeLabelFilterType);
  itkGetObjectMacro(InputImageList, VectorToImageListType);
  itkGetObjectMacro(Writer, WriterType);

  /** Feature image filename */
  itkSetMacro(FeatureFileName, std:: string);
  itkGetMacro(FeatureFileName, std:: string);

  /** Output filename */
  itkSetMacro(OutputFileName, std:: string);
  itkGetMacro(OutputFileName, std:: string);

  /** Input filename */
  itkSetMacro(InputFileName, std:: string);
  itkGetMacro(InputFileName, std:: string);

  /** Model Filename */
  itkSetMacro(ModelFileName, std:: string);
  itkGetMacro(ModelFileName, std:: string);

  /** Is New Image */
  /*itkSetMacro(IsNewImage, bool);
  itkGetMacro(IsNewImage, bool); */

  /** Shrinker accessor*/
  itkGetObjectMacro(Shrinker, ShrinkFilterType);

  /** Shrink factor */
  itkGetMacro(ShrinkFactor, int);

  /** Selected output filters. */
  void SetSelectedFilters( std::vector<bool> vect)
  {
    m_SelectedFilters = vect;
    this->Modified();
  };

  std::vector<bool> GetSelectedFilters()
  {
    return m_SelectedFilters;
  };

  void ChangeSelectedFilterStatus(unsigned int i)
  {
    if (i>m_SelectedFilters.size())
      return;
    m_SelectedFilters[i] = !m_SelectedFilters[i];
    this->Modified();
  };

  itkGetMacro(HasInput, bool);
  itkGetMacro(HasOutput, bool);

  itkSetMacro( NumberOfComponent, unsigned int);
  itkGetMacro( NumberOfComponent, unsigned int);

  void SetQuicklookSize( int x, int y)
  {
    m_QuicklookSize[0] = x;
    m_QuicklookSize[1] = y;
    this->Modified();
  }

  itkGetObjectMacro(Rescaler, RescaleFilterType);
  ImageListToVectorImageFilterType::Pointer GetIL2VI()
  {
    return m_IL2VI;
  };

  /** Open Input Image */
  void SetInputImage(std::string strfilename);
  /** Load Model */
  void LoadModel(const char * modelFilename);
  /** Save the output */
  void SaveResult();
  /** Save the computed feature image */
  void SaveFeatureImage();
  /** Wire the whole pipeline */
  void WirePipeline();
  /** If a descriptor was not selected, replace its ouput with a black image */
  void CheckSelected();
  /** Shrink input image*/
  void ShrinkInputImage();
  /** Get the mins and the maxs */
  PixelType GetMin(int ch) const;
  PixelType GetMax(int ch) const;
protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  LandCoverMapModel();
  /** Destructor */
  virtual ~LandCoverMapModel();

  /** Filter Wiring */
  void WireNDVI();
  void WireNDVIGR();
  void WireARVI();
  void InitColor();

private:
  LandCoverMapModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  void SetupProcessing();

  /** Imput Image Pointer */
  InputImagePointerType                     m_InputImage;
  /** Input Image List*/
  VectorToImageListType::Pointer            m_InputImageList;
  /** Output Image */
  LabeledVectorImageType::Pointer           m_OutputImage;
  /** Black Image*/
  BlackImageType::Pointer                   m_BlackGenerator;
  /** The instance singleton */
  static Pointer                            Instance;
  /** Output filename*/
  std:: string                              m_OutputFileName;
  /** Feature image filename*/
  std:: string                              m_FeatureFileName;
  /** Model filename */
  std::string                               m_ModelFileName;
  /** Input filename*/
  std:: string                              m_InputFileName;
  /** Output filters selected to generate the output image. */
  std::vector<bool>                         m_SelectedFilters;
  /** Reader pointer*/
  ReaderPointerType                         m_Reader;
  /** bool to know if an input is selected*/
  bool                                      m_HasInput;
  /** bool to know if an output has been computed*/
  bool                                      m_HasOutput;
  /** bool to know if a new image has been opened*/
  bool                                      m_IsNewImage;
  /** number of bands (4)*/
  unsigned int                              m_NumberOfComponent;
  /** SVM classifier*/
  ClassificationFilterType::Pointer         m_Classifier;
  /** Model ponter for SVM classif*/
  ModelType::Pointer                        m_Model;

  // Filter List
  /** Mean Shift filter*/
  MeanShiftFilterType::Pointer              m_MeanShift;
  /** Mean Shift Image List*/
  VectorToImageListType::Pointer            m_MeanShiftImageList;
  /** NDVI filter*/
  NDVIFilterType::Pointer                   m_NDVI;
  /** ARVI filter*/
  ARVIFilterType::Pointer                   m_ARVI;
  /** Water index*/
  NDVIFilterType::Pointer                   m_NDVIGR;

  /** List of output image filter*/
  ImageListType::Pointer                    m_OutputList;
  /** Filter transforming outputlist to vector list*/
  ImageListToVectorImageFilterType::Pointer m_IL2VI;
  /** Rescale input image */
  RescaleFilterType::Pointer                m_InputRescaler;
  VectorRescaleFilterType::Pointer          m_InputVectorRescaler;
  /** Rescaler between -1 and 1 */
  RescaleFilterType::Pointer                m_Rescaler;
  VectorRescaleFilterType::Pointer          m_VectorRescaler;
  /** Stores the index of each starting output filter in output list*/
  std::vector<unsigned int>                 m_Index;
  /** Color the SVM output image*/
  ChangeLabelFilterType::Pointer            m_ChangeLabelFilter;
  /** Colormap used to color the SVM output image*/
  ColorMapType                              m_ColorMap;
  /** Maximum valuye used to normaiiezd the descriptor image*/
  OutputImagePixelType                      m_Maximum;
  /** Minimum valuye used to normaiiezd the descriptor image*/
  OutputImagePixelType                      m_Minimum;
  WriterType::Pointer                       m_Writer;
  ShrinkFilterPointerType                   m_Shrinker;
  SizeType                                  m_QuicklookSize;
  int                                       m_ShrinkFactor;
  /** Histogram list */
  std::vector<HistogramPointerType> m_Histogram;
  /** percentage used for normalization */
  double m_ClampThreshold;
};
}
#endif
