/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#ifndef _otbMinimumMaximumVectorImageFilter_txx
#define _otbMinimumMaximumVectorImageFilter_txx

#include "otbMinimumMaximumVectorImageFilter.h"

#include "itkImageRegionIterator.h"
#include "itkImageRegionConstIterator.h"
#include "itkNumericTraits.h"
#include "itkProgressReporter.h"

#include <vector>

namespace otb
{

template<class TInputImage>
MinimumMaximumVectorImageFilter<TInputImage>
::MinimumMaximumVectorImageFilter()
{
  this->SetNumberOfRequiredOutputs(3);
  for (int i=1; i < 3; ++i)
  {
    typename PixelObjectType::Pointer output = static_cast<PixelObjectType*>(this->MakeOutput(i).GetPointer());
    this->itk::ProcessObject::SetNthOutput(i, output.GetPointer());
  }
  //this->GetMinimumOutput()->Set( NumericTraits<PixelType>::max() );
  //this->GetMaximumOutput()->Set( NumericTraits<PixelType>::NonpositiveMin() );
}


template<class TInputImage>
itk::DataObject::Pointer
MinimumMaximumVectorImageFilter<TInputImage>
::MakeOutput(unsigned int output)
{
  switch (output)
  {
  case 0:
    return static_cast<itk::DataObject*>(TInputImage::New().GetPointer());
    break;
  case 1:
  case 2:
    return static_cast<itk::DataObject*>(PixelObjectType::New().GetPointer());
    break;
  default:
    // might as well make an image
    return static_cast<itk::DataObject*>(TInputImage::New().GetPointer());
    break;
  }
}


template<class TInputImage>
typename MinimumMaximumVectorImageFilter<TInputImage>::PixelObjectType*
MinimumMaximumVectorImageFilter<TInputImage>
::GetMinimumOutput()
{
  return static_cast<PixelObjectType*>(this->itk::ProcessObject::GetOutput(1));
}

template<class TInputImage>
const typename MinimumMaximumVectorImageFilter<TInputImage>::PixelObjectType*
MinimumMaximumVectorImageFilter<TInputImage>
::GetMinimumOutput() const
{
  return static_cast<const PixelObjectType*>(this->itk::ProcessObject::GetOutput(1));
}


template<class TInputImage>
typename MinimumMaximumVectorImageFilter<TInputImage>::PixelObjectType*
MinimumMaximumVectorImageFilter<TInputImage>
::GetMaximumOutput()
{
  return static_cast<PixelObjectType*>(this->itk::ProcessObject::GetOutput(2));
}

template<class TInputImage>
const typename MinimumMaximumVectorImageFilter<TInputImage>::PixelObjectType*
MinimumMaximumVectorImageFilter<TInputImage>
::GetMaximumOutput() const
{
  return static_cast<const PixelObjectType*>(this->itk::ProcessObject::GetOutput(2));
}

template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::GenerateInputRequestedRegion()
{
  Superclass::GenerateInputRequestedRegion();
  if ( this->GetInput() )
  {
    InputImagePointer image =
      const_cast< typename Superclass::InputImageType * >( this->GetInput() );
    image->SetRequestedRegionToLargestPossibleRegion();
  }
}

template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::EnlargeOutputRequestedRegion(itk::DataObject *data)
{
  Superclass::EnlargeOutputRequestedRegion(data);
  data->SetRequestedRegionToLargestPossibleRegion();
}


template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::AllocateOutputs()
{
  // Pass the input through as the output
  InputImagePointer image = const_cast< TInputImage * >( this->GetInput() );
  this->GraftOutput( image );
  // Nothing that needs to be allocated for the remaining outputs
}

template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::BeforeThreadedGenerateData()
{
  int numberOfThreads = this->GetNumberOfThreads();
  InputImagePointer inputPtr = const_cast< InputImageType * >( this->GetInput() );
  unsigned int nbChannels = inputPtr->GetNumberOfComponentsPerPixel();

  // Create the thread temporaries
  PixelType threadMin;
  threadMin.SetSize(nbChannels);
  threadMin.Fill( itk::NumericTraits<InternalPixelType>::max() );
  m_ThreadMin = VectorPixelType(numberOfThreads, threadMin );
  PixelType threadMax;
  threadMax.SetSize(nbChannels);
  threadMax.Fill( itk::NumericTraits<InternalPixelType>::NonpositiveMin() );
  m_ThreadMax = VectorPixelType(numberOfThreads, threadMax );

  PixelType pix;
  pix.SetSize( nbChannels );
  pix.Fill(itk::NumericTraits<InternalPixelType>::max());
  this->GetMinimumOutput()->Set( threadMax );
  pix.Fill(itk::NumericTraits<InternalPixelType>::NonpositiveMin());
  this->GetMaximumOutput()->Set( threadMin );
}

template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::AfterThreadedGenerateData()
{
  int i;
  int numberOfThreads = this->GetNumberOfThreads();
  InputImagePointer inputPtr = const_cast< InputImageType * >( this->GetInput() );
  unsigned int nbChannels = inputPtr->GetNumberOfComponentsPerPixel();

  PixelType minimum = m_ThreadMin[0];
  PixelType maximum = m_ThreadMax[0];

  for ( i = 1; i < numberOfThreads; i++)
  {
    for ( unsigned int j = 0; j < nbChannels; j++)
    {
      if (m_ThreadMin[i][j] < minimum[j])
      {
        minimum[j] = m_ThreadMin[i][j];
      }
      if (m_ThreadMax[i][j] > maximum[j])
      {
        maximum[j] = m_ThreadMax[i][j];
      }
    }
  }

  // Set the outputs
  this->GetMinimumOutput()->Set( minimum );
  this->GetMaximumOutput()->Set( maximum );
}

template<class TInputImage>
void
MinimumMaximumVectorImageFilter<TInputImage>
::ThreadedGenerateData(const RegionType& outputRegionForThread, int threadId)
{
  InputImagePointer inputPtr = const_cast< InputImageType * >( this->GetInput() );

  unsigned int nbChannels = inputPtr->GetNumberOfComponentsPerPixel();

  PixelType value;
  itk::ImageRegionConstIterator<TInputImage> it (inputPtr, outputRegionForThread);

  // support progress methods/callbacks
  itk::ProgressReporter progress(this, threadId, outputRegionForThread.GetNumberOfPixels());

  // do the work
  while (!it.IsAtEnd())
  {
    value = static_cast<PixelType>(it.Get());
    for ( unsigned int j = 0; j < nbChannels; j++)
    {
      if (value[j] < m_ThreadMin[threadId][j])
      {
        m_ThreadMin[threadId][j] = value[j];
      }
      if (value[j] > m_ThreadMax[threadId][j])
      {
        m_ThreadMax[threadId][j] = value[j];
      }
    }
    ++it;
    progress.CompletedPixel();
  }
}

template <class TImage>
void
MinimumMaximumVectorImageFilter<TImage>
::PrintSelf(std::ostream& os, itk::Indent indent) const
{
  Superclass::PrintSelf(os, indent);

  os << indent << "Minimum: "<< std::endl;
  for (unsigned int i = 0; i<this->GetMinimum().Size(); i++)
  {
    os << indent  << "  " << static_cast<typename itk::NumericTraits<InternalPixelType>::PrintType>(this->GetMinimum()[i]) << std::endl;
  }
  os << indent << "Maximum: "<< std::endl;
  for (unsigned int i = 0; i<this->GetMaximum().Size(); i++)
  {
    os << indent  << "  " << static_cast<typename itk::NumericTraits<InternalPixelType>::PrintType>(this->GetMaximum()[i]) << std::endl;
  }
}


}// end namespace itk
#endif
