/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbLandCoverMapViewGUI.h"
#include <FL/Fl_File_Chooser.H>
#include <FL/fl_ask.H>
#include <FL/Enumerations.H>
#include <FL/Fl_Text_Buffer.H>
#include "otbMultiToMonoChannelExtractROI.h"
#include "otbMacro.h"
#include "itkExceptionObject.h"
#include "itkVectorCastImageFilter.h"
#include "otbFltkWriterWatcher.h"

#include "otbConfigureAppli.h"

namespace otb
{

LandCoverMapViewGUI
::LandCoverMapViewGUI()
{
  // Model pointer and listener registration
  m_LandCoverMapModel = LandCoverMapModel::GetInstance();
  m_LandCoverMapModel->RegisterListener(this);
  // Initialisation
  m_FileName = "";
  m_ViewerName =  "";
  m_DisplayedLabel = "+ ";
  m_UndisplayedLabel = "+ ";
  m_DisplayStatus = true;

  this->CreateGUI();

  this->guiSelection->add("NDVI");
  this->guiSelection->add("Water Index");
  this->guiSelection->add("ARVI");

  guiSelection->check_all();
  guiSelection->redraw();

  this->guiImageName->value("");

  // CHANGE HERE FOR A NEW DEFAULT PATH
  std::string filename(OTB_Applications_SOURCE_DIR);
  filename += "/LandCoverMap/Data/LinearModel1.svm";
  if (itksys::SystemTools::FileExists(filename.c_str()) )
  {
    m_LastPath = filename;
    guiModelName->value(filename.c_str());
    guiModelName->activate();
  }
}


LandCoverMapViewGUI
::~LandCoverMapViewGUI()
{
}

void
LandCoverMapViewGUI
::RefreshViewer()
{
  if ( !m_Viewer.IsNull() && (m_LandCoverMapModel->GetHasInput() ||  m_LandCoverMapModel->GetHasOutput()) )
  {
    if (m_Viewer->GetUseScroll())
      gScroll->remove(m_Viewer->GetScrollWidget());
    gFull->remove(m_Viewer->GetFullWidget());
  }

  if (m_Viewer.IsNull() || m_LandCoverMapModel->GetHasInput() ||  m_LandCoverMapModel->GetHasOutput())
  {
    m_Viewer = ViewerType::New();
    m_Viewer->SetImage(m_LandCoverMapModel->GetInputImage());
    if (m_LandCoverMapModel->GetInputImage()->GetNumberOfComponentsPerPixel()>3)
    {
      m_Viewer->SetRedChannelIndex(2);
      m_Viewer->SetGreenChannelIndex(1);
      m_Viewer->SetBlueChannelIndex(0);
    }
    unsigned int maxfullsize = std::min(gFull->w(), gFull->h());
    unsigned int maxscrollsize = std::min(gScroll->w(), gScroll->h());
    m_Viewer->SetFullMaxInitialSize(maxfullsize);
    m_Viewer->SetScrollMaxInitialSize(maxscrollsize);
    m_Viewer->SetScrollLimitSize(maxscrollsize);
    m_Viewer->SetShowZoomWidget(false);
    m_Viewer->SetShowHistograms(false);
    if ( m_LandCoverMapModel->GetHasOutput() )
    {
      m_Viewer->SetUseImageOverlay( true );
      m_Viewer->SetImageOverlay(m_LandCoverMapModel->GetOutputImage());
    }
    m_Viewer->Build();
    m_Viewer->GetFullWidget()->SetEventsInterface(m_LandCoverMapController);
    m_Viewer->GetFullWidget()->position(gFull->x(), gFull->y());
    m_Viewer->GetFullWidget()->resize(gFull->x(), gFull->y(), gFull->w(), gFull->h());
    gFull->add(m_Viewer->GetFullWidget());
    gFull->box(FL_NO_BOX);
    if (m_Viewer->GetUseScroll())
    {
      gScroll->show();
      m_Viewer->GetScrollWidget()->position(gScroll->x(), gScroll->y());
      m_Viewer->GetScrollWidget()->resize(gScroll->x(), gScroll->y(), gScroll->w(), gScroll->h());
      gScroll->add(m_Viewer->GetScrollWidget());
      gScroll->box(FL_NO_BOX);
    }
    else
    {
      gScroll->hide();
    }

  }

  if ( m_LandCoverMapModel->GetHasOutput() )
  {
    m_Viewer->SetImageOverlayOpacity( static_cast<unsigned char>( guiOpacity->value() ));
  }

  m_Viewer->Show();
  m_Viewer->Update();

}


void
LandCoverMapViewGUI
::Notify()
{
  if (m_LandCoverMapModel->GetHasInput())
  {
    guiImageName->value( m_LandCoverMapModel->GetInputFileName().c_str() );
    this->guiOpacity->value(0);
    this->guiOpacity->deactivate();
  }
  else
  {
    if ( m_LandCoverMapModel->GetHasOutput() )// output computed
    {
      this->guiOpacity->value(120);
      this->guiOpacity->activate();
    }
  }
  this->RefreshViewer();
}


void
LandCoverMapViewGUI
::OpenImage()
{
  const char * cfname = fl_file_chooser("Pick an image file", "*.*", m_LastPath.c_str());
  Fl::check();
  guiMainWindow->redraw();
  if (cfname == NULL || strlen(cfname)<1)
  {
    return;
  }

  m_LastPath = itksys::SystemTools::GetFilenamePath(cfname);

  this->guiOpacity->deactivate();
  m_LandCoverMapController->OpenInputImage(cfname);
}


void
LandCoverMapViewGUI
::Show()
{
  guiMainWindow->show();
}


void
LandCoverMapViewGUI
::OK()
{
  const char * cfname = fl_file_chooser("Save as...", "*.*", m_LastPath.c_str());

  if (cfname == NULL || strlen(cfname)<1)
  {
    return;
  }
  m_LandCoverMapController->SetOutputFileName(cfname);

  try
  {
    m_LandCoverMapController->LoadModel( guiModelName->value() );
  }
  catch ( itk::ExceptionObject & err )
  {
    itk::OStringStream oss;
    oss<<"ERROR LOADING MODEL: "<<std::endl;
    oss<<guiModelName->value()<<std::endl;
    oss<<"Following Exception caught : ";
    oss<<err;
    const char* msg = oss.str().c_str();
    fl_alert("%s", msg);
    return;
  }
  catch ( ... )
  {
    itk::OStringStream oss;
    oss<<"ERROR LOADING MODEL: "<<std::endl<<std::endl;
    oss<<guiModelName->value();
    const char* msg = oss.str().c_str();
    fl_alert("%s", msg);
    return;
  }


  try
  {
    typedef FltkWriterWatcher WatcherType;
    WatcherType watcher(m_LandCoverMapModel->GetWriter(), m_LandCoverMapModel->GetInputImageList(), 100, 100, 200, 40, "Generating Output...");
    m_LandCoverMapController->SaveOutput();
  }
  catch ( itk::ExceptionObject & err )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Following Exception caught : ";
    oss<<err;
    const char* msg = oss.str().c_str();
    fl_alert("%s", msg);
    return;
  }
  catch ( std::bad_alloc & err )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Bad alloc, following exception caught : ";
    oss<<(char*)err.what();
    const char* msg = oss.str().c_str();
    fl_alert("%s", msg);
    return;
  }
  catch ( ... )
  {
    itk::OStringStream oss;
    oss<<"ERROR FOUND"<<std::endl<<std::endl;
    oss<<"Unknown Exception found !";
    const char* msg = oss.str().c_str();
    fl_alert("%s", msg);
    return;
  }
}

void
LandCoverMapViewGUI
::ChangeOpacity()
{
  m_Viewer->SetImageOverlayOpacity( static_cast<unsigned char>( guiOpacity->value() ));
  m_Viewer->Update();
  gScroll->redraw();
  gFull->redraw();
}


void
LandCoverMapViewGUI
::LoadModel()
{
  const char * cfname = fl_file_chooser("Pick a SVM Model", "*.*", m_LastPath.c_str());
  Fl::check();
  guiMainWindow->redraw();
  if (cfname == NULL || strlen(cfname)<1)
  {
    return;
  }

  m_LastPath = itksys::SystemTools::GetFilenamePath(cfname);
  guiModelName->value(cfname);
  guiModelName->activate();
}


void
LandCoverMapViewGUI
::Quit()
{
  guiMainWindow->hide();
}


}
