/*=========================================================================

  $Author: allain $
  $Date: 2005/06/28 11:28:58 $
  $Revision: 1.5 $
  $Id: GenericRegistration.h, v 1.5 2005/06/28 11:28:58 allain Exp $
  $Source: /usr/local/stok/CVS_REPO/Pireo/Code/Registration/NewClassNames/GenericRegistration.h, v $
  $Log: GenericRegistration.h, v $
  Revision 1.5  2005/06/28 11:28:58  allain
  Update

  Revision 1.4  2005/06/10 15:56:42  allain
  Add some functionalities for registration

  Revision 1.3  2005/06/06 16:16:56  allain
  Gave new method names, add commentary and implement new params

  Revision 1.2  2005/05/09 15:52:57  allain
  Add pause and stop buttons for generic registration

  Revision 1.1  2005/05/04 09:16:09  allain
  Add generic registration class


=========================================================================*/

#ifndef __GenericRegistration_H__
#define __GenericRegistration_H__

/* ITK Librairies */
#include "otbImage.h"
#include "itkMultiResolutionImageRegistrationMethod.h"
#include "itkMultiResolutionPyramidImageFilter.h"
#include "itkTransform.h"
#include "itkImageToImageMetric.h"
#include "itkInterpolateImageFunction.h"
#include "itkSingleValuedNonLinearOptimizer.h"
#include "itkResampleImageFilter.h"
#include "otbImageFileWriter.h"
#include "itkRescaleIntensityImageFilter.h"


/* Class GenericRegistration
   Do the images registration.
   Cannot be used by itself: indeed, only generic registation
   parameters are declared in this class. So it is necessary to set up
   their value, in particular for:
   - the metric: SetMetric(...) method
   - the transform: SetTransform(...) method
   - the interpolator: SetInterpolator(...) method
   - the optimizer: SetOptimizer(...) method
   Anyway, the registration process is set to a multi-resolution
   registration.
   We also need to set the two images to register with SetFixedImage()
   and SetMovingImage() methods.
*/
class GenericRegistration
{
public:
  /** Registration image pixel type (can only be double or float).*/
  typedef double PixelType;
  /** Registration image type.*/
  typedef otb::Image<PixelType, 2> ImageType2D;

  /** Generic registration transform.*/
  typedef itk::Transform<double, 2, 2>       TransformType;

  /** Generic registration Metric.*/
  typedef itk::ImageToImageMetric<ImageType2D, ImageType2D>      MetricType;

  /** Generic registration Optimizer.*/
  typedef itk::SingleValuedNonLinearOptimizer         OptimizerType;

  /** Generic registration Interpolator.*/
  typedef itk::InterpolateImageFunction<ImageType2D, double>     InterpolatorType;

  /** Multi-resolution registration.*/
  typedef itk::MultiResolutionImageRegistrationMethod< ImageType2D, ImageType2D >    RegistrationType;
  /** Filter that generates image pyramids for multi-resolution registration.*/
  typedef itk::MultiResolutionPyramidImageFilter< ImageType2D, ImageType2D > FixedImagePyramidType;
  /** Filter that generates image pyramids for multi-resolution registration.*/
  typedef itk::MultiResolutionPyramidImageFilter< ImageType2D, ImageType2D > MovingImagePyramidType;
  /** Generic registration transform parameters.*/
  typedef RegistrationType::ParametersType ParametersType;

  /** Resampling filter (used to generate deformed image after
  registration process).*/
  typedef itk::ResampleImageFilter< ImageType2D, ImageType2D >    ResampleFilterType;

  /** Constructor.*/
  GenericRegistration();
  /** Destructor.*/
  ~GenericRegistration();

  /** Set fixed image tp register.*/
  void SetFixedImage(ImageType2D::Pointer m_Im);
  /** Set moving image to register.*/
  void SetMovingImage(ImageType2D::Pointer m_Im);
  /** Register (main method that launch the registration process).*/
  ImageType2D::Pointer DoRegister();
  /** Return fixed image to register.*/
  ImageType2D::Pointer GetFixedImage();
  /** Return moving image to register.*/
  ImageType2D::Pointer GetMovingImage();
  /** Return registration metric.*/
  MetricType* GetMetric()
  {
    return metric;
  }
  /** Return registration interpolator.*/
  InterpolatorType* GetInterpolator()
  {
    return interpolator;
  }
  /** Return registration transform.*/
  TransformType* GetTransform()
  {
    return transform;
  }
  /** Return registration optimizer.*/
  OptimizerType* GetOptimizer()
  {
    return optimizer;
  }
  /** Return registration final transform parameters.*/
  TransformType::Pointer GetFinalTransform();
  /** Return deformed image after registration.*/
  ImageType2D::Pointer GetRegistrationResult();
  /** Return parameters list */
  ParametersType GetParams();

  /** Set registration transform.*/
  void SetTransform(TransformType::Pointer T);
  /** Set registration metric.*/
  void SetMetric(MetricType::Pointer M);
  /** Set registration optimizer.*/
  void SetOptimizer(OptimizerType::Pointer O);
  /** Set registration interpolator.*/
  void SetInterpolator(InterpolatorType::Pointer I);
  /** Set number of levels of the multi-resolution registration.*/
  void SetNumberOfLevels(int l);
  /** Set registration transform parameters (used to track iteratively
  the transform parameters).*/
  void SetRegistrationTransformParameters(ParametersType t);

  /** Write resample filter output*/
  void WriteResult(ResampleFilterType::Pointer r);

protected:
  /** Fixed image.*/
  ImageType2D::Pointer m_FixedIm;
  /** Moving image.*/
  ImageType2D::Pointer m_MovingIm;
  /** Multi-resolution registration number of levels.*/
  int m_NumberOfLevels;

  /** Resample image filter.*/
  //ResampleFilterType::Pointer resample;
  /** Multi-resolution registration.*/

  RegistrationType::Pointer   registration;
  /** Registration interpolator.*/
  InterpolatorType::Pointer   interpolator;
  /** Registration metric.*/
  MetricType::Pointer         metric;
  /** Registration optimizer.*/
  OptimizerType::Pointer      optimizer;
  /** Registration transform.*/
  TransformType::Pointer      transform;
  /** Fixed pyramidal image.*/
  FixedImagePyramidType::Pointer  fixedPyramid;
  /** Moving pyramidal image.*/
  MovingImagePyramidType::Pointer movingPyramid;

  /** Final transform (get at the end of the registration process).*/
  TransformType::Pointer finalTransform;
  /** Initial registration transform parameters (used to reload
  initial transform parameters for each new registration).*/
  ParametersType initialParams;
  /** Useful to set the registration transform parameters from
  outside.*/
  ParametersType Params;


  int g;
};

#endif
