/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: itkLBFGSModifiedOptimizer.cxx, v $
  Language:  C++
  Date:      $Date: 2005/06/24 15:40:09 $
  Version:   $Revision: 1.1 $

  Copyright (c) Insight Software Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#ifndef _itkLBFGSModifiedOptimizer_txx
#define _itkLBFGSModifiedOptimizer_txx

#include "itkLBFGSModifiedOptimizer.h"

namespace itk
{

/**
 * Constructor
 */
LBFGSModifiedOptimizer
::LBFGSModifiedOptimizer()
{
  m_OptimizerInitialized    = false;
  m_VnlOptimizer            = 0;
  m_Trace                              = false;
  m_MaximumNumberOfFunctionEvaluations = 2000;
  m_GradientConvergenceTolerance       = 1e-5;
  m_LineSearchAccuracy                 = 0.9;
  m_DefaultStepLength                  = 1.0;
}


/**
 * Destructor
 */
LBFGSModifiedOptimizer
::~LBFGSModifiedOptimizer()
{
  delete m_VnlOptimizer;
}

/**
 * PrintSelf
 */
void
LBFGSModifiedOptimizer
::PrintSelf(std::ostream& os, Indent indent) const
{
  Superclass::PrintSelf( os, indent );
  os << indent << "Trace: ";
  if ( m_Trace )
  {
    os << "On";
  }
  else
  {
    os << "Off";
  }
  os << std::endl;
  os << indent << "MaximumNumberOfFunctionEvaluations: ";
  os << m_MaximumNumberOfFunctionEvaluations << std::endl;
  os << indent << "GradientConvergenceTolerance: ";
  os << m_GradientConvergenceTolerance << std::endl;
  os << indent << "LineSearchAccuracy: ";
  os << m_LineSearchAccuracy << std::endl;
  os << indent << "DefaultStepLength: ";
  os << m_DefaultStepLength << std::endl;

}

/**
 * Set the optimizer trace flag
 */
void
LBFGSModifiedOptimizer
::SetTrace( bool flag )
{
  if ( flag == m_Trace )
  {
    return;
  }

  m_Trace = flag;
  if ( m_OptimizerInitialized )
  {
    m_VnlOptimizer->set_trace( m_Trace );
  }

  this->Modified();
}

/**
 * Set the maximum number of function evalutions
 */
void
LBFGSModifiedOptimizer
::SetMaximumNumberOfFunctionEvaluations( unsigned int n )
{
  if ( n == m_MaximumNumberOfFunctionEvaluations )
  {
    return;
  }

  m_MaximumNumberOfFunctionEvaluations = n;
  if ( m_OptimizerInitialized )
  {
    m_VnlOptimizer->set_max_function_evals(
      static_cast<int>( m_MaximumNumberOfFunctionEvaluations ) );
  }

  this->Modified();
}

/**
 * Set the gradient convergence tolerance
 */
void
LBFGSModifiedOptimizer
::SetGradientConvergenceTolerance( double f )
{
  if ( f == m_GradientConvergenceTolerance )
  {
    return;
  }

  m_GradientConvergenceTolerance = f;
  if ( m_OptimizerInitialized )
  {
    m_VnlOptimizer->set_g_tolerance( m_GradientConvergenceTolerance );
  }

  this->Modified();
}

/**
 * Set the line search accuracy
 */
void
LBFGSModifiedOptimizer
::SetLineSearchAccuracy( double f )
{
  if ( f == m_LineSearchAccuracy )
  {
    return;
  }

  m_LineSearchAccuracy = f;
  if ( m_OptimizerInitialized )
  {
    m_VnlOptimizer->line_search_accuracy = m_LineSearchAccuracy;
  }

  this->Modified();
}

/**
 * Set the default step length
 */
void
LBFGSModifiedOptimizer
::SetDefaultStepLength( double f )
{
  if ( f == m_DefaultStepLength )
  {
    return;
  }

  m_DefaultStepLength = f;
  if ( m_OptimizerInitialized )
  {
    m_VnlOptimizer->default_step_length = m_DefaultStepLength;
  }

  this->Modified();
}

/** Return Current Value */
LBFGSModifiedOptimizer::MeasureType
LBFGSModifiedOptimizer
::GetValue()
{
  return this->GetCostFunctionAdaptor()->f(this->GetCurrentPosition());
}

void LBFGSModifiedOptimizer::AddIterationObserver(Command* cmd)
{
  observer = cmd;
}


/**
 * Connect a Cost Function
 */
void
LBFGSModifiedOptimizer
::SetCostFunction( SingleValuedCostFunction * costFunction )
{
  const unsigned int numberOfParameters =
    costFunction->GetNumberOfParameters();

  CostFunctionAdaptorType * adaptor =
    new CostFunctionAdaptorType( numberOfParameters );

  adaptor->SetCostFunction( costFunction );

  if ( m_OptimizerInitialized )
  {
    delete m_VnlOptimizer;
  }

  this->SetCostFunctionAdaptor( adaptor );

  m_VnlOptimizer = new VnlModifiedOptimizer( *adaptor );
//  m_VnlOptimizer->AddObserver(itk::IterationEvent(), (itk::Command*)observer);

  // set the optimizer parameters
  m_VnlOptimizer->set_trace( m_Trace );
  m_VnlOptimizer->set_max_function_evals(
    static_cast<int>( m_MaximumNumberOfFunctionEvaluations ) );
  m_VnlOptimizer->set_g_tolerance( m_GradientConvergenceTolerance );
  m_VnlOptimizer->line_search_accuracy = m_LineSearchAccuracy;
  m_VnlOptimizer->default_step_length  = m_DefaultStepLength;

  m_OptimizerInitialized = true;

}


/**
 * Start the optimization
 */
void
LBFGSModifiedOptimizer
::StartOptimization( void )
{

  this->InvokeEvent( StartEvent() );

  ParametersType initialPosition = GetInitialPosition();
  ParametersType parameters( initialPosition );

  // If the user provides the scales then we set otherwise we don't
  // for computation speed
  if (m_ScalesInitialized)
  {
    this->GetCostFunctionAdaptor()->SetScales(this->GetScales());
  }

  // vnl optimizers return the solution by reference
  // in the variable provided as initial position
  m_VnlOptimizer->minimize( parameters );

  if ( parameters.size() != initialPosition.size() )
  {
    // set current position to initial position and throw an exception
    this->SetCurrentPosition( initialPosition );
    itkExceptionMacro( << "Error occured in optimization" );
  }

  // we scale the parameters down if scales are defined
  if (m_ScalesInitialized)
  {
    ScalesType scales = this->GetScales();
    for (unsigned int i=0; i<parameters.size(); i++)
    {
      parameters[i] /= scales[i];
    }
  }

  this->SetCurrentPosition( parameters );

  this->InvokeEvent( EndEvent() );

}


/**
 * Get the Optimizer
 */
VnlModifiedOptimizer *
LBFGSModifiedOptimizer
::GetOptimizer()
{
  return m_VnlOptimizer;
}


} // end namespace itk

#endif
