/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#if defined(_MSC_VER)
#pragma warning ( disable : 4786 )
#endif

#include <iostream>

#include "itkExceptionObject.h"
#include "otbCommandLineArgumentParser.h"
#include "otbStreamingImageFileWriter.h"
#include "otbConfigure.h"
#include "otbMacro.h"

#include "otbPolarimetricSynthesisManager.h"


int main(int argc, char* argv[] )
{

  try
  {
    typedef std::complex <double>                   InputPixelType;
    typedef otb::Image< InputPixelType,  2 >        InputImageType;
    typedef otb::VectorImage< InputPixelType,  2 >  InputVectorType;
    typedef otb::Image< double,  2 >                OutputImageType;

    typedef otb::PolarimetricSynthesisManager< InputVectorType, InputImageType, InputImageType, InputImageType, InputImageType,
    OutputImageType >  PolarimetricManagerType;

    // Parse command line parameters
    typedef otb::CommandLineArgumentParser ParserType;
    ParserType::Pointer parser = ParserType::New();
    int nbParams;

    parser->AddOption("--ImageHH","InputImage HH ","-inHH", 1, false);
    parser->AddOption("--ImageHV","InputImage HV ","-inHV", 1, false);
    parser->AddOption("--ImageVH","InputImage VH ","-inVH", 1, false);
    parser->AddOption("--ImageVV","InputImage VV ","-inVV", 1, false);
    parser->AddOption("--ImageVector","InputVectorImage ","-in", 1, false);
    parser->AddOption("--OutputFileName","Output Image File Name ","-out", 1, true);
    parser->AddOption("--Extension","Extension type of the output image (tif by default)","-ext", 1, false);
    parser->AddOptionNParams("--Pol","For only one synthesis : PsiI KhiI [ PsiR KhiR ]","-Pol", false);
    parser->AddOptionNParams("--PolRed","For only one synthesis : PsiI KhiI [ PsiR KhiR ]","-Pol", false);
    parser->AddOptionNParams("--PolGreen","For only one synthesis : PsiI KhiI [ PsiR KhiR ]","-Pol", false);
    parser->AddOptionNParams("--PolBlue","For only one synthesis : PsiI KhiI [ PsiR KhiR ]","-Pol", false);
    parser->AddOptionNParams("--PolBegin","Set the start parameters : PsiI KhiI [ PsiR KhiR ]","-PolBegin", false);
    parser->AddOptionNParams("--PolEnd","Set the finish parameters : PsiI KhiI [ PsiR KhiR ]","-PolEnd", false);
    parser->AddOptionNParams("--PolBeginRed","Set the start parameters for red the channel : PsiI KhiI [ PsiR KhiR ]","-PolBeginRed", false);
    parser->AddOptionNParams("--PolEndRed","Set the finish parameters for red the channel : PsiI KhiI [ PsiR KhiR ]","-PolEndRed", false);
    parser->AddOptionNParams("--PolBeginGreen","Set the start parameters for the green channel : PsiI KhiI [ PsiR KhiR ]","-PolBeginGreen", false);
    parser->AddOptionNParams("--PolEndGreen","Set the finish parameters for the green channel : PsiI KhiI [ PsiR KhiR ]","-PolEndGreen", false);
    parser->AddOptionNParams("--PolBeginBlue","Set the start parameters for the blue channel : PsiI KhiI [ PsiR KhiR ]","-PolBeginBlue", false);
    parser->AddOptionNParams("--PolEndBlue","Set the finish parameters for the blue channel : PsiI KhiI [ PsiR KhiR ]","-PolEndBlue", false);
    parser->AddOption("--EmissionH","Set EmissionH to true ","-emH", 1, false);
    parser->AddOption("--EmissionV","Set EmissionV to true ","-emV", 1, false);
    parser->AddOption("--Copolar","Set the copolar mode","-co", 0, false);
    parser->AddOption("--Crosspolar","Set the crosspolar node","-cross", 0, false);
    parser->AddOption("--Gain","Set the gain","-gain", 1, false);

    // The user have to choose between the three following values
    parser->AddOption("--NbImages","number of images to produce ","-nbi", 1, false);

    typedef otb::CommandLineArgumentParseResult ParserResultType;
    ParserResultType::Pointer  parseResult = ParserResultType::New();

    try
    {
      parser->ParseCommandLine(argc, argv, parseResult);
    }
    catch ( itk::ExceptionObject & err )
    {
      std::string descriptionException = err.GetDescription();
      if (descriptionException.find("ParseCommandLine(): Help Parser") != std::string::npos)
      {
        std::cout << "WARNING : output file pixels are converted in 'unsigned char'" << std::endl;
        return EXIT_SUCCESS;
      }
      if (descriptionException.find("ParseCommandLine(): Version Parser") != std::string::npos)
      {
        return EXIT_SUCCESS;
      }
      return EXIT_FAILURE;
    }


    // Creation of the polarimetric synthesis manager
    PolarimetricManagerType::Pointer polarimetricSynthesisManager = PolarimetricManagerType::New();

    // Analyse of the parameters
    if ( parseResult->IsOptionPresent("--OutputFileName") )
    {
      polarimetricSynthesisManager->SetOutputFileName(parseResult->GetParameterString("--OutputFileName"));
    }
    if ( parseResult->IsOptionPresent("--ImageHH") )
    {
      polarimetricSynthesisManager->SetInputHH(parseResult->GetParameterString("--ImageHH"));
    }
    if ( parseResult->IsOptionPresent("--ImageHV") )
    {
      polarimetricSynthesisManager->SetInputHV(parseResult->GetParameterString("--ImageHV"));
    }
    if ( parseResult->IsOptionPresent("--ImageVH") )
    {
      polarimetricSynthesisManager->SetInputVH(parseResult->GetParameterString("--ImageVH"));
    }
    if ( parseResult->IsOptionPresent("--ImageVV") )
    {
      polarimetricSynthesisManager->SetInputVV(parseResult->GetParameterString("--ImageVV"));
    }
    if ( parseResult->IsOptionPresent("--ImageVector") )
    {
      polarimetricSynthesisManager->SetInputVector(parseResult->GetParameterString("--ImageVector"));
    }
    if ( parseResult->IsOptionPresent("--Extension") )
    {
      polarimetricSynthesisManager->SetExtension(parseResult->GetParameterString("--Extension"));
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--Pol") )
    {
      nbParams = parseResult->GetNumberOfParameters("--Pol");
      polarimetricSynthesisManager->SetPsiI_A(0, parseResult->GetParameterDouble("--Pol", 0));
      polarimetricSynthesisManager->SetKhiI_A(0, parseResult->GetParameterDouble("--Pol", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(0, parseResult->GetParameterDouble("--Pol", 2));
        polarimetricSynthesisManager->SetKhiR_A(0, parseResult->GetParameterDouble("--Pol", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolBegin") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolBegin");
      polarimetricSynthesisManager->SetPsiI_A(0, parseResult->GetParameterDouble("--PolBegin", 0));
      polarimetricSynthesisManager->SetKhiI_A(0, parseResult->GetParameterDouble("--PolBegin", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(0, parseResult->GetParameterDouble("--PolBegin", 2));
        polarimetricSynthesisManager->SetKhiR_A(0, parseResult->GetParameterDouble("--PolBegin", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolEnd") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolEnd");
      polarimetricSynthesisManager->SetPsiI_B(0, parseResult->GetParameterDouble("--PolEnd", 0));
      polarimetricSynthesisManager->SetKhiI_B(0, parseResult->GetParameterDouble("--PolEnd", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_B(0, parseResult->GetParameterDouble("--PolEnd", 2));
        polarimetricSynthesisManager->SetKhiR_B(0, parseResult->GetParameterDouble("--PolEnd", 3));
      }
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--PolRed") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolRed");
      polarimetricSynthesisManager->SetPsiI_A(0, parseResult->GetParameterDouble("--PolRed", 0));
      polarimetricSynthesisManager->SetKhiI_A(0, parseResult->GetParameterDouble("--PolRed", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(0, parseResult->GetParameterDouble("--PolRed", 2));
        polarimetricSynthesisManager->SetKhiR_A(0, parseResult->GetParameterDouble("--PolRed", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolGreen") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolGreen");
      polarimetricSynthesisManager->SetPsiI_A(1, parseResult->GetParameterDouble("--PolGreen", 0));
      polarimetricSynthesisManager->SetKhiI_A(1, parseResult->GetParameterDouble("--PolGreen", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(1, parseResult->GetParameterDouble("--PolGreen", 2));
        polarimetricSynthesisManager->SetKhiR_A(1, parseResult->GetParameterDouble("--PolGreen", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolBlue") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolBlue");
      polarimetricSynthesisManager->SetPsiI_A(2, parseResult->GetParameterDouble("--PolBlue", 0));
      polarimetricSynthesisManager->SetKhiI_A(2, parseResult->GetParameterDouble("--PolBlue", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(2, parseResult->GetParameterDouble("--PolBlue", 2));
        polarimetricSynthesisManager->SetKhiR_A(2, parseResult->GetParameterDouble("--PolBlue", 3));
      }
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--PolBeginRed") )
    {
      polarimetricSynthesisManager->SetRGB(true);
      nbParams = parseResult->GetNumberOfParameters("--PolBeginRed");
      polarimetricSynthesisManager->SetPsiI_A(0, parseResult->GetParameterDouble("--PolBeginRed", 0));
      polarimetricSynthesisManager->SetKhiI_A(0, parseResult->GetParameterDouble("--PolBeginRed", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(0, parseResult->GetParameterDouble("--PolBeginRed", 2));
        polarimetricSynthesisManager->SetKhiR_A(0, parseResult->GetParameterDouble("--PolBeginRed", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolEndRed") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolEndRed");
      polarimetricSynthesisManager->SetPsiI_B(0, parseResult->GetParameterDouble("--PolEndRed", 0));
      polarimetricSynthesisManager->SetKhiI_B(0, parseResult->GetParameterDouble("--PolEndRed", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_B(0, parseResult->GetParameterDouble("--PolEndRed", 2));
        polarimetricSynthesisManager->SetKhiR_B(0, parseResult->GetParameterDouble("--PolEndRed", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolBeginGreen") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolBeginGreen");
      polarimetricSynthesisManager->SetPsiI_A(1, parseResult->GetParameterDouble("--PolBeginGreen", 0));
      polarimetricSynthesisManager->SetKhiI_A(1, parseResult->GetParameterDouble("--PolBeginGreen", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(1, parseResult->GetParameterDouble("--PolBeginGreen", 2));
        polarimetricSynthesisManager->SetKhiR_A(1, parseResult->GetParameterDouble("--PolBeginGreen", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolEndGreen") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolEndGreen");
      polarimetricSynthesisManager->SetPsiI_B(1, parseResult->GetParameterDouble("--PolEndGreen", 0));
      polarimetricSynthesisManager->SetKhiI_B(1, parseResult->GetParameterDouble("--PolEndGreen", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_B(1, parseResult->GetParameterDouble("--PolEndGreen", 2));
        polarimetricSynthesisManager->SetKhiR_B(1, parseResult->GetParameterDouble("--PolEndGreen", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolBeginBlue") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolBeginBlue");
      polarimetricSynthesisManager->SetPsiI_A(2, parseResult->GetParameterDouble("--PolBeginBlue", 0));
      polarimetricSynthesisManager->SetKhiI_A(2, parseResult->GetParameterDouble("--PolBeginBlue", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_A(2, parseResult->GetParameterDouble("--PolBeginBlue", 2));
        polarimetricSynthesisManager->SetKhiR_A(2, parseResult->GetParameterDouble("--PolBeginBlue", 3));
      }
    }
    if ( parseResult->IsOptionPresent("--PolEndBlue") )
    {
      nbParams = parseResult->GetNumberOfParameters("--PolEndBlue");
      polarimetricSynthesisManager->SetPsiI_B(2, parseResult->GetParameterDouble("--PolEndBlue", 0));
      polarimetricSynthesisManager->SetKhiI_B(2, parseResult->GetParameterDouble("--PolEndBlue", 1));
      if (nbParams==4)
      {
        polarimetricSynthesisManager->SetPsiR_B(2, parseResult->GetParameterDouble("--PolEndBlue", 2));
        polarimetricSynthesisManager->SetKhiR_B(2, parseResult->GetParameterDouble("--PolEndBlue", 3));
      }
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--Copolar") )
    {
      std::cout<<"Mode Copola "<<std::endl;
      polarimetricSynthesisManager->ForceCoPolar(0);
      polarimetricSynthesisManager->ForceCoPolar(1);
      polarimetricSynthesisManager->ForceCoPolar(2);
    }
    if ( parseResult->IsOptionPresent("--Crosspolar") )
    {
      std::cout<<"Mode Crosspola "<<std::endl;
      polarimetricSynthesisManager->ForceCrossPolar(0);
      polarimetricSynthesisManager->ForceCrossPolar(1);
      polarimetricSynthesisManager->ForceCrossPolar(2);
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--EmissionH") )
    {
      polarimetricSynthesisManager->SetEmissionH(true);
    }
    if ( parseResult->IsOptionPresent("--EmissionV") )
    {
      polarimetricSynthesisManager->SetEmissionV(true);
    }
    if ( parseResult->IsOptionPresent("--Gain") )
    {
      polarimetricSynthesisManager->SetGain(true);
    }
//----------------------------
    if ( parseResult->IsOptionPresent("--NbImages") )
    {
      polarimetricSynthesisManager->SetNbImages(parseResult->GetParameterULong("--NbImages"));
      std::cout<<"Number of images : "<<parseResult->GetParameterULong("--NbImages")<<std::endl;
    }
//----------------------------
    // Execution
    polarimetricSynthesisManager->SynthesisFactory();

  }
  catch ( itk::ExceptionObject & err )
  {
    std::cout << "Following otbException catch :" << std::endl;
    std::cout << err << std::endl;
    return EXIT_FAILURE;
  }
  catch ( std::bad_alloc & err )
  {
    std::cout << "Exception bad_alloc : "<<(char*)err.what()<< std::endl;
    return EXIT_FAILURE;
  }
  catch ( ... )
  {
    std::cout << "Unknown Exception found !" << std::endl;
    return EXIT_FAILURE;
  }


  return EXIT_SUCCESS;

}

