/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#if defined(_MSC_VER)
#pragma warning ( disable : 4786 )
#endif

#include <iostream>

#include "otbImage.h"
#include "otbVectorImage.h"
#include "itkExceptionObject.h"

#include "otbImageFileReader.h"
#include "otbImageFileWriter.h"
#include "otbStreamingImageFileWriter.h"
#include "otbConfigure.h"
#include "otbMacro.h"
#include "otbImageList.h"
#include "otbImageListToVectorImageFilter.h"

#include "otbPolarimetricSynthesisFilter.h"
#include "otbMultiChannelsPolarimetricSynthesisFilter.h"

namespace otb
{

/** \class otbPolarimetricSynthesisManager
 * \brief
 *
 */

template <class TInputVectorImage, class TInputImageHH, class TInputImageHV, class TInputImageVH, class TInputImageVV, class TOutputImage >
class ITK_EXPORT PolarimetricSynthesisManager: public itk::ProcessObject
{

public:

  /** Standard typedefs */
  typedef PolarimetricSynthesisManager      Self;
  typedef itk::SmartPointer<Self>           Pointer;
  typedef itk::SmartPointer<const Self>     ConstPointer;

  /** Template parameters typedefs */
  typedef std::complex <double>                     InputPixelType;
  typedef otb::Image< InputPixelType,  2 >          InputImageType;
  typedef otb::VectorImage< InputPixelType,  2 >    InputVectorType;
  typedef otb::Image< double,  2 >                  OutputImageType;
  typedef otb::VectorImage< double,  2 >            OutputVectorType;
  typedef otb::ImageFileReader< InputImageType >    ReaderImageType;
  typedef otb::ImageFileReader< InputVectorType >   ReaderVectorType;
  typedef otb::ImageFileWriter< OutputImageType >   ImagesWriterType;
  typedef otb::ImageFileWriter< OutputVectorType >  VectorWriterType;
  typedef otb::ImageList< OutputImageType >         ImageListType;
  typedef otb::ImageListToVectorImageFilter<ImageListType, OutputVectorType>
  ImageListToVectorImageFilterType;
  typedef otb::PolarimetricSynthesisFilter< InputImageType, InputImageType, InputImageType, InputImageType,
  OutputImageType >  PolarimetricFilterType;
  typedef otb::MultiChannelsPolarimetricSynthesisFilter< InputVectorType, OutputImageType > MultiChanPolarimetricFilterType;


  /** Type macro */
  itkNewMacro(Self);

  /** Constructor */
  PolarimetricSynthesisManager()
  {
    m_DeltaPsiI[0]=0;
    m_DeltaKhiI[0]=0;
    m_DeltaPsiR[0]=0;
    m_DeltaKhiR[0]=0;
    m_DeltaPsiI[1]=0;
    m_DeltaKhiI[1]=0;
    m_DeltaPsiR[1]=0;
    m_DeltaKhiR[1]=0;
    m_DeltaPsiI[2]=0;
    m_DeltaKhiI[2]=0;
    m_DeltaPsiR[2]=0;
    m_DeltaKhiR[2]=0;
    m_NbImages=1;
    m_VectorMode=false;
    m_Extension="tif";
    m_RGB=false;

    m_polarimetricSynthesisFilter[0] = PolarimetricFilterType::New();
    m_polarimetricSynthesisFilter[1] = PolarimetricFilterType::New();
    m_polarimetricSynthesisFilter[2] = PolarimetricFilterType::New();
    m_multiChannelsPolarimetricSynthesisFilter[0] = MultiChanPolarimetricFilterType::New();
    m_multiChannelsPolarimetricSynthesisFilter[1] = MultiChanPolarimetricFilterType::New();
    m_multiChannelsPolarimetricSynthesisFilter[2] = MultiChanPolarimetricFilterType::New();
  }
  virtual ~PolarimetricSynthesisManager() {};


  /** Compute de the number of images if the user had set one of the two step */
  /*  void ComputeNbOfImages()
    {

          if (m_NbImages==1)
          {
                  if (m_DeltaKhi!=0)
                  {
                          m_NbImages = (int)( 1 + (vcl_abs( m_KhiI_B - m_KhiI_A )) / vcl_abs(m_DeltaKhi));

                  }
                  else if (m_DeltaPsi!=0)
                  {
                          m_NbImages = (int)( 1 + (vcl_abs( m_PsiI_B - m_PsiI_A )) / vcl_abs(m_DeltaPsi));
                  }
                  std::cout<<"Computed number of images : "<<m_NbImages<<std::endl;
          }
    }
  */
  /** Compute de the Delta PsiI considering the number of images */
  void ComputeDeltaPsiI()
  {

    int i, NbColor=1;
    if (m_RGB)
    {
      NbColor=3;
    }

    if (m_NbImages!=1)
    {

      for (i=0; i<NbColor; i++)
      {
        if (m_DeltaPsiI[i]==0)
        {
          m_DeltaPsiI[i] = ( m_PsiI_B[i] - m_PsiI_A[i]) / ( m_NbImages - 1);
          std::cout<<" PsiI_A["<<i<<"] : "<<m_PsiI_A[i]<<" PsiI_B["<<i<<"] : "<<m_PsiI_B[i]<<std::endl;
          std::cout<<"Computed DeltaPsiI["<<i<<"] : "<<m_DeltaPsiI[i]<<std::endl;
        }
      }
    }
  }

  /** Compute de the Delta KhiI considering the number of images */
  void ComputeDeltaKhiI()
  {

    int i, NbColor=1;
    if (m_RGB)
    {
      NbColor=3;
    }

    if (m_NbImages!=1)
    {
      for (i=0; i<NbColor; i++)
      {
        if (m_DeltaKhiI[i]==0)
        {
          m_DeltaKhiI[i] = ( m_KhiI_B[i] - m_KhiI_A[i] ) / ( m_NbImages - 1 );
          std::cout<<" KhiI_A["<<i<<"] : "<<m_KhiI_A[i]<<" KhiI_B["<<i<<"] : "<<m_KhiI_B[i]<<std::endl;
          std::cout<<"Computed DeltaKhiI["<<i<<"] : "<<m_DeltaKhiI[i]<<std::endl;
        }
      }
    }
  }

  /** Compute de the Delta PsiR considering the number of images */
  void ComputeDeltaPsiR()
  {
    int i, NbColor=1;
    if (m_RGB)
    {
      NbColor=3;
    }

    if (m_NbImages!=1)
    {
      for (i=0; i<NbColor; i++)
      {
        if (m_DeltaPsiR[i]==0)
        {
          m_DeltaPsiR[i] = ( m_PsiR_B[i] - m_PsiR_A[i] ) / ( m_NbImages - 1 );
          std::cout<<" PsiR_A["<<i<<"] : "<<m_PsiR_A[i]<<" PsiR_B["<<i<<"] : "<<m_PsiR_B[i]<<std::endl;
          std::cout<<"Computed DeltaPsiR["<<i<<"] : "<<m_DeltaPsiR[i]<<std::endl;
        }
      }
    }
  }

  /** Compute de the Delta KhiR considering the number of images */
  void ComputeDeltaKhiR()
  {
    int i, NbColor=1;
    if (m_RGB)
    {
      NbColor=3;
    }

    if (m_NbImages!=1)
    {
      for (i=0; i<NbColor; i++)
      {
        if (m_DeltaKhiR[i]==0)
        {
          m_DeltaKhiR[i] = ( m_KhiR_B[i] - m_KhiR_A[i] ) / ( m_NbImages - 1 );
          std::cout<<" KhiR_A["<<i<<"] : "<<m_KhiR_A[i]<<" KhiR_B["<<i<<"] : "<<m_KhiR_B[i]<<std::endl;
          std::cout<<"Computed DeltaKhiR["<<i<<"] : "<<m_DeltaKhiR[i]<<std::endl;
        }
      }
    }
  }

  /** Force copolar mode for the filters */
  void ForceCoPolar(int i)
  {
    SetPsiR_A(i, m_PsiI_A[i]);
    SetKhiR_A(i, m_KhiI_A[i]);
    SetPsiR_B(i, m_PsiI_B[i]);
    SetKhiR_B(i, m_KhiI_B[i]);
    m_polarimetricSynthesisFilter[i]->ForceCoPolar();
    m_multiChannelsPolarimetricSynthesisFilter[i]->ForceCoPolar();
  }

  /** Force crosspolar mode for the filters */
  void ForceCrossPolar(int i)
  {
    SetPsiR_A(i, m_PsiI_A[i]+90);
    SetKhiR_A(i, -m_KhiI_A[i]);
    SetPsiR_B(i, m_PsiI_B[i]+90);
    SetKhiR_B(i, -m_KhiI_B[i]);
    m_polarimetricSynthesisFilter[i]->ForceCrossPolar();
    m_multiChannelsPolarimetricSynthesisFilter[i]->ForceCrossPolar();
  }

  /** Launch the polarimetric synthesis with images and write the result */
  void LaunchSynthesisWithImages(double *psiI, double *khiI, double *psiR, double *khiR, std::string outputFileName )
  {

    int i, NbColors=1;
    if (m_RGB)
    {
      NbColors=3;
    }

    for (i=0; i<NbColors; i++)
    {
      m_polarimetricSynthesisFilter[i]->SetPsiI(psiI[i]);
      m_polarimetricSynthesisFilter[i]->SetKhiI(khiI[i]);
      m_polarimetricSynthesisFilter[i]->SetPsiR(psiR[i]);
      m_polarimetricSynthesisFilter[i]->SetKhiR(khiR[i]);
    }

    // Fusion of the images into a RGBVectorImage
    if (m_RGB)
    {
      ImageListType::Pointer imageList = ImageListType::New();
      ImageListToVectorImageFilterType::Pointer imageListToRGBVector = ImageListToVectorImageFilterType::New();

      imageList->Clear();
      imageList->PushBack(m_polarimetricSynthesisFilter[0]->GetOutput());
      imageList->PushBack(m_polarimetricSynthesisFilter[1]->GetOutput());
      imageList->PushBack(m_polarimetricSynthesisFilter[2]->GetOutput());

      imageListToRGBVector->SetInput(imageList);
      imageListToRGBVector->UpdateOutputInformation();

      // Creation of the writer
      VectorWriterType::Pointer writer = VectorWriterType::New();

      // Execution
      writer->SetFileName( outputFileName );
      writer->SetInput( imageListToRGBVector->GetOutput() );
      writer->Update();
    }
    // Grayscale image
    else
    {
      // Creation of the writer
      ImagesWriterType::Pointer writer = ImagesWriterType::New();

      // Execution
      writer->SetFileName( outputFileName );
      writer->SetInput( m_polarimetricSynthesisFilter[0]->GetOutput() );
      writer->Update();
    }
  }

  /** Launch the polarimetric synthesis with a vector and write the result */
  void LaunchSynthesisWithVector(double *psiI, double *khiI, double *psiR, double *khiR, string outputFileName )
  {

    int i, NbColors=1;
    if (m_RGB)
    {
      NbColors=3;
    }

    for (i=0; i<NbColors; i++)
    {
      m_multiChannelsPolarimetricSynthesisFilter[i]->SetPsiI(psiI[i]);
      m_multiChannelsPolarimetricSynthesisFilter[i]->SetKhiI(khiI[i]);
      m_multiChannelsPolarimetricSynthesisFilter[i]->SetPsiR(psiR[i]);
      m_multiChannelsPolarimetricSynthesisFilter[i]->SetKhiR(khiR[i]);
    }

    // Fusion of the images into a RGBVectorImage
    if (m_RGB)
    {
      ImageListType::Pointer imageList = ImageListType::New();
      ImageListToVectorImageFilterType::Pointer imageListToRGBVector = ImageListToVectorImageFilterType::New();

      imageList->Clear();
      imageList->PushBack(m_multiChannelsPolarimetricSynthesisFilter[0]->GetOutput());
      imageList->PushBack(m_multiChannelsPolarimetricSynthesisFilter[1]->GetOutput());
      imageList->PushBack(m_multiChannelsPolarimetricSynthesisFilter[2]->GetOutput());

      imageListToRGBVector->SetInput(imageList);
      imageListToRGBVector->UpdateOutputInformation();

      // Creation of the writer
      VectorWriterType::Pointer writer = VectorWriterType::New();

      // Execution
      writer->SetFileName( outputFileName );
      writer->SetInput( imageListToRGBVector->GetOutput() );
      writer->Update();
    }
    // Grayscale image
    else
    {
      // Creation of the writer
      ImagesWriterType::Pointer writer = ImagesWriterType::New();

      // Execution
      writer->SetFileName( outputFileName );
      writer->SetInput( m_multiChannelsPolarimetricSynthesisFilter[0]->GetOutput() );
      writer->Update();
    }

  }

  /** Synthesis Factory (main) */
  void SynthesisFactory()
  {
    int imageIndex;
    double psiI[3], psiR[3], khiI[3], khiR[3];
    std::string outputFileName;
    int i, NbColor=1;

    if (m_RGB)
    {
      NbColor=3;
    }

    // Initialisation
    for (i=0; i<NbColor; i++)
    {
      psiI[i]=this->GetPsiI_A(i);
      khiI[i]=this->GetKhiI_A(i);
      psiR[i]=this->GetPsiR_A(i);
      khiR[i]=this->GetKhiR_A(i);
    }

    // Computation of the number of images
    /*        ComputeNbOfImages(); */

    // Computation of the steps
    ComputeDeltaPsiI();
    ComputeDeltaKhiI();
    ComputeDeltaPsiR();
    ComputeDeltaKhiR();

    for (imageIndex=0; imageIndex<m_NbImages; imageIndex++)
    {

      outputFileName=GenerateOutputfilenames(imageIndex, psiI, khiI, psiR, khiR);

      if (m_VectorMode)
      {
        LaunchSynthesisWithVector( psiI, khiI, psiR, khiR, outputFileName );
      }
      else
      {
        LaunchSynthesisWithImages( psiI, khiI, psiR, khiR, outputFileName );
      }

      for (i=0; i<NbColor; i++)
      {
        psiI[i]+= m_DeltaPsiI[i];
        khiI[i]+= m_DeltaKhiI[i];
        psiR[i]+= m_DeltaPsiR[i];
        khiR[i]+= m_DeltaKhiR[i];
      }
    }
  }

  /** Generate outputfile names */
  std::string GenerateOutputfilenames(int num, double *psiI, double *khiI, double *psiR, double *khiR)
  {
    std::string extension;
    std::string outputFileName;
    itk::OStringStream oss;

    // put the root name
    outputFileName=m_OutputFileName;

    if ( m_Extension.at(0)=='.'  )
    {
      extension=m_Extension;
    }
    else
    {
      extension=".";
      extension.append(m_Extension);
    }
    outputFileName.append("_");

    // Number of the produced image list
    oss.str("");
    oss<<num;
    outputFileName.append(oss.str().c_str());
    oss.str("");
    if (m_RGB)
    {
      outputFileName.append("_Red");
    }
    outputFileName.append("_I_");
    oss.str("");
    oss<<psiI[0];
    outputFileName.append(oss.str().c_str());
    outputFileName.append("_");
    oss.str("");
    oss<<khiI[0];
    outputFileName.append(oss.str().c_str());
    outputFileName.append("_R_");
    oss.str("");
    oss<<psiR[0];
    outputFileName.append(oss.str().c_str());
    outputFileName.append("_");
    oss.str("");
    oss<<khiR[0];
    outputFileName.append(oss.str().c_str());

    if (m_RGB)
    {
      // Green
      outputFileName.append("_Green_I_");
      oss.str("");
      oss<<psiI[1];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_");
      oss.str("");
      oss<<khiI[1];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_R_");
      oss.str("");
      oss<<psiR[1];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_");
      oss.str("");
      oss<<khiR[1];
      outputFileName.append(oss.str().c_str());

      // Blue
      outputFileName.append("_Blue_I_");
      oss.str("");
      oss<<psiI[2];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_");
      oss.str("");
      oss<<khiI[2];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_R_");
      oss.str("");
      oss<<psiR[2];
      outputFileName.append(oss.str().c_str());
      outputFileName.append("_");
      oss.str("");
      oss<<khiR[2];
      outputFileName.append(oss.str().c_str());

    }

    // Add the extension type
    outputFileName.append(extension);

    std::cout<<"GenerateOutputfileName : "<<outputFileName<<std::endl;
    return outputFileName;

  }

  /* Getters/Setters */
  itkSetMacro(NbImages, int);
  itkSetMacro(VectorMode, bool);
  itkSetMacro(RGB, bool);
  void SetPsiI_A(int ind, double val)
  {
    m_PsiI_A[ind]=val;
  }
  void SetKhiI_A(int ind, double val)
  {
    m_KhiI_A[ind]=val;
  }
  void SetPsiR_A(int ind, double val)
  {
    m_PsiR_A[ind]=val;
  }
  void SetKhiR_A(int ind, double val)
  {
    m_KhiR_A[ind]=val;
  }
  void SetPsiI_B(int ind, double val)
  {
    m_PsiI_B[ind]=val;
  }
  void SetKhiI_B(int ind, double val)
  {
    m_KhiI_B[ind]=val;
  }
  void SetPsiR_B(int ind, double val)
  {
    m_PsiR_B[ind]=val;
  }
  void SetKhiR_B(int ind, double val)
  {
    m_KhiR_B[ind]=val;
  }
  void SetDeltaPsiI(int ind, double val)
  {
    m_DeltaPsiI[ind]=val;
  }
  void SetDeltaKhiI(int ind, double val)
  {
    m_DeltaKhiI[ind]=val;
  }
  void SetDeltaPsiR(int ind, double val)
  {
    m_DeltaPsiR[ind]=val;
  }
  void SetDeltaKhiR(int ind, double val)
  {
    m_DeltaKhiR[ind]=val;
  }
  double GetPsiI_A(int ind)
  {
    return m_PsiI_A[ind];
  }
  double GetKhiI_A(int ind)
  {
    return m_KhiI_A[ind];
  }
  double GetPsiR_A(int ind)
  {
    return m_PsiR_A[ind];
  }
  double GetKhiR_A(int ind)
  {
    return m_KhiR_A[ind];
  }
  double GetPsiI_B(int ind)
  {
    return m_PsiI_B[ind];
  }
  double GetKhiI_B(int ind)
  {
    return m_KhiI_B[ind];
  }
  double GetPsiR_B(int ind)
  {
    return m_PsiR_B[ind];
  }
  double GetKhiR_B(int ind)
  {
    return m_KhiR_B[ind];
  }
  void SetOutputFileName( std::string outputFileName )
  {
    m_OutputFileName=outputFileName;
  }
  void SetExtension( std::string extension )
  {
    m_Extension=extension;
  }
  void SetInputHH( std::string inputFileNameHH )
  {
    m_OutputFileName.append("_HH");
    reader1 = ReaderImageType::New();
    reader1->SetFileName( inputFileNameHH );
    m_polarimetricSynthesisFilter[0]->SetInputHH(reader1->GetOutput());
    m_polarimetricSynthesisFilter[1]->SetInputHH(reader1->GetOutput());
    m_polarimetricSynthesisFilter[2]->SetInputHH(reader1->GetOutput());
  }
  void SetInputHV( std::string inputFileNameHV )
  {
    m_OutputFileName.append("_HV");
    reader2 = ReaderImageType::New();
    reader2->SetFileName( inputFileNameHV );
    m_polarimetricSynthesisFilter[0]->SetInputHV(reader2->GetOutput());
    m_polarimetricSynthesisFilter[1]->SetInputHV(reader2->GetOutput());
    m_polarimetricSynthesisFilter[2]->SetInputHV(reader2->GetOutput());
  }
  void SetInputVH( std::string inputFileNameVH )
  {
    m_OutputFileName.append("_VH");
    reader3 = ReaderImageType::New();
    reader3->SetFileName( inputFileNameVH );
    m_polarimetricSynthesisFilter[0]->SetInputVH(reader3->GetOutput());
    m_polarimetricSynthesisFilter[1]->SetInputVH(reader3->GetOutput());
    m_polarimetricSynthesisFilter[2]->SetInputVH(reader3->GetOutput());
  }
  void SetInputVV( std::string inputFileNameVV )
  {
    m_OutputFileName.append("_VV");
    reader4 = ReaderImageType::New();
    reader4->SetFileName( inputFileNameVV );
    m_polarimetricSynthesisFilter[0]->SetInputVV(reader4->GetOutput());
    m_polarimetricSynthesisFilter[1]->SetInputVV(reader4->GetOutput());
    m_polarimetricSynthesisFilter[2]->SetInputVV(reader4->GetOutput());

  }
  void SetInputVector( std::string inputFileNameVector )
  {
    SetVectorMode(true);
    vectorReader = ReaderVectorType::New();
    vectorReader->SetFileName( inputFileNameVector );
    m_multiChannelsPolarimetricSynthesisFilter[0]->SetInput(vectorReader->GetOutput());
    m_multiChannelsPolarimetricSynthesisFilter[1]->SetInput(vectorReader->GetOutput());
    m_multiChannelsPolarimetricSynthesisFilter[2]->SetInput(vectorReader->GetOutput());
  }
  void SetEmissionV(bool value)
  {
    m_multiChannelsPolarimetricSynthesisFilter[0]->SetEmissionV(value);
    m_multiChannelsPolarimetricSynthesisFilter[1]->SetEmissionV(value);
    m_multiChannelsPolarimetricSynthesisFilter[2]->SetEmissionV(value);
  }
  void SetEmissionH(bool value)
  {
    m_multiChannelsPolarimetricSynthesisFilter[0]->SetEmissionH(value);
    m_multiChannelsPolarimetricSynthesisFilter[1]->SetEmissionH(value);
    m_multiChannelsPolarimetricSynthesisFilter[2]->SetEmissionH(value);
  }
  void SetGain(double value)
  {
    m_polarimetricSynthesisFilter[0]->SetGain(value);
    m_polarimetricSynthesisFilter[1]->SetGain(value);
    m_polarimetricSynthesisFilter[2]->SetGain(value);
    m_multiChannelsPolarimetricSynthesisFilter[0]->SetGain(value);
    m_multiChannelsPolarimetricSynthesisFilter[1]->SetGain(value);
    m_multiChannelsPolarimetricSynthesisFilter[2]->SetGain(value);
  }

protected:

private:

  /** Filters */
  PolarimetricFilterType::Pointer m_polarimetricSynthesisFilter[3];
  MultiChanPolarimetricFilterType::Pointer m_multiChannelsPolarimetricSynthesisFilter[3];

  /** Readers */
  ReaderImageType::Pointer reader1;
  ReaderImageType::Pointer reader2;
  ReaderImageType::Pointer reader3;
  ReaderImageType::Pointer reader4;
  ReaderVectorType::Pointer vectorReader;

  /** Psi Incident */
  double m_PsiI_A[3];
  /** Khi Incident */
  double m_KhiI_A[3];
  /** Psi Incident */
  double m_PsiI_B[3];
  /** Khi Incident */
  double m_KhiI_B[3];
  /** Psi Reflected */
  double m_PsiR_A[3];
  /** Khi Reflected */
  double m_KhiR_A[3];
  /** Psi Reflected */
  double m_PsiR_B[3];
  /** Khi Reflected */
  double m_KhiR_B[3];

  /** Steps */
  double m_DeltaPsiI[3];
  double m_DeltaKhiI[3];
  double m_DeltaPsiR[3];
  double m_DeltaKhiR[3];

  /** Gain */
  double m_Gain;

  /** Number of images result */
  int m_NbImages;

  /** Outputfilename */
  std::string m_OutputFileName;
  /** Extension */
  std::string m_Extension;

  /** Vector or Images Mode*/
  bool m_VectorMode;

  /** RGB Mode or grayscale */
  bool m_RGB;

};  //end class

} // end namespace otb
