/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/


#include "otbRoadExtractionController.h"
#include "otbMsgReporter.h"

#include <FL/fl_ask.H>

namespace otb
{

RoadExtractionController::RoadExtractionController()
{
  m_Computing=FALSE;
}

RoadExtractionController::~RoadExtractionController()
{
  MsgReporter::GetInstance()->Hide();
}


void RoadExtractionController::SetView(ViewPointerType view)
{
  m_View = view;
}

void RoadExtractionController::ViewRegionChanged()
{
  try
  {
    m_Model->SetRegion(m_View->GetViewer()->GetFullWidget()->GetViewedRegion());
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void RoadExtractionController::PixelClicked(const IndexType & index)
{

  if (!m_Computing)
  {
    m_Computing = TRUE;

    //this->ViewRegionChanged();
    try
    {
      m_Model->SetReferencePixel(m_Model->GetImage()->GetPixel(index));
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }

    if (m_Model->GetRoadExtractionInput() == SPECTRAL_ANGLE)
    {
      itk::OStringStream oss;
      oss << m_Model->GetReferencePixel();
      m_View->vReferencePixel->value(oss.str().c_str());
    }
    m_Computing = FALSE;
  }
}

void RoadExtractionController::SelectFileName(std::string filename)
{
  try
  {
    m_Model->SetFileName(filename);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
  if (m_Model->GetValid())
  {
    m_View->bRun->activate();
    m_View->bClear->activate();
    m_View->bUseImageChannel->activate();
    m_View->bUseImageIntensity->activate();
    m_View->bUseSpectralAngle->activate();
    m_View->bUseWaterIndex->activate();
    m_View->vReferencePixel->activate();
    this->ViewRegionChanged();

  }
}

void RoadExtractionController::ChangeChannel(unsigned int channel)
{
  m_Model->SetChannelIndex(channel);
}
void RoadExtractionController::ChangeRedChannel(unsigned int channel)
{
  m_Model->SetRedChannelIndex(channel);
}
void RoadExtractionController::ChangeGreenChannel(unsigned int channel)
{
  m_Model->SetGreenChannelIndex(channel);
}
void RoadExtractionController::ChangeAlpha(double alpha)
{
  m_Model->SetAlpha(alpha);
}
void RoadExtractionController::ChangeResolution(double resolution)
{
  m_Model->SetResolution(resolution);
}
void RoadExtractionController::ChangeTolerance(double tolerance)
{
  m_Model->SetTolerance(tolerance);
}
void RoadExtractionController::ChangeMaxAngle(double maxAngle)
{
  m_Model->SetMaxAngle(maxAngle);
}
void RoadExtractionController::ChangeAngularThreshold(double angularThreshold)
{
  m_Model->SetAngularThreshold(angularThreshold);
}
void RoadExtractionController::ChangeAmplitudeThreshold(double amplitudeThreshold)
{
  m_Model->SetAmplitudeThreshold(amplitudeThreshold);
}
void RoadExtractionController::ChangeDistanceThreshold(double distanceThreshold)
{
  m_Model->SetDistanceThreshold(distanceThreshold);
}
void RoadExtractionController::ChangeFirstMeanDistanceThreshold(double firstMeanDistanceThreshold)
{
  m_Model->SetFirstMeanDistanceThreshold(firstMeanDistanceThreshold);
}
void RoadExtractionController::ChangeSecondMeanDistanceThreshold(double secondMeanDistanceThreshold)
{
  m_Model->SetSecondMeanDistanceThreshold(secondMeanDistanceThreshold);
}


void RoadExtractionController::ChangeUseImageIntensity()
{
  try
  {
    m_Model->SetRoadExtractionInput(INTENSITY);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
  m_View->bUseImageChannel->clear();
  m_View->bUseImageIntensity->set();
  m_View->bUseSpectralAngle->clear();
  m_View->bUseWaterIndex->clear();
  m_View->vChannel->deactivate();
  m_View->vRedChannel->deactivate();
  m_View->vGreenChannel->deactivate();
  m_View->vReferencePixel->deactivate();
  m_View->vReferencePixel->value("");
}

void RoadExtractionController::ChangeUseSpectralAngle()
{
  try
  {
    m_Model->SetRoadExtractionInput(SPECTRAL_ANGLE);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
  m_View->bUseImageChannel->clear();
  m_View->bUseImageIntensity->clear();
  m_View->bUseSpectralAngle->set();
  m_View->bUseWaterIndex->clear();
  m_View->vChannel->deactivate();
  m_View->vRedChannel->deactivate();
  m_View->vGreenChannel->deactivate();
  m_View->vReferencePixel->activate();
  itk::OStringStream oss;
  oss << m_Model->GetReferencePixel();
  m_View->vReferencePixel->value(oss.str().c_str());
}

void RoadExtractionController::ChangeUseImageChannel()
{
  try
  {
    m_Model->SetRoadExtractionInput(CHANNEL);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
  m_View->bUseImageChannel->set();
  m_View->bUseImageIntensity->clear();
  m_View->bUseSpectralAngle->clear();
  m_View->bUseWaterIndex->clear();
  m_View->vChannel->activate();
  m_View->vRedChannel->deactivate();
  m_View->vGreenChannel->deactivate();
  m_View->vReferencePixel->deactivate();
  m_View->vReferencePixel->value("");
}

void RoadExtractionController::ChangeUseWaterIndex()
{
  try
  {
    m_Model->SetRoadExtractionInput(WATER_INDEX);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
  m_View->bUseImageChannel->clear();
  m_View->bUseImageIntensity->clear();
  m_View->bUseSpectralAngle->clear();
  m_View->bUseWaterIndex->set();
  m_View->vChannel->deactivate();
  m_View->vRedChannel->activate();
  m_View->vGreenChannel->activate();
  m_View->vReferencePixel->deactivate();
  m_View->vReferencePixel->value("");

}

void RoadExtractionController::SaveResults(std::string filename)
{
  try
  {
    m_Model->SaveResults(filename);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void RoadExtractionController::ImportSegments(std::string filename)
{
  try
  {
    m_Model->ImportSegments(filename);
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void RoadExtractionController::ClearSegments()
{
  try
  {
    m_Model->ClearPathList();
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

void RoadExtractionController::Run()
{
  this->ViewRegionChanged();
  try
  {
    m_Model->GenerateData();
  }
  catch (itk::ExceptionObject & err)
  {
    MsgReporter::GetInstance()->SendError(err.GetDescription());
  }
}

} // end namespace otb


