/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.

    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbRoadExtractionModel.h"
#include "itkMacro.h"
#include "otbImageFileWriter.h"
#include "otbMath.h"
#include "otbVectorDataProjectionFilter.h"

namespace otb
{

/** Initialize the singleton */
RoadExtractionModel::Pointer RoadExtractionModel::Instance = NULL;

/** Manage the singleton */
RoadExtractionModel::Pointer
RoadExtractionModel::GetInstance()
{
  if (!Instance)
  {
    Instance = RoadExtractionModel::New();
  }
  return Instance;
}

void RoadExtractionModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

RoadExtractionModel::RoadExtractionModel()
{
  // filters
  m_Reader = ReaderType::New();
  m_GenericRoadExtractionFilter = GenericRoadExtractionFilterType::New();
  m_MultiChannelExtractROIFilter = MultiChannelExtractROIFilterType::New();
  m_SpectralAngleFilter = SpectralAngleDistanceImageFilterType::New();
  m_WaterIndexFilter = WaterIndexFilterType::New();
  m_ShiftScaleFilter = ShiftScaleFilterType::New();
  m_IntensityFilter = IntensityFilterType::New();
  m_MultiToMonoChannelExtractROIFilter = MultiToMonoChannelExtractROIFilterType::New();

  m_Valid = false;
  m_ReferencePixelValid = false;
  m_IsNewImage = false;

  m_ChannelIndex = 0;
  m_RedChannelIndex = 0;
  m_GreenChannelIndex = 0;

  VectorPixelType lReferencePixel;
  lReferencePixel.SetSize(1);
  lReferencePixel.Fill(0.);
  m_ReferencePixel = lReferencePixel;
  m_RegionValid = false;
  m_FileName = "";

  m_AmplitudeThreshold = static_cast<AmplitudeThresholdType>(0.00005 );
  m_Tolerance = static_cast<ToleranceType>(1.);
  m_MaxAngle = static_cast<MaxAngleType>(CONST_PI_8);
  m_FirstMeanDistanceThreshold = static_cast<MeanDistanceThresholdType>(1.);
  m_SecondMeanDistanceThreshold = static_cast<MeanDistanceThresholdType>(10.);
  m_AngularThreshold = static_cast<LinkRealType>(CONST_PI_8);
  m_DistanceThreshold = 25.;
  m_Alpha = 1.0;
  m_Resolution = 1.;

  IndexType index;
  index.Fill(0);
  SizeType size;
  size.Fill(0);
  m_Region.SetIndex(index);
  m_Region.SetSize(size);

  m_PathList = PathListType::New();

  m_RoadExtractionInput = SPECTRAL_ANGLE;

}

RoadExtractionModel::~RoadExtractionModel()
{}

void RoadExtractionModel::SetFileName(std::string file)
{
  m_FileName = file;
  m_Reader->SetFileName(m_FileName);

  this->SetIsNewImage(true);
  /// TODO: try/catch missing there (should be done in the controler ?)
  m_Reader->GenerateOutputInformation();
  m_Valid = true;
  this->SetupRoadExtraction();
  this->NotifyAll();

  this->SetIsNewImage(false);
}

RoadExtractionModel::VectorImageType * RoadExtractionModel::GetImage()
{
  return m_Reader->GetOutput();
}

void RoadExtractionModel::SetReferencePixel(const VectorPixelType & reference)
{
  m_ReferencePixelValid = true;
  m_ReferencePixel = reference;
}

void RoadExtractionModel::SetRegion(const RegionType & region)
{
  m_RegionValid = true;
  m_Region = region;
  this->Modified();
}

void RoadExtractionModel::SetFilterParameters()
{

  m_GenericRoadExtractionFilter->SetAlpha(m_Alpha);
  m_GenericRoadExtractionFilter->SetResolution(m_Resolution);
  m_GenericRoadExtractionFilter->SetTolerance(m_Tolerance);
  m_GenericRoadExtractionFilter->SetMaxAngle(m_MaxAngle);
  m_GenericRoadExtractionFilter->SetAngularThreshold(m_AngularThreshold);
  m_GenericRoadExtractionFilter->SetAmplitudeThreshold(m_AmplitudeThreshold);
  m_GenericRoadExtractionFilter->SetDistanceThreshold(m_DistanceThreshold);
  m_GenericRoadExtractionFilter->SetFirstMeanDistanceThreshold(m_FirstMeanDistanceThreshold);
  m_GenericRoadExtractionFilter->SetSecondMeanDistanceThreshold(m_SecondMeanDistanceThreshold);
}

void RoadExtractionModel::SetupRoadExtraction()
{

  m_GenericRoadExtractionFilter = GenericRoadExtractionFilterType::New();
  this->SetFilterParameters();


  switch (m_RoadExtractionInput)
  {
  case SPECTRAL_ANGLE:
  {
    m_MultiChannelExtractROIFilter->SetInput(m_Reader->GetOutput());
    m_SpectralAngleFilter->SetInput(m_MultiChannelExtractROIFilter->GetOutput());
    m_GenericRoadExtractionFilter->SetInput(m_SpectralAngleFilter->GetOutput());
    break;
  }
  case INTENSITY:
  {
    m_MultiChannelExtractROIFilter->SetInput(m_Reader->GetOutput());
    m_IntensityFilter->SetInput(m_MultiChannelExtractROIFilter->GetOutput());
    m_GenericRoadExtractionFilter->SetInput(m_IntensityFilter->GetOutput());
    break;
  }
  case CHANNEL:
  {
    m_MultiToMonoChannelExtractROIFilter->SetInput(m_Reader->GetOutput());
    m_MultiToMonoChannelExtractROIFilter->SetChannel(m_ChannelIndex+1);
    m_GenericRoadExtractionFilter->SetInput(m_MultiToMonoChannelExtractROIFilter->GetOutput());

    break;
  }
  case WATER_INDEX:
  {
    m_MultiChannelExtractROIFilter->SetInput(m_Reader->GetOutput());


    m_WaterIndexFilter->SetRedIndex(m_RedChannelIndex+1);
    m_WaterIndexFilter->SetNIRIndex(m_GreenChannelIndex+1);
    m_WaterIndexFilter->SetInput(m_MultiChannelExtractROIFilter->GetOutput());


    //SEEEMS TO BE USELESS -> water-> 1
    m_ShiftScaleFilter->SetInput(m_WaterIndexFilter->GetOutput());
    m_ShiftScaleFilter->SetShift(-1);
    m_ShiftScaleFilter->SetScale(-CONST_PI_2);

    //m_GenericRoadExtractionFilter->SetInput(m_WaterIndexFilter->GetOutput());
    m_GenericRoadExtractionFilter->SetInput(m_ShiftScaleFilter->GetOutput());

    break;
  }
  default:
  {
    itkExceptionMacro(<<"m_RoadExtractionInput variable is not initialized ! ");
    break;
  }
  }
}

void RoadExtractionModel::GenerateData()
{

  this->SetupRoadExtraction();
  if (!m_RegionValid)
  {
    itkExceptionMacro(<<"The region is not initialized ! ");
  }

  switch (m_RoadExtractionInput)
  {
  case SPECTRAL_ANGLE:
  {
    if (!m_ReferencePixelValid)
    {
      itkExceptionMacro(<<"The reference pixel is not initialized !");
    }
    m_SpectralAngleFilter->SetReferencePixel(m_ReferencePixel);
    m_MultiChannelExtractROIFilter->SetExtractionRegion(m_Region);

    break;
  }
  case INTENSITY:
  {
    m_MultiChannelExtractROIFilter->SetExtractionRegion(m_Region);
    break;
  }
  case CHANNEL:
  {
    m_MultiToMonoChannelExtractROIFilter->SetExtractionRegion(m_Region);
    break;
  }
  case WATER_INDEX:
  {
    m_MultiChannelExtractROIFilter->SetExtractionRegion(m_Region);
    break;
  }
  default:
  {
    itkExceptionMacro(<<"m_RoadExtractionInput variable is not initialized ! ");
    break;
  }
  }

  m_GenericRoadExtractionFilter->Update();

  PathListPointerType newPathList = m_GenericRoadExtractionFilter->GetOutput();
  PathListType::Iterator pit = newPathList->Begin();
  while (pit != newPathList->End())
  {
    PathPointerType newPath = PathType::New();

    for (PathType::VertexListConstIteratorType vit = pit.Get()->GetVertexList()->Begin();
         vit!=pit.Get()->GetVertexList()->End(); ++vit)
    {
      PathType::VertexType newVertex = vit.Value();
      newVertex[0] += m_Region.GetIndex()[0];
      newVertex[1] += m_Region.GetIndex()[1];
      newPath->AddVertex(newVertex);
    }
    if (newPath->GetVertexList()->Size()>0)
    {
      m_PathList->PushBack(newPath);
    }
    ++pit;
  }
  /*
        std::cout<<"Image: " <<m_MultiChannelExtractROIFilter->GetOutput()->GetBufferedRegion()<<std::endl;
        std::cout<<"Reference pixel: "<<m_ReferencePixel<<std::endl;
        std::cout<<"Region: "<<m_Region<<std::endl;
        std::cout<<"newPathList->Size(): "<<newPathList->Size()<<std::endl;
        std::cout<<"Extraction done. Number of path: "<<m_PathList->Size()<<std::endl;
        std::cout<<"RE: "<< m_GenericRoadExtractionFilter <<std::endl;
  */
  this->NotifyAll();


}

void RoadExtractionModel::ClearPathList()
{
  m_PathList->Clear();
  this->NotifyAll();
}


void RoadExtractionModel::SaveResults(std::string filename)
{

  VectorDataPointerType vectorData = VectorDataType::New();

  DataNodePointerType document = DataNodeType::New();
  document->SetNodeType(DOCUMENT);
  document->SetNodeId("Path");
  DataNodePointerType folder = DataNodeType::New();
  folder->SetNodeType(FOLDER);

  DataTreePointerType tree = vectorData->GetDataTree();
  DataNodePointerType root = tree->GetRoot()->Get();

  tree->Add(document, root);
  tree->Add(folder, document);

  if ( m_PathList->Size() == 0)
  {
    itkExceptionMacro(<<"The List of polygons is empty !");
  }

  PathListType::Iterator pit = m_PathList->Begin();

  while (pit != m_PathList->End())
  {
    DataNodePointerType newPolyLine = DataNodeType::New();
    newPolyLine->SetLine(pit.Get());
    tree->Add(newPolyLine, folder);
    ++pit;
  }

  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  ProjectionFilterType::Pointer vectorDataProjection = ProjectionFilterType::New();
  vectorDataProjection->SetInput(vectorData);
  vectorDataProjection->SetInputOrigin(m_Reader->GetOutput()->GetOrigin());
  vectorDataProjection->SetInputSpacing(m_Reader->GetOutput()->GetSpacing());

  std::string projectionRef;
  itk::ExposeMetaData<std::string>(m_Reader->GetOutput()->GetMetaDataDictionary(),
                                   MetaDataKey::ProjectionRefKey, projectionRef );
  vectorDataProjection->SetInputProjectionRef(projectionRef);

  vectorDataProjection->SetInputKeywordList(m_Reader->GetOutput()->GetImageKeywordlist());

  VectorDataFileWriterPointerType writer = VectorDataFileWriterType::New();
  writer->SetInput(vectorDataProjection->GetOutput());
  writer->SetFileName(filename);
  writer->Update();
}

void RoadExtractionModel::ImportSegments(std::string filename)
{
  VectorDataFileReaderPointerType reader = VectorDataFileReaderType::New();
  reader->SetFileName(filename);

  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  ProjectionFilterType::Pointer vectorDataProjection = ProjectionFilterType::New();
  vectorDataProjection->SetInput(reader->GetOutput());

  if (m_Valid)
  {
    vectorDataProjection->SetOutputOrigin(m_Reader->GetOutput()->GetOrigin());
    vectorDataProjection->SetOutputSpacing(m_Reader->GetOutput()->GetSpacing());

    std::string projectionRef;
    itk::ExposeMetaData<std::string>(m_Reader->GetOutput()->GetMetaDataDictionary(),
                                     MetaDataKey::ProjectionRefKey, projectionRef );
    vectorDataProjection->SetOutputProjectionRef(projectionRef);

    vectorDataProjection->SetOutputKeywordList(m_Reader->GetOutput()->GetImageKeywordlist());
  }

  vectorDataProjection->Update();

  TreeIteratorType it(vectorDataProjection->GetOutput()->GetDataTree());

  it.GoToBegin();

  while (!it.IsAtEnd())
  {
    if (it.Get()->IsLineFeature())
    {
      m_PathList->PushBack(it.Get()->GetLine());
    }
    ++it;
  }
  this->NotifyAll();
}

}
