/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRoadExtractionModel_h
#define __otbRoadExtractionModel_h

#include "otbMVCModel.h"

#include "otbImage.h"
#include "otbImageFileReader.h"
#include "otbMultiToMonoChannelExtractROI.h"
#include "otbMultiChannelExtractROI.h"
#include "otbExtractROI.h"
#include "otbVectorImageToIntensityImageFilter.h"
#include "otbPolyLineParametricPathWithValue.h"
#include "otbObjectList.h"
#include "otbVectorDataFileWriter.h"
#include "otbGenericRoadExtractionFilter.h"
#include "otbSpectralAngleDistanceImageFilter.h"
#include "otbVegetationIndicesFunctor.h"
#include "itkShiftScaleImageFilter.h"
#include "otbMultiChannelRAndNIRIndexImageFilter.h"

#include "otbListenerBase.h"


#include "otbVectorData.h"
#include "otbVectorDataFileWriter.h"
#include "otbVectorDataFileReader.h"
#include "itkPreOrderTreeIterator.h"


namespace otb
{

typedef enum {SPECTRAL_ANGLE, INTENSITY, CHANNEL, WATER_INDEX} RoadExtractionInputType;

/** \class RoadExtractionModel
 *
 *
 *
 */
class ITK_EXPORT RoadExtractionModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef RoadExtractionModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                              Pointer;
  typedef itk::SmartPointer<const Self>                        ConstPointer;

  /** Standard type macro */
  itkTypeMacro(RoadExtractionModel, Object);

  typedef double                                                     PixelType;
  typedef double                                                     LabelType;
  typedef otb::VectorImage<PixelType, 2>                              VectorImageType;
  typedef VectorImageType::Pointer                                   VectorImagePointerType;
  typedef VectorImageType::PixelType                                 VectorPixelType;
  typedef VectorImageType::SizeType                                  SizeType;
  typedef VectorImageType::IndexType                                 IndexType;
  typedef VectorImageType::RegionType                                RegionType;
  typedef Image<PixelType, 2>                                         ImageType;
  typedef ImageType::Pointer                                         ImagePointerType;
  typedef ImageFileReader<VectorImageType>                           ReaderType;
  typedef ReaderType::Pointer                                        ReaderPointerType;
  typedef VectorImageToIntensityImageFilter<VectorImageType, ImageType> IntensityFilterType;
  typedef itk::ShiftScaleImageFilter<ImageType, ImageType>              ShiftScaleFilterType;
  typedef Functor::NDVI< PixelType, PixelType, PixelType >                                                  NDVIFunctorType;
  typedef MultiChannelRAndNIRIndexImageFilter<VectorImageType, ImageType, NDVIFunctorType>     WaterIndexFilterType;
  typedef IntensityFilterType::Pointer                               IntensityFilterPointerType;
  typedef WaterIndexFilterType::Pointer                              WaterIndexFilterPointerType;
  typedef ShiftScaleFilterType::Pointer                              ShiftScaleFilterPointerType;
  typedef ExtractROI<PixelType, PixelType>                            ExtractROIFilterType;
  typedef ExtractROIFilterType::Pointer                              ExtractROIFilterPointerType;
  typedef MultiToMonoChannelExtractROI<PixelType, PixelType>          MultiToMonoChannelExtractROIFilterType;
  typedef MultiToMonoChannelExtractROIFilterType::Pointer            MultiToMonoChannelExtractROIFilterPointerType;
  typedef MultiChannelExtractROI<PixelType, PixelType>                MultiChannelExtractROIFilterType;
  typedef MultiChannelExtractROIFilterType::Pointer                  MultiChannelExtractROIFilterPointerType;
  typedef PolyLineParametricPathWithValue<double>                                       PathType;
  typedef PathType::Pointer                                       PathPointerType;
  typedef PathType::ContinuousIndexType                           ContinuousIndexType;
  typedef ObjectList<PathType>                                    PathListType;
  typedef PathListType::Pointer                                   PathListPointerType;


  typedef SpectralAngleDistanceImageFilter<
  VectorImageType,
  ImageType>                      SpectralAngleDistanceImageFilterType;
  typedef SpectralAngleDistanceImageFilterType::Pointer SpectralAngleDistanceImageFilterPointerType;
  typedef GenericRoadExtractionFilter<ImageType, PathType>
  GenericRoadExtractionFilterType;
  typedef GenericRoadExtractionFilterType::AmplitudeThresholdType AmplitudeThresholdType;
  typedef GenericRoadExtractionFilterType::ToleranceType ToleranceType;
  typedef GenericRoadExtractionFilterType::MaxAngleType MaxAngleType;
  typedef GenericRoadExtractionFilterType::LinkRealType LinkRealType;
  typedef GenericRoadExtractionFilterType::MeanDistanceThresholdType MeanDistanceThresholdType;
  typedef GenericRoadExtractionFilterType::Pointer GenericRoadExtractionFilterPointerType;

  /** Vector data handling */
  typedef VectorData<double, 2>                                       VectorDataType;
  typedef VectorDataType::Pointer                                    VectorDataPointerType;
  typedef VectorDataFileWriter<VectorDataType>                       VectorDataFileWriterType;
  typedef VectorDataFileWriterType::Pointer                          VectorDataFileWriterPointerType;
  typedef VectorDataType::DataNodeType                               DataNodeType;
  typedef DataNodeType::Pointer                                      DataNodePointerType;
  typedef VectorDataType::DataTreeType                               DataTreeType;
  typedef DataTreeType::Pointer                                      DataTreePointerType;
  typedef VectorDataFileReader<VectorDataType>                       VectorDataFileReaderType;
  typedef VectorDataFileReaderType::Pointer                          VectorDataFileReaderPointerType;
  typedef itk::PreOrderTreeIterator<DataTreeType>                    TreeIteratorType;

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Setters */
  itkGetConstReferenceMacro(Region, RegionType);

  /** Getters */
  itkGetMacro(Valid, bool);
  itkGetObjectMacro(PathList, PathListType);
  itkGetStringMacro(FileName);
  itkGetMacro(Region, RegionType);

  itkGetMacro(ReferencePixel, VectorPixelType);
  itkGetMacro(RoadExtractionInput, RoadExtractionInputType);
  itkSetMacro(RoadExtractionInput, RoadExtractionInputType);
  itkGetMacro(GenericRoadExtractionFilter, GenericRoadExtractionFilterPointerType);

  itkGetMacro(ChannelIndex, unsigned int);
  itkSetMacro(ChannelIndex, unsigned int);
  itkGetMacro(RedChannelIndex, unsigned int);
  itkSetMacro(RedChannelIndex, unsigned int);
  itkGetMacro(GreenChannelIndex, unsigned int);
  itkSetMacro(GreenChannelIndex, unsigned int);

  itkGetMacro(Reader, ReaderPointerType);
  itkSetMacro(Reader, ReaderPointerType);
  //Todo UseSpectralAngle

  itkGetMacro(IsNewImage, bool);
  itkSetMacro(IsNewImage, bool);

  /** Get the image */
  VectorImageType * GetImage();

  /** Set the reference pixel */
  void SetReferencePixel(const VectorPixelType & reference);

  /** Set the reference pixel */
  void SetRegion(const RegionType & region);

  /** Get the spectral angle image */
  ImageType * GetSpectralAngleImage();

  /** Set up the road extraction pipeline */
  void SetupRoadExtraction();

  /** Get the filename */
  void SetFileName(std::string file);

  /** Clear the Path list */
  void ClearPathList();

  /** retrieve the road extraction results */
  void GenerateData();

  /** Save segmentation result to filename */
  void SaveResults(std::string filename);

  /** Import a shapefile into the segmented regions */
  void ImportSegments(std::string filename);

  /** Set the GenericRoadExtractionFilter paramters*/
  void SetFilterParameters();

  /** Set the GenericRoadExtraction parameters */
  itkSetMacro(Alpha, double);
  itkSetMacro(AmplitudeThreshold, AmplitudeThresholdType);
  itkSetMacro(Tolerance, ToleranceType);
  itkSetMacro(Resolution, double);
  itkSetMacro(MaxAngle, MaxAngleType);
  itkSetMacro(FirstMeanDistanceThreshold, MeanDistanceThresholdType);
  itkSetMacro(SecondMeanDistanceThreshold, MeanDistanceThresholdType);

  /** Get/Set the angular threshold (use by LinkPathFilter)*/
  itkSetMacro(AngularThreshold, LinkRealType);
  itkGetMacro(AngularThreshold, LinkRealType);
  /** Get/Set the distance threshold (use by LinkPathFilter)*/
  itkSetMacro(DistanceThreshold, LinkRealType);
  itkGetMacro(DistanceThreshold, LinkRealType);


protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  RoadExtractionModel();
  /** Destructor */
  ~RoadExtractionModel();

private:
  RoadExtractionModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented


  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Filters */
  ReaderPointerType                             m_Reader;
  WaterIndexFilterPointerType                   m_WaterIndexFilter;
  ShiftScaleFilterPointerType                   m_ShiftScaleFilter;
  IntensityFilterPointerType                    m_IntensityFilter;
  ExtractROIFilterPointerType                   m_ExtractROIFilter;
  MultiToMonoChannelExtractROIFilterPointerType m_MultiToMonoChannelExtractROIFilter;
  MultiChannelExtractROIFilterPointerType       m_MultiChannelExtractROIFilter;
  SpectralAngleDistanceImageFilterPointerType   m_SpectralAngleFilter;
  GenericRoadExtractionFilterPointerType        m_GenericRoadExtractionFilter;

  /** Whether the input image is loaded or not */
  bool m_Valid;
  bool m_IsNewImage;
  bool m_ReferencePixelValid;
  bool m_RegionValid;


  /** The instance singleton */
  static Pointer Instance;

  /** Parameters */
  std::string             m_FileName;
  RegionType              m_Region;
  unsigned int            m_ChannelIndex;
  unsigned int            m_RedChannelIndex;
  unsigned int            m_GreenChannelIndex;
  VectorPixelType         m_ReferencePixel;
  RoadExtractionInputType m_RoadExtractionInput;

  /** Road extraction parameters */
  AmplitudeThresholdType m_AmplitudeThreshold;
  ToleranceType m_Tolerance;
  MaxAngleType m_MaxAngle;
  MeanDistanceThresholdType m_FirstMeanDistanceThreshold;
  MeanDistanceThresholdType m_SecondMeanDistanceThreshold;
  LinkRealType m_AngularThreshold;
  double m_DistanceThreshold;
  double m_Alpha;
  double m_Resolution;

  /** Paths */
  PathListPointerType        m_PathList;
};
}
#endif
