/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbRoadExtractionViewGUI.h"
#include <FL/Fl_File_Chooser.H>
#include "otbMacro.h"
#include "otbVisualizationMacro.h"


namespace otb
{

RoadExtractionViewGUI::RoadExtractionViewGUI()
{
  m_RoadExtractionModel = RoadExtractionModel::GetInstance();
  m_RoadExtractionModel->RegisterListener(this);

  m_Color[0]=1.0;
  m_Color[1]=0.;
  m_Color[2]=0.;
  m_Color[3]=0.75;

  this->Build();

  switch (m_RoadExtractionModel->GetRoadExtractionInput())
  {
  case SPECTRAL_ANGLE:
    bUseImageIntensity->clear();
    bUseImageChannel->clear();
    bUseSpectralAngle->set();
    bUseWaterIndex->clear();
    break;
  case INTENSITY:
    bUseImageIntensity->set();
    bUseImageChannel->clear();
    bUseSpectralAngle->clear();
    bUseWaterIndex->clear();
    break;
  case CHANNEL:
    bUseImageIntensity->clear();
    bUseImageChannel->set();
    bUseSpectralAngle->clear();
    bUseWaterIndex->clear();
    break;
  case WATER_INDEX:
    bUseImageIntensity->clear();
    bUseImageChannel->clear();
    bUseSpectralAngle->clear();
    bUseWaterIndex->set();
    break;
  }

  /* Init the IHM values */
  this->InitViewValues();

  /* Start running */
  this->StartRunning();
  wMainWindow->show();
}

void RoadExtractionViewGUI::SetRoadExtractionController(RoadExtractionControllerInterface * controller)
{
  m_RoadExtractionController = controller;
}

void RoadExtractionViewGUI::RefreshViewer()
{
  if (m_RoadExtractionModel->GetIsNewImage() && !m_Viewer.IsNull())
  {
    if (m_Viewer->GetUseScroll())
      gScroll->remove(m_Viewer->GetScrollWidget());
    gFullResolution->remove(m_Viewer->GetFullWidget());
  }

  if (m_Viewer.IsNull() || (m_RoadExtractionModel->GetIsNewImage()))
  {
    m_Viewer = ImageViewerBaseType::New();
    m_Viewer->SetImage(m_RoadExtractionModel->GetImage());
    if (m_RoadExtractionModel->GetImage()->GetNumberOfComponentsPerPixel()>3)
    {
      m_Viewer->SetRedChannelIndex(2);
      m_Viewer->SetGreenChannelIndex(1);
      m_Viewer->SetBlueChannelIndex(0);
    }

    unsigned int maxfullsize = std::min(gFullResolution->w(), gFullResolution->h());
    unsigned int maxscrollsize = std::min(gScroll->w(), gScroll->h());
    m_Viewer->SetFullMaxInitialSize(maxfullsize);
    m_Viewer->SetScrollMaxInitialSize(maxscrollsize);
    m_Viewer->SetShowZoomWidget(false);
    m_Viewer->SetShowHistograms(false);
    m_Viewer->SetPathList(m_RoadExtractionModel->GetPathList());
    m_Viewer->SetDefaultROIColor(m_Color);
    /*m_Viewer->SetLineWidth(5.0); */

    m_Viewer->Build();
    m_Viewer->GetFullWidget()->SetEventsInterface(m_RoadExtractionController);
    m_Viewer->GetFullWidget()->position(gFullResolution->x(), gFullResolution->y());
    m_Viewer->GetFullWidget()->resize(gFullResolution->x(), gFullResolution->y(), gFullResolution->w(), gFullResolution->h());
    gFullResolution->add(m_Viewer->GetFullWidget());
    gFullResolution->box(FL_NO_BOX);
    if (m_Viewer->GetUseScroll())
    {
      m_Viewer->GetScrollWidget()->position(gScroll->x(), gScroll->y());
      m_Viewer->GetScrollWidget()->resize(gScroll->x(), gScroll->y(), gScroll->w(), gScroll->h());
      gScroll->add(m_Viewer->GetScrollWidget());
      gScroll->show();
      gScroll->box(FL_NO_BOX);
    }
    else
    {
      gScroll->hide();
    }
  }

  double alpha = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetAlpha();
  m_Viewer->SetLineWidth(alpha*4.0);
  m_Viewer->Update();
  m_Viewer->Show();

}


void RoadExtractionViewGUI::Notify()
{
  if (m_RoadExtractionModel->GetValid())
  {
    this->RefreshViewer();
    this->UpdateInterface();
  }
}

void RoadExtractionViewGUI::OpenImageCallback()
{
  const char * filename = NULL;

  filename = fl_file_chooser("Pick an image file", "*.*",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_RoadExtractionController->SelectFileName(filename);
  this->ClearCallback();

}
void RoadExtractionViewGUI::QuitCallback()
{
  this->StopRunning();
}

void RoadExtractionViewGUI::UseSpectralAngleCallback()
{
  m_RoadExtractionController->ChangeUseSpectralAngle();
}

void RoadExtractionViewGUI::UseImageIntensityCallback()
{
  m_RoadExtractionController->ChangeUseImageIntensity();
}

void RoadExtractionViewGUI::UseImageChannelCallback()
{
  m_RoadExtractionController->ChangeUseImageChannel();
}

void RoadExtractionViewGUI::UseWaterIndexCallback()
{
  m_RoadExtractionController->ChangeUseWaterIndex();
}

void RoadExtractionViewGUI::InputAlphaCallback()
{
  m_RoadExtractionController->ChangeAlpha(iAlpha->value());
}

void RoadExtractionViewGUI::InputResolutionCallback()
{
  m_RoadExtractionController->ChangeResolution(iResolution->value());
}

void RoadExtractionViewGUI::InputToleranceCallback()
{
  m_RoadExtractionController->ChangeTolerance(iTolerance->value());
}

void RoadExtractionViewGUI::InputMaxAngleCallback()
{
  m_RoadExtractionController->ChangeMaxAngle(iMaxAngle->value());
}

void RoadExtractionViewGUI::InputAngularThresholdCallback()
{
  m_RoadExtractionController->ChangeAngularThreshold(iAngularThreshold->value());
}

void RoadExtractionViewGUI::InputAmplitudeThresholdCallback()
{
  m_RoadExtractionController->ChangeAmplitudeThreshold(iAmplitudeThreshold->value());
}

void RoadExtractionViewGUI::InputDistanceThresholdCallback()
{
  m_RoadExtractionController->ChangeDistanceThreshold(iDistanceThreshold->value());
}

void RoadExtractionViewGUI::InputFirstMeanDistanceThresholdCallback()
{
  m_RoadExtractionController->ChangeFirstMeanDistanceThreshold(iFirstMeanDistanceThreshold->value());
}

void RoadExtractionViewGUI::InputSecondMeanDistanceThresholdCallback()
{
  m_RoadExtractionController->ChangeSecondMeanDistanceThreshold(iSecondMeanDistanceThreshold->value());
}

void RoadExtractionViewGUI::ChannelCallback()
{
  m_RoadExtractionController->ChangeChannel(vChannel->value());
}

void RoadExtractionViewGUI::GreenChannelCallback()
{
  m_RoadExtractionController->ChangeGreenChannel(vGreenChannel->value());
}

void RoadExtractionViewGUI::RedChannelCallback()
{
  m_RoadExtractionController->ChangeRedChannel(vRedChannel->value());
}

void RoadExtractionViewGUI::ClearCallback()
{
  m_RoadExtractionController->ClearSegments();
}

void RoadExtractionViewGUI::SaveResultsCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output data file", "*.shp\t*.kml",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_RoadExtractionController->SaveResults(filename);
}

void RoadExtractionViewGUI::ImportSegmentsCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an input data file", "*.shp\t*.kml",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_RoadExtractionController->ImportSegments(filename);
}

void RoadExtractionViewGUI::InitViewValues()
{
  double alpha = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetAlpha();
  double resolution = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetResolution();
  double tolerance = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetTolerance();
  double maxAngle = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetMaxAngle();
  double angularThreshold = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetAngularThreshold();
  double amplitudeThreshold = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetAmplitudeThreshold();
  double distanceThreshold = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetDistanceThreshold();
  double firstMeanDistanceThreshold = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetFirstMeanDistanceThreshold();
  double secondMeanDistanceThreshold = m_RoadExtractionModel->GetGenericRoadExtractionFilter()->GetSecondMeanDistanceThreshold();

  iAlpha->value(alpha);
  iResolution->value(resolution);
  iTolerance->value(tolerance);
  iMaxAngle->value(maxAngle);
  iAngularThreshold->value(angularThreshold);
  iAmplitudeThreshold->value(amplitudeThreshold);
  iDistanceThreshold->value(distanceThreshold);
  iFirstMeanDistanceThreshold->value(firstMeanDistanceThreshold);
  iSecondMeanDistanceThreshold->value(secondMeanDistanceThreshold);

}

void RoadExtractionViewGUI::RunCallback()
{
  m_RoadExtractionController->Run();
}

void RoadExtractionViewGUI::UpdateInterface()
{

  if ( !m_RoadExtractionModel->GetReader().IsNull() )
  {
    int nbOfChannels = m_RoadExtractionModel->GetImage()->GetNumberOfComponentsPerPixel();
    itk::OStringStream oss;

    vChannel->clear();
    vRedChannel->clear();
    vGreenChannel->clear();
    for (int i=1; i<=nbOfChannels; i++)
    {
      oss.str("");
      oss << "Channel "<<i;
      vChannel->add(oss.str().c_str());
      vRedChannel->add(oss.str().c_str());
      vGreenChannel->add(oss.str().c_str());
    }

    if ( nbOfChannels == 1 )
    {
      bUseSpectralAngle->deactivate();
      vReferencePixel->deactivate();
      bUseWaterIndex->deactivate();
      bUseImageChannel->deactivate();
      m_RoadExtractionController->ChangeUseImageIntensity();
    }
    else
    {
      bUseSpectralAngle->activate();
      vReferencePixel->activate();
      bUseWaterIndex->activate();
      vRedChannel->value(2);
      vGreenChannel->value(1);
      bUseImageChannel->activate();
      this->GreenChannelCallback();
      this->RedChannelCallback();
    }

    vChannel->value(0);
    this->ChannelCallback();
    vChannel->redraw();
    vRedChannel->redraw();
    vGreenChannel->redraw();

  }
}

}
