/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRoadExtractionViewer_h
#define __otbRoadExtractionViewer_h

#include "otbRoadExtractionViewerGUI.h"
#include "otbRoadExtractionViewerFullWidget.h"
#include "otbRoadExtractionViewerScrollWidget.h"
#include "otbImageWidgetBoxForm.h"
#include "otbObjectList.h"
#include "itkObject.h"

namespace otb
{
/** \class RoadExtractionViewer
 *  \brief This class is the main class of the road extraction application.
 */
template <class TPixel>
class ITK_EXPORT RoadExtractionViewer
      : public otbRoadExtractionViewerGUI, public itk::Object
{
public:
  /// Standard typedefs
  typedef RoadExtractionViewer                       Self;
  typedef otbRoadExtractionViewerGUI                    Superclass;
  typedef itk::SmartPointer<Self>                    Pointer;
  typedef itk::SmartPointer<const Self>              ConstPointer;

  /// Type macro
  itkNewMacro(Self);

  /// Creation through object factory macro
  itkTypeMacro(RoadExtractionViewer, otbRoadExtractionViewerGUI);

  /// Standards typedef
  typedef TPixel PixelType;


  /// Widget definitions
  typedef RoadExtractionViewerFullWidget<PixelType> FullWidgetType;
  typedef RoadExtractionViewerScrollWidget<PixelType> ScrollWidgetType;
  typedef typename FullWidgetType::Pointer FullWidgetPointerType;
  typedef typename ScrollWidgetType::Pointer ScrollWidgetPointerType;
  typedef typename ScrollWidgetType::TransferFunctionListType TransferFunctionListType;
  typedef typename TransferFunctionListType::Pointer TransferFunctionListPointerType;
  typedef typename ScrollWidgetType::AffineTransferFunctionType AffineTransferFunctionType;

  typedef typename FullWidgetType::ImageType ImageType;
  typedef typename FullWidgetType::OverlayImageType OverlayImageType;
  typedef typename OverlayImageType::Pointer OverlayImagePointerType;
  typedef typename ImageType::Pointer ImagePointerType;
  typedef typename ImageType::IndexType IndexType;
  typedef typename ImageType::SizeType SizeType;
  typedef typename ImageType::RegionType RegionType;

  /// Definition of box form to represent displayed regions.
  typedef otb::ImageWidgetBoxForm BoxType;
  typedef typename BoxType::Pointer BoxPointerType;
  typedef typename BoxType::ColorType ColorType;

  /// Set the input image
  itkSetObjectMacro(InputImage, ImageType);
  itkGetObjectMacro(InputImage, ImageType);

  /// Set the Alpha parameter
  itkSetMacro(Alpha, double);
  itkGetMacro(Alpha, double);

  /// Build the GUI
  virtual void Build();
  /// Show the GUI
  virtual void Show();

  /// Interactions from the widgets
  /// Select a road
  virtual void RoadSelected(IndexType index);
  /// Change the viewed region
  virtual void ChangeFullViewedRegion(IndexType index);

protected:
  /// Constructor
  RoadExtractionViewer();
  /// Destructor
  virtual ~RoadExtractionViewer() {};
  /// Build road mask
  virtual void BuildRoadMask(IndexType index);
  /// Clear the road mask
  virtual void ClearRoadMask(void);

  /// Sub widget creation
  /// Create the full resolution widget
  virtual void CreateFullResolutionWindow(Fl_Gl_Window * w);
  /// Create the scroll the scroll widget
  virtual void CreateScrollWindow(Fl_Gl_Window * w);

  ///
  virtual void Update();
  /// Update the scroll widget
  virtual void UpdateScrollWidget(void);
  /// Compute the constrained region of region smallRegion by region bigRegion
  virtual RegionType ComputeConstrainedRegion(RegionType smallRegion, RegionType bigRegion);

  /// Hook for the Quit button
  virtual void Quit();
  /// Hook for the erase button
  virtual void Erase();

  /// Compute the normalization factors
  virtual void ComputeNormalizationFactors();

private:
  RoadExtractionViewer(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /// The input image
  ImagePointerType m_InputImage;
  /// The road mask
  OverlayImagePointerType m_RoadMask;

  /// The full resolution widget
  FullWidgetPointerType m_FullWidget;
  /// The scroll widget
  ScrollWidgetPointerType m_ScrollWidget;


  /// parameters for the road detection algo
  double m_Alpha;
  double m_AmplitudeThreshold;
  double m_Tolerance;
  double m_MaxAngle;
  double m_AngularThreshold;
  double m_FirstMeanDistanceThreshold;
  double m_SecondMeanDistanceThreshold;
  double m_DistanceThreshold;
  double m_NormalizationFactor;
  TransferFunctionListPointerType m_TransferFunctionList;


  /// True if the GUI has been built
  bool m_Built;
};
}// End namespace otb
#ifndef OTB_MANUAL_INSTANTIATION
#include "otbRoadExtractionViewer.txx"
#endif

#endif
