/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRoadExtractionViewer_txx
#define __otbRoadExtractionViewer_txx

#include "otbRoadExtractionViewer.h"

#include <FL/Fl.H>

#include "otbPolyLineParametricPathWithValue.h"
#include "otbRoadExtractionFilter.h"
#include "otbDrawPathListFilter.h"
#include "itkRescaleIntensityImageFilter.h"
#include "itkImageRegionIterator.h"
#include "itkListSampleToHistogramGenerator.h"
#include "itkListSample.h"
#include "itkGrayscaleDilateImageFilter.h"
#include "itkBinaryBallStructuringElement.h"

namespace otb
{

template <class TImage>
RoadExtractionViewer<TImage>
::RoadExtractionViewer()
{
  m_Alpha = 1.2;
  m_AmplitudeThreshold = 0.00005;
  m_Tolerance = 1.0;
  m_MaxAngle = 0.39269;
  m_AngularThreshold = 0.39269;
  m_FirstMeanDistanceThreshold =1.0;
  m_SecondMeanDistanceThreshold =10.0;
  m_DistanceThreshold =25.;
  m_NormalizationFactor = 0.02;
  m_Built = false;
  m_TransferFunctionList = TransferFunctionListType::New();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::Update(void)
{
  UpdateScrollWidget();
  m_FullWidget->redraw();
  Fl::check();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::CreateFullResolutionWindow(Fl_Gl_Window * original)
{
  int x = original->x();
  int y = original->y();
  int w = original->w();
  int h = original->h();
  const char * label = original->label();
  m_FullWidget = FullWidgetType::New();
  m_FullWidget->SetInput(m_InputImage);
  m_FullWidget->SetInputOverlay(m_RoadMask);
  m_FullWidget->Init(x, y, w, h, label);
  m_FullWidget->box( FL_EMBOSSED_BOX );
  m_FullWidget->SetImageOverlayVisible(true);
  m_FullWidget->SetInputOverlay(m_RoadMask);
  m_FullWidget->SetImageOverlayOpacity(255);
  m_FullWidget->SetBlackTransparency(true);
  m_FullWidget->SetParent(this);
  m_FullWidget->resizable(winMain);
  m_FullWidget->SetRedChannelIndex(2);
  m_FullWidget->SetGreenChannelIndex(1);
  m_FullWidget->SetBlueChannelIndex(0);
  winMain->remove(original);
  delete original;
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::CreateScrollWindow(Fl_Gl_Window * original )
{
  int x = original->x();
  int y = original->y();
  int w = original->w();
  int h = original->h();
  const char * label = original->label();
  m_ScrollWidget = ScrollWidgetType::New();
  m_ScrollWidget->SetInput(m_InputImage);
  m_ScrollWidget->Init(x, y, w, h, label);
  m_ScrollWidget->box( FL_EMBOSSED_BOX );
  m_ScrollWidget->SetFormOverlayVisible(true);
  m_ScrollWidget->SetImageOverlayVisible(true);
  m_ScrollWidget->SetBlackTransparency(true);
  m_ScrollWidget->SetInputOverlay(m_RoadMask);
  m_ScrollWidget->SetImageOverlayOpacity(255);
  m_ScrollWidget->SetParent(this);
  m_ScrollWidget->SetRedChannelIndex(2);
  m_ScrollWidget->SetGreenChannelIndex(1);
  m_ScrollWidget->SetBlueChannelIndex(0);
  winMain->remove(original);
  delete original;
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::Show(void)
{
  if (!m_Built)
    this->Build();
  winMain->show();
  m_ScrollWidget->show();
  m_FullWidget->Show();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::Build(void)
{
  ClearRoadMask();
  CreateGUI();
  ComputeNormalizationFactors();
  BoxPointerType box = BoxType::New();
  SizeType scrollBoxSize;
  IndexType scrollBoxIndex;
  scrollBoxSize[0]=(m_FullWidget->GetViewedRegion().GetSize()[0])-1;
  scrollBoxSize[1]=(m_FullWidget->GetViewedRegion().GetSize()[1])-1;
  scrollBoxIndex[0]=(m_FullWidget->GetViewedRegion().GetIndex()[0])+1;
  scrollBoxIndex[1]=(m_FullWidget->GetViewedRegion().GetIndex()[1])+1;
  box->SetSize(scrollBoxSize);
  box->SetIndex(scrollBoxIndex);
  ColorType color;
  color.Fill(0);
  color[0]=1;
  color[3]=1;
  box->SetColor(color);
  m_ScrollWidget->GetFormList()->PushBack(box);
  m_Built = true;
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::ClearRoadMask(void)
{
  typename ImageType::PixelType black;
  black.SetSize(this->GetInputImage()->GetNumberOfComponentsPerPixel());
  black.Fill(0);
  m_RoadMask = OverlayImageType::New();
  m_RoadMask->SetRegions(this->GetInputImage()->GetLargestPossibleRegion());
  m_RoadMask->SetNumberOfComponentsPerPixel(this->GetInputImage()->GetNumberOfComponentsPerPixel());
  m_RoadMask->Allocate();
  m_RoadMask->FillBuffer(black);
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::Quit(void)
{
  exit(0);
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::Erase(void)
{
  ClearRoadMask();
  m_ScrollWidget->SetInputOverlay(m_RoadMask);
  m_FullWidget->SetInputOverlay(m_RoadMask);
  Update();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::RoadSelected(IndexType index)
{

  BuildRoadMask(index);
  m_ScrollWidget->SetInputOverlay(m_RoadMask);
  m_FullWidget->SetInputOverlay(m_RoadMask);
  Update();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::ChangeFullViewedRegion(IndexType index)
{
  RegionType region = m_FullWidget->GetViewedRegion();
  IndexType newIndex;
  newIndex[0]=index[0]-region.GetSize()[0]/2;
  newIndex[1]=index[1]-region.GetSize()[1]/2;
  region.SetIndex(newIndex);
  RegionType newRegion = ComputeConstrainedRegion(region, m_InputImage->GetLargestPossibleRegion());
  m_FullWidget->SetUpperLeftCorner(newRegion.GetIndex());
  this->Update();
}

template <class TImage>
typename RoadExtractionViewer<TImage>
::RegionType
RoadExtractionViewer<TImage>
::ComputeConstrainedRegion(RegionType smallRegion, RegionType bigRegion)
{
  RegionType resp;
// This function assumes that smallRegion is inside huge region
  if (smallRegion.GetSize()[0]>bigRegion.GetSize()[0]
      ||smallRegion.GetSize()[1]>bigRegion.GetSize()[1])
  {
    itkExceptionMacro("Small region not inside big region !");
  }
  else
  {
    IndexType index = smallRegion.GetIndex();
    SizeType size = smallRegion.GetSize();

    if (smallRegion.GetIndex()[0]<bigRegion.GetIndex()[0])
    {
      index[0]=bigRegion.GetIndex()[0];
    }
    if (smallRegion.GetIndex()[1]<bigRegion.GetIndex()[1])
    {
      index[1]=bigRegion.GetIndex()[1];
    }
    if (index[0]+size[0]>=bigRegion.GetIndex()[0]+bigRegion.GetSize()[0])
    {
      index[0]=bigRegion.GetIndex()[0]+bigRegion.GetSize()[0]-size[0]-2;
    }
    if (index[1]+size[1]>=bigRegion.GetIndex()[1]+bigRegion.GetSize()[1])
    {
      index[1]=bigRegion.GetIndex()[1]+bigRegion.GetSize()[1]-size[1]-2;
    }
    resp.SetSize(size);
    resp.SetIndex(index);
  }
  return resp;
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::UpdateScrollWidget(void)
{
  BoxPointerType box = BoxType::New();
  SizeType scrollBoxSize;
  IndexType scrollBoxIndex;
  scrollBoxSize[0]=(m_FullWidget->GetViewedRegion().GetSize()[0])-1;
  scrollBoxSize[1]=(m_FullWidget->GetViewedRegion().GetSize()[1])-1;
  scrollBoxIndex[0]=(m_FullWidget->GetViewedRegion().GetIndex()[0])+1;
  scrollBoxIndex[1]=(m_FullWidget->GetViewedRegion().GetIndex()[1])+1;
  box->SetSize(scrollBoxSize);
  box->SetIndex(scrollBoxIndex);
  ColorType color;
  color.Fill(0);
  color[0]=1;
  color[3]=1;
  box->SetColor(color);
  m_ScrollWidget->GetFormList()->SetNthElement(0, box);
  m_ScrollWidget->redraw();
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::BuildRoadMask(IndexType index)
{
  typedef PolyLineParametricPathWithValue<PixelType, ImageType::ImageDimension> PathType;
  typedef RoadExtractionFilter<ImageType, PathType> RoadExtractionFilterType;
  typedef Image<PixelType, ImageType::ImageDimension> SingleImageType;
  typedef DrawPathListFilter<SingleImageType, PathType, SingleImageType> DrawPathFilterType;
  typedef itk::RescaleIntensityImageFilter<SingleImageType, SingleImageType> RescalerType;
  typedef itk::BinaryBallStructuringElement<PixelType, ImageType::ImageDimension> StructuringElementType;
  typedef itk::GrayscaleDilateImageFilter<SingleImageType, SingleImageType, StructuringElementType> DilateFilterType;

  typename RoadExtractionFilterType::Pointer roadExtractionFilter = RoadExtractionFilterType::New();
  typename DrawPathFilterType::Pointer drawingFilter = DrawPathFilterType::New();
  typename RescalerType::Pointer rescaleFilter = RescalerType::New();
  typename DilateFilterType::Pointer dilater = DilateFilterType::New();

  roadExtractionFilter->SetReferencePixel(this->GetInputImage()->GetPixel(index));
  roadExtractionFilter->SetAlpha(m_Alpha);
  roadExtractionFilter->SetAmplitudeThreshold(m_AmplitudeThreshold);
  roadExtractionFilter->SetTolerance(m_Tolerance);
  roadExtractionFilter->SetMaxAngle(m_MaxAngle);
  roadExtractionFilter->SetAngularThreshold(m_AngularThreshold);
  roadExtractionFilter->SetFirstMeanDistanceThreshold(m_FirstMeanDistanceThreshold);
  roadExtractionFilter->SetSecondMeanDistanceThreshold(m_SecondMeanDistanceThreshold);
  roadExtractionFilter->SetDistanceThreshold(m_DistanceThreshold);

  typename SingleImageType::Pointer singleRoadMap = SingleImageType::New();
  singleRoadMap->SetRegions(this->GetInputImage()->GetLargestPossibleRegion());
  singleRoadMap->Allocate();
  singleRoadMap->FillBuffer(0);

  drawingFilter->UseInternalPathValueOn();
  rescaleFilter->SetOutputMinimum(0);
  rescaleFilter->SetOutputMaximum(1);
  roadExtractionFilter->SetInput(this->GetInputImage());
  drawingFilter->SetInput(singleRoadMap);
  drawingFilter->SetInputPath(roadExtractionFilter->GetOutput());
  rescaleFilter->SetInput(drawingFilter->GetOutput());
  StructuringElementType se;
  se.SetRadius(1);
  se.CreateStructuringElement();
  dilater->SetInput(rescaleFilter->GetOutput());
  dilater->SetKernel(se);
  dilater->Update();

  if (!bKeep->value())
  {
    ClearRoadMask();
  }

  typedef itk::ImageRegionIterator<OverlayImageType> VectorIteratorType;
  typedef itk::ImageRegionIterator<SingleImageType> IteratorType;

  VectorIteratorType vit(m_RoadMask, m_RoadMask->GetLargestPossibleRegion());
  IteratorType it(dilater->GetOutput(), dilater->GetOutput()->GetLargestPossibleRegion());

  vit.GoToBegin();
  it.GoToBegin();

  while (!vit.IsAtEnd()&&!it.IsAtEnd())
  {
    typename ImageType::PixelType pixel = vit.Get();
    if (it.Get()>0)
    {
      pixel[0] = 127+128*it.Get();
      vit.Set(pixel);
    }
    ++it;
    ++vit;
  }
}

template <class TImage>
void
RoadExtractionViewer<TImage>
::ComputeNormalizationFactors(void)
{
  typedef itk::ImageRegionConstIterator< ImageType >  InputIteratorType;
  typedef itk::Vector<typename ImageType::InternalPixelType, 1> MeasurementVectorType;
  typedef itk::Statistics::ListSample<MeasurementVectorType> ListSampleType;
  typedef float HistogramMeasurementType;
  typedef itk::Statistics::ListSampleToHistogramGenerator<ListSampleType, HistogramMeasurementType,
  itk::Statistics::DenseFrequencyContainer, 1> HistogramGeneratorType;
  typedef otb::ObjectList<ListSampleType> ListSampleListType;

  typename ListSampleListType::Pointer sl =  ListSampleListType::New();

  sl->Reserve(this->GetInputImage()->GetNumberOfComponentsPerPixel());

  for (unsigned int i = 0; i<this->GetInputImage()->GetNumberOfComponentsPerPixel(); ++i)
  {
    sl->PushBack(ListSampleType::New());
  }
  InputIteratorType it = InputIteratorType(this->GetInputImage(), this->GetInputImage()->GetRequestedRegion());
  it.GoToBegin();

  while ( !it.IsAtEnd() )
  {
    typename ImageType::PixelType pixel = it.Get();
    for (unsigned int i = 0; i<this->GetInputImage()->GetNumberOfComponentsPerPixel(); ++i)
    {
      MeasurementVectorType measure[0] = pixel[i];
      sl->GetNthElement(i)->PushBack(pixel[i]);
    }
    ++it;
  }

  for (unsigned int i = 0; i<this->GetInputImage()->GetNumberOfComponentsPerPixel(); ++i)
  {
    typename HistogramGeneratorType::Pointer generator = HistogramGeneratorType::New();
    generator->SetListSample(sl->GetNthElement(i));
    typename HistogramGeneratorType::HistogramType::SizeType size;
    size.Fill(static_cast<unsigned int>(vcl_ceil(1/m_NormalizationFactor)*10));
    generator->SetNumberOfBins(size);
    generator->Update();
    m_TransferFunctionList->PushBack(AffineTransferFunctionType::New());
    double min = generator->GetOutput()->Quantile(0, m_NormalizationFactor);
    double max = generator->GetOutput()->Quantile(0, 1-m_NormalizationFactor);
    m_TransferFunctionList->Back()->SetLowerBound(min);
    m_TransferFunctionList->Back()->SetUpperBound(max);
  }
  m_FullWidget->SetTransferFunctionList(m_TransferFunctionList);
  m_ScrollWidget->SetTransferFunctionList(m_TransferFunctionList);
}
} // End namespace otb
#endif
