/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbSegmentationApplicationController.h"
#include "otbPreprocessingController.h"
#include "otbRegionGrowingSegmentationController.h"
#include "otbLevelSetSegmentationController.h"
#include "otbVectorizationController.h"
#include <FL/fl_ask.H>

namespace otb
{

SegmentationApplicationController::SegmentationApplicationController()
{
  m_PreprocessingController             = PreprocessingController::New();
  m_RegionGrowingSegmentationController = RegionGrowingSegmentationController::New();
  m_LevelSetSegmentationController      = LevelSetSegmentationController::New();
  m_SVMBasedRegionGrowingSegmentationController = SVMBasedRegionGrowingSegmentationController::New();
  m_VectorizationController             = VectorizationController::New();
}

SegmentationApplicationController::~SegmentationApplicationController()
{}


void SegmentationApplicationController::SetView(ViewPointerType view)
{
  m_View = view;
  m_PreprocessingController->SetView(view->GetPreprocessingView());
  m_RegionGrowingSegmentationController->SetView(view->GetRegionGrowingView());
  m_LevelSetSegmentationController->SetView(view->GetLevelSetView());
  m_SVMBasedRegionGrowingSegmentationController->SetView(view->GetSVMBasedRegionGrowingView());
  m_VectorizationController->SetView(view->GetVectorizationView());
}

PreprocessingControllerInterface * SegmentationApplicationController:: GetPreprocessingController()
{
  return m_PreprocessingController;
}

RegionGrowingSegmentationControllerInterface * SegmentationApplicationController::GetRegionGrowingSegmentationController()
{
  return m_RegionGrowingSegmentationController;
}

LevelSetSegmentationControllerInterface * SegmentationApplicationController::GetLevelSetSegmentationController()
{
  return m_LevelSetSegmentationController;
}

SVMBasedRegionGrowingSegmentationControllerInterface * SegmentationApplicationController::GetSVMBasedRegionGrowingSegmentationController()
{
  return m_SVMBasedRegionGrowingSegmentationController;
}

VectorizationControllerInterface * SegmentationApplicationController::GetVectorizationController()
{
  return m_VectorizationController;
}


void SegmentationApplicationController::RegionSelected(const RegionType& region)
{
  m_Model->SetRegion(region);
  m_View->bSegment->activate();
}

void SegmentationApplicationController::SelectFileName(std::string filename)
{
  m_Model->SetFileName(filename);
}

void SegmentationApplicationController::ChangeAlgorithm(SegmentationAlgorithm algorithm)
{
  m_Model->SetAlgorithm(algorithm);
}
void SegmentationApplicationController::ChangeChannel(unsigned int channel)
{
  m_Model->SetChannelIndex(channel);
}
void SegmentationApplicationController::ChangeUseImageIntensity()
{
  m_Model->SetUseIntensity();
  m_View->bUseImageChannel->clear();
  m_View->bUseImageIntensity->set();
  m_View->vChannel->deactivate();
}
void SegmentationApplicationController::ChangeUseImageChannel()
{
  m_Model->SetUseChannel();
  m_View->bUseImageChannel->set();
  m_View->bUseImageIntensity->clear();
  m_View->vChannel->activate();
}
void SegmentationApplicationController::TriggerSegmentation()
{
  try
  {
    m_Model->SetupSegmentation();
    m_View->SetupSegmentationWindow();
  }
  catch (itk::ExceptionObject & err)
  {
    const char* msg = err.GetDescription();
    fl_alert("%s", msg);
  }
}

void SegmentationApplicationController::EndSegmentation()
{
  m_View->bSegment->deactivate();
  m_Model->RetrieveSegmentationResult();
}

void SegmentationApplicationController::SaveResults(std::string filename)
{
  m_Model->SaveResults(filename);
}

void SegmentationApplicationController::ImportSegments(std::string filename)
{
  m_Model->ImportSegments(filename);
}

void SegmentationApplicationController::ClearSegments()
{
  m_Model->ClearPolygonList();
}

void SegmentationApplicationController::DeleteSegment(unsigned int index)
{
  m_Model->DeletePolygon(index);
}
} // end namespace otb
