/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbLevelSetSegmentationModel.h"

namespace otb
{

/** Initialize the singleton */
LevelSetSegmentationModel::Pointer LevelSetSegmentationModel::Instance = NULL;

/** Manage the singleton */
LevelSetSegmentationModel::Pointer
LevelSetSegmentationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = LevelSetSegmentationModel::New();
  }
  return Instance;
}

void LevelSetSegmentationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

LevelSetSegmentationModel::LevelSetSegmentationModel()
{
  m_Valid                  = false;
  m_NodeContainer          = NodeContainerType::New();
  m_Sigma                  = 1.0;
  m_Alpha                  = -1.0;
  m_Beta                   = 0.0;
  m_StoppingTime           = 200.;
  m_TimeThreshold          = 100.;

  /** Filters declarations */
  m_GradientFilter     = GradientFilterType::New();
  m_SigmoidFilter      = SigmoidFilterType::New();
  m_FastMarchingFilter = FastMarchingFilterType::New();
  m_ThresholdFilter    = ThresholdFilterType::New();

  /** static parameters */
  m_ThresholdFilter->SetLowerThreshold( 0.0);
  m_ThresholdFilter->SetOutsideValue(0);
  m_ThresholdFilter->SetInsideValue(255);
  m_SigmoidFilter->SetOutputMinimum(0);
  m_SigmoidFilter->SetOutputMaximum(1.0);

  m_NodeContainer->Initialize();
  m_FastMarchingFilter->SetTrialPoints(m_NodeContainer);


  m_GradientFilter->SetSigma(m_Sigma);
  m_SigmoidFilter->SetAlpha(m_Alpha);
  m_SigmoidFilter->SetBeta(m_Beta);
  m_FastMarchingFilter->SetStoppingValue(m_StoppingTime);
  m_ThresholdFilter->SetUpperThreshold(m_TimeThreshold);

  /** pipeline wiring */
  m_SigmoidFilter->SetInput(m_GradientFilter->GetOutput());
  m_FastMarchingFilter->SetInput(m_SigmoidFilter->GetOutput());
  m_ThresholdFilter->SetInput(m_FastMarchingFilter->GetOutput());
}

LevelSetSegmentationModel::~LevelSetSegmentationModel()
{}

void LevelSetSegmentationModel::SetInputImage(ImagePointerType image)
{
  // Togle the valid flag
  m_Valid = true;
  m_GradientFilter->SetInput(image);
  m_FastMarchingFilter->SetOutputSize(image->GetBufferedRegion().GetSize());
  m_NodeContainer->Initialize();
  this->UpdatePipeline();
  this->NotifyAll();
}

LevelSetSegmentationModel::ImagePointerType LevelSetSegmentationModel::GetInputImage()
{
  return const_cast<ImageType *>(m_GradientFilter->GetInput());
}

LevelSetSegmentationModel::ImagePointerType LevelSetSegmentationModel::GetGradientMagnitudeImage()
{
  return m_GradientFilter->GetOutput();
}

LevelSetSegmentationModel::ImagePointerType LevelSetSegmentationModel::GetSpeedMapImage()
{
  return m_SigmoidFilter->GetOutput();
}

LevelSetSegmentationModel::ImagePointerType LevelSetSegmentationModel::GetTimeCrossingImage()
{
  return m_FastMarchingFilter->GetOutput();
}

LevelSetSegmentationModel::UCharImagePointerType LevelSetSegmentationModel::GetOutputImage()
{
  return m_ThresholdFilter->GetOutput();
}

void LevelSetSegmentationModel::SetSigma(double sigma)
{
  if (sigma == m_Sigma)
  {
    // nothing to do, return
    return;
  }
  m_Sigma = sigma;
  m_GradientFilter->SetSigma(m_Sigma);
  this->UpdatePipeline();
  this->NotifyAll();
}

double LevelSetSegmentationModel::GetSigma()
{
  return m_Sigma;
}

void LevelSetSegmentationModel::SetSigmoidAlpha(double alpha)
{
  if (alpha == m_Alpha)
  {
    // nothing to do, return
    return;
  }
  m_Alpha = alpha;
  m_SigmoidFilter->SetAlpha(m_Alpha);
  this->UpdatePipeline();
  this->NotifyAll();
}

double LevelSetSegmentationModel::GetSigmoidAlpha()
{
  return m_Alpha;
}

void LevelSetSegmentationModel::SetSigmoidBeta(double beta)
{
  if (beta == m_Beta)
  {
    // nothing to do, return
    return;
  }
  m_Beta = beta;
  m_SigmoidFilter->SetBeta(m_Beta);
  this->UpdatePipeline();
  this->NotifyAll();
}

double LevelSetSegmentationModel::GetSigmoidBeta()
{
  return m_Beta;
}

void LevelSetSegmentationModel::SetTimeThreshold(double timeThreshold)
{
  if (timeThreshold == m_TimeThreshold)
  {
    // nothing to do, return
    return;
  }
  m_TimeThreshold = timeThreshold;
  m_ThresholdFilter->SetUpperThreshold(m_TimeThreshold);
  this->UpdatePipeline();
  this->NotifyAll();
}
double LevelSetSegmentationModel::GetTimeThreshold()
{
  return m_TimeThreshold;

}
void LevelSetSegmentationModel::SetStoppingTime(double stoppingTime)
{
  if (stoppingTime == m_StoppingTime)
  {
    // nothing to do, return
    return;
  }
  m_StoppingTime = stoppingTime;
  m_FastMarchingFilter->SetStoppingValue(m_StoppingTime);
  this->UpdatePipeline();
  this->NotifyAll();
}
double LevelSetSegmentationModel::GetStoppingTime()
{
  return m_StoppingTime;
}

void LevelSetSegmentationModel::AddSeed(IndexType seed)
{
  NodeType node;
  node.SetValue(0.0);
  node.SetIndex(seed);
  m_NodeContainer->InsertElement(m_NodeContainer->Size(), node);
  m_FastMarchingFilter->Modified();
  this->UpdatePipeline();
  this->NotifyAll();
}

void LevelSetSegmentationModel::ClearSeeds()
{
  m_NodeContainer->Initialize();
  m_FastMarchingFilter->Modified();
  this->UpdatePipeline();
  this->NotifyAll();
}

void LevelSetSegmentationModel::UpdatePipeline()
{
  if (m_Valid)
  {
    m_GradientFilter->Modified();
    m_GradientFilter->GetOutput()->UpdateOutputInformation();
    m_GradientFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
    m_GradientFilter->Update();
    m_SigmoidFilter->GetOutput()->UpdateOutputInformation();
    m_SigmoidFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
    m_SigmoidFilter->Update();
    m_FastMarchingFilter->GetOutput()->UpdateOutputInformation();
    m_FastMarchingFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
    m_FastMarchingFilter->Update();
    m_ThresholdFilter->GetOutput()->UpdateOutputInformation();
    m_ThresholdFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
    m_ThresholdFilter->Update();
  }
}

void LevelSetSegmentationModel::Notify()
{
  m_GradientFilter->Modified();
  this->UpdatePipeline();
  this->NotifyAll();
}
}
