/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbLevelSetSegmentationModel_h
#define __otbLevelSetSegmentationModel_h

#include "otbMVCModel.h"
#include "otbImage.h"
#include "itkFastMarchingImageFilter.h"
#include "itkGradientMagnitudeRecursiveGaussianImageFilter.h"
#include "itkSigmoidImageFilter.h"
#include "itkBinaryThresholdImageFilter.h"

#include "otbListenerBase.h"
#include "otbListenerBase.h"

namespace otb
{
/** \class LevelSetSegmentationModel
 *
 *
 *
 */
class LevelSetSegmentationModel
      : public MVCModel<ListenerBase>, public ListenerBase, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef LevelSetSegmentationModel                         Self;
  typedef otb::MVCModel<ListenerBase>         Superclass;
  typedef itk::SmartPointer<Self>                           Pointer;
  typedef itk::SmartPointer<const Self>                     ConstPointer;

  /** Standard type macro */
  itkTypeMacro(LevelSetSegmentationModel, Object);

  typedef otb::Image<double, 2>                                                    ImageType;
  typedef ImageType::Pointer                                                      ImagePointerType;
  typedef ImageType::IndexType                                                    IndexType;
  typedef otb::Image<unsigned char, 2>                                             UCharImageType;
  typedef UCharImageType::Pointer                                                 UCharImagePointerType;
  typedef itk::GradientMagnitudeRecursiveGaussianImageFilter<ImageType, ImageType> GradientFilterType;
  typedef GradientFilterType::Pointer                                             GradientFilterPointerType;
  typedef itk::SigmoidImageFilter<ImageType, ImageType>                            SigmoidFilterType;
  typedef SigmoidFilterType::Pointer                                              SigmoidFilterPointerType;
  typedef itk::FastMarchingImageFilter<ImageType, ImageType>                       FastMarchingFilterType;
  typedef FastMarchingFilterType::Pointer                                         FastMarchingFilterPointerType;
  typedef itk::BinaryThresholdImageFilter<ImageType, UCharImageType>               ThresholdFilterType;
  typedef ThresholdFilterType::Pointer                                            ThresholdFilterPointerType;
  typedef FastMarchingFilterType::NodeContainer                                   NodeContainerType;
  typedef NodeContainerType::Pointer                                              NodeContainerPointerType;
  typedef FastMarchingFilterType::NodeType                                        NodeType;

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Get the output image */
  UCharImagePointerType GetOutputImage();

  /** Get the gradient magnitude image */
  ImagePointerType GetGradientMagnitudeImage();

  /** Get the speed map image */
  ImagePointerType GetSpeedMapImage();

  /** Get the time crossing image */
  ImagePointerType GetTimeCrossingImage();

  /** Get the flags to activate/deactivate the preprocessings */
  itkGetMacro(Valid, bool);

  /** Get the output image */
  void SetInputImage(ImagePointerType image);
  ImagePointerType GetInputImage();

  /** Set/get the sigma gradient parameter */
  void   SetSigma(double sigma);
  double GetSigma();

  /** Set/get the sigmoid alpha parameter */
  void   SetSigmoidAlpha(double alpha);
  double GetSigmoidAlpha();

  /** Set/get the sigmoid beta parameter */
  void   SetSigmoidBeta(double beta);
  double GetSigmoidBeta();

  /** Set/get the time threshold parameter */
  void   SetTimeThreshold(double timeThreshold);
  double GetTimeThreshold();

  /** Set/get the stopping time parameter */
  void   SetStoppingTime(double stoppingTime);
  double GetStoppingTime();

  /** Manage seeds */
  void AddSeed(IndexType seed);
  void ClearSeeds();

  /** Get the Seed container */
  itkGetObjectMacro(NodeContainer, NodeContainerType);

  /** Receive events from the preprocessing model */
  virtual void Notify();

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  LevelSetSegmentationModel();
  /** Destructor */
  ~LevelSetSegmentationModel();

private:
  LevelSetSegmentationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Update the pipeline */
  void UpdatePipeline(void);

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Filters */
  GradientFilterPointerType     m_GradientFilter;
  SigmoidFilterPointerType      m_SigmoidFilter;
  FastMarchingFilterPointerType m_FastMarchingFilter;
  ThresholdFilterPointerType    m_ThresholdFilter;

  /** Flags to activate/deactivate the preprocessings */
  bool m_Valid;

  /** Node Container */
  NodeContainerPointerType m_NodeContainer;

  /** The instance singleton */
  static Pointer Instance;

  /** Since the itk::GradientMagnitudeRecursiveGaussianImageFilter
   *does not provide a GetSigma(), we must store it here
   */
  double m_Sigma;
  double m_Alpha;
  double m_Beta;
  double m_StoppingTime;
  double m_TimeThreshold;
};
}
#endif
