/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbPreprocessingModel.h"

namespace otb
{

/** Initialize the singleton */
PreprocessingModel::Pointer PreprocessingModel::Instance = NULL;

/** Manage the singleton */
PreprocessingModel::Pointer
PreprocessingModel::GetInstance()
{
  if (!Instance)
  {
    Instance = PreprocessingModel::New();
  }
  return Instance;
}

void PreprocessingModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

PreprocessingModel::PreprocessingModel()
{
  /** Initial parameters */
  m_TimeStep               = 0.125;
  m_NumberOfIterations     = 10;
  m_Alpha                  = 0.5;
  m_UseSmoothing           = true;
  m_UseEdgeEnhancement     = true;
  m_Valid                  = false;
  m_InputImage             = NULL;

  /** Filters declarations */
  m_CurvatureFlowFilter    = CurvatureFlowFilterType::New();
  m_XDerivativeFilter      = DerivativeFilterType::New();
  m_YDerivativeFilter      = DerivativeFilterType::New();
  m_MultiplyByScalarFilter = MultiplyByScalarFilterType::New();
  m_AddFilter1             = AddFilterType::New();
  m_AddFilter2             = AddFilterType::New();
  m_IdentityFilter         = IdentityFilterType::New();

  m_OutputImage = m_IdentityFilter->GetOutput();

  /** static parameters */
  m_XDerivativeFilter->SetDirection(0);
  m_YDerivativeFilter->SetDirection(1);

  // Initial parameters */
  m_CurvatureFlowFilter->SetNumberOfIterations(m_NumberOfIterations);
  m_CurvatureFlowFilter->SetTimeStep(m_TimeStep);
  m_MultiplyByScalarFilter->SetCoef(m_Alpha);
}

PreprocessingModel::~PreprocessingModel()
{}

PreprocessingModel::ImagePointerType PreprocessingModel::GetSmoothedImage()
{
  return m_CurvatureFlowFilter->GetOutput();
}

void PreprocessingModel::SetInputImage(ImagePointerType image)
{
  // Set the input image
  m_InputImage = image;
  // Togle the valid flag
  m_Valid = true;
  // Notify the observers
  this->RewirePipeline();
  this->UpdatePipeline();
  this->NotifyAll();
}

void PreprocessingModel::RewirePipeline()
{

  //TODO: check this
  if (m_UseSmoothing)
  {
    m_CurvatureFlowFilter->SetInput(m_InputImage);

    if (m_UseEdgeEnhancement)
    {
      m_XDerivativeFilter->SetInput(m_CurvatureFlowFilter->GetOutput());
      m_YDerivativeFilter->SetInput(m_CurvatureFlowFilter->GetOutput());
      m_AddFilter1->SetInput1(m_XDerivativeFilter->GetOutput());
      m_AddFilter1->SetInput2(m_YDerivativeFilter->GetOutput());
      m_MultiplyByScalarFilter->SetInput(m_AddFilter1->GetOutput());
      m_AddFilter2->SetInput2(m_MultiplyByScalarFilter->GetOutput());
      m_AddFilter2->SetInput1(m_CurvatureFlowFilter->GetOutput());
      m_IdentityFilter->SetInput(m_AddFilter2->GetOutput());
    }
    else
    {
      m_IdentityFilter->SetInput(m_CurvatureFlowFilter->GetOutput());
    }
  }
  else
  {
    if (m_UseEdgeEnhancement)
    {
      m_XDerivativeFilter->SetInput(m_InputImage);
      m_YDerivativeFilter->SetInput(m_InputImage);
      m_AddFilter1->SetInput1(m_XDerivativeFilter->GetOutput());
      m_AddFilter1->SetInput2(m_YDerivativeFilter->GetOutput());
      m_MultiplyByScalarFilter->SetInput(m_AddFilter1->GetOutput());
      m_AddFilter2->SetInput2(m_MultiplyByScalarFilter->GetOutput());
      m_AddFilter2->SetInput1(m_InputImage);
      m_IdentityFilter->SetInput( m_AddFilter2->GetOutput());
    }
    else
    {
      m_IdentityFilter->SetInput(m_InputImage);
    }
  }
}

void PreprocessingModel::UpdatePipeline()
{
  m_IdentityFilter->GetOutput()->UpdateOutputInformation();
  m_IdentityFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
  m_IdentityFilter->Update();
}

void PreprocessingModel::SetUseSmoothing(bool flag)
{
  if (m_UseSmoothing == flag)
  {
    // nothing to do, exit
    return;
  }
  m_UseSmoothing = flag;
  this->RewirePipeline();
  this->UpdatePipeline();
  this->NotifyAll();
}

void PreprocessingModel::SetUseEdgeEnhancement(bool flag)
{
  if (m_UseEdgeEnhancement == flag)
  {
    // nothing to do, exit
    return;
  }
  m_UseEdgeEnhancement = flag;
  this->RewirePipeline();
  this->UpdatePipeline();
  this->NotifyAll();
}

void PreprocessingModel::SetNumberOfIterations(unsigned int numberOfIterations)
{
  if (numberOfIterations == m_NumberOfIterations)
  {
    //nothing to do, return
    return;
  }
  m_NumberOfIterations = numberOfIterations;
  m_CurvatureFlowFilter->SetNumberOfIterations(m_NumberOfIterations);
  this->UpdatePipeline();
  this->NotifyAll();
}

void PreprocessingModel::SetTimeStep(double timeStep)
{
  if (timeStep == m_TimeStep)
  {
    //nothing to do, return
    return;
  }
  m_TimeStep = timeStep;
  m_CurvatureFlowFilter->SetTimeStep(m_TimeStep);
  this->UpdatePipeline();
  this->NotifyAll();
}

void PreprocessingModel::SetAlpha(double alpha)
{
  if (alpha == m_Alpha)
  {
    //nothing to do, return
    return;
  }
  m_Alpha = alpha;
  m_MultiplyByScalarFilter->SetCoef(m_Alpha);
  this->UpdatePipeline();
  this->NotifyAll();
}
}
