/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbPreprocessingModel_h
#define __otbPreprocessingModel_h

#include "otbMVCModel.h"
#include "otbImage.h"
#include "itkCurvatureFlowImageFilter.h"
#include "itkDerivativeImageFilter.h"
#include "otbMultiplyByScalarImageFilter.h"
#include "itkAddImageFilter.h"
#include "itkCastImageFilter.h"

#include "otbListenerBase.h"

namespace otb
{
/** \class PreprocessingModel
 *
 * This class modelize a preprocessing chain using CurvatureFlowFilter
 * for edge-preserving smoothing and a simple edge enhancement method.
 *
 */


class PreprocessingModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef PreprocessingModel                        Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                   Pointer;
  typedef itk::SmartPointer<const Self>             ConstPointer;

  /** Standard type macro */
  itkTypeMacro(PreprocessModel, Object);

  /** Images and filters typedefs */
  typedef double PixelType;
  typedef otb::Image<PixelType, 2> ImageType;
  typedef ImageType::Pointer ImagePointerType;
  typedef itk::CurvatureFlowImageFilter<ImageType, ImageType> CurvatureFlowFilterType;
  typedef CurvatureFlowFilterType::Pointer CurvatureFlowFilterPointerType;
  typedef itk::DerivativeImageFilter<ImageType, ImageType> DerivativeFilterType;
  typedef DerivativeFilterType::Pointer DerivativeFilterPointerType;
  typedef MultiplyByScalarImageFilter<ImageType, ImageType> MultiplyByScalarFilterType;
  typedef MultiplyByScalarFilterType::Pointer MultiplyByScalarFilterPointerType;
  typedef itk::AddImageFilter<ImageType, ImageType, ImageType> AddFilterType;
  typedef AddFilterType::Pointer AddFilterPointerType;
  typedef itk::CastImageFilter<ImageType, ImageType> IdentityFilterType;
  typedef IdentityFilterType::Pointer IdentityFilterPointerType;

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Get the smoothed image */
  ImagePointerType GetSmoothedImage();

  /** Get the output image */
  itkGetObjectMacro(OutputImage, ImageType);

  /** Get the input image */
  itkGetObjectMacro(InputImage, ImageType);

  /** Get the flags to activate/deactivate the preprocessings */
  itkGetMacro(UseSmoothing, bool);
  itkGetMacro(UseEdgeEnhancement, bool);
  itkGetMacro(Valid, bool);
  itkGetMacro(NumberOfIterations, unsigned int);
  itkGetMacro(TimeStep, double);
  itkGetMacro(Alpha, double);

  /** Get the output image */
  void SetInputImage(ImagePointerType image);

  /** Set the flags to activate/deactivate the preprocessings */
  void SetUseSmoothing(bool flag);
  void SetUseEdgeEnhancement(bool flag);

  /** Set the parameters */
  void SetNumberOfIterations(unsigned int numberOfIterations);
  void SetTimeStep(double timeStep);
  void SetAlpha(double alpha);

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  PreprocessingModel();
  /** Destructor */
  ~PreprocessingModel();

private:
  PreprocessingModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Re do pipeline wiring */
  void RewirePipeline(void);
  /** Update the pipeline */
  void UpdatePipeline(void);

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Input Image */
  ImagePointerType m_InputImage;
  /** Output Image */
  ImagePointerType m_OutputImage;

  /** Filters */
  CurvatureFlowFilterPointerType    m_CurvatureFlowFilter;
  DerivativeFilterPointerType       m_XDerivativeFilter;
  DerivativeFilterPointerType       m_YDerivativeFilter;
  MultiplyByScalarFilterPointerType m_MultiplyByScalarFilter;
  AddFilterPointerType              m_AddFilter1;
  AddFilterPointerType              m_AddFilter2;
  /** Used to present an homogeneous interface */
  IdentityFilterPointerType         m_IdentityFilter;

  /** Flags to activate/deactivate the preprocessings */
  bool m_UseSmoothing;
  bool m_UseEdgeEnhancement;
  bool m_Valid;

  /** Parameters */
  double m_TimeStep;
  unsigned int m_NumberOfIterations;
  double m_Alpha;

  /** The instance singleton */
  static Pointer Instance;
};
}
#endif
