/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbRegionGrowingSegmentationModel.h"

namespace otb
{

/** Initialize the singleton */
RegionGrowingSegmentationModel::Pointer RegionGrowingSegmentationModel::Instance = NULL;

/** Manage the singleton */
RegionGrowingSegmentationModel::Pointer
RegionGrowingSegmentationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = RegionGrowingSegmentationModel::New();
  }
  return Instance;
}

void RegionGrowingSegmentationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

RegionGrowingSegmentationModel::RegionGrowingSegmentationModel()
{


  m_Valid                  = false;
  m_LowerThreshold         = 0.0;
  m_UpperThreshold         = 100.0;
  m_Radius.Fill(2);
  m_Tolerance              = 2.5;
  m_NumberOfIterations     = 10;
  m_Algorithm              = THRESHOLD_CONNECTED;

  /** Filters declarations */
  m_ConnectedThresholdFilter    = ConnectedThresholdFilterType::New();
  m_NeighborhoodConnectedFilter = NeighborhoodConnectedFilterType::New();
  m_ConfidenceConnectedFilter   = ConfidenceConnectedImageFilterType::New();
  m_IsolatedConnectedFilter     = IsolatedConnectedFilterType::New();
  m_IdentityFilter              = IdentityFilterType::New();

  /** static parameters */
  m_ConnectedThresholdFilter->SetReplaceValue(255);
  m_NeighborhoodConnectedFilter->SetReplaceValue(255);
  m_ConfidenceConnectedFilter->SetReplaceValue(255);
  m_IsolatedConnectedFilter->SetReplaceValue(255);

  m_ConnectedThresholdFilter->SetLower(m_LowerThreshold);
  m_ConnectedThresholdFilter->SetUpper(m_UpperThreshold);

  m_NeighborhoodConnectedFilter->SetLower(m_LowerThreshold);
  m_NeighborhoodConnectedFilter->SetUpper(m_UpperThreshold);
  m_NeighborhoodConnectedFilter->SetRadius(m_Radius);

  m_ConfidenceConnectedFilter->SetMultiplier(m_Tolerance);
  m_ConfidenceConnectedFilter->SetInitialNeighborhoodRadius(m_Radius[0]);

  m_IsolatedConnectedFilter->SetLower(m_LowerThreshold);
  m_IdentityFilter->SetInput(m_ConnectedThresholdFilter->GetOutput());
}

RegionGrowingSegmentationModel::~RegionGrowingSegmentationModel()
{}

void RegionGrowingSegmentationModel::SetInputImage(ImagePointerType image)
{
  // Togle the valid flag
  m_Valid = true;
  m_InputImage = image;
  m_ConnectedThresholdFilter->SetInput(image);
  m_NeighborhoodConnectedFilter->SetInput(image);
  m_ConfidenceConnectedFilter->SetInput(image);
  m_IsolatedConnectedFilter->SetInput(image);
  // Clearing the seeds
  m_InsideSeeds.clear();
  m_OutsideSeeds.clear();
  m_ConfidenceConnectedFilter->ClearSeeds();
  m_ConnectedThresholdFilter->ClearSeeds();
  m_NeighborhoodConnectedFilter->ClearSeeds();
  m_IsolatedConnectedFilter->ClearSeeds1();
  m_IsolatedConnectedFilter->ClearSeeds2();
  m_ConnectedThresholdFilter->Modified();
  m_ConfidenceConnectedFilter->Modified();
  m_NeighborhoodConnectedFilter->Modified();
  m_IsolatedConnectedFilter->Modified();
  this->UpdatePipeline();
  this->NotifyAll();
}

RegionGrowingSegmentationModel::ImagePointerType RegionGrowingSegmentationModel::GetInputImage()
{
  return m_InputImage;
}

RegionGrowingSegmentationModel::UCharImagePointerType RegionGrowingSegmentationModel::GetOutputImage()
{
  return m_IdentityFilter->GetOutput();
}

void RegionGrowingSegmentationModel::SetLowerThreshold(double thresh)
{
  if (thresh == m_LowerThreshold)
  {
    // nothing to do, return
    return;
  }
  m_LowerThreshold = thresh;
  m_ConnectedThresholdFilter->SetLower(m_LowerThreshold);
  m_NeighborhoodConnectedFilter->SetLower(m_LowerThreshold);
  m_IsolatedConnectedFilter->SetLower(m_LowerThreshold);
  this->UpdatePipeline();
  this->NotifyAll();
}

double RegionGrowingSegmentationModel::GetLowerThreshold()
{
  return m_LowerThreshold;
}

void RegionGrowingSegmentationModel::SetUpperThreshold(double thresh)
{
  if (thresh == m_UpperThreshold)
  {
    // nothing to do, return
    return;
  }
  m_UpperThreshold = thresh;
  m_ConnectedThresholdFilter->SetUpper(m_UpperThreshold);
  m_NeighborhoodConnectedFilter->SetUpper(m_UpperThreshold);
  this->UpdatePipeline();
  this->NotifyAll();
}

double RegionGrowingSegmentationModel::GetUpperThreshold()
{
  return m_UpperThreshold;
}

void RegionGrowingSegmentationModel::SetRadius(unsigned int radius)
{
  if (radius  == m_Radius[0])
  {
    // nothing to do, return
    return;
  }
  m_Radius.Fill(radius);
  m_NeighborhoodConnectedFilter->SetRadius(m_Radius);
  m_ConfidenceConnectedFilter->SetInitialNeighborhoodRadius(m_Radius[0]);
  this->UpdatePipeline();
  this->NotifyAll();
}

unsigned int RegionGrowingSegmentationModel::GetRadius()
{
  return m_Radius[0];
}

void RegionGrowingSegmentationModel::SetNumberOfIterations(unsigned int num)
{
  if (num == m_NumberOfIterations)
  {
    // nothing to do, return
    return;
  }
  m_NumberOfIterations = num;
  m_ConfidenceConnectedFilter->SetNumberOfIterations(m_NumberOfIterations);
  this->UpdatePipeline();
  this->NotifyAll();
}

unsigned int RegionGrowingSegmentationModel::GetNumberOfIterations()
{
  return m_NumberOfIterations;
}

void RegionGrowingSegmentationModel::SetTolerance(double tolerance)
{
  if (tolerance == m_Tolerance)
  {
    // nothing to do, return
    return;
  }
  m_Tolerance = tolerance;
  m_ConfidenceConnectedFilter->SetMultiplier(m_Tolerance);
  this->UpdatePipeline();
  this->NotifyAll();
}
double RegionGrowingSegmentationModel::GetTolerance()
{
  return m_Tolerance;

}
void RegionGrowingSegmentationModel::SetRegionGrowingAlgorithm(RegionGrowingAlgorithm algorithm)
{
  if (algorithm == m_Algorithm)
  {
    // nothing to do, return
    return;
  }
  m_Algorithm = algorithm;

  switch (m_Algorithm)
  {
  case THRESHOLD_CONNECTED:
    m_IdentityFilter->SetInput(m_ConnectedThresholdFilter->GetOutput());
    break;
  case NEIGHBORHOOD_CONNECTED:
    m_IdentityFilter->SetInput(m_NeighborhoodConnectedFilter->GetOutput());
    break;
  case CONFIDENCE_CONNECTED:
    m_IdentityFilter->SetInput(m_ConfidenceConnectedFilter->GetOutput());
    break;
  case ISOLATED_CONNECTED:
    m_IdentityFilter->SetInput(m_IsolatedConnectedFilter->GetOutput());
    break;
  }
  this->UpdatePipeline();
  this->NotifyAll();
}
RegionGrowingAlgorithm RegionGrowingSegmentationModel::GetRegionGrowingAlgorithm()
{
  return m_Algorithm;
}

void RegionGrowingSegmentationModel::AddInsideSeed(IndexType seed)
{
  m_ConnectedThresholdFilter->AddSeed(seed);
  m_ConfidenceConnectedFilter->AddSeed(seed);
  m_NeighborhoodConnectedFilter->AddSeed(seed);
  m_IsolatedConnectedFilter->AddSeed1(seed);
  m_InsideSeeds.push_back(seed);
  this->UpdatePipeline();
  this->NotifyAll();
}

void RegionGrowingSegmentationModel::AddOutsideSeed(IndexType seed)
{
  m_IsolatedConnectedFilter->AddSeed2(seed);
  m_OutsideSeeds.push_back(seed);
  this->UpdatePipeline();
  this->NotifyAll();
}

RegionGrowingSegmentationModel::SeedsContainerType & RegionGrowingSegmentationModel::GetInsideSeeds()
{
  return m_InsideSeeds;
}

RegionGrowingSegmentationModel::SeedsContainerType & RegionGrowingSegmentationModel::GetOutsideSeeds()
{
  return m_OutsideSeeds;
}

void RegionGrowingSegmentationModel::ClearSeeds()
{
  m_ConfidenceConnectedFilter->ClearSeeds();
  m_ConnectedThresholdFilter->ClearSeeds();
  m_NeighborhoodConnectedFilter->ClearSeeds();
  m_IsolatedConnectedFilter->ClearSeeds1();
  m_IsolatedConnectedFilter->ClearSeeds2();
  m_ConnectedThresholdFilter->Modified();
  m_ConfidenceConnectedFilter->Modified();
  m_NeighborhoodConnectedFilter->Modified();
  m_IsolatedConnectedFilter->Modified();
  m_InsideSeeds.clear();
  m_OutsideSeeds.clear();
  this->UpdatePipeline();
  this->NotifyAll();
}

void RegionGrowingSegmentationModel::UpdatePipeline()
{
  if (m_Valid)
  {
    // This is done because isolated connected does not check if OutsideSeeds is empty before dereferencing, leading to a segmentation fault
    if ((m_Algorithm != ISOLATED_CONNECTED) || !m_OutsideSeeds.empty())
    {
      m_IdentityFilter->GetOutput()->UpdateOutputInformation();
      m_IdentityFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
      m_IdentityFilter->Update();
    }

  }
}

void RegionGrowingSegmentationModel::Notify()
{
  m_ConnectedThresholdFilter->Modified();
  m_ConfidenceConnectedFilter->Modified();
  m_NeighborhoodConnectedFilter->Modified();
  m_IsolatedConnectedFilter->Modified();
  this->UpdatePipeline();
  this->NotifyAll();
}
}

