/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbRegionGrowingSegmentationModel_h
#define __otbRegionGrowingSegmentationModel_h

#include "otbMVCModel.h"
#include "otbImage.h"

#include "otbListenerBase.h"
#include "otbListenerBase.h"

#include "itkConnectedThresholdImageFilter.h"
#include "itkNeighborhoodConnectedImageFilter.h"
#include "itkConfidenceConnectedImageFilter.h"
#include "itkIsolatedConnectedImageFilter.h"
#include "itkCastImageFilter.h"


namespace otb
{

typedef enum
{
  THRESHOLD_CONNECTED,
  NEIGHBORHOOD_CONNECTED,
  ISOLATED_CONNECTED,
  CONFIDENCE_CONNECTED
} RegionGrowingAlgorithm;

/** \class RegionGrowingSegmentationModel
 *
 *
 *
 */
class ITK_EXPORT RegionGrowingSegmentationModel
      : public MVCModel<ListenerBase>, public ListenerBase, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef RegionGrowingSegmentationModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                                Pointer;
  typedef itk::SmartPointer<const Self>                          ConstPointer;

  /** Standard type macro */
  itkTypeMacro(RegionGrowingSegmentationModel, Object);

  typedef otb::Image<double, 2>                                                    ImageType;
  typedef ImageType::Pointer                                                      ImagePointerType;
  typedef ImageType::SizeType                                                     SizeType;
  typedef ImageType::IndexType                                                    IndexType;
  typedef otb::Image<unsigned char, 2>                                             UCharImageType;
  typedef UCharImageType::Pointer                                                 UCharImagePointerType;
  typedef itk::ConnectedThresholdImageFilter<ImageType, UCharImageType>            ConnectedThresholdFilterType;
  typedef ConnectedThresholdFilterType::Pointer                                   ConnectedThresholdFilterPointerType;
  typedef itk::NeighborhoodConnectedImageFilter<ImageType, UCharImageType>         NeighborhoodConnectedFilterType;
  typedef NeighborhoodConnectedFilterType::Pointer                                NeighborhoodConnectedFilterPointerType;
  typedef itk::IsolatedConnectedImageFilter<ImageType, UCharImageType>             IsolatedConnectedFilterType;
  typedef IsolatedConnectedFilterType::Pointer            IsolatedConnectedFilterPointerType;
  typedef itk::ConfidenceConnectedImageFilter<ImageType, UCharImageType>           ConfidenceConnectedImageFilterType;
  typedef ConfidenceConnectedImageFilterType::Pointer                             ConfidenceConnectedImageFilterPointerType;
  typedef itk::CastImageFilter<UCharImageType, UCharImageType>                     IdentityFilterType;
  typedef  IdentityFilterType::Pointer                                            IdentityFilterPointerType;
  typedef std::vector<IndexType>                                                  SeedsContainerType;

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Get the output image */
  UCharImagePointerType GetOutputImage();

  /** Get the flags to activate/deactivate the preprocessings */
  itkGetMacro(Valid, bool);

  /** Get the output image */
  void SetInputImage(ImagePointerType image);
  ImagePointerType GetInputImage();

  /** Set/get the upper threshold parameter */
  void   SetUpperThreshold(double thresh);
  double GetUpperThreshold();

  /** Set/get the lower threshold parameter */
  void   SetLowerThreshold(double thresh);
  double GetLowerThreshold();

  /** Set/get the radius parameter */
  void SetRadius(unsigned int radius);
  unsigned int GetRadius();

  /** Set/get the number of iterations parameter */
  void SetNumberOfIterations(unsigned int num);
  unsigned int GetNumberOfIterations();

  /** Set/get the tolerance parameter */
  void   SetTolerance(double tolerance);
  double GetTolerance();

  /** Set/get the region growing algorithm */
  void SetRegionGrowingAlgorithm(RegionGrowingAlgorithm algorithm);
  RegionGrowingAlgorithm GetRegionGrowingAlgorithm();

  /** Manage seeds */
  void AddInsideSeed(IndexType seed);
  void AddOutsideSeed(IndexType seed);
  void ClearSeeds();
  SeedsContainerType & GetInsideSeeds();
  SeedsContainerType & GetOutsideSeeds();

  /** Receive events from the preprocessing model */
  virtual void Notify();

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  RegionGrowingSegmentationModel();
  /** Destructor */
  ~RegionGrowingSegmentationModel();

private:
  RegionGrowingSegmentationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Update the pipeline */
  void UpdatePipeline(void);

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Filters */
  ConnectedThresholdFilterPointerType       m_ConnectedThresholdFilter;
  NeighborhoodConnectedFilterPointerType    m_NeighborhoodConnectedFilter;
  ConfidenceConnectedImageFilterPointerType m_ConfidenceConnectedFilter;
  IsolatedConnectedFilterPointerType        m_IsolatedConnectedFilter;
  IdentityFilterPointerType                 m_IdentityFilter;

  /** Flags to activate/deactivate the preprocessings */
  bool m_Valid;

  /** The instance singleton */
  static Pointer Instance;

  /** Parameters */
  double                 m_LowerThreshold;
  double                 m_UpperThreshold;
  SizeType               m_Radius;
  unsigned int           m_NumberOfIterations;
  double                 m_Tolerance;
  RegionGrowingAlgorithm m_Algorithm;
  SeedsContainerType     m_InsideSeeds;
  SeedsContainerType     m_OutsideSeeds;
  ImagePointerType       m_InputImage;
};
}
#endif
