/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbSVMBasedRegionGrowingSegmentationModel.h"
#include "otbMultiToMonoChannelExtractROI.h"
#include "otbRAndNIRIndexImageFilter.h"
#include "otbRAndBAndNIRIndexImageFilter.h"
#include "otbObjectList.h"
#include "otbSpectralAngleDistanceImageFilter.h"
#include "itkSqrtImageFilter.h"
#include "itkCurvatureFlowImageFilter.h"
#include "otbMultiplyByScalarImageFilter.h"
#include "itkShiftScaleImageFilter.h"
#include "itkLaplacianRecursiveGaussianImageFilter.h"
#include "otbImageListToImageListApplyFilter.h"
#include "itkImageRegionConstIterator.h"
#include "itkImageRegionIterator.h"
#include "otbThresholdImageToPointSetFilter.h"
#include "itkAddImageFilter.h"
#include "otbSVMPointSetModelEstimator.h"
#include "otbSVMModel.h"
#include "itkMinimumMaximumImageCalculator.h"
#include "itkLineIterator.h"
#include "itkGradientMagnitudeImageFilter.h"
#include "itkBinaryThresholdImageFilter.h"
#include "itkScalarImageToHistogramGenerator.h"
#include "itkOtsuMultipleThresholdsCalculator.h"
#include "itkMinimumMaximumImageCalculator.h"
#include "otbMacro.h"


namespace otb
{

/** Initialize the singleton */
SVMBasedRegionGrowingSegmentationModel::Pointer SVMBasedRegionGrowingSegmentationModel::Instance = NULL;

/** Manage the singleton */
SVMBasedRegionGrowingSegmentationModel::Pointer
SVMBasedRegionGrowingSegmentationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = SVMBasedRegionGrowingSegmentationModel::New();
  }
  return Instance;
}

void SVMBasedRegionGrowingSegmentationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

SVMBasedRegionGrowingSegmentationModel::SVMBasedRegionGrowingSegmentationModel()
{
  m_Valid = false;
  m_UseLaplacian = false;
  m_NearInfraRedChannelIndex = 3;
  m_RedChannelIndex          = 2;
  m_GreenChannelIndex        = 1;
  m_BlueChannelIndex         = 0;
  m_NumberOfSpectralAngleDistanceImages = 10;
  m_NumberOfThresholds = 4;
  m_ConnectedFilter = ConnectedFilterType::New();
  m_FeatureImagesList = InternImageListType::New();
  m_LuminanceImage = InternImageType::New();
  m_ExtractedShadows = InternImageType::New();
  m_HyperplanDistanceImage = InternImageType::New();

  m_PixelsInPS = InputPSType::New();
  m_PixelsOutPS = InputPSType::New();
  m_PixelsInPS->SetPoints(InputPSType::PointsContainer::New());
  m_PixelsOutPS->SetPoints(InputPSType::PointsContainer::New());
}

SVMBasedRegionGrowingSegmentationModel::~SVMBasedRegionGrowingSegmentationModel()
{}

void SVMBasedRegionGrowingSegmentationModel::SetInputImage(InputMSImageType::Pointer image)
{
  if (image->GetNumberOfComponentsPerPixel()<4)
  {
    itkExceptionMacro(<<"This segmentation method requires at least a 4 channels image!");
  }

  this->ClearSeeds();
  m_FeatureImagesList->Clear();
  m_InputImage = image;
  m_LuminanceImage->SetRegions(image->GetLargestPossibleRegion());
  m_LuminanceImage->Allocate();
  m_LuminanceImage->FillBuffer(0);
  m_HyperplanDistanceImage->SetRegions(image->GetLargestPossibleRegion());
  m_HyperplanDistanceImage->Allocate();
  m_HyperplanDistanceImage->FillBuffer(-1);
  m_ExtractedShadows->SetRegions(image->GetLargestPossibleRegion());
  m_ExtractedShadows->Allocate();
  m_ExtractedShadows->FillBuffer(0);
  m_Valid = true;
  this->ComputeStaticFeaturesVector();
  this->UpdatePipeline();
  this->NotifyAll();
}

void SVMBasedRegionGrowingSegmentationModel::SetNumberOfThresholds(unsigned int nb)
{
  m_NumberOfThresholds = nb;
}

void SVMBasedRegionGrowingSegmentationModel::SetNumberOfSpectralAngleDistanceImages(unsigned int nb)
{
  m_NumberOfSpectralAngleDistanceImages = nb;
}

SVMBasedRegionGrowingSegmentationModel::OutputImagePointerType SVMBasedRegionGrowingSegmentationModel::GetOutputImage()
{
  return m_ConnectedFilter->GetOutput();
}

void SVMBasedRegionGrowingSegmentationModel::Update()
{
  this->NotifyAll();
  this->UpdatePipeline();
  this->NotifyAll();
}

void SVMBasedRegionGrowingSegmentationModel::AddInsidePoint(PointType & point)
{
  m_PixelsInPS->SetPoint(m_PixelsInPS->GetNumberOfPoints(), point);
}

void SVMBasedRegionGrowingSegmentationModel::AddOutsidePoint(PointType & point)
{
  m_PixelsOutPS->SetPoint(m_PixelsOutPS->GetNumberOfPoints(), point);
}

void SVMBasedRegionGrowingSegmentationModel::ClearSeeds()
{
  m_PixelsInPS->GetPoints()->Initialize();
  m_PixelsOutPS->GetPoints()->Initialize();
}

void SVMBasedRegionGrowingSegmentationModel::UpdatePipeline()
{
  if (m_PixelsInPS->GetNumberOfPoints()>0 && m_PixelsOutPS->GetNumberOfPoints()>0)
  {
    this->ComputeFeaturesVector();
    this->ComputeHyperplanDistanceImage();
  }
  this->ComputeRegionGrowing();
}

void SVMBasedRegionGrowingSegmentationModel::ComputeFeaturesVector()
{
  // check if we need to compute static feature vector
  if (m_FeatureImagesList->Size()<8)
  {
    this->ComputeStaticFeaturesVector();
  }
  this->ComputePointDependentFeaturesVector();
}

void SVMBasedRegionGrowingSegmentationModel::ComputeStaticFeaturesVector()
{
  typedef otb::MultiToMonoChannelExtractROI<InternPixelType, InternPixelType> ExtractorType;

  // Blue
  ExtractorType::Pointer blue = ExtractorType::New();
  blue->SetChannel(m_BlueChannelIndex+1);
  blue->SetInput(m_InputImage);
  blue->Update();
  // Green
  ExtractorType::Pointer green = ExtractorType::New();
  green->SetChannel(m_GreenChannelIndex+1);
  green->SetInput(m_InputImage);
  green->Update();
  // Red
  ExtractorType::Pointer red = ExtractorType::New();
  red->SetChannel(m_RedChannelIndex+1);
  red->SetInput(m_InputImage);
  red->Update();
  // IR
  ExtractorType::Pointer infraRed = ExtractorType::New();
  infraRed->SetChannel(m_NearInfraRedChannelIndex+1);
  infraRed->SetInput(m_InputImage);
  infraRed->Update();

  // the 4 bands are rescaled between -1 and 1, for the learning
  InternImageListType::Pointer RGBImageList = InternImageListType::New();
  RGBImageList->PushBack(blue->GetOutput());
  RGBImageList->PushBack(green->GetOutput());
  RGBImageList->PushBack(red->GetOutput());
  RGBImageList->PushBack(infraRed->GetOutput());
  typedef itk::ShiftScaleImageFilter<InternImageType, InternImageType> ShiftScaleType;
  ShiftScaleType::Pointer shiftScaleImageFilter = ShiftScaleType::New();
  double shiftValue = -4096.0/2.0;  // 4096 = 2**12; 12 = number of bits of a pixel of the input image
  double scaleValue = 2.0/4096.0;
  shiftScaleImageFilter->SetShift(shiftValue);
  shiftScaleImageFilter->SetScale(scaleValue);
  typedef otb::ImageListToImageListApplyFilter<InternImageListType, InternImageListType, ShiftScaleType> ImageListRescalerType;
  ImageListRescalerType::Pointer imageListRescaler = ImageListRescalerType::New();
  imageListRescaler->SetFilter(shiftScaleImageFilter);
  imageListRescaler->SetInput(RGBImageList);
  imageListRescaler->Update();

  // Calculating radiometric indices
  typedef otb::Functor::NDVI<InternPixelType, InternPixelType, InternPixelType> NDVIFunctorType;
  typedef otb::RAndNIRIndexImageFilter <InternImageType, InternImageType, InternImageType, NDVIFunctorType> NDVIType;
  NDVIType::Pointer ndvi = NDVIType::New();
  ndvi->SetInputR(red->GetOutput());
  ndvi->SetInputNIR(infraRed->GetOutput());
  ndvi->Update();

  // Like the NDVI, but with  green instead of NIR
  NDVIType::Pointer ndvi2 = NDVIType::New();
  ndvi2->SetInputR(red->GetOutput());
  ndvi2->SetInputNIR(green->GetOutput());
  ndvi2->Update();

  // a spectral angle distance with a reference pixel representative of a  pixel of water
  typedef otb::SpectralAngleDistanceImageFilter<InputMSImageType, InternImageType> SpectralAngleType;
  SpectralAngleType::Pointer waterIndice = SpectralAngleType::New();
  InputMSImageType::PixelType reference;
  reference.SetSize(4);
  reference[0] = 136.0;
  reference[1] = 132.0;
  reference[2] = 47.0;
  reference[3] = 24.0;
  waterIndice->SetInput(m_InputImage);
  waterIndice->SetReferencePixel(reference);
  waterIndice->Update();

  // Computing the luminance image
  typedef itk::ImageRegionConstIterator<InternImageType> ImageRegionConstIteratorType;
  typedef itk::ImageRegionIterator<InternImageType> ImageRegionIteratorType;
  ImageRegionConstIteratorType itr(red->GetOutput(), m_InputImage->GetLargestPossibleRegion());
  ImageRegionConstIteratorType itv(green->GetOutput(), m_InputImage->GetLargestPossibleRegion());
  ImageRegionConstIteratorType itb(blue->GetOutput(), m_InputImage->GetLargestPossibleRegion());
  ImageRegionConstIteratorType itir(infraRed->GetOutput(), m_InputImage->GetLargestPossibleRegion());

  ImageRegionIteratorType itLuminanceImage(m_LuminanceImage, m_InputImage->GetLargestPossibleRegion());

  for (itr.GoToBegin(), itv.GoToBegin(), itb.GoToBegin(), itir.GoToBegin(), itLuminanceImage.GoToBegin(); !itr.IsAtEnd(); ++itr, ++itv, ++itb, ++itir, ++itLuminanceImage)
  {
    itLuminanceImage.Set((itr.Get() + itv.Get() + itb.Get() + itir.Get())/4.0);
  }

  // Extracting shadows
  typedef itk::CurvatureFlowImageFilter<InternImageType, InternImageType> CurvatureFlowImageFilterType;
  CurvatureFlowImageFilterType::Pointer smoothed = CurvatureFlowImageFilterType::New();
  smoothed->SetInput(m_LuminanceImage);
  smoothed->SetNumberOfIterations(5);
  smoothed->SetTimeStep(0.125);
  smoothed->Update();

  short threshold = 255-100;
  short lowerThreshold = -35;
  short upperThreshold = 100;

  // shadows are dark, so a thresholding with a low threshold is applied to the m_LuminanceImage image
  typedef otb::MultiplyByScalarImageFilter<InternImageType, InternImageType> MultiplyFilterType;
  typedef itk::ShiftScaleImageFilter<InternImageType, InternImageType> AddFilterType;
  MultiplyFilterType::Pointer multiplyFilter1 = MultiplyFilterType::New();
  multiplyFilter1->SetCoef(-1.0);
  multiplyFilter1->SetInput(smoothed->GetOutput());
  AddFilterType::Pointer addFilter = AddFilterType::New();
  addFilter->SetScale(1.0);
  addFilter->SetShift(255.0);
  addFilter->SetInput(multiplyFilter1->GetOutput());
  typedef itk::PointSet<InternPixelType, 2>  ShadowPointSetType;
  typedef otb::ThresholdImageToPointSetFilter<InternImageType, ShadowPointSetType> FilterThresholdType;
  FilterThresholdType::Pointer filterThreshold = FilterThresholdType::New();
  ShadowPointSetType::Pointer shadowPointSet = ShadowPointSetType::New();
  int iSeed;
  for (iSeed = 0; shadowPointSet->GetNumberOfPoints() < 1000; iSeed++)
  {
    upperThreshold += 5;
    filterThreshold->SetLowerThreshold(lowerThreshold);
    filterThreshold->SetUpperThreshold(upperThreshold);
    filterThreshold->SetInput(0, smoothed->GetOutput());
    shadowPointSet = filterThreshold->GetOutput();
    filterThreshold->Update();
  }
  threshold -= 5*iSeed;
  unsigned int nbOfPix = shadowPointSet->GetNumberOfPoints() / 50;
  unsigned int iCount = 0;
  typedef ShadowPointSetType::PointsContainer ShadowContainerType;
  ShadowContainerType* shadowPointsContainer = shadowPointSet->GetPoints();
  typedef ShadowContainerType::Iterator ShadowIteratorType;
  ShadowIteratorType itList = shadowPointsContainer->Begin();
  double replaceValue = 1;
  typedef itk::ConnectedThresholdImageFilter<InternImageType, InternImageType> InternConnectedFilterType;
  typedef itk::AddImageFilter<InternImageType, InternImageType, InternImageType> AddImageFilterType;
  typedef ShadowPointSetType::PointType ShadowsPointType;
  addFilter->Update();
  // then, a region growing is applied to the m_LuminanceImage image, using the seeds
  while (itList != shadowPointsContainer->End())
  {
    InternImageType::IndexType index;
    ShadowsPointType pCoordinate = (itList.Value());
    index[0] = static_cast <long> (pCoordinate[0]);
    index[1] = static_cast <long> (pCoordinate[1]);
    if (m_ExtractedShadows->GetPixel(index) == 0)
    {
      InternConnectedFilterType::Pointer connectedThreshold = InternConnectedFilterType::New();
      connectedThreshold->SetInput(addFilter->GetOutput());
      connectedThreshold->SetLower(threshold+lowerThreshold);
      connectedThreshold->SetUpper(threshold+upperThreshold);
      connectedThreshold->SetReplaceValue(replaceValue);
      connectedThreshold->SetSeed(index);
      AddImageFilterType::Pointer addImage = AddImageFilterType::New();
      addImage->SetInput1(m_ExtractedShadows);
      addImage->SetInput2(connectedThreshold->GetOutput());
      addImage->Update();
      m_ExtractedShadows = addImage->GetOutput();
      replaceValue += 0.0001;
    }
    iCount++;
    if (iCount == nbOfPix)
    {
      iCount = 0;
    }
    ++itList;
  }
  for (unsigned int i = 0; i<4; ++i)
  {
    m_FeatureImagesList->PushBack(imageListRescaler->GetOutput()->GetNthElement(i));
  }
  m_FeatureImagesList->PushBack(ndvi->GetOutput());
  m_FeatureImagesList->PushBack(ndvi2->GetOutput());
  m_FeatureImagesList->PushBack(waterIndice->GetOutput());
  m_FeatureImagesList->PushBack(m_ExtractedShadows);
}

void SVMBasedRegionGrowingSegmentationModel::ComputePointDependentFeaturesVector()
{
  //Erasing previous point dependent features
  for (unsigned int i = 8; i<m_FeatureImagesList->Size(); ++i)
  {
    m_FeatureImagesList->Erase(i);
  }

  // If less than 15 points in the pixelsInContainer, the object must be little.
  // Laplacian
  if (m_PixelsInPS->GetNumberOfPoints() < 10)
  {
    m_UseLaplacian = true;
    typedef itk::LaplacianRecursiveGaussianImageFilter<InternImageType, InternImageType> LRGType;
    LRGType::Pointer LRG = LRGType::New();
    LRG->SetInput(m_LuminanceImage);
    LRG->SetSigma(1);

    // inversing the image. with the LRG, pixels in the object are black, but for the next step, they must be white.
    typedef otb::MultiplyByScalarImageFilter<InternImageType, InternImageType> MultiplyFilterType;
    MultiplyFilterType::Pointer inversor = MultiplyFilterType::New();
    inversor->SetInput(LRG->GetOutput());
    inversor->SetCoef(-1);
    inversor->Update();
    m_FeatureImagesList->PushBack(inversor->GetOutput());
  }
  else
  {
    m_UseLaplacian = false;
  }

  // sorting references pixels by Kmean
  typedef InputMSImageType::PixelType MSPixel; // it is a variable length vector
  typedef itk::PointSet<MSPixel, 2> InputVectorPSType;
  // container for the pixels to sort
  typedef InputVectorPSType::PointDataContainer SortContainerType;
  SortContainerType::Pointer sortContainer = SortContainerType::New();
  // a class is a vector that contain the pixels inside of it
  typedef std::vector<MSPixel> ClasseType;
  // Omega is the list of all classes
  typedef std::vector<ClasseType> OmegaType;
  OmegaType omega(m_NumberOfSpectralAngleDistanceImages);
  // the prototype is the list of choosed reference pixels
  ClasseType prototype(m_NumberOfSpectralAngleDistanceImages); // prototype is a std::vector<MSPixel> too
  ContainerIteratorType pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
  ContainerIteratorType end = m_PixelsInPS->GetPoints()->End();
  unsigned int iSortContainer = 0;
  double zero = 0.0;
  MSPixel zerO(&zero, 4);
  // Initializing prototype
  for (unsigned int iProto = 0; pixelsInContainerIt != end; ++pixelsInContainerIt , iProto++)
  {
    InputMSImageType::IndexType sortIndex;
    InputPointType pp = pixelsInContainerIt.Value();
    sortIndex[0] = static_cast <long> (pp[0]);
    sortIndex[1] = static_cast <long> (pp[1]);
    sortContainer->InsertElement(iSortContainer++, m_InputImage->GetPixel(sortIndex));
    if (iProto < m_NumberOfSpectralAngleDistanceImages) //the NumberOfSpectralAngleDistanceImages first points are put in the prototype
    {
      prototype[iProto] = m_InputImage->GetPixel(sortIndex);
    }
  }
  int nbOfSAD(0);
  // if there is less input points than NB_OF_SAD, Kmean is useless
  if (m_PixelsInPS->GetNumberOfPoints() > m_NumberOfSpectralAngleDistanceImages)
  {
    nbOfSAD = m_NumberOfSpectralAngleDistanceImages;
    unsigned int nombreIterationKmean = 0;
    while (nombreIterationKmean < 10) // There is no convergence test, a limite of 10 iterations
    {
      typedef SortContainerType::Iterator SortContainerItType;
      SortContainerItType sortContainerIt;
      // Browsing points for classification
      for (sortContainerIt = sortContainer->Begin(); sortContainerIt != sortContainer->End(); ++sortContainerIt)
      {
        MSPixel sortPixie = sortContainerIt.Value();
        double mini = itk::NumericTraits<double>::max();
        int iMini = 0;
        // Browsing prototype for classification of this point by spectral angle
        for (unsigned int iClassif = 0; iClassif < m_NumberOfSpectralAngleDistanceImages; iClassif++)
        {
          if (prototype[iClassif] != zerO)
          {
            double dist = 0.0;
            double scalarProd = 0.0;
            double normProd = 0.0;
            double normProd1 = 0.0;
            double normProd2 = 0.0;
            for (unsigned int i = 0; i < sortPixie.Size(); i++)
            {
              scalarProd += sortPixie[i]*prototype[iClassif][i];
              normProd1 += sortPixie[i]*sortPixie[i];
              normProd2 += prototype[iClassif][i]*prototype[iClassif][i];
            }
            normProd = normProd1 * normProd2;
            if (normProd == 0.0)
            {
              dist = itk::NumericTraits<double>::max();
            }
            else
            {
              dist = vcl_acos(scalarProd/vcl_sqrt(normProd));
            }
            if (dist < mini)
            {
              mini = dist;
              iMini = iClassif;
            }
          }
        } // end browsing prototype
        // once the good class has been found, value is added to the good class.
        omega[iMini].push_back(sortPixie);
      } // end browsing points

      // actualisation of prototype
      for (unsigned int iActu = 0; iActu < m_NumberOfSpectralAngleDistanceImages; iActu++)
      {
        MSPixel sum = zerO;
        unsigned int nbElem = omega[iActu].size();
        if (nbElem != 0)
        {
          for (unsigned int iSum = 0; iSum < nbElem; iSum++)
          {
            sum += omega[iActu][iSum];
          }
          sum /= nbElem;
        }
        prototype[iActu] = sum;
      } // end actualisation
      nombreIterationKmean++;
    }
  } // end Kmean
  else   // there is less input points than m_NumberOfSpectralDistanceImages
  {
    pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
    end = m_PixelsInPS->GetPoints()->End();
    unsigned int iSortContainer = 0;
    for (unsigned int iProto = 0; pixelsInContainerIt != end; ++pixelsInContainerIt , iProto++)
    {
      nbOfSAD = m_PixelsInPS->GetNumberOfPoints();
      InputMSImageType::IndexType sortIndex;
      InputPointType pp = pixelsInContainerIt.Value();
      sortIndex[0] = static_cast <long> (pp[0]);
      sortIndex[1] = static_cast <long> (pp[1]);
      sortContainer->InsertElement(iSortContainer++, m_InputImage->GetPixel(sortIndex));
      //points are put in the prototype
      prototype[iProto] = m_InputImage->GetPixel(sortIndex);
    }
  }
  typedef otb::ObjectList<SpectralAngleType> SpectralAngleListType;
  SpectralAngleListType::Pointer spectralAngleList = SpectralAngleListType::New();
  typedef otb::ObjectList<SqrtFilterType> SqrtFilterListType;
  SqrtFilterListType::Pointer sqrtFilterList = SqrtFilterListType::New();
  for (int iSAD = 0; iSAD < nbOfSAD; iSAD++) // computing m_NumberOfSpectralDistanceImages spectral angle distances
  {
    SpectralAngleType::Pointer spectralAngleDistance = SpectralAngleType::New();
    spectralAngleDistance->SetInput(m_InputImage);
    spectralAngleDistance->SetReferencePixel(prototype[iSAD]);
    spectralAngleList->PushBack(spectralAngleDistance);

    SqrtFilterType::Pointer sqrtFilter = SqrtFilterType::New();
    sqrtFilter->SetInput(spectralAngleDistance->GetOutput());
    sqrtFilter->Update();
    sqrtFilterList->PushBack(sqrtFilter);
    m_FeatureImagesList->PushBack(sqrtFilter->GetOutput());
  }
}

void SVMBasedRegionGrowingSegmentationModel::ComputeHyperplanDistanceImage()
{
  const unsigned int syntheseSize = m_FeatureImagesList->Size();
  unsigned int pointID = 0;
  typedef itk::PointSet<VectorType, 2> featurePSType;
  typedef featurePSType::PointDataContainer PointDataContainer;
  typedef LabelPSType::PointDataContainer LabelContainerType;
  LabelContainerType::Pointer labelContainer = LabelContainerType::New();
  InputContainerType::Pointer learnPointContainer = InputContainerType::New();
  PointDataContainer::Pointer learnValueContainer = PointDataContainer::New();
  ContainerIteratorType pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
  ContainerIteratorType end = m_PixelsInPS->GetPoints()->End();
  // browsing input pointsets to create learning pointsets
  while (pixelsInContainerIt != end)
  {
    InputPointType pCoordinate = (pixelsInContainerIt.Value());
    learnPointContainer->InsertElement(pointID, pCoordinate);
    InternImageType::IndexType index;
    index[0] = static_cast <long> (pCoordinate[0]);
    index[1] = static_cast <long> (pCoordinate[1]);
    VectorType pixie(syntheseSize);
    for (unsigned int iLoop = 0; iLoop < syntheseSize; iLoop++)
    {
      pixie[iLoop] = m_FeatureImagesList->GetNthElement(iLoop)->GetPixel(index);
    }
    learnValueContainer->InsertElement(pointID, pixie);
    labelContainer->InsertElement(pointID, 1);
    pointID++;
    pixelsInContainerIt++;
  }
  ContainerIteratorType pixelsOutContainerIt = m_PixelsOutPS->GetPoints()->Begin();
  end = m_PixelsOutPS->GetPoints()->End();
  while (pixelsOutContainerIt != end)
  {
    InputPointType pCoordinate = (pixelsOutContainerIt.Value());
    learnPointContainer->InsertElement(pointID, pCoordinate);
    InternImageType::IndexType index;
    index[0] = static_cast <long> (pCoordinate[0]);
    index[1] = static_cast <long> (pCoordinate[1]);
    VectorType pixie(syntheseSize);
    for (unsigned int iLoop = 0; iLoop < syntheseSize; iLoop++)
    {
      pixie[iLoop] = m_FeatureImagesList->GetNthElement(iLoop)->GetPixel(index);
    }
    learnValueContainer->InsertElement(pointID, pixie);
    labelContainer->InsertElement(pointID, -1);
    pointID++;
    pixelsOutContainerIt++;
  }


  featurePSType::Pointer featuresPS = featurePSType::New();
  featuresPS->SetPoints(learnPointContainer);
  featuresPS->SetPointData(learnValueContainer);
  LabelPSType::Pointer labelPS = LabelPSType::New();
  labelPS->SetPoints(learnPointContainer);
  labelPS->SetPointData(labelContainer);

  // Estimating SVM model
  typedef otb::SVMPointSetModelEstimator<featurePSType, LabelPSType> EstimatorType;
  EstimatorType::Pointer estimator = EstimatorType::New();
  estimator->SetInputPointSet(featuresPS);
  estimator->SetTrainingPointSet(labelPS);
  estimator->Update();
  typedef otb::SVMModel<InternPixelType, int> ModelType;
  ModelType::Pointer model = estimator->GetModel();

  // Segmenting synthese image by SVM
  typedef ModelType::DistancesVectorType SVMOutputPixelType;
  typedef ModelType::MeasurementType MeasurementType;
  typedef InternImageListType::Iterator InternImageListItType;
  // image that will containe the result of evaluation of the distance to the hyperplan for eatch pixels
  typedef itk::ImageRegionIterator<InternImageType> InternImageItType;
  InternImageItType syntheticImageIt(m_HyperplanDistanceImage, m_InputImage->GetLargestPossibleRegion());
  for (syntheticImageIt.GoToBegin(); !syntheticImageIt.IsAtEnd(); ++syntheticImageIt)
  {
    MeasurementType svmNode;

    InternImageListItType syntheseImageListIt = m_FeatureImagesList->Begin();
    InternImageListItType end = m_FeatureImagesList->End();
    while (syntheseImageListIt != end)
    {
      svmNode.push_back(syntheseImageListIt.Get()->GetPixel(syntheticImageIt.GetIndex()));
      ++syntheseImageListIt;
    }

    SVMOutputPixelType pixie2;
    pixie2 = model->EvaluateHyperplanesDistances(svmNode);
    syntheticImageIt.Set(pixie2[0]);
  }
  // amplification of shadows edges in hyperplan distance map
  // first, finding shadows that lies at the north-west of the object. inside pixels are used.
  typedef itk::LineIterator<InternImageType> LineIteratorType;
  typedef itk::ConnectedThresholdImageFilter<InternImageType, InternImageType> InternConnectedFilterType;
  InternConnectedFilterType::Pointer bonneOmbre = InternConnectedFilterType::New();
  pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
  end = m_PixelsInPS->GetPoints()->End();
  while (pixelsInContainerIt != end)
  {
    InputPointType pCoordinate = (pixelsInContainerIt.Value());
    learnPointContainer->InsertElement(pointID, pCoordinate);
    InternImageType::IndexType endLine;
    InternImageType::IndexType selectedPixel;
    selectedPixel[0] = static_cast <long> (pCoordinate[0]);
    selectedPixel[1] = static_cast <long> (pCoordinate[1]);
    // creating a line that go from the pixel to the north-west
    if (selectedPixel[0] <= selectedPixel[1])
    {
      endLine[0] = 0;
      endLine[1] = selectedPixel[1] - selectedPixel[0];
    }
    else
    {
      endLine[1] = 0;
      endLine[0] = selectedPixel[0] - selectedPixel[1];
    }
    LineIteratorType itLigneOmbre(m_ExtractedShadows, selectedPixel, endLine);
    itLigneOmbre.GoToBegin();
    // and finding the first shadow
    while ((!itLigneOmbre.IsAtEnd()) && (itLigneOmbre.Get() == 0))
    {
      ++itLigneOmbre;
    }
    // then, region growing
    if (!itLigneOmbre.IsAtEnd())
    {
      bonneOmbre->AddSeed(itLigneOmbre.GetIndex());
    }
    pixelsInContainerIt++;
  }
  bonneOmbre->SetInput(m_ExtractedShadows);
  bonneOmbre->SetLower(1);
  bonneOmbre->SetUpper(itk::NumericTraits<InternPixelType>::max());
  bonneOmbre->SetReplaceValue(10);
  bonneOmbre->Update();

  // Values of the edges of the good shadows will be setted to the minimum value of the hyperplan distance map
  typedef itk::MinimumMaximumImageCalculator<InternImageType> MinMaxType;
  MinMaxType::Pointer minMax = MinMaxType::New();
  minMax->SetImage(m_HyperplanDistanceImage);
  minMax->ComputeMinimum();
  typedef itk::GradientMagnitudeImageFilter<InternImageType, InternImageType> GradientMagnitudeType;
  GradientMagnitudeType::Pointer contoursOmbres = GradientMagnitudeType::New();
  contoursOmbres->SetInput(bonneOmbre->GetOutput());
  typedef itk::BinaryThresholdImageFilter<InternImageType, InternImageType> BinaryThresholdImageType;
  BinaryThresholdImageType::Pointer shadowsEdges = BinaryThresholdImageType::New();
  shadowsEdges->SetInput(contoursOmbres->GetOutput());
  shadowsEdges->SetLowerThreshold(3);
  shadowsEdges->SetUpperThreshold(itk::NumericTraits<InternPixelType>::max());
  shadowsEdges->SetOutsideValue(0);
  shadowsEdges->SetInsideValue(minMax->GetMinimum());
  typedef itk::AddImageFilter<InternImageType, InternImageType, InternImageType> AddImageFilterType;
  AddImageFilterType::Pointer gradSyntheticImage1 = AddImageFilterType::New();
  gradSyntheticImage1->SetInput1(m_HyperplanDistanceImage);
  gradSyntheticImage1->SetInput2(shadowsEdges->GetOutput());
  gradSyntheticImage1->Update();
  m_HyperplanDistanceImage = gradSyntheticImage1->GetOutput();
  pixelsOutContainerIt = m_PixelsOutPS->GetPoints()->Begin();
  end = m_PixelsOutPS->GetPoints()->End();
  while (pixelsOutContainerIt != end)
  {
    InputPointType pCoordinate = pixelsOutContainerIt.Value();
    InternImageType::IndexType index;
    index[0] = static_cast <long> (pCoordinate[0]);
    index[1] = static_cast <long> (pCoordinate[1]);
    m_HyperplanDistanceImage->SetPixel(index, minMax->GetMinimum());
    ++pixelsOutContainerIt;
  }
}


void SVMBasedRegionGrowingSegmentationModel::ComputeRegionGrowing()
{
  if (m_PixelsInPS->GetNumberOfPoints()>0 && m_PixelsOutPS->GetNumberOfPoints()>0)
  {
    typedef itk::Statistics::ScalarImageToHistogramGenerator<InternImageType> HistogramGeneratorType;
    HistogramGeneratorType::Pointer histogram = HistogramGeneratorType::New();
    histogram->SetInput(m_HyperplanDistanceImage);
    histogram->SetNumberOfBins(128);
    histogram->Compute();
    typedef itk::OtsuMultipleThresholdsCalculator<HistogramGeneratorType::HistogramType> OtsuType;
    OtsuType::Pointer thresholdCalculator = OtsuType::New();
    thresholdCalculator->SetInputHistogram(histogram->GetOutput());
    thresholdCalculator->SetNumberOfThresholds(m_NumberOfThresholds);
    try
    {
      thresholdCalculator->Update();
    }
    catch ( itk::ExceptionObject & excp )
    {
      std::cerr << "Exception thrown " << excp << std::endl;
    }
    const OtsuType::OutputType &thresholds = thresholdCalculator->GetOutput();


    // selecting the best threshold by comparing them with the inside pixels
    int iThreshold = m_NumberOfThresholds - 1;
    ContainerIteratorType pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
    ContainerIteratorType end = m_PixelsInPS->GetPoints()->End();
    while (pixelsInContainerIt != end)
    {
      InputPointType pCoordinate = (pixelsInContainerIt.Value());
      InternImageType::IndexType index;
      index[0] = static_cast <long> (pCoordinate[0]);
      index[1] = static_cast <long> (pCoordinate[1]);
      // if the threshold is too high, the next is selected
      while ((m_HyperplanDistanceImage->GetPixel(index) < thresholds[iThreshold]) && iThreshold != 0)
      {
        iThreshold--;
      }
      ++pixelsInContainerIt;
    }
    pixelsInContainerIt = m_PixelsInPS->GetPoints()->Begin();
    end = m_PixelsInPS->GetPoints()->End();
    while (pixelsInContainerIt != end)
    {
      InputPointType pCoordinate = pixelsInContainerIt.Value();
      InternImageType::IndexType index;
      index[0] = static_cast <long> (pCoordinate[0]);
      index[1] = static_cast <long> (pCoordinate[1]);
      m_ConnectedFilter->AddSeed(index);
      ++pixelsInContainerIt;
    }
    m_ConnectedFilter->SetLower(thresholds[iThreshold]);
  }
  else
  {
    m_ConnectedFilter->SetLower(itk::NumericTraits<InternPixelType>::max());
  }

  m_ConnectedFilter->SetInput(m_HyperplanDistanceImage);
  m_ConnectedFilter->SetReplaceValue(255);
  m_ConnectedFilter->SetUpper(itk::NumericTraits<InternPixelType>::max());
  m_ConnectedFilter->Update();
}
}

