/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbSVMBasedRegionGrowingSegmentationModel_h
#define __otbSVMBasedRegionGrowingSegmentationModel_h

#include "otbMVCModel.h"
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageList.h"
#include "itkPointSet.h"
#include "itkConnectedThresholdImageFilter.h"
#include "otbSpectralAngleDistanceImageFilter.h"
#include "itkSqrtImageFilter.h"

#include "otbListenerBase.h"

namespace otb
{


/** \class SVMBasedRegionGrowingSegmentationModel
 *
 *
 *
 */
class ITK_EXPORT SVMBasedRegionGrowingSegmentationModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef SVMBasedRegionGrowingSegmentationModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                                Pointer;
  typedef itk::SmartPointer<const Self>                          ConstPointer;

  /** Standard type macro */
  itkTypeMacro(SVMBasedRegionGrowingSegmentationModel, Object);

  // pixel type of the input images
  typedef double InputPixelType;
  // pixel type of the intern images
  typedef double InternPixelType;
  // pixel type for classification
  typedef std::vector<InternPixelType> VectorType;
  // image type of the multispectral input image
  typedef otb::VectorImage<InputPixelType, 2> InputMSImageType;
  // image type of the intern image
  typedef otb::Image<InternPixelType, 2> InternImageType;
  typedef InternImageType::Pointer      InternImagePointerType;
  typedef InternImageType::IndexType    IndexType;
  /// Output image type
  typedef unsigned char                  UCharPixelType;
  typedef otb::Image<UCharPixelType, 2>   OutputImageType;
  typedef OutputImageType::Pointer       OutputImagePointerType;
  // List that will contain the extracted informations for the SVM
  typedef otb::ImageList<InternImageType> InternImageListType;
  // Pointset that will contain the labels for the SVM
  typedef itk::PointSet<int, 2> LabelPSType;
  // Type of the Pointsets that will contain the markers
  typedef itk::PointSet<InternPixelType, 2> InputPSType;
  typedef InputPSType::PointType PointType;
  typedef itk::PointSet<VectorType, 2> FeaturePSType;
  typedef InputPSType::PointsContainer InputContainerType;
  typedef InputContainerType::Iterator ContainerIteratorType;
  typedef InputPSType::PointType InputPointType;

  // Useful filters
  typedef otb::SpectralAngleDistanceImageFilter<InputMSImageType, InternImageType> SpectralAngleType;
  typedef itk::SqrtImageFilter<InternImageType, InternImageType> SqrtFilterType;
  typedef itk::ConnectedThresholdImageFilter<InternImageType, OutputImageType> ConnectedFilterType;


  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Get the flags to activate/deactivate the preprocessings */
  itkGetMacro(Valid, bool);
  itkGetMacro(UseLaplacian, bool);
  itkSetMacro(NearInfraRedChannelIndex, unsigned int);
  itkGetMacro(NearInfraRedChannelIndex, unsigned int);
  itkSetMacro(RedChannelIndex, unsigned int);
  itkGetMacro(RedChannelIndex, unsigned int);
  itkSetMacro(GreenChannelIndex, unsigned int);
  itkGetMacro(GreenChannelIndex, unsigned int);
  itkSetMacro(BlueChannelIndex, unsigned int);
  itkGetMacro(BlueChannelIndex, unsigned int);
  itkGetMacro(NumberOfThresholds, unsigned int);
  itkGetMacro(NumberOfSpectralAngleDistanceImages, unsigned int);
  itkGetObjectMacro(PixelsInPS, InputPSType);
  itkGetObjectMacro(PixelsOutPS, InputPSType);

  void SetNumberOfSpectralAngleDistanceImages(unsigned int nb);
  void SetNumberOfThresholds(unsigned int nb);

  /** Get the output image */
  void SetInputImage(InputMSImageType::Pointer image);
  itkGetObjectMacro(InputImage, InputMSImageType);
  itkGetObjectMacro(HyperplanDistanceImage, InternImageType);
  itkGetObjectMacro(FeatureImagesList, InternImageListType);


  /** Get the output image */
  OutputImagePointerType GetOutputImage();

  void AddInsidePoint(PointType & point);
  void AddOutsidePoint(PointType & point);
  void ClearSeeds();
  void Update();

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  SVMBasedRegionGrowingSegmentationModel();
  /** Destructor */
  ~SVMBasedRegionGrowingSegmentationModel();

private:
  SVMBasedRegionGrowingSegmentationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Update the pipeline */
  void UpdatePipeline(void);

  /** Compute the feature vectors */
  void ComputeFeaturesVector();
  void ComputeStaticFeaturesVector();
  void ComputePointDependentFeaturesVector();
  void ComputeHyperplanDistanceImage();
  void ComputeRegionGrowing();

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Data */
  InputPSType::Pointer m_PixelsInPS;
  InputPSType::Pointer m_PixelsOutPS;

  /** Channel indices */
  unsigned int m_NearInfraRedChannelIndex;
  unsigned int m_RedChannelIndex;
  unsigned int m_GreenChannelIndex;
  unsigned int m_BlueChannelIndex;

  /** Number of spectral angle distance images in features vector */
  unsigned int m_NumberOfSpectralAngleDistanceImages;

  /** Number of Otsu thresholds */
  unsigned int m_NumberOfThresholds;

  /** Filters */
  ConnectedFilterType::Pointer m_ConnectedFilter;

  /** Flags to activate/deactivate the preprocessings */
  bool m_Valid;
  bool m_UseLaplacian;

  /** The instance singleton */
  static Pointer Instance;

  /** Input image */
  InputMSImageType::Pointer    m_InputImage;
  InternImageType::Pointer     m_HyperplanDistanceImage;
  InternImageListType::Pointer m_FeatureImagesList;
  InternImageType::Pointer     m_LuminanceImage;
  InternImageType::Pointer     m_ExtractedShadows;

};
}
#endif
