/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbSegmentationApplicationModel.h"
#include "otbVectorDataProjectionFilter.h"

namespace otb
{

/** Initialize the singleton */
SegmentationApplicationModel::Pointer SegmentationApplicationModel::Instance = NULL;

/** Manage the singleton */
SegmentationApplicationModel::Pointer
SegmentationApplicationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = SegmentationApplicationModel::New();
  }
  return Instance;
}

void SegmentationApplicationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

SegmentationApplicationModel::SegmentationApplicationModel()
{
  // filters
  m_Reader = ReaderType::New();
  m_IntensityFilter  = IntensityFilterType::New();

  // Sub models
  m_PreprocessingModel = PreprocessingModel::GetInstance();
  m_VectorizationModel = VectorizationModel::GetInstance();
  m_RegionGrowingModel = RegionGrowingSegmentationModel::GetInstance();
  m_LevelSetModel      = LevelSetSegmentationModel::GetInstance();
  m_SVMBasedModel         = SVMBasedRegionGrowingSegmentationModel::GetInstance();

  m_Valid = false;
  m_FileName = "";
  m_UseIntensity = true;
  m_Algorithm = REGION_GROWING;

  IndexType index;
  index.Fill(0);
  SizeType size;
  size.Fill(0);
  m_Region.SetIndex(index);
  m_Region.SetSize(size);

  m_PolygonList = PolygonListType::New();
}

SegmentationApplicationModel::~SegmentationApplicationModel()
{}

void SegmentationApplicationModel::SetFileName(std::string file)
{
  m_FileName = file;
  m_Reader->SetFileName(m_FileName);

  this->SetIsNewImage(true);
  /// TODO: try/catch missing there (should be done in the view ?)
  m_Reader->GenerateOutputInformation();
  m_Valid = true;
  this->NotifyAll();

  this->SetIsNewImage(false);
}

SegmentationApplicationModel::VectorImageType * SegmentationApplicationModel::GetImage()
{
  return m_Reader->GetOutput();
}

void SegmentationApplicationModel::SetUseChannel()
{
  m_UseIntensity = false;
}

void SegmentationApplicationModel::SetUseIntensity()
{
  m_UseIntensity = true;
}

void SegmentationApplicationModel::SetupSegmentation()
{
  m_ExtractROIFilter = ExtractROIFilterType::New();
  m_MultiToMonoChannelExtractROIFilter = MultiToMonoChannelExtractROIFilterType::New();
  m_IntensityFilter->SetInput(m_Reader->GetOutput());
  m_ExtractROIFilter->SetInput(m_IntensityFilter->GetOutput());
  m_MultiToMonoChannelExtractROIFilter->SetInput(m_Reader->GetOutput());
  m_ExtractROIFilter->SetExtractionRegion(m_Region);
  m_MultiToMonoChannelExtractROIFilter->SetExtractionRegion(m_Region);
  m_MultiToMonoChannelExtractROIFilter->SetChannel(m_ChannelIndex+1);

  m_MultiChannelExtractROIFilter = MultiChannelExtractROIFilterType::New();
  m_MultiChannelExtractROIFilter->SetInput(m_Reader->GetOutput());
  m_MultiChannelExtractROIFilter->SetExtractionRegion(m_Region);

  if (m_UseIntensity)
  {
    m_ExtractROIFilter->Update();
    m_PreprocessingModel->SetInputImage(m_ExtractROIFilter->GetOutput());
    m_VectorizationModel->SetReferenceImage(m_ExtractROIFilter->GetOutput());
  }
  else
  {
    m_MultiToMonoChannelExtractROIFilter->Update();
    m_PreprocessingModel->SetInputImage(m_MultiToMonoChannelExtractROIFilter->GetOutput());
    m_VectorizationModel->SetReferenceImage(m_MultiToMonoChannelExtractROIFilter->GetOutput());
  }

  m_MultiChannelExtractROIFilter->Update();

  m_PreprocessingModel->UnRegisterListener(m_LevelSetModel);
  m_PreprocessingModel->UnRegisterListener(m_RegionGrowingModel);
  m_LevelSetModel->UnRegisterListener(m_VectorizationModel);
  m_RegionGrowingModel->UnRegisterListener(m_VectorizationModel);
  m_SVMBasedModel->UnRegisterListener(m_VectorizationModel);

  switch (m_Algorithm)
  {
  case LEVEL_SET:
    m_LevelSetModel->SetInputImage(m_PreprocessingModel->GetOutputImage());
    m_VectorizationModel->SetSegmentationImage(m_LevelSetModel->GetOutputImage());
    m_PreprocessingModel->RegisterListener(m_LevelSetModel);
    m_LevelSetModel->RegisterListener(m_VectorizationModel);
    break;
  case REGION_GROWING:
    m_RegionGrowingModel->SetInputImage(m_PreprocessingModel->GetOutputImage());
    m_VectorizationModel->SetSegmentationImage(m_RegionGrowingModel->GetOutputImage());
    m_PreprocessingModel->RegisterListener(m_RegionGrowingModel);
    m_RegionGrowingModel->RegisterListener(m_VectorizationModel);
    break;
  case SVM_BASED_REGION_GROWING:
    m_SVMBasedModel->SetInputImage(m_MultiChannelExtractROIFilter->GetOutput());
    m_VectorizationModel->SetSegmentationImage(m_SVMBasedModel->GetOutputImage());
    m_SVMBasedModel->RegisterListener(m_VectorizationModel);
  }
}

void SegmentationApplicationModel::RetrieveSegmentationResult()
{
  PolygonListPointerType newPolygonList = m_VectorizationModel->GetPolygonList();
  PolygonListType::Iterator pit = newPolygonList->Begin();
  while (pit != newPolygonList->End())
  {
    PolygonPointerType newPolygon = PolygonType::New();

    for (PolygonType::VertexListConstIteratorType vit = pit.Get()->GetVertexList()->Begin();
         vit!=pit.Get()->GetVertexList()->End(); ++vit)
    {
      PolygonType::VertexType newVertex = vit.Value();
      newVertex[0]+=m_Region.GetIndex()[0];
      newVertex[1]+=m_Region.GetIndex()[1];
      newPolygon->AddVertex(newVertex);
    }
    if (newPolygon->GetVertexList()->Size()>0)
    {
      m_PolygonList->PushBack(newPolygon);
    }
    ++pit;
  }
  this->NotifyAll();
}

void SegmentationApplicationModel::ClearPolygonList()
{
  m_PolygonList->Clear();
  this->NotifyAll();
}


void SegmentationApplicationModel::DeletePolygon(unsigned int index)
{
  m_PolygonList->Erase(index);
  this->NotifyAll();
}

void SegmentationApplicationModel::SaveResults(std::string filename)
{
  VectorDataPointerType vectorData = VectorDataType::New();

  DataNodePointerType document = DataNodeType::New();
  document->SetNodeType(DOCUMENT);
  document->SetNodeId("polygon");
  DataNodePointerType folder = DataNodeType::New();
  folder->SetNodeType(FOLDER);

  DataTreePointerType tree = vectorData->GetDataTree();
  DataNodePointerType root = tree->GetRoot()->Get();

  tree->Add(document, root);
  tree->Add(folder, document);

  PolygonListType::Iterator pit = m_PolygonList->Begin();

  while (pit != m_PolygonList->End())
  {
    DataNodePointerType newPolygon = DataNodeType::New();
    newPolygon->SetPolygonExteriorRing(pit.Get());
    tree->Add(newPolygon, folder);
    ++pit;
  }

  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  ProjectionFilterType::Pointer vectorDataProjection = ProjectionFilterType::New();
  vectorDataProjection->SetInput(vectorData);
  vectorDataProjection->SetInputOrigin(m_Reader->GetOutput()->GetOrigin());
  vectorDataProjection->SetInputSpacing(m_Reader->GetOutput()->GetSpacing());

  std::string projectionRef;
  itk::ExposeMetaData<std::string>(m_Reader->GetOutput()->GetMetaDataDictionary(),
                                   MetaDataKey::ProjectionRefKey, projectionRef );
  vectorDataProjection->SetInputProjectionRef(projectionRef);

  vectorDataProjection->SetInputKeywordList(m_Reader->GetOutput()->GetImageKeywordlist());

  VectorDataFileWriterPointerType writer = VectorDataFileWriterType::New();
  writer->SetInput(vectorDataProjection->GetOutput());
  writer->SetFileName(filename);

  try
  {
    writer->Update();
  }
  catch ( itk::ExceptionObject & err )
  {
    std::cerr<<"Error while writing data file: "<<err<<std::endl;
  }
}

void SegmentationApplicationModel::ImportSegments(std::string filename)
{
  VectorDataFileReaderPointerType reader = VectorDataFileReaderType::New();
  reader->SetFileName(filename);

  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  ProjectionFilterType::Pointer vectorDataProjection = ProjectionFilterType::New();
  vectorDataProjection->SetInput(reader->GetOutput());

  if (m_Valid)
  {
    vectorDataProjection->SetOutputOrigin(m_Reader->GetOutput()->GetOrigin());
    vectorDataProjection->SetOutputSpacing(m_Reader->GetOutput()->GetSpacing());

    std::string projectionRef;
    itk::ExposeMetaData<std::string>(m_Reader->GetOutput()->GetMetaDataDictionary(),
                                     MetaDataKey::ProjectionRefKey, projectionRef );
    vectorDataProjection->SetOutputProjectionRef(projectionRef);

    vectorDataProjection->SetOutputKeywordList(m_Reader->GetOutput()->GetImageKeywordlist());
  }

  vectorDataProjection->Update();

  TreeIteratorType it(vectorDataProjection->GetOutput()->GetDataTree());

  it.GoToBegin();

  while (!it.IsAtEnd())
  {
    if (it.Get()->IsPolygonFeature())
    {
      m_PolygonList->PushBack(it.Get()->GetPolygonExteriorRing());
    }
    ++it;
  }
  this->NotifyAll();
}

}
