/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbSegmentationApplicationModel_h
#define __otbSegmentationApplicationModel_h

#include "otbMVCModel.h"

#include "otbImage.h"
#include "otbImageFileReader.h"
#include "otbMultiToMonoChannelExtractROI.h"
#include "otbMultiChannelExtractROI.h"
#include "otbExtractROI.h"
#include "otbVectorImageToIntensityImageFilter.h"
#include "otbPolygon.h"
#include "otbObjectList.h"
#include "otbVectorDataFileWriter.h"

#include "otbListenerBase.h"

#include "otbPreprocessingModel.h"
#include "otbLevelSetSegmentationModel.h"
#include "otbRegionGrowingSegmentationModel.h"
#include "otbSVMBasedRegionGrowingSegmentationModel.h"
#include "otbVectorizationModel.h"

#include "otbVectorData.h"
#include "otbVectorDataFileWriter.h"
#include "otbVectorDataFileReader.h"
#include "itkPreOrderTreeIterator.h"


namespace otb
{

typedef enum
{
  LEVEL_SET,
  REGION_GROWING,
  SVM_BASED_REGION_GROWING
} SegmentationAlgorithm;

/** \class SegmentationApplicationModel
 *
 *
 *
 */
class ITK_EXPORT SegmentationApplicationModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef SegmentationApplicationModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                              Pointer;
  typedef itk::SmartPointer<const Self>                        ConstPointer;

  /** Standard type macro */
  itkTypeMacro(SegmentationApplicationModel, Object);

  typedef double                                                     PixelType;
  typedef double                                                     LabelType;
  typedef otb::VectorImage<PixelType, 2>                              VectorImageType;
  typedef VectorImageType::Pointer                                   VectorImagePointerType;
  typedef VectorImageType::SizeType                                  SizeType;
  typedef VectorImageType::IndexType                                 IndexType;
  typedef VectorImageType::RegionType                                RegionType;
  typedef otb::Image<PixelType, 2>                                    ImageType;
  typedef ImageType::Pointer                                         ImagePointerType;
  typedef otb::ImageFileReader<VectorImageType>                      ReaderType;
  typedef ReaderType::Pointer                                        ReaderPointerType;
  typedef otb::VectorImageToIntensityImageFilter<VectorImageType, ImageType> IntensityFilterType;
  typedef IntensityFilterType::Pointer                               IntensityFilterPointerType;
  typedef otb::ExtractROI<PixelType, PixelType>                       ExtractROIFilterType;
  typedef ExtractROIFilterType::Pointer                              ExtractROIFilterPointerType;
  typedef otb::MultiToMonoChannelExtractROI<PixelType, PixelType>     MultiToMonoChannelExtractROIFilterType;
  typedef MultiToMonoChannelExtractROIFilterType::Pointer            MultiToMonoChannelExtractROIFilterPointerType;
  typedef MultiChannelExtractROI<PixelType, PixelType>                MultiChannelExtractROIFilterType;
  typedef MultiChannelExtractROIFilterType::Pointer                  MultiChannelExtractROIFilterPointerType;
  typedef PreprocessingModel::Pointer                                PreprocessingModelPointerType;
  typedef VectorizationModel::Pointer                                VectorizationmodelPointerType;
  typedef RegionGrowingSegmentationModel::Pointer                    RegionGrowingModelPointerType;
  typedef SVMBasedRegionGrowingSegmentationModel::Pointer               SVMBasedRegionGrowingSegmentationModelPointerType;
  typedef LevelSetSegmentationModel::Pointer                         LevelSetModelPointerType;
  typedef otb::Polygon<double>                                       PolygonType;
  typedef PolygonType::Pointer                                       PolygonPointerType;
  typedef PolygonType::ContinuousIndexType                           ContinuousIndexType;
  typedef otb::ObjectList<PolygonType>                               PolygonListType;
  typedef PolygonListType::Pointer                                   PolygonListPointerType;

  /** Vector data handling */
  typedef VectorData<double, 2>                                       VectorDataType;
  typedef VectorDataType::Pointer                                    VectorDataPointerType;
  typedef VectorDataFileWriter<VectorDataType>                       VectorDataFileWriterType;
  typedef VectorDataFileWriterType::Pointer                          VectorDataFileWriterPointerType;
  typedef VectorDataType::DataNodeType                               DataNodeType;
  typedef DataNodeType::Pointer                                      DataNodePointerType;
  typedef VectorDataType::DataTreeType                               DataTreeType;
  typedef DataTreeType::Pointer                                      DataTreePointerType;
  typedef VectorDataFileReader<VectorDataType>                       VectorDataFileReaderType;
  typedef VectorDataFileReaderType::Pointer                          VectorDataFileReaderPointerType;
  typedef itk::PreOrderTreeIterator<DataTreeType>                    TreeIteratorType;

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  /** Setters */
  itkSetMacro(Region, RegionType);
  itkGetConstReferenceMacro(Region, RegionType);

  /** Getters */
  itkGetMacro(Valid, bool);
  itkGetObjectMacro(PolygonList, PolygonListType);
  itkGetStringMacro(FileName);
  itkGetMacro(Region, RegionType);
  itkGetMacro(UseIntensity, bool);
  itkGetMacro(ChannelIndex, unsigned int);
  itkSetMacro(ChannelIndex, unsigned int);
  itkGetMacro(Algorithm, SegmentationAlgorithm);
  itkSetMacro(Algorithm, SegmentationAlgorithm);
  itkGetMacro(IsNewImage, bool);
  itkSetMacro(IsNewImage, bool);

  /** Get the image */
  VectorImageType * GetImage();

  /** Get the filename */
  void SetFileName(std::string file);

  /** Set the segmentation input to a single channel */
  void SetUseChannel();

  /** Set the segmentation input to the intensity image */
  void SetUseIntensity();

  /** wire the model with the segmentation */
  void SetupSegmentation();

  /** Retrieve the segmented polygon */
  void RetrieveSegmentationResult();

  /** Clear the polygon list */
  void ClearPolygonList();

  /** Delete a given polygon */
  void DeletePolygon(unsigned int index);

  /** Save segmentation result to filename */
  void SaveResults(std::string filename);

  /** Import a shapefile into the segmented regions */
  void ImportSegments(std::string filename);

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  SegmentationApplicationModel();
  /** Destructor */
  ~SegmentationApplicationModel();

private:
  SegmentationApplicationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented


  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  /** Filters */
  ReaderPointerType                             m_Reader;
  IntensityFilterPointerType                    m_IntensityFilter;
  ExtractROIFilterPointerType                   m_ExtractROIFilter;
  MultiToMonoChannelExtractROIFilterPointerType m_MultiToMonoChannelExtractROIFilter;
  MultiChannelExtractROIFilterPointerType       m_MultiChannelExtractROIFilter;

  /** Flags to activate/deactivate the preprocessings */
  bool m_Valid;
  bool m_IsNewImage;

  /** The instance singleton */
  static Pointer Instance;

  /** Parameters */
  std::string           m_FileName;
  RegionType            m_Region;
  bool                  m_UseIntensity;
  unsigned int          m_ChannelIndex;
  SegmentationAlgorithm m_Algorithm;

  /** Sub models */
  PreprocessingModelPointerType m_PreprocessingModel;
  VectorizationmodelPointerType m_VectorizationModel;
  RegionGrowingModelPointerType m_RegionGrowingModel;
  SVMBasedRegionGrowingSegmentationModelPointerType m_SVMBasedModel;
  LevelSetModelPointerType      m_LevelSetModel;

  /** Polygons */
  PolygonListPointerType        m_PolygonList;

};
}
#endif
