/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbVectorizationModel.h"

namespace otb
{

/** Initialize the singleton */
VectorizationModel::Pointer VectorizationModel::Instance = NULL;

/** Manage the singleton */
VectorizationModel::Pointer
VectorizationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = VectorizationModel::New();
  }
  return Instance;
}

void VectorizationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

VectorizationModel::VectorizationModel()
{
  m_Valid                  = false;
  m_Tolerance              = 1.0;

  /** Filters declarations */
  m_SimplifyPolygonFilter     = SimplifyPolygonFilterType::New();

  /** static parameters */
  m_SimplifyPolygonFilter->GetFunctor().SetTolerance(m_Tolerance);


}

VectorizationModel::~VectorizationModel()
{}

void VectorizationModel::SetSegmentationImage(UCharImagePointerType image)
{
  // Togle the valid flag
  m_SegmentationImage = image;
  if (m_ReferenceImage.IsNotNull())
  {
    m_Valid = true;
    this->NotifyAll();
  }
}

void VectorizationModel::SetReferenceImage(ImagePointerType image)
{
  // Togle the valid flag
  m_ReferenceImage = image;
  if (m_SegmentationImage.IsNotNull())
  {
    m_Valid = true;
    this->NotifyAll();
  }
}

void VectorizationModel::SetTolerance(double tolerance)
{
  if (tolerance == m_Tolerance)
  {
    // nothing to do, return
    return;
  }
  m_Tolerance = tolerance;
  this->UpdatePipeline();
  this->NotifyAll();
}

VectorizationModel::PolygonListPointerType VectorizationModel::GetPolygonList()
{
  return m_SimplifyPolygonFilter->GetOutput();
}

void VectorizationModel::UpdatePipeline()
{
  if (m_Valid)
  {

    PolygonListPointerType polygonList = PolygonListType::New();

    ConnectedComponentsFilterPointerType connectedComponentsFilter = ConnectedComponentsFilterType::New();
    connectedComponentsFilter->SetInput(m_SegmentationImage);
    connectedComponentsFilter->Update();

    for (LabeledPixelType label = 1; label<=connectedComponentsFilter->GetObjectCount(); ++label)
    {
      PolygonFilterPointerType polygonFilter = PolygonFilterType::New();
      polygonFilter->SetInput(connectedComponentsFilter->GetOutput());
      polygonFilter->SetForegroundValue(label);
      polygonFilter->Update();
      polygonList->PushBack(polygonFilter->GetOutput());
    }

    m_SimplifyPolygonFilter->SetInput(polygonList);
    m_SimplifyPolygonFilter->GetFunctor().SetTolerance(m_Tolerance);
    m_SimplifyPolygonFilter->Update();
  }
}

void VectorizationModel::Notify()
{
  this->UpdatePipeline();
  this->NotifyAll();
}
}
