/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbPreprocessingViewGUI.h"


namespace otb
{

PreprocessingViewGUI::PreprocessingViewGUI()
    : PreprocessingViewGroup(0, 0, 670, 200,"Preprocessing")
{
  m_SmoothedVectorCastFilter = VectorCastFilterType::New();
  m_EdgeEnhancedVectorCastFilter = VectorCastFilterType::New();

  m_PreprocessingModel = PreprocessingModel::GetInstance();

  vNumberOfIterations->value(m_PreprocessingModel->GetNumberOfIterations());
  vTimeStep->value(m_PreprocessingModel->GetTimeStep());
  vAlpha->value(m_PreprocessingModel->GetAlpha());


  m_PreprocessingModel->RegisterListener(this);

  m_WMargin = 25;
  m_HMargin = 25;
}

void PreprocessingViewGUI::RefreshViewers()
{
  if (m_PreprocessingModel->GetUseSmoothing())
  {
    MinMaxCalculatorType::Pointer minMax = MinMaxCalculatorType::New();
    minMax->SetImage(m_PreprocessingModel->GetSmoothedImage());
    minMax->Compute();

    m_SmoothedVectorCastFilter->SetInput(m_PreprocessingModel->GetSmoothedImage());
    m_SmoothedVectorCastFilter->GetOutput()->UpdateOutputInformation();
    m_SmoothedVectorCastFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();
    m_SmoothedVectorCastFilter->Update();

    if (m_SmoothedWidget.IsNull())
    {
      m_SmoothedWidget  = ImageWidgetType::New();
      m_SmoothedWidget->SetInput(m_SmoothedVectorCastFilter->GetOutput());
      m_SmoothedWidget->Init(gSmoothedWidgetGroup->x(), gSmoothedWidgetGroup->y(), gSmoothedWidgetGroup->w(), gSmoothedWidgetGroup->h(),"Smoothed image");
      m_SmoothedWidget->SetViewModel(ImageWidgetType::GRAYSCALE);
      m_SmoothedWidget->GetTransferFunctionList()->GetNthElement(0)->SetLowerBound(minMax->GetMinimum());
      m_SmoothedWidget->GetTransferFunctionList()->GetNthElement(0)->SetUpperBound(minMax->GetMaximum());
      gSmoothedWidgetGroup->add(m_SmoothedWidget);
      gSmoothedWidgetGroup->box(FL_NO_BOX);
    }
    else
    {
      m_SmoothedWidget->ClearBufferedRegion();
      m_SmoothedWidget->GetTransferFunctionList()->GetNthElement(0)->SetLowerBound(minMax->GetMinimum());
      m_SmoothedWidget->GetTransferFunctionList()->GetNthElement(0)->SetUpperBound(minMax->GetMaximum());
    }
    m_SmoothedWidget->activate();
    m_SmoothedWidget->redraw();
  }
  else
  {
    if (m_SmoothedWidget.IsNotNull())
    {
      m_SmoothedWidget->deactivate();
    }
  }
  if (m_PreprocessingModel->GetUseEdgeEnhancement())
  {
    MinMaxCalculatorType::Pointer minMax = MinMaxCalculatorType::New();
    minMax->SetImage(m_PreprocessingModel->GetOutputImage());
    minMax->Compute();


    m_EdgeEnhancedVectorCastFilter->SetInput(m_PreprocessingModel->GetOutputImage());
    m_EdgeEnhancedVectorCastFilter->GetOutput()->UpdateOutputInformation();
    m_EdgeEnhancedVectorCastFilter->GetOutput()->SetRequestedRegionToLargestPossibleRegion();

    m_EdgeEnhancedVectorCastFilter->Update();

    if (m_EdgeEnhancedWidget.IsNull())
    {
      m_EdgeEnhancedWidget  = ImageWidgetType::New();
      m_EdgeEnhancedWidget->SetInput(m_EdgeEnhancedVectorCastFilter->GetOutput());
      m_EdgeEnhancedWidget->Init(gEdgeEnhancedWidgetGroup->x(), gEdgeEnhancedWidgetGroup->y(), gEdgeEnhancedWidgetGroup->w(), gEdgeEnhancedWidgetGroup->h(),"Edge enhanced image");
      m_EdgeEnhancedWidget->SetViewModel(ImageWidgetType::GRAYSCALE);
      m_EdgeEnhancedWidget->GetTransferFunctionList()->GetNthElement(0)->SetLowerBound(minMax->GetMinimum());
      m_EdgeEnhancedWidget->GetTransferFunctionList()->GetNthElement(0)->SetUpperBound(minMax->GetMaximum());
      gEdgeEnhancedWidgetGroup->add(m_EdgeEnhancedWidget);
      gEdgeEnhancedWidgetGroup->box(FL_NO_BOX);
    }
    else
    {
      m_EdgeEnhancedWidget->GetTransferFunctionList()->GetNthElement(0)->SetLowerBound(minMax->GetMinimum());
      m_EdgeEnhancedWidget->GetTransferFunctionList()->GetNthElement(0)->SetUpperBound(minMax->GetMaximum());
      m_EdgeEnhancedWidget->ClearBufferedRegion();
    }
    m_EdgeEnhancedWidget->activate();
    m_EdgeEnhancedWidget->redraw();
  }
  else
  {
    if (m_EdgeEnhancedWidget.IsNotNull())
    {
      m_EdgeEnhancedWidget->deactivate();
    }
  }
}

void PreprocessingViewGUI::show()
{
  this->Fl_Group::show();
  if (m_PreprocessingModel->GetUseEdgeEnhancement()&&m_EdgeEnhancedWidget.IsNotNull())
  {
    m_EdgeEnhancedWidget->Show();
  }
  if (m_PreprocessingModel->GetUseSmoothing()&&m_SmoothedWidget.IsNotNull())
  {
    m_SmoothedWidget->Show();
  }

}

void PreprocessingViewGUI::CenterSmoothedWidget(int x, int y, int w, int h)
{
  ImageType::SizeType size = m_SmoothedVectorCastFilter->GetOutput()->GetLargestPossibleRegion().GetSize();
  ImageType::IndexType upperLeft;
  upperLeft.Fill(0);

  ImageType::IndexType offset;
  offset.Fill(0);

  ImageType::SizeType finalSize;
  finalSize.Fill(0);

  if (static_cast<int>(size[0])>w)
  {
    upperLeft[0] = static_cast<int>(size[0])/2-w/2;
    finalSize[0] = w;
  }
  else
  {
    offset[0]=w/2-size[0]/2;
    finalSize[0]=size[0];
  }
  if (static_cast<int>(size[1])>h)
  {
    upperLeft[1] = static_cast<int>(size[1])/2-h/2;
    finalSize[1]= h;
  }
  else
  {
    offset[1]=w/2-size[1]/2;
    finalSize[1]=size[1];
  }
  m_SmoothedWidget->SetUpperLeftCorner(upperLeft);
  m_SmoothedWidget->resize(x+offset[0], y+offset[1], finalSize[0], finalSize[1]);
}

void PreprocessingViewGUI::CenterEdgeEnhancedWidget(int x, int y, int w, int h)
{
  ImageType::SizeType size = m_EdgeEnhancedVectorCastFilter->GetOutput()->GetLargestPossibleRegion().GetSize();
  ImageType::IndexType upperLeft;
  upperLeft.Fill(0);

  ImageType::IndexType offset;
  offset.Fill(0);

  ImageType::SizeType finalSize;
  finalSize.Fill(0);

  if (static_cast<int>(size[0])>w)
  {
    upperLeft[0] = static_cast<int>(size[0])/2-w/2;
    finalSize[0] = w;
  }
  else
  {
    offset[0]=w/2-size[0]/2;
    finalSize[0]=size[0];
  }
  if (static_cast<int>(size[1])>h)
  {
    upperLeft[1] = static_cast<int>(size[1])/2-h/2;
    finalSize[1]= h;
  }
  else
  {
    offset[1]=w/2-size[1]/2;
    finalSize[1]=size[1];
  }
  m_EdgeEnhancedWidget->SetUpperLeftCorner(upperLeft);
  m_EdgeEnhancedWidget->resize(x+offset[0], y+offset[1], finalSize[0], finalSize[1]);

}

PreprocessingViewGUI::SizeType PreprocessingViewGUI::GetIdealSize()
{
  ImageType::SizeType size;

  if (m_PreprocessingModel->GetValid())
  {
    size = m_PreprocessingModel->GetOutputImage()->GetLargestPossibleRegion().GetSize();
  }
  else
  {
    SizeType defaultSize;
    defaultSize[0]=this->w();
    defaultSize[1]= this->h();
    return defaultSize;
  }

  ImageType::SizeType newSize;
  newSize[0] = 2*size[0]+4*m_WMargin + gInterface->w();
  newSize[1] = std::max(static_cast<int>(size[1]+2*m_HMargin), 2*m_HMargin+gInterface->h());

  return newSize;
}


void PreprocessingViewGUI::resize(int x, int y, int w, int h)
{
  int m_WMargin = 25;
  int m_HMargin = 25;

  int interfacew = gInterface->w();
  int interfaceh = gInterface->h();

  this->Fl_Group::resize(x, y, w, h);
  int imagew = (w-4*m_WMargin-interfacew)/2;
  int imageh = h-2*m_HMargin;

  gSmoothedWidgetGroup->resize(x+m_WMargin, y+m_HMargin, imagew, imageh);
  if (m_SmoothedWidget.IsNotNull())
  {
    m_SmoothedWidget->resize(x+m_WMargin, y+m_HMargin, imagew, imageh);
  }
  gEdgeEnhancedWidgetGroup->resize(x+2*m_WMargin+imagew, y+m_HMargin, imagew, imageh);
  if (m_EdgeEnhancedWidget.IsNotNull())
  {
    m_EdgeEnhancedWidget->resize(x+2*m_WMargin+imagew, y+m_HMargin, imagew, imageh);
  }

  int interfacex = 3*m_WMargin + 2 * imagew;
  int interfacey = (h-interfaceh)/2;

  gInterface->resize(x+interfacex, y+interfacey, interfacew, interfaceh);
}

void PreprocessingViewGUI::Notify()
{
  if (m_PreprocessingModel->GetValid())
  {
    this->RefreshViewers();
  }
  this->redraw();
}

void PreprocessingViewGUI::UseSmoothingCallback()
{
  if (bUseSmoothing->value())
  {
    m_PreprocessingController->EnableSmoothing();
  }
  else
  {
    m_PreprocessingController->DisableSmoothing();
  }
}

void PreprocessingViewGUI::UseEdgeEnhancementCallback()
{
  if (bUseEdgeEnhancement->value())
  {
    m_PreprocessingController->EnableEdgeEnhancement();
  }
  else
  {
    m_PreprocessingController->DisableEdgeEnhancement();
  }
}

void PreprocessingViewGUI::NumberOfIterationsCallback()
{
  m_PreprocessingController->ChangeNumberOfIterations(static_cast<unsigned int>(vNumberOfIterations->value()));
}

void PreprocessingViewGUI::TimeStepCallback()
{
  m_PreprocessingController->ChangeTimeStep(vTimeStep->value());
}

void PreprocessingViewGUI::AlphaCallback()
{
  m_PreprocessingController->ChangeAlpha(vAlpha->value());
}
}

