/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbSegmentationApplicationViewGUI.h"
#include <FL/Fl_File_Chooser.H>
#include "otbMacro.h"


namespace otb
{

SegmentationApplicationViewGUI::SegmentationApplicationViewGUI()
{
  m_SegmentationApplicationModel = SegmentationApplicationModel::GetInstance();
  m_SegmentationApplicationModel->RegisterListener(this);

  m_PolygonColor[0]=1.0;
  m_PolygonColor[1]=0.;
  m_PolygonColor[2]=0.;
  m_PolygonColor[3]=0.75;

  m_RegionColor[0]=0.;
  m_RegionColor[1]=1.0;
  m_RegionColor[2]=0.;
  m_RegionColor[3]=0.5;

  m_PreprocessingView = PreprocessingViewGUI::New();
  m_RegionGrowingView = RegionGrowingSegmentationViewGUI::New();
  m_LevelSetView      = LevelSetSegmentationViewGUI::New();
  m_SVMBasedRegionGrowingView         = SVMBasedRegionGrowingSegmentationViewGUI::New();
  m_VectorizationView = VectorizationViewGUI::New();

  m_PreprocessingView->hide();
  m_RegionGrowingView->hide();
  m_LevelSetView->hide();
  m_VectorizationView->hide();

  m_PreprocessingView->align(FL_ALIGN_TOP);
  m_RegionGrowingView->align(FL_ALIGN_TOP);
  m_LevelSetView->align(FL_ALIGN_TOP);
  m_VectorizationView->align(FL_ALIGN_TOP);

  this->Build();
  wSegmentationWindow->hide();
  bUseImageIntensity->set();
  bUseImageChannel->clear();
  vAlgorithm->add("Region growing");
  vAlgorithm->add("Level set");
  vAlgorithm->add("SVM based region growing");
  vAlgorithm->value(0);
  wMainWindow->show();
}


void SegmentationApplicationViewGUI::SetSegmentationApplicationController(SegmentationApplicationControllerInterface * controller)
{
  m_SegmentationApplicationController = controller;
  m_PreprocessingView->SetPreprocessingController(controller->GetPreprocessingController());
  m_RegionGrowingView->SetRegionGrowingSegmentationController(controller->GetRegionGrowingSegmentationController());
  m_LevelSetView->SetLevelSetSegmentationController(controller->GetLevelSetSegmentationController());
  m_SVMBasedRegionGrowingView->SetSVMBasedRegionGrowingSegmentationController(controller->GetSVMBasedRegionGrowingSegmentationController());
  m_VectorizationView->SetVectorizationController(controller->GetVectorizationController());
}

void SegmentationApplicationViewGUI::RefreshViewer()
{

  if (m_SegmentationApplicationModel->GetIsNewImage() && !m_Viewer.IsNull())
  {
    if (m_Viewer->GetUseScroll())
      gScroll->remove(m_Viewer->GetScrollWidget());
    gFullResolution->remove(m_Viewer->GetFullWidget());
  }


  if (m_Viewer.IsNull() || (m_SegmentationApplicationModel->GetIsNewImage()))
  {

    m_Viewer = ImageViewerBaseType::New();
    m_Viewer->SetImage(m_SegmentationApplicationModel->GetImage());
    if (m_SegmentationApplicationModel->GetImage()->GetNumberOfComponentsPerPixel()>3)
    {
      m_Viewer->SetRedChannelIndex(2);
      m_Viewer->SetGreenChannelIndex(1);
      m_Viewer->SetBlueChannelIndex(0);
    }
    unsigned int maxfullsize = std::min(gFullResolution->w(), gFullResolution->h());
    unsigned int maxscrollsize = std::min(gScroll->w(), gScroll->h());
    m_Viewer->SetFullMaxInitialSize(maxfullsize);
    m_Viewer->SetScrollMaxInitialSize(maxscrollsize);
    m_Viewer->SetShowZoomWidget(false);
    m_Viewer->SetShowHistograms(false);
    m_Viewer->SetRectangularROISelectionMode(true);
    m_Viewer->SetPolygonROIList(m_SegmentationApplicationModel->GetPolygonList());
    m_Viewer->Build();
    m_Viewer->GetFullWidget()->SetEventsInterface(m_SegmentationApplicationController);
    m_Viewer->GetFullWidget()->position(gFullResolution->x(), gFullResolution->y());
    m_Viewer->GetFullWidget()->resize(gFullResolution->x(), gFullResolution->y(), gFullResolution->w(), gFullResolution->h());
    gFullResolution->add(m_Viewer->GetFullWidget());
    gFullResolution->box(FL_NO_BOX);
    if (m_Viewer->GetUseScroll())
    {
      m_Viewer->GetScrollWidget()->position(gScroll->x(), gScroll->y());
      m_Viewer->GetScrollWidget()->resize(gScroll->x(), gScroll->y(), gScroll->w(), gScroll->h());
      gScroll->add(m_Viewer->GetScrollWidget());
      gScroll->show();
      gScroll->box(FL_NO_BOX);
    }
    else
    {
      gScroll->hide();
    }
  }

  m_Viewer->Show();
  m_Viewer->Update();
}


void SegmentationApplicationViewGUI::UpdateInterface()
{
  vAlgorithm->activate();
  bUseImageIntensity->activate();
  bUseImageChannel->activate();
  vChannel->clear();

  /// Updating channel list
  for (unsigned int channel = 1; channel <= m_SegmentationApplicationModel->GetImage()->GetNumberOfComponentsPerPixel(); ++channel)
  {
    itk::OStringStream oss;
    oss<<"Channel "<<channel;
    vChannel->add(oss.str().c_str());
  }
  vChannel->value(m_SegmentationApplicationModel->GetChannelIndex());

  if ( m_SegmentationApplicationModel->GetUseIntensity())
  {
    bUseImageIntensity->set();
    bUseImageChannel->clear();
    vChannel->deactivate();
  }
  else
  {
    bUseImageIntensity->clear();
    bUseImageChannel->set();
    vChannel->activate();
  }

  // Updating polygon list
  lSegmentedRegions->clear();

  SegmentationApplicationModel::PolygonListType::Iterator pit = m_SegmentationApplicationModel->GetPolygonList()->Begin();

  unsigned int pIndex = 1;
  itk::OStringStream  oss;
  oss.str("");

  while (pit!=m_SegmentationApplicationModel->GetPolygonList()->End())
  {
    oss<<"Region "<<pIndex<<" ("<<pit.Get()->GetArea()<<" pixels)";
    lSegmentedRegions->add(oss.str().c_str());
    oss.str("");
    ++pit;
    ++pIndex;
  }
  lSegmentedRegions->redraw();
}

void SegmentationApplicationViewGUI::SetupSegmentationWindow()
{
  SizeType idealSize;

  idealSize.Fill(0);

  tSegmentationTabs->remove(m_PreprocessingView);
  tSegmentationTabs->remove(m_RegionGrowingView);
  tSegmentationTabs->remove(m_LevelSetView);
  tSegmentationTabs->remove(m_SVMBasedRegionGrowingView);
  tSegmentationTabs->remove(m_VectorizationView);

  switch (m_SegmentationApplicationModel->GetAlgorithm())
  {
  case REGION_GROWING:
    idealSize = m_RegionGrowingView->GetIdealSize();
    m_RegionGrowingView->resize(0, 20, idealSize[0], idealSize[1]);
    wSegmentationWindow->label("Region growing segmentation");
    break;
  case LEVEL_SET:
    idealSize = m_LevelSetView->GetIdealSize();
    m_LevelSetView->resize(0, 20, idealSize[0], idealSize[1]);
    wSegmentationWindow->label("Level set segmentation");
    break;
  case SVM_BASED_REGION_GROWING:
    idealSize = m_SVMBasedRegionGrowingView->GetIdealSize();
    m_SVMBasedRegionGrowingView->resize(0, 20, idealSize[0], idealSize[1]);
    wSegmentationWindow->label("SVM based region growing segmentation");
    break;
  }
  wSegmentationWindow->resize(wSegmentationWindow->x(), wSegmentationWindow->y(), idealSize[0], idealSize[1]+bSegmentationOk->h()+30);
  bSegmentationOk->position(3*idealSize[0]/4-bSegmentationOk->w()/2, idealSize[1]+5+20);
  bSegmentationCancel->position(idealSize[0]/4-bSegmentationOk->w()/2, idealSize[1]+5+20);
  tSegmentationTabs->resize(0, 0, idealSize[0], idealSize[1]+20);
  m_PreprocessingView->resize(0, 20, idealSize[0], idealSize[1]);
  m_VectorizationView->resize(0, 20, idealSize[0], idealSize[1]);

  if (m_SegmentationApplicationModel->GetAlgorithm()!=SVM_BASED_REGION_GROWING)
  {
    tSegmentationTabs->add(m_PreprocessingView);
    tSegmentationTabs->value(m_PreprocessingView);
  }

  switch (m_SegmentationApplicationModel->GetAlgorithm())
  {
  case REGION_GROWING:
    tSegmentationTabs->add(m_RegionGrowingView);
    break;
  case LEVEL_SET:
    tSegmentationTabs->add(m_LevelSetView);
    break;
  case SVM_BASED_REGION_GROWING:
    tSegmentationTabs->add(m_SVMBasedRegionGrowingView);
    tSegmentationTabs->value(m_SVMBasedRegionGrowingView);
    break;
  }

  tSegmentationTabs->add(m_VectorizationView);
  wSegmentationWindow->show();
}

void SegmentationApplicationViewGUI::Notify()
{
  if (m_SegmentationApplicationModel->GetValid())
  {
    this->RefreshViewer();
    this->UpdateInterface();
  }
}

void SegmentationApplicationViewGUI::OpenImageCallback()
{
  const char * filename = NULL;

  filename = fl_file_chooser("Pick an image file", "*.*",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_SegmentationApplicationController->SelectFileName(filename);
  m_SegmentationApplicationController->ClearSegments();

}
void SegmentationApplicationViewGUI::QuitCallback()
{
  wMainWindow->hide();
  wSegmentationWindow->hide();
}

void SegmentationApplicationViewGUI::SegmentationOkCallback()
{
  wSegmentationWindow->hide();
  m_SegmentationApplicationController->EndSegmentation();
  // Fixing bug #10: forbid ROI selection when segmentation is on progress
  m_Viewer->SetRectangularROISelectionMode(true);
}

void SegmentationApplicationViewGUI::SegmentationCancelCallback()
{
  wSegmentationWindow->hide();
  // Fixing bug #10: forbid ROI selection when segmentation is on progress
  m_Viewer->SetRectangularROISelectionMode(true);
}

void SegmentationApplicationViewGUI::UseImageIntensityCallback()
{
  m_SegmentationApplicationController->ChangeUseImageIntensity();
}
void SegmentationApplicationViewGUI::UseImageChannelCallback()
{
  m_SegmentationApplicationController->ChangeUseImageChannel();
}
void SegmentationApplicationViewGUI::ChannelCallback()
{
  m_SegmentationApplicationController->ChangeChannel(vChannel->value());
}
void SegmentationApplicationViewGUI::AlgorithmCallback()
{
  switch (vAlgorithm->value())
  {
  case 0:
    m_SegmentationApplicationController->ChangeAlgorithm(REGION_GROWING);
    break;
  case 1:
    m_SegmentationApplicationController->ChangeAlgorithm(LEVEL_SET);
    break;
  case 2:
    m_SegmentationApplicationController->ChangeAlgorithm(SVM_BASED_REGION_GROWING);
  }
}

void SegmentationApplicationViewGUI::SegmentCallback()
{
  m_RegionGrowingView->SetOffset(m_SegmentationApplicationModel->GetRegion().GetIndex());
  // Fixing bug #10: forbid ROI selection when segmentation is on progress
  m_Viewer->SetRectangularROISelectionMode(false);
  m_SegmentationApplicationController->TriggerSegmentation();
}


void SegmentationApplicationViewGUI::FocusCallback()
{

  int index = lSegmentedRegions->value()-1;
  // No region selected, return
  if (index < 0)
  {
    return;
  }
  // This should maybe be moved into the controller, but much more easier to implement here
  SegmentationApplicationModel::RegionType boundingRegion = m_SegmentationApplicationModel->GetPolygonList()->GetNthElement(index)->GetBoundingRegion().GetImageRegion();
  SegmentationApplicationModel::IndexType upperLeft;
  upperLeft[0] = boundingRegion.GetIndex()[0] + boundingRegion.GetSize()[0]/2;
  upperLeft[1] = boundingRegion.GetIndex()[1] + boundingRegion.GetSize()[1]/2;
  m_Viewer->ChangeFullViewedRegion(upperLeft);
  m_Viewer->Update();

}

void SegmentationApplicationViewGUI::DeleteCallback()
{
  int index = lSegmentedRegions->value()-1;
  // No region selected, return
  if (index < 0)
  {
    return;
  }
  m_SegmentationApplicationController->DeleteSegment(index);
}

void SegmentationApplicationViewGUI::ClearCallback()
{
  m_SegmentationApplicationController->ClearSegments();
}

void SegmentationApplicationViewGUI::SaveResultsCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output vector file", "*.shp\t*.kml",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_SegmentationApplicationController->SaveResults(filename);
}

void SegmentationApplicationViewGUI::ImportSegmentsCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an input vector file", "*.shp\t*.kml",".");

  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_SegmentationApplicationController->ImportSegments(filename);
}
}
