/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbStereoscopicApplicationModel.h"

namespace otb
{

/** Initialize the singleton */
StereoscopicApplicationModel::Pointer StereoscopicApplicationModel::Instance = NULL;

/** Manage the singleton */
StereoscopicApplicationModel::Pointer
StereoscopicApplicationModel::GetInstance()
{
  if (!Instance)
  {
    Instance = StereoscopicApplicationModel::New();
  }
  return Instance;
}

void StereoscopicApplicationModel::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

StereoscopicApplicationModel::StereoscopicApplicationModel()
{
  m_LeftReaders = ReaderListType::New();
  m_RightReaders = ReaderListType::New();
  m_AnaglyhFilters = AnaglyphFilterListType::New();
  m_LeftShrinkers = ShrinkFilterListType::New();
  m_RightShrinkers = ShrinkFilterListType::New();
  m_ShrinkAnaglyphFilters = AnaglyphFilterListType::New();
  m_HistogramListOfList = HistogramListOfListType::New();
  m_QuicklookSize.Fill(256);
  m_ClampThreshold = 0.01;
}

StereoscopicApplicationModel::~StereoscopicApplicationModel()
{}

StereoscopicApplicationModel::ImageType *
StereoscopicApplicationModel::GetLeftImage(unsigned int index)
{
  return m_LeftReaders->GetNthElement(index)->GetOutput();
}

const char *
StereoscopicApplicationModel::GetLeftImageFilename(unsigned int index)
{
  return m_LeftReaders->GetNthElement(index)->GetFileName();
}

const char *
StereoscopicApplicationModel::GetRightImageFilename(unsigned int index)
{
  return m_RightReaders->GetNthElement(index)->GetFileName();
}

StereoscopicApplicationModel::ImageType *
StereoscopicApplicationModel::GetRightImage(unsigned int index)
{
  return m_RightReaders->GetNthElement(index)->GetOutput();
}

StereoscopicApplicationModel:: ImageType *
StereoscopicApplicationModel::GetAnaglyph(unsigned int index)
{
  return m_AnaglyhFilters->GetNthElement(index)->GetOutput();
}

StereoscopicApplicationModel::ImageType *
StereoscopicApplicationModel::GetLeftQuicklook(unsigned int index)
{
  return m_LeftShrinkers->GetNthElement(index)->GetOutput();
}

StereoscopicApplicationModel::ImageType *
StereoscopicApplicationModel::GetRightQuicklook(unsigned int index)
{
  return m_RightShrinkers->GetNthElement(index)->GetOutput();
}

StereoscopicApplicationModel:: ImageType *
StereoscopicApplicationModel::GetAnaglyphQuicklook(unsigned int index)
{
  return m_ShrinkAnaglyphFilters->GetNthElement(index)->GetOutput();
}


StereoscopicApplicationModel::ShrinkFilterType *
StereoscopicApplicationModel::GetLeftShrinkFilter(unsigned int index)
{
  return m_LeftShrinkers->GetNthElement(index);
}


StereoscopicApplicationModel::ShrinkFilterType *
StereoscopicApplicationModel::GetRightShrinkFilter(unsigned int index)
{
  return m_RightShrinkers->GetNthElement(index);
}

unsigned int StereoscopicApplicationModel::GetShrinkFactor(unsigned int index)
{
  return m_ShrinkFactors[index];
}

StereoscopicApplicationModel::PixelType StereoscopicApplicationModel::GetMin(unsigned int index)
{
  unsigned int nbComp = m_LeftShrinkers->GetNthElement(index)->GetOutput()->GetNumberOfComponentsPerPixel();
  PixelType min;
  min.SetSize(nbComp);
  min.Fill(0);


  for (unsigned int i = 0; i<nbComp; ++i)
  {
    min[i]=m_HistogramListOfList->GetNthElement(index)->GetNthElement(i)->Quantile(0, m_ClampThreshold);
  }
  if (nbComp<3)
  {
    double minValue = min[0];
    min.SetSize(3);
    min.Fill(minValue);
  }
  return min;
}

StereoscopicApplicationModel::PixelType StereoscopicApplicationModel::GetMax(unsigned int index)
{
  unsigned int nbComp = m_LeftShrinkers->GetNthElement(index)->GetOutput()->GetNumberOfComponentsPerPixel();
  PixelType max;
  max.SetSize(nbComp);
  max.Fill(0);

  for (unsigned int i = 0; i<nbComp; ++i)
  {
    max[i]=m_HistogramListOfList->GetNthElement(index)->GetNthElement(i)->Quantile(0, 1-m_ClampThreshold);
  }
  if (nbComp<3)
  {
    double maxValue = max[0];
    max.SetSize(3);
    max.Fill(maxValue);
  }
  return max;

}

unsigned int StereoscopicApplicationModel::GetNumberOfStereoscopicCouples()
{
  return m_AnaglyhFilters->Size();
}


unsigned int StereoscopicApplicationModel::AddCouple(const char * leftName, const char * rightName)
{
  // Trying to read left image
  ReaderType::Pointer lReader = ReaderType::New();
  lReader->SetFileName(leftName);
  lReader->GenerateOutputInformation();

  // Trying to read right image
  ReaderType::Pointer rReader = ReaderType::New();
  rReader->SetFileName(rightName);
  rReader->GenerateOutputInformation();

  if (lReader->GetOutput()->GetLargestPossibleRegion()
      != rReader->GetOutput()->GetLargestPossibleRegion()
      || lReader->GetOutput()->GetNumberOfComponentsPerPixel()
      != rReader->GetOutput()->GetNumberOfComponentsPerPixel())
  {
    itkExceptionMacro(<<"Left and right images do not have the same size or number of bands. Can not generate anaglyph");
  }

  // Shrink factor
  SizeType largestRegionSize = lReader->GetOutput()->GetLargestPossibleRegion().GetSize();
  unsigned int maxQuicklook = std::max(m_QuicklookSize[0], m_QuicklookSize[1]);
  unsigned int maxLargest   = std::max(largestRegionSize[0], largestRegionSize[1]);
  unsigned int shrinkFactor = maxLargest/maxQuicklook;

  if (shrinkFactor == 0)
  {
    shrinkFactor = 1;
  }

  // Left shrinker
  ShrinkFilterType::Pointer lShrinker = ShrinkFilterType::New();
  lShrinker->SetInput(lReader->GetOutput());
  lShrinker->SetShrinkFactor(shrinkFactor);

  // Right shrinker
  ShrinkFilterType::Pointer rShrinker = ShrinkFilterType::New();
  rShrinker->SetInput(rReader->GetOutput());
  rShrinker->SetShrinkFactor(shrinkFactor);

  // shrinked anaglyph
  AnaglyphFilterPointerType shrinkAnaglyph = AnaglyphFilterType::New();
  shrinkAnaglyph->SetInput1(lShrinker->GetOutput());
  shrinkAnaglyph->SetInput2(rShrinker->GetOutput());

  // If still no exceptions thrown files are valid
  // and we can proceed to the anaglyph filter
  AnaglyphFilterPointerType anaglyph = AnaglyphFilterType::New();
  anaglyph->SetInput1(lReader->GetOutput());
  anaglyph->SetInput2(rReader->GetOutput());

  if (lReader->GetOutput()->GetNumberOfComponentsPerPixel()<3)
  {
    std::cout<<"Grayscale mode"<<std::endl;
    anaglyph->SetRedChannelIndex(0);
    anaglyph->SetGreenChannelIndex(0);
    anaglyph->SetBlueChannelIndex(0);
    shrinkAnaglyph->SetRedChannelIndex(0);
    shrinkAnaglyph->SetGreenChannelIndex(0);
    shrinkAnaglyph->SetBlueChannelIndex(0);
  }

  // saving everyone
  m_LeftReaders->PushBack(lReader);
  m_RightReaders->PushBack(rReader);
  m_AnaglyhFilters->PushBack(anaglyph);
  m_LeftShrinkers->PushBack(lShrinker);
  m_RightShrinkers->PushBack(rShrinker);
  m_ShrinkAnaglyphFilters->PushBack(shrinkAnaglyph);
  m_ShrinkFactors.push_back(shrinkFactor);

  return  (m_LeftReaders->Size()-1);
}

void StereoscopicApplicationModel
::GenerateQuicklooksAndNormalizationFactors(unsigned int index)
{
  m_ShrinkAnaglyphFilters->GetNthElement(index)->Update();

  // Compute the normalization factors
  ImageType::Pointer leftQuicklook = m_LeftShrinkers->GetNthElement(index)->GetOutput();
  ImageType::Pointer rightQuicklook = m_RightShrinkers->GetNthElement(index)->GetOutput();

  m_HistogramListOfList->PushBack(HistogramListType::New());

  ListSampleListType::Pointer sl =  ListSampleListType::New();

  sl->Reserve(leftQuicklook->GetNumberOfComponentsPerPixel());


  for (unsigned int i = 0; i<leftQuicklook->GetNumberOfComponentsPerPixel(); ++i)
  {
    sl->PushBack(ListSampleType::New());
  }

  IteratorType leftIt(leftQuicklook, leftQuicklook->GetLargestPossibleRegion());
  leftIt.GoToBegin();
  while ( !leftIt.IsAtEnd() )
  {
    PixelType pixel = leftIt.Get();
    for (unsigned int i = 0; i<leftQuicklook->GetNumberOfComponentsPerPixel(); ++i)
    {
      sl->GetNthElement(i)->PushBack(pixel[i]);
    }
    ++leftIt;
  }

  IteratorType rightIt(rightQuicklook, rightQuicklook->GetLargestPossibleRegion());
  rightIt.GoToBegin();
  while ( !rightIt.IsAtEnd() )
  {
    PixelType pixel = rightIt.Get();
    for (unsigned int i = 0; i<rightQuicklook->GetNumberOfComponentsPerPixel(); ++i)
    {
      sl->GetNthElement(i)->PushBack(pixel[i]);
    }
    ++rightIt;
  }


  for (unsigned int i = 0; i<sl->Size(); ++i)
  {
    HistogramGeneratorType::Pointer generator = HistogramGeneratorType::New();
    generator->SetListSample(sl->GetNthElement(i));
    HistogramGeneratorType::HistogramType::SizeType size;
    size.Fill(256);
    generator->SetNumberOfBins(size);
    generator->Update();
    m_HistogramListOfList->GetNthElement(index)->PushBack(const_cast<HistogramType *>(generator->GetOutput()));
  }
  this->NotifyAll();
}

}
