/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbStereoscopicApplicationModel_h
#define __otbStereoscopicApplicationModel_h

#include "otbMVCModel.h"

#include "otbVectorImage.h"
#include "otbImageFileReader.h"
#include "otbVectorImageToColorAnaglyphVectorImageFilter.h"
#include "otbStreamingShrinkImageFilter.h"
#include "otbListenerBase.h"
#include "otbObjectList.h"
#include "itkImageRegionIterator.h"
#include "itkListSample.h"
#include "itkListSampleToHistogramGenerator.h"

namespace otb
{

/** \class StereoscopicApplicationModel
 *
 *
 *
 */
class ITK_EXPORT StereoscopicApplicationModel
      : public MVCModel<ListenerBase>, public itk::Object
{

public:
  /** Standard class typedefs */
  typedef StereoscopicApplicationModel                         Self;
  typedef otb::MVCModel<ListenerBase> Superclass;
  typedef itk::SmartPointer<Self>                              Pointer;
  typedef itk::SmartPointer<const Self>                        ConstPointer;

  /** Standard type macro */
  itkTypeMacro(StereoscopicApplicationModel, Object);

  /** Get the unique instance of the model */
  static Pointer GetInstance();

  typedef double                                                                          ScalarPixelType;
  typedef otb::VectorImage<ScalarPixelType, 2>                                              ImageType;
  typedef ImageType::PixelType                                                            PixelType;
  typedef ImageType::SizeType                                                             SizeType;
  typedef otb::ImageFileReader<ImageType>                                                 ReaderType;
  typedef ReaderType::Pointer                                                             ReaderPointerType;
  typedef otb::VectorImageToColorAnaglyphVectorImageFilter<ImageType, ImageType, ImageType> AnaglyphFilterType;
  typedef AnaglyphFilterType::Pointer                                                     AnaglyphFilterPointerType;
  typedef otb::ObjectList<ReaderType>                                                     ReaderListType;
  typedef ReaderListType::Pointer                                                         ReaderListPointerType;
  typedef otb::ObjectList<AnaglyphFilterType>                                             AnaglyphFilterListType;
  typedef AnaglyphFilterListType::Pointer                                                 AnaglyphFilterListPointerType;
  typedef otb::StreamingShrinkImageFilter<ImageType, ImageType>                            ShrinkFilterType;
  typedef otb::ObjectList<ShrinkFilterType>                                               ShrinkFilterListType;
  typedef ShrinkFilterListType::Pointer                                                   ShrinkFilterListPointerType;
  typedef std::vector<unsigned int>                                                       UnsignedIntVectorType;
  typedef itk::ImageRegionConstIterator< ImageType >                                      IteratorType;
  typedef itk::Vector<itk::NumericTraits<ScalarPixelType>::RealType, 1>                    MeasurementVectorType;
  typedef itk::Statistics::ListSample<MeasurementVectorType>                              ListSampleType;
  typedef float                                                                           HistogramMeasurementType;
  typedef itk::Statistics::ListSampleToHistogramGenerator<ListSampleType, HistogramMeasurementType,
  itk::Statistics::DenseFrequencyContainer, 1>                                           HistogramGeneratorType;
  typedef HistogramGeneratorType::HistogramType                                           HistogramType;
  typedef otb::ObjectList<HistogramType>                                                  HistogramListType;
  typedef otb::ObjectList<HistogramListType>                                              HistogramListOfListType;
  typedef otb::ObjectList<ListSampleType>                                                 ListSampleListType;


  /** Add a stereoscopic couple to the application */
  unsigned int AddCouple(const char * leftName, const char * rightName);

  /** Generate the nth quicklook and normalization factor */
  void GenerateQuicklooksAndNormalizationFactors(unsigned int index);

  /** Get the number of stereoscopic couples loaded */
  unsigned int GetNumberOfStereoscopicCouples();

  /** Get the nth left image */
  ImageType * GetLeftImage(unsigned int index);

  /** Get the nth left image filename */
  const char * GetLeftImageFilename (unsigned int index);

  /** Get the nth right image filename */
  const char * GetRightImageFilename (unsigned int index);

  /** Get the nth right image */
  ImageType * GetRightImage(unsigned int index);

  /** Get the nth anagyph image */
  ImageType * GetAnaglyph(unsigned int index);

  /** Get the nth left quicklook */
  ImageType * GetLeftQuicklook(unsigned int index);

  /** Get the nth right quicklook */
  ImageType * GetRightQuicklook(unsigned int index);

  /** Get the nth anagyph quicklook */
  ImageType * GetAnaglyphQuicklook(unsigned int index);

  /** Get the nth Shrink filter for progress reporting */
  ShrinkFilterType * GetLeftShrinkFilter(unsigned int index);

  /** Get the nth Shrink filter for progress reporting */
  ShrinkFilterType * GetRightShrinkFilter(unsigned int index);

  /** Get the nth shrink factor */
  unsigned int GetShrinkFactor(unsigned int index);

  /** Get the mins and the maxs */
  PixelType GetMin(unsigned int index);
  PixelType GetMax(unsigned int index);

  /** Set the size of the quicklooks to generate */
  itkSetMacro(QuicklookSize, SizeType);

  /** Set the percentage used for normalization */
  itkSetMacro(ClampThreshold, double);
  itkGetMacro(ClampThreshold, double);

protected:
  /** This is protected for the singleton. Use GetInstance() instead. */
  itkNewMacro(Self);

  /** Constructor */
  StereoscopicApplicationModel();
  /** Destructor */
  ~StereoscopicApplicationModel();

private:
  StereoscopicApplicationModel(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** The instance singleton */
  static Pointer Instance;

  /** Notify a given listener of changes */
  virtual void NotifyListener(ListenerBase * listener);

  // Left readers list
  ReaderListPointerType m_LeftReaders;
  // Right readers list
  ReaderListPointerType m_RightReaders;
  // Anaglyph filters list
  AnaglyphFilterListPointerType m_AnaglyhFilters;
  /** Left shrink filters list */
  ShrinkFilterListPointerType m_LeftShrinkers;
  /** Right shrink filters list */
  ShrinkFilterListPointerType m_RightShrinkers;
  /** Srhink anaglyph filters */
  AnaglyphFilterListPointerType m_ShrinkAnaglyphFilters;
  /** Size of the quicklook to generate */
  SizeType m_QuicklookSize;
  /** Shrink factors */
  UnsignedIntVectorType m_ShrinkFactors;
  /** Histogram list */
  HistogramListOfListType::Pointer m_HistogramListOfList;
  /** percentage used for normalization */
  double m_ClampThreshold;
};
}
#endif
