/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbStereoscopicApplicationViewGUI.h"

#include <FL/Fl_File_Chooser.H>
#include "otbMacro.h"


namespace otb
{
StereoscopicApplicationViewGUI::StereoscopicApplicationViewGUI()
{
  m_StereoscopicApplicationModel = StereoscopicApplicationModel::GetInstance();
  m_StereoscopicApplicationModel->RegisterListener(this);
  this->Build();
  wMainWindow->show();
  wOpenCouplesWindow->hide();
  wRenameWindow->hide();

  // Slider/display coherency
  double factor = vcl_pow(2., vResample->value());
  vResampleValue->value(factor);
  vMagnifyValue->value(vMagnify->value());

  // Setting the right quicklook size
  StereoscopicApplicationModel::SizeType qlSize;
  qlSize[0]=gScroll->w();
  qlSize[1]=gScroll->h();
  m_StereoscopicApplicationModel->SetQuicklookSize(qlSize);

  // Scroll rectangle color
  m_InterfaceColor[0]=1;
  m_InterfaceColor[1]=0;
  m_InterfaceColor[2]=0;
  m_InterfaceColor[3]=1;

  // Interpolators choice
  vZoomInInterpolator->add("Nearest neighbors");
  vZoomInInterpolator->add("Bilinear");
  vZoomInInterpolator->add("BSpline (2nd order)");
  vZoomInInterpolator->add("BSpline (3rd order)");
  vZoomInInterpolator->add("BSpline (4th order)");
  vZoomInInterpolator->add("BSpline (5th order)");
  vZoomInInterpolator->add("Cardinal sinus (Hamming windowed, radius 2)");
  vZoomInInterpolator->add("Cardinal sinus (Hamming windowed, radius 4)");
  vZoomInInterpolator->add("Cardinal sinus (Hamming windowed, radius 8)");
  vZoomInInterpolator->add("Cardinal sinus (Hamming windowed, radius 16)");
  vZoomInInterpolator->add("Cardinal sinus (Hamming windowed, radius 32)");
  vZoomInInterpolator->add("Cardinal sinus (Gaussian windowed, radius 2)");
  vZoomInInterpolator->add("Cardinal sinus (Gaussian windowed, radius 4)");
  vZoomInInterpolator->add("Cardinal sinus (Gaussian windowed, radius 8)");
  vZoomInInterpolator->add("Cardinal sinus (Gaussian windowed, radius 16)");
  vZoomInInterpolator->add("Cardinal sinus (Gaussian windowed, radius 32)");
  vZoomInInterpolator->add("Prolate (radius 2)");
  vZoomInInterpolator->add("Prolate (radius 4)");
  vZoomInInterpolator->add("Prolate (radius 8)");
  vZoomInInterpolator->add("Prolate (radius 16)");
  vZoomInInterpolator->add("Prolate (radius 32)");

  vZoomOutInterpolator->add("Nearest neighbors");
  vZoomOutInterpolator->add("Bilinear");
  vZoomOutInterpolator->add("BSpline (2nd order)");
  vZoomOutInterpolator->add("BSpline (3rd order)");
  vZoomOutInterpolator->add("BSpline (4th order)");
  vZoomOutInterpolator->add("BSpline (5th order)");
  vZoomOutInterpolator->add("Cardinal sinus (Hamming windowed, radius 2)");
  vZoomOutInterpolator->add("Cardinal sinus (Hamming windowed, radius 4)");
  vZoomOutInterpolator->add("Cardinal sinus (Hamming windowed, radius 8)");
  vZoomOutInterpolator->add("Cardinal sinus (Hamming windowed, radius 16)");
  vZoomOutInterpolator->add("Cardinal sinus (Hamming windowed, radius 32)");
  vZoomOutInterpolator->add("Cardinal sinus (Gaussian windowed, radius 2)");
  vZoomOutInterpolator->add("Cardinal sinus (Gaussian windowed, radius 4)");
  vZoomOutInterpolator->add("Cardinal sinus (Gaussian windowed, radius 8)");
  vZoomOutInterpolator->add("Cardinal sinus (Gaussian windowed, radius 16)");
  vZoomOutInterpolator->add("Cardinal sinus (Gaussian windowed, radius 32)");
  vZoomOutInterpolator->add("Prolate (radius 2)");
  vZoomOutInterpolator->add("Prolate (radius 4)");
  vZoomOutInterpolator->add("Prolate (radius 8)");
  vZoomOutInterpolator->add("Prolate (radius 16)");
  vZoomOutInterpolator->add("Prolate (radius 32)");

  // Setting default interpolator
  vZoomInInterpolator->value(1);
  vZoomOutInterpolator->value(1);

  bNormalization->value(m_StereoscopicApplicationModel->GetClampThreshold()*100);

  // We do not have to reset the viewer for now
  m_NeedToReset = false;
  m_RenameIndex = 0;
  m_Tooltip = "";
  lMainCouples->tooltip(m_Tooltip.c_str());
  lInsightCouples->tooltip(m_Tooltip.c_str());
}

void StereoscopicApplicationViewGUI::RefreshViewer()
{
  ImageType::Pointer masterImage;
  ImageType::Pointer masterQuicklook;
  ImageType::Pointer slaveImage;

  int masterIndex = lMainCouples->value()-1;
  int slaveIndex = lInsightCouples->value()-1;

  if (masterIndex<0)
  {
    return;
  }
  if (slaveIndex<0)
  {
    slaveIndex = masterIndex;
  }

  // Main image setup
  if (bMainShowLeftImage->value())
  {
    masterImage=m_StereoscopicApplicationModel->GetLeftImage(masterIndex);
    masterQuicklook=m_StereoscopicApplicationModel->GetLeftQuicklook(masterIndex);
  }
  else if (bMainShowRightImage->value())
  {
    masterImage = m_StereoscopicApplicationModel->GetRightImage(masterIndex);
    masterQuicklook=m_StereoscopicApplicationModel->GetRightQuicklook(masterIndex);
  }
  else
  {
    masterImage = m_StereoscopicApplicationModel->GetAnaglyph(masterIndex);
    masterQuicklook=m_StereoscopicApplicationModel->GetAnaglyphQuicklook(masterIndex);
  }


  VectorPixelType min, max;
  min = m_StereoscopicApplicationModel->GetMin(masterIndex);
  max = m_StereoscopicApplicationModel->GetMax(masterIndex);

  // Insight setup
  if (slaveIndex >= 0)
  {
    if (bInsightShowLeftImage->value())
    {
      slaveImage=m_StereoscopicApplicationModel->GetLeftImage(slaveIndex);
    }
    else if (bInsightShowRightImage->value())
    {
      slaveImage = m_StereoscopicApplicationModel->GetRightImage(slaveIndex);
    }
    else
    {
      slaveImage = m_StereoscopicApplicationModel->GetAnaglyph(slaveIndex);
    }
  }

  // draw the viewer
  if (m_Viewer.IsNull())
  {
    m_Viewer = ViewerType::New();
    m_Viewer->SetImage(masterImage);

    ImageType::IndexType center;

    center[0]=masterImage->GetLargestPossibleRegion().GetIndex()[0]+masterImage->GetLargestPossibleRegion().GetSize()[0]/2;
    center[1]=masterImage->GetLargestPossibleRegion().GetIndex()[1]+masterImage->GetLargestPossibleRegion().GetSize()[1]/2;
    m_Viewer->SetViewedRegionCenter(center);
    if (slaveIndex >= 0)
    {
      m_Viewer->SetSecondImage(slaveImage);
      if (slaveImage->GetNumberOfComponentsPerPixel()<3)
      {
        m_Viewer->SetInsightViewModelToGrayscale();
      }
      else
      {
        m_Viewer->SetInsightViewModelToRGB();
      }
    }
    m_Viewer->SetMinComponentValues(min);
    m_Viewer->SetMaxComponentValues(max);
    m_Viewer->Init(gViewer->x(), gViewer->y(), gViewer->w(), gViewer->h(),"");
    if (masterImage->GetNumberOfComponentsPerPixel()<3)
    {
      m_Viewer->SetViewModelToGrayscale();
    }
    else
    {
      m_Viewer->SetViewModelToRGB();
    }


    gViewer->box(FL_NO_BOX);
    gViewer->add(m_Viewer);
    gViewer->show();
    m_Viewer->Show();
  }
  else
  {
    if (masterImage->GetNumberOfComponentsPerPixel()<3)
    {
      m_Viewer->SetViewModelToGrayscale();
    }
    else
    {
      m_Viewer->SetViewModelToRGB();
    }

    m_Viewer->SetMinComponentValues(min);
    m_Viewer->SetMaxComponentValues(max);
    m_Viewer->SetImage(masterImage);
    if (slaveIndex>=0)
    {
      m_Viewer->SetSecondImage(slaveImage);
      if (slaveImage->GetNumberOfComponentsPerPixel()<3)
      {
        m_Viewer->SetInsightViewModelToGrayscale();
      }
      else
      {
        m_Viewer->SetInsightViewModelToRGB();
      }
    }
    if (m_NeedToReset)
    {
      m_Viewer->Reset();
      m_NeedToReset = false;
    }
    else
    {
      m_Viewer->redraw();
    }
  }


  // Drawing the scroll box
  BoxType::IndexType boxIndex;
  BoxType::SizeType boxSize;
  double factor = m_Viewer->GetSpacingZoomFactor()/m_Viewer->GetOpenGlIsotropicZoom();
  boxSize[0]= static_cast<unsigned int>(m_Viewer->GetDisplayExtent().GetSize()[0]*factor);
  boxSize[1]= static_cast<unsigned int>(m_Viewer->GetDisplayExtent().GetSize()[1]*factor);
  boxIndex[0]=m_Viewer->GetViewedRegionCenter()[0]-boxSize[0]/2;
  boxIndex[1]=m_Viewer->GetViewedRegionCenter()[1]-boxSize[1]/2;

  BoxType::Pointer box = BoxType::New();
  box->SetSize(boxSize);
  box->SetIndex(boxIndex);
  box->SetColor(m_InterfaceColor);

  // Draw the scroll widget
  if (m_Scroll.IsNull())
  {
    m_Scroll  = ScrollWidgetType::New();
    m_Scroll->SetController(m_StereoscopicApplicationController);
    m_Scroll->SetInput(masterQuicklook);
    m_Scroll->Init(gScroll->x(), gScroll->y(), gScroll->w(), gScroll->h(),"Scroll");
    m_Scroll->SetSubSamplingRate(m_StereoscopicApplicationModel->GetShrinkFactor(masterIndex));
    m_Scroll->SetFormOverlayVisible(true);
    gScroll->add(m_Scroll);
    gScroll->box(FL_NO_BOX);

    for (unsigned int i = 0; i<masterQuicklook->GetNumberOfComponentsPerPixel(); ++i)
    {
      m_Scroll->GetTransferFunctionList()->GetNthElement(i)->SetLowerBound(min[i]);
      m_Scroll->GetTransferFunctionList()->GetNthElement(i)->SetUpperBound(max[i]);
    }

    if (masterQuicklook->GetNumberOfComponentsPerPixel()<3)
    {
      m_Scroll->SetViewModel(ScrollWidgetType::GRAYSCALE);
    }
    else
    {
      m_Scroll->SetViewModel(ScrollWidgetType::RGB);
    }

    m_Scroll->GetFormList()->PushBack(box);
    m_Scroll->Show();
  }
  else

  {
    m_Scroll->GetFormList()->SetNthElement(0, box);

    for (unsigned int i = 0; i<masterQuicklook->GetNumberOfComponentsPerPixel(); ++i)
    {
      m_Scroll->GetTransferFunctionList()->GetNthElement(i)->SetLowerBound(min[i]);
      m_Scroll->GetTransferFunctionList()->GetNthElement(i)->SetUpperBound(max[i]);
    }
    m_Scroll->SetInput(masterQuicklook);

    if (masterQuicklook->GetNumberOfComponentsPerPixel()<3)
    {
      m_Scroll->SetViewModel(ScrollWidgetType::GRAYSCALE);
    }
    else
    {
      m_Scroll->SetViewModel(ScrollWidgetType::RGB);
    }

    m_Scroll->ClearBufferedRegion();
    m_Scroll->redraw();
  }
}

void StereoscopicApplicationViewGUI::ResetViewer()
{
  m_NeedToReset = true;
}

void StereoscopicApplicationViewGUI::Notify()
{}

void
StereoscopicApplicationViewGUI
::SetStereoscopicApplicationController(StereoscopicApplicationControllerInterface * controller)
{
  m_StereoscopicApplicationController = controller;
}

void StereoscopicApplicationViewGUI::ActivateMainCouplesInput(bool active)
{
  if (active)
  {
    bMainShowLeftImage->activate();
    bMainShowRightImage->activate();
    bMainShowAnaglyph->activate();
  }
  else
  {
    bMainShowLeftImage->deactivate();
    bMainShowRightImage->deactivate();
    bMainShowAnaglyph->deactivate();
  }
}

void StereoscopicApplicationViewGUI::ActivateInsightCouplesInput(bool active)
{
  if (active)
  {
    bInsightShowLeftImage->activate();
    bInsightShowRightImage->activate();
    bInsightShowAnaglyph->activate();
  }
  else
  {
    bInsightShowLeftImage->deactivate();
    bInsightShowRightImage->deactivate();
    bInsightShowAnaglyph->deactivate();
  }
}


void StereoscopicApplicationViewGUI::MainCouplesCallback()
{
  if (lMainCouples->value()>0)
  {
    itk::OStringStream oss;
    oss.str("");
    oss<<"Left  image file: "<<m_StereoscopicApplicationModel->GetLeftImageFilename(lMainCouples->value()-1)<<std::endl;
    oss<<"Right image file: "<<m_StereoscopicApplicationModel->GetRightImageFilename(lMainCouples->value()-1);

    m_Tooltip = oss.str();
    lMainCouples->tooltip(m_Tooltip.c_str());

    this->ActivateMainCouplesInput(true);
    m_NeedToReset=true;
    m_RenameIndex = lMainCouples->value();
    vRenameField->value(lMainCouples->text(m_RenameIndex));
    this->RefreshViewer();
  }
  else
  {
    this->ActivateMainCouplesInput(false);
  }
}

void StereoscopicApplicationViewGUI::MainShowLeftImageCallback()
{
  bMainShowLeftImage->set();
  bMainShowRightImage->clear();
  bMainShowAnaglyph->clear();
  m_NeedToReset = true;
  this->RefreshViewer();
}

void StereoscopicApplicationViewGUI::MainShowRightImageCallback()
{
  bMainShowLeftImage->clear();
  bMainShowRightImage->set();
  bMainShowAnaglyph->clear();
  m_NeedToReset = true;
  this->RefreshViewer();
}
void StereoscopicApplicationViewGUI::MainShowAnaglyphCallback()
{
  bMainShowLeftImage->clear();
  bMainShowRightImage->clear();
  bMainShowAnaglyph->set();
  m_NeedToReset = true;
  this->RefreshViewer();

}
void StereoscopicApplicationViewGUI::ZoomInInterpolatorCallback()
{
  m_Viewer->SetZoomInInterpolator(this->GetInterpolator(vZoomInInterpolator->value()));
  m_NeedToReset=true;
  this->RefreshViewer();
}
void StereoscopicApplicationViewGUI::ZoomOutInterpolatorCallback()
{
  m_Viewer->SetZoomOutInterpolator(this->GetInterpolator(vZoomOutInterpolator->value()));
  m_NeedToReset=true;
  this->RefreshViewer();
}

StereoscopicApplicationViewGUI::InterpolatorPointerType StereoscopicApplicationViewGUI::GetInterpolator(unsigned int index)
{
  InterpolatorPointerType resp;
  BSplineInterpolatorType::Pointer spline2, spline3, spline4, spline5;
  GaussianWindowedSincInterpolatorType::Pointer sinc;
  ProlateInterpolatorType::Pointer prolate;

  switch (index)
  {
  case 1:
    resp = LinearInterpolatorType::New();
    break;
  case 2:
    spline2 = BSplineInterpolatorType::New();
    spline2->SetSplineOrder(2);
    resp = spline2;
    break;
  case 3:
    spline3 = BSplineInterpolatorType::New();
    spline3->SetSplineOrder(3);
    resp = spline3;
    break;
  case 4:
    spline4 = BSplineInterpolatorType::New();
    spline4->SetSplineOrder(4);
    resp = spline4;
    break;
  case 5:
    spline5 = BSplineInterpolatorType::New();
    spline5->SetSplineOrder(5);
    resp = spline5;
    break;
  case 6:
    resp = Hamming2WindowedSincInterpolatorType::New();
    break;
  case 7:
    resp = Hamming4WindowedSincInterpolatorType::New();
    break;
  case 8:
    resp = Hamming8WindowedSincInterpolatorType::New();
    break;
  case 9:
    resp = Hamming16WindowedSincInterpolatorType::New();
    break;
  case 10:
    resp = Hamming32WindowedSincInterpolatorType::New();
    break;
  case 11:
    sinc = GaussianWindowedSincInterpolatorType::New();
    sinc->SetRadius(2);
    resp = sinc;
    break;
  case 12:
    sinc = GaussianWindowedSincInterpolatorType::New();
    sinc->SetRadius(4);
    resp = sinc;
    break;
  case 13:
    sinc = GaussianWindowedSincInterpolatorType::New();
    sinc->SetRadius(8);
    resp = sinc;
    break;
  case 14:
    sinc = GaussianWindowedSincInterpolatorType::New();
    sinc->SetRadius(16);
    resp = sinc;
    break;
  case 15:
    sinc = GaussianWindowedSincInterpolatorType::New();
    sinc->SetRadius(32);
    resp = sinc;
    break;
  case 16:
    prolate = ProlateInterpolatorType::New();
    prolate->SetRadius(2);
    resp = prolate;
    break;
  case 17:
    prolate = ProlateInterpolatorType::New();
    prolate->SetRadius(4);
    resp = prolate;
    break;
  case 18:
    prolate = ProlateInterpolatorType::New();
    prolate->SetRadius(8);
    resp = prolate;
    break;
  case 19:
    prolate = ProlateInterpolatorType::New();
    prolate->SetRadius(16);
    resp = prolate;
    break;
  case 20:
    prolate = ProlateInterpolatorType::New();
    prolate->SetRadius(32);
    resp = prolate;
    break;
  default:
    resp = NearestNeighborInterpolatorType::New();
    break;
  }
  return resp;
}


void StereoscopicApplicationViewGUI::InsightCouplesCallback()
{
  if (lInsightCouples->value()>0)
  {
    itk::OStringStream oss;
    oss.str("");
    oss<<"Left  image file: "<<m_StereoscopicApplicationModel->GetLeftImageFilename(lInsightCouples->value()-1)<<std::endl;
    oss<<"Right image file: "<<m_StereoscopicApplicationModel->GetRightImageFilename(lInsightCouples->value()-1);
    m_Tooltip = oss.str();

    lInsightCouples->tooltip(m_Tooltip.c_str());

    this->ActivateInsightCouplesInput(true);
    m_RenameIndex = lInsightCouples->value();
    vRenameField->value(lInsightCouples->text(m_RenameIndex));
    this->RefreshViewer();
  }
  else
  {
    this->ActivateInsightCouplesInput(false);
  }
}


void StereoscopicApplicationViewGUI::InsightShowLeftImageCallback()
{
  bInsightShowLeftImage->set();
  bInsightShowRightImage->clear();
  bInsightShowAnaglyph->clear();
  this->RefreshViewer();
}
void StereoscopicApplicationViewGUI::InsightShowRightImageCallback()
{
  bInsightShowLeftImage->clear();
  bInsightShowRightImage->set();
  bInsightShowAnaglyph->clear();
  this->RefreshViewer();
}
void StereoscopicApplicationViewGUI::InsightShowAnaglyphCallback()
{
  bInsightShowLeftImage->clear();
  bInsightShowRightImage->clear();
  bInsightShowAnaglyph->set();
  this->RefreshViewer();
}
void StereoscopicApplicationViewGUI::OpenStereoscopicCoupleCallback()
{
  wOpenCouplesWindow->show();

}
void StereoscopicApplicationViewGUI::QuitCallback()
{
  wMainWindow->hide();
  wOpenCouplesWindow->hide();
}
void StereoscopicApplicationViewGUI::OpenCoupleCancelCallback()
{
  wOpenCouplesWindow->hide();
}
void StereoscopicApplicationViewGUI::OpenCoupleOkCallback()
{
  m_StereoscopicApplicationController->OpenStereoCouple(vLeftImagePath->value(), vRightImagePath->value());
  wOpenCouplesWindow->hide();
}
void StereoscopicApplicationViewGUI::OpenCoupleLeftBrowserCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Choose the left image file ...", "*.*",".");
  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  vLeftImagePath->value(filename);
}
void StereoscopicApplicationViewGUI::OpenCoupleRightBrowserCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Choose the right image file ...", "*.*",".");
  if (filename == NULL || !strcmp(filename, ""))
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  vRightImagePath->value(filename);
}

void StereoscopicApplicationViewGUI::MagnifyCallback()
{
  vMagnifyValue->value(vMagnify->value());
  if (m_Viewer.IsNotNull())
  {
    m_Viewer->SetOpenGlIsotropicZoom(vMagnify->value());
    this->RefreshViewer();
  }
}

void StereoscopicApplicationViewGUI::ResampleCallback()
{
  double factor = vcl_pow(2., -vResample->value());
  vResampleValue->value(factor);
  vResampleValue->redraw();
  if (m_Viewer.IsNotNull())
  {
    m_Viewer->SetSpacingZoomFactor(factor);
    vResample->deactivate();
    vResample->redraw();
    this->RefreshViewer();
    vResample->activate();
    vResample->redraw();
  }
}

void StereoscopicApplicationViewGUI::NormalizationCallback()
{
  m_StereoscopicApplicationController->ChangeNormalizationFactor(bNormalization->value());
}

void StereoscopicApplicationViewGUI::RenameCallback()
{
  wRenameWindow->show();
}

void StereoscopicApplicationViewGUI::RenameOkCallback()
{
  lMainCouples->text(m_RenameIndex, vRenameField->value());
  lInsightCouples->text(m_RenameIndex, vRenameField->value());
  wRenameWindow->hide();
}

void StereoscopicApplicationViewGUI::RenameCancelCallback()
{
  wRenameWindow->hide();
}

void StereoscopicApplicationViewGUI::UseInsightCallback()
{


}

}
