/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbStereoscopicApplicationViewGUI_h
#define __otbStereoscopicApplicationViewGUI_h

// Disabling deprecation warning
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4996)
#endif
#include "otbStereoscopicApplicationViewGroup.h"
#ifdef _MSC_VER
#pragma warning(pop)
#endif
#include "otbListenerBase.h"
#include "otbStereoscopicApplicationModel.h"
#include "otbStereoscopicApplicationControllerInterface.h"

#include "itkObject.h"

#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Tabs.H>

#include "otbImageAlternateViewer.h"
#include "otbInteractiveScrollWidget.h"
#include "otbImageWidgetBoxForm.h"

// Interpolators
#include "itkInterpolateImageFunction.h"
#include "itkLinearInterpolateImageFunction.h"
#include "otbBSplineInterpolateImageFunction.h"
#include "itkWindowedSincInterpolateImageFunction.h"
#include "otbWindowedSincInterpolateImageGaussianFunction.h"
#include "otbWindowedSincInterpolateImageHammingFunction.h"
#include "itkZeroFluxNeumannBoundaryCondition.h"
#include "otbProlateInterpolateImageFunction.h"
#include "itkNearestNeighborInterpolateImageFunction.h"

namespace otb
{
/** \class StereoscopicApplicationViewGUI
 *
 */
class StereoscopicApplicationViewGUI
      : public ListenerBase, public StereoscopicApplicationViewGroup, public itk::Object
{
public:
  /** Standard class typedefs */
  typedef StereoscopicApplicationViewGUI Self;
  typedef itk::Object                    Superclass;
  typedef itk::SmartPointer<Self>        Pointer;
  typedef itk::SmartPointer<const Self>  ConstPointer;

  // Pixel type
  typedef double                               PixelType;
  typedef otb::ImageAlternateViewer<PixelType> ViewerType;
  typedef ViewerType::SingleImageType          SingleImageType;
  typedef ViewerType::ImageType                ImageType;
  typedef ImageType::PixelType                 VectorPixelType;
  typedef otb::InteractiveScrollWidget<PixelType, StereoscopicApplicationControllerInterface>  ScrollWidgetType;
  typedef ImageWidgetBoxForm                   BoxType;
  typedef BoxType::ColorType                   ColorType;

  // Interpolators
  typedef itk::InterpolateImageFunction<SingleImageType, double> InterpolatorType;
  typedef InterpolatorType::Pointer InterpolatorPointerType;
  typedef itk::NearestNeighborInterpolateImageFunction<SingleImageType, double> NearestNeighborInterpolatorType;
  typedef itk::LinearInterpolateImageFunction<SingleImageType, double> LinearInterpolatorType;
  typedef BSplineInterpolateImageFunction<SingleImageType, double, double> BSplineInterpolatorType;
  typedef itk::Function::HammingWindowFunction<4> WindowFunctionType;
  typedef otb::Function::GaussianWindowFunction<> GaussianWindowType;
  typedef itk::ZeroFluxNeumannBoundaryCondition<SingleImageType> ConditionType;
  typedef itk::WindowedSincInterpolateImageFunction<SingleImageType, 2, WindowFunctionType, ConditionType , double> Hamming2WindowedSincInterpolatorType;
  typedef itk::WindowedSincInterpolateImageFunction<SingleImageType, 4, WindowFunctionType, ConditionType , double> Hamming4WindowedSincInterpolatorType;
  typedef itk::WindowedSincInterpolateImageFunction<SingleImageType, 8, WindowFunctionType, ConditionType , double> Hamming8WindowedSincInterpolatorType;
  typedef itk::WindowedSincInterpolateImageFunction<SingleImageType, 16, WindowFunctionType, ConditionType , double> Hamming16WindowedSincInterpolatorType;
  typedef itk::WindowedSincInterpolateImageFunction<SingleImageType, 32, WindowFunctionType, ConditionType , double> Hamming32WindowedSincInterpolatorType;
  typedef WindowedSincInterpolateImageGaussianFunction<SingleImageType, ConditionType> GaussianWindowedSincInterpolatorType;
  typedef ProlateInterpolateImageFunction<SingleImageType, ConditionType> ProlateInterpolatorType;


  /** Standards macros */
  itkNewMacro(Self);
  itkTypeMacro(StereoscopicApplicationViewGUI, Object);

  /** Set the controller */
  virtual void SetStereoscopicApplicationController(StereoscopicApplicationControllerInterface * controller);


  /** Event from the model */
  virtual void Notify();

  /** Buttons callback */
  virtual void MainCouplesCallback();
  virtual void MainShowLeftImageCallback();
  virtual void MainShowRightImageCallback();
  virtual void MainShowAnaglyphCallback();
  virtual void ZoomInInterpolatorCallback();
  virtual void ZoomOutInterpolatorCallback();
  virtual void InsightCouplesCallback();
  virtual void InsightShowLeftImageCallback();
  virtual void InsightShowRightImageCallback();
  virtual void InsightShowAnaglyphCallback();
  virtual void OpenStereoscopicCoupleCallback();
  virtual void QuitCallback();
  virtual void OpenCoupleCancelCallback();
  virtual void OpenCoupleOkCallback();
  virtual void OpenCoupleLeftBrowserCallback();
  virtual void OpenCoupleRightBrowserCallback();
  virtual void MagnifyCallback();
  virtual void ResampleCallback();
  virtual void NormalizationCallback();
  virtual void RenameCallback();
  virtual void RenameOkCallback();
  virtual void RenameCancelCallback();
  virtual void UseInsightCallback();

  /** Refresh the image widgets */
  virtual void RefreshViewer();
  virtual void ResetViewer();

  /** Get a given interpolator */
  InterpolatorPointerType GetInterpolator(unsigned int index);


  /** Pointer to the viewer */
  ViewerType::Pointer m_Viewer;
  /** Pointer to the scroll image */
  ScrollWidgetType::Pointer m_Scroll;

protected:
  /** Constructor */
  StereoscopicApplicationViewGUI();
  /** Destructor */
  ~StereoscopicApplicationViewGUI() {};

  // Activate or deactivate the input blocks
  void ActivateMainCouplesInput(bool active);
  void ActivateInsightCouplesInput(bool active);

private:
  StereoscopicApplicationViewGUI(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  /** Pointer to the model */
  StereoscopicApplicationModel::Pointer m_StereoscopicApplicationModel;
  /** Pointer to the controller */
  StereoscopicApplicationControllerInterface::Pointer m_StereoscopicApplicationController;
  /** Interface color */
  ColorType m_InterfaceColor;
  /** Do we have to reset the main viewer ? */
  bool m_NeedToReset;
  /** Keep trace of the couple to rename */
  unsigned int m_RenameIndex;
  /** Tooltip  to display filenames */
  std::string m_Tooltip;


};
}//end namespace otb

#endif
