/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbUrbanAreaExtractionController.h"
#include "otbMsgReporter.h"

namespace otb
{

  UrbanAreaExtractionController
  ::UrbanAreaExtractionController()
  {
    // Build a focusing controller
    m_FocusingController = FocusingControllerType::New();
    m_PixelModel = PixelDescriptionModelType::New();
    m_PixelView = PixelDescriptionViewType::New();
    m_PixelModelWorking = PixelDescriptionModelType::New();
    m_PixelViewWorking = PixelDescriptionViewType::New();
  }

  UrbanAreaExtractionController
  ::~UrbanAreaExtractionController()
  {}

  void
  UrbanAreaExtractionController
  ::OpenInputImage(const char * filename)
  {
    try
    {
      std::string strFilename = filename;
      m_Model->SetImageFilename( strFilename );
      m_Model->OpenImage();
      m_Model->GetPolygonsManager()->ClearAllPolygons();
      m_Model->GetFocusingModel()->ClearAllVectorData();
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::OpenInputVector( const char * filename )
  {
    try
    {
      m_Model->OpenVector( filename );
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }
  
  void
  UrbanAreaExtractionController
  ::SaveImageOnExtract(const char * filename)
  {
    try
    {
      m_Model->SaveImage(filename);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }
  
  void
  UrbanAreaExtractionController
  ::SaveVectorOnExtract(const char * filename)
  {
    try
    {
      m_Model->SaveVectorOnExtract(filename);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SaveImageOnFull(const char * filename)
  {
    try
    {
      m_Model->SaveImageOnFull(filename); ;
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }
  
  void
  UrbanAreaExtractionController
  ::SaveVectorOnFull(const char * filename)
  {
    try
    {
      m_Model->SaveVectorOnFull(filename);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::Run()
  {
    try
    {
      m_Model->GetPolygonsManager()->ClearAllPolygons();
      m_Model->GetFocusingModel()->ClearAllVectorData();

      // Re add the Loaded Vector if it exists
      if (m_Model->GetLoadedVDIndex() != itk::NumericTraits<unsigned int>::max())
      {
        ColorType lBlue;
        lBlue.Fill(0);
        lBlue[2]=1.;
        lBlue[3]=1;
        m_Model->GetFocusingModel()->SetColor(lBlue);
        m_Model->GetFocusingModel()->AddVectorData(m_Model->GetLoadedVector());
        unsigned int lIndex = m_Model->GetFocusingModel()->GetVectorDataList()->Size() -1;
        m_Model->SetLoadedVDIndex(lIndex);
      }
      m_Model->ExtractUrbanAreas();

    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetThreshold(double value)
  {
    try
    {
      m_Model->SetThresholdValue(value);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetThresholdDensity(double value)
  {
    try
    {
      m_Model->SetThresholdDensity(value);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetSobelLowerThreshold(double value)
  {
    try
    {
      m_Model->SetSobelLowerThreshold(value);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetSobelUpperThreshold(double value)
  {
    try
    {
      m_Model->SetSobelUpperThreshold(value);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetCurrentPolygonIndex(unsigned int index)
  {
    try
    {
      m_Model->GetPolygonsManager()->SetCurrentPolygonIndex(index);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::ClearAllPolygons()
  {
    try
    {
      m_Model->InitVector();
      m_Model->GetPolygonsManager()->ClearAllPolygons();
      m_Model->SetLoadedVDIndex(itk::NumericTraits<unsigned int>::max());
      m_Model->GetFocusingModel()->ClearAllVectorData();
      m_Model->GetFocusingModel()->Update();
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::DeletePolygon(unsigned int index)
  {
    try
    {

      if(m_Model->GetLoadedVDIndex()<=index)
       m_Model->GetFocusingModel()->ClearVectorData(index+1);
      else
       m_Model->GetFocusingModel()->ClearVectorData(index);
    
      m_Model->GetPolygonsManager()->DeletePolygon(index);
      m_Model->DisplayNewLayer();
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetAlpha(double alpha)
  {
    try
    {
      m_Model->SetAlpha(alpha);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetIndices(unsigned int green, unsigned int red , unsigned int nir)
  {
    try
    {
      m_Model->SetGreenIndex(green);
      m_Model->SetRedIndex(red);
      m_Model->SetNIRIndex(nir);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetMinDetail(unsigned int val)
  {
    try
    {
      m_Model->SetMinRegionSize(val);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetSubSampleLevel(unsigned int val)
  {
    try
    {
      m_Model->SetSubSampleLevel(val);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::SetVectorDisplayed(bool value)
  {
    try
    {
      m_Model->SetVectorDisplayed(value);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::UpdateRGBChannelOrder(unsigned int red, unsigned int green, unsigned int blue)
  {
    try
    {
      m_Model->GetFocusingModel()->UpdateRGBChannelOrder(red, green, blue);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::UpdateGrayscaleChannel(unsigned int channel)
  {
    try
    {
      m_Model->GetFocusingModel()->UpdateGrayscaleChannel(channel);
    }
    catch (itk::ExceptionObject & err)
    {
      MsgReporter::GetInstance()->SendError(err.GetDescription());
    }
  }

  void
  UrbanAreaExtractionController
  ::AddHandlers()
  {

    m_FocusingController->SetFocusingModel(m_Model->GetFocusingModel());
    m_FocusingController->SetFocusingView(m_View->GetFocusingView());
    // Add the select area handler
    m_FocusingController->AddSelectAreaHandler();

    // Add the resizing handler to the Master viewer
    ResizingHandlerType::Pointer resizingHandlerM = ResizingHandlerType::New();
    resizingHandlerM->SetModel(m_Model->GetMasterModel());
    resizingHandlerM->SetView(m_View->GetMasterView());
    this->GetMasterController()->AddActionHandler(resizingHandlerM);

    // Add the resizing handler to the Working viewer
    ResizingHandlerType::Pointer resizingHandlerW = ResizingHandlerType::New();
    resizingHandlerW->SetModel(m_Model->GetWorkingModel());
    resizingHandlerW->SetView(m_View->GetWorkingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(resizingHandlerW);

    // Add the change scaled region handler
    ChangeScaledRegionHandlerType::Pointer changeScaledHandler =ChangeScaledRegionHandlerType::New();
    changeScaledHandler->SetModel(m_Model->GetWorkingModel());
    changeScaledHandler->SetView(m_View->GetWorkingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(changeScaledHandler);

    // Add the change scaled handler
    ChangeScaleHandlerType::Pointer changeScaleHandler =ChangeScaleHandlerType::New();
    changeScaleHandler->SetModel(m_Model->GetWorkingModel());
    changeScaleHandler->SetView(m_View->GetWorkingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(changeScaleHandler);

    // Add the change region action handler
    ChangeRegionHandlerType::Pointer changeRegionHandler = ChangeRegionHandlerType::New();
    changeRegionHandler->SetModel(m_Model->GetWorkingModel());
    changeRegionHandler->SetView(m_View->GetWorkingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(changeRegionHandler);

    // Add the arrow keys move action handler
    ArrowKeyMoveActionHandlerType::Pointer arrowKeyMoveHandler = ArrowKeyMoveActionHandlerType::New();
    arrowKeyMoveHandler->SetModel(m_Model->GetWorkingModel());
    arrowKeyMoveHandler->SetView(m_View->GetWorkingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(arrowKeyMoveHandler);

    // Add the Opacity management
    ChangeOpacityHandlerType::Pointer changeOpacityHandler = ChangeOpacityHandlerType::New();
    changeOpacityHandler->SetModel(m_Model->GetFocusingModel());
    changeOpacityHandler->SetView(m_View->GetFocusingView());
    m_FocusingController->GetWorkingController()->AddActionHandler(changeOpacityHandler);
    m_FocusingController->GetMasterController()->AddActionHandler(changeOpacityHandler);

    // Pixel description
    // master scroll
    m_PixelModel->SetLayers(m_Model->GetMasterModel()->GetLayers());
    PixelDescriptionActionHandlerType::Pointer pixelActionHandler = PixelDescriptionActionHandlerType::New();
    pixelActionHandler->SetView(m_View->GetMasterView());
    pixelActionHandler->SetModel(m_PixelModel);
    m_FocusingController->GetMasterController()->AddActionHandler(pixelActionHandler);
    m_PixelView->SetModel(m_PixelModel);
    // Build a pixel description view
    m_PixelView->SetModel(m_PixelModel);
    m_View->gPixelValue->add(m_PixelView->GetPixelDescriptionWidget());
    m_PixelView->GetPixelDescriptionWidget()->show();
    m_PixelView->GetPixelDescriptionWidget()->resize(m_View->gPixelValue->x(), m_View->gPixelValue->y(), m_View->gPixelValue->w(), m_View->gPixelValue->h());

    // working scroll
    m_PixelModelWorking->SetLayers(m_Model->GetWorkingModel()->GetLayers());
    PixelDescriptionActionHandlerType::Pointer pixelActionHandler2 = PixelDescriptionActionHandlerType::New();
    pixelActionHandler2->SetView(m_View->GetWorkingView());
    pixelActionHandler2->SetModel(m_PixelModelWorking);
    m_FocusingController->GetWorkingController()->AddActionHandler(pixelActionHandler2);
    m_PixelViewWorking->SetModel(m_PixelModelWorking);
    // Build a pixel description view
    m_PixelViewWorking->SetModel(m_PixelModelWorking);
    m_View->gPixelValueW->add(m_PixelViewWorking->GetPixelDescriptionWidget());
    //m_View->gPixelValue->resizable(m_PixelView->GetPixelDescriptionWidget());
    //m_View->gPixelValue->show();
    m_PixelViewWorking->GetPixelDescriptionWidget()->show();
    m_PixelViewWorking->GetPixelDescriptionWidget()->resize(m_View->gPixelValueW->x(), m_View->gPixelValueW->y(), m_View->gPixelValueW->w(), m_View->gPixelValueW->h());

  }

} // end namespace otb

