/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#include "otbUrbanAreaExtractionModel.h"
#include "otbFltkFilterWatcher.h"
#include <FL/fl_ask.H>
#include "itkExceptionObject.h"
#include "otbMacro.h"

#include "otbStreamingImageFileWriter.h"

#include "otbUniformAlphaBlendingFunction.h"
#include "otbBlendingFunction.h"

#include "itkExceptionObject.h"
#include "itkPreOrderTreeIterator.h"


namespace otb
{

/** Initialize the singleton */
UrbanAreaExtractionModel::Pointer UrbanAreaExtractionModel::Instance = NULL;

UrbanAreaExtractionModel::UrbanAreaExtractionModel()
{
  // Viewer
  m_FocusingModel = FocusingModelType::New();

  m_LoadedVector = VectorDataType::New();
  m_Reader = ReaderType::New();
  m_PolygonsManager = PolygonsManagerType::New();

  this->InitPipeline();
  this->InitVector();

  m_HasNewInput = false;
  m_VectorDisplayed = false;
  m_ThresholdValue = 0.8;
  m_ThresholdDensity = 0.01;
  m_SobelLowerThreshold = -100.0;
  m_SobelUpperThreshold = 200.0;
  m_SubSampleLevel = 1;
  m_MinRegionSize = 40;
  m_Alpha = 0.0;

  // Channels Indices
  m_GreenIndex = 2;
  m_RedIndex = 3;
  m_NIRIndex = 4;

  m_UrbanAreaDetectionFilter->SetGreenIndex(m_GreenIndex);
  m_UrbanAreaDetectionFilter->SetRedIndex(m_RedIndex);
  m_UrbanAreaDetectionFilter->SetNIRIndex(m_NIRIndex);

  m_State = NO_IMAGE;
  m_WhichImage = EXTRACT;

}


UrbanAreaExtractionModel
::~UrbanAreaExtractionModel() {}


/** Manage the singleton */
UrbanAreaExtractionModel::Pointer
UrbanAreaExtractionModel::GetInstance()
{
  if (!Instance)
  {
    Instance = UrbanAreaExtractionModel::New();
  }
  return Instance;
}

void
UrbanAreaExtractionModel
::NotifyListener(ListenerBase * listener)
{
  listener->Notify();
}

void
UrbanAreaExtractionModel
::OpenImage()
{
  if(m_ImageFilename != "" )
  {
    m_Reader = ReaderType::New();
    m_Reader->SetFileName(m_ImageFilename);
    this->InitPipeline();

    // Watcher
    FltkFilterWatcher qlwatcher(m_Reader, 0, 0, 250, 20,"Loading image...");
    m_Reader->GenerateOutputInformation();

    m_PolygonsManager->ClearAllPolygons();

    m_FocusingModel->Init();
    m_FocusingModel->SetInputImage( m_Reader->GetOutput() );
    m_FocusingModel->UpdateModels();

    // Notify the observers
    m_HasNewInput = true;
    m_State = IMAGE_OPENED;
    this->NotifyAll();
    m_HasNewInput = false;
    m_HasOutputImage = false;
    m_LoadedVector = VectorDataType::New();
    m_LoadedVDIndex = itk::NumericTraits<unsigned int>::max();
  }
}

void
UrbanAreaExtractionModel
::OpenVector(std::string name)
{
  VectorDataReaderPointerType vectorDataReader = VectorDataReaderType::New();
  vectorDataReader->SetFileName(name);
  vectorDataReader->Update();
  m_LoadedVector = vectorDataReader->GetOutput();

  ColorType lBlue;

  lBlue.Fill(0);
  lBlue[2]=1.;
  lBlue[3]=1;
  m_FocusingModel->SetColor(lBlue);
  m_FocusingModel->AddVectorData(m_LoadedVector);
  m_LoadedVDIndex = m_FocusingModel->GetVectorDataList()->Size() -1;

  // to display the vector
  m_VectorDisplayed = true;
  this->NotifyAll();
}

void
UrbanAreaExtractionModel
::ExtractUrbanAreas()
{
    m_WhichImage = EXTRACT;
    if(m_FocusingModel->GetWorkingModel()->GetHasExtract())
    {
      unsigned int lNbOfBands = m_Reader->GetOutput()->GetNumberOfComponentsPerPixel();
      if( lNbOfBands > 2 )
      {
        /** Compute urban areas on an image */
        this->FindUrbanAreas();
        m_VectorDisplayed = true;
        this->NotifyAll();
      }
      else
      {
        itkExceptionMacro(<<"The input image only has "<<lNbOfBands<< " band(s), but at least three are required.");
      }
    }
    else
    {
      itkExceptionMacro(<<"Select a working area before running!");
    }
}

void
UrbanAreaExtractionModel
::InitPipeline()
{

  // Urban Area detection : generate a binary image
  m_UrbanAreaDetectionFilter = UrbanAreaDetectionFilterType::New();
  // Polygons Manager : use the binary image to produce a list of polygons
  m_PolygonsManager = PolygonsManagerType::New();

  m_InputImage = InputImageType::New();
  m_CurrentImage = InputImageType::New();
  m_OutputImage = OutputImageType::New();
  m_ShrinkFilter =  StreamingShrinkImageFilterType::New();
  m_ExpandFilter =  ExpandImageFilterType::New();
  m_ExtractROIFilter = ExtractROIFilterType::New();

  this->SetInputImage( m_Reader->GetOutput() );
  RegionType lRegion;

  // Choose between working on an extract or on the full image (using banners)
  if(m_WhichImage == EXTRACT)
  {

    lRegion =  m_FocusingModel->GetWorkingModel()->GetExtractRegion();
    m_ExtractROIFilter->SetInput(m_FocusingModel->GetExtractedImage());
    m_SavedIndex =  this->GetFocusingModel()->GetExtractRegion().GetIndex();

  }
  else
  {
    lRegion = m_StreamingRegion;
    m_ExtractROIFilter->SetInput(m_InputImage);
  }

  m_ExtractROIFilter->SetExtractionRegion(lRegion);
  m_ShrinkFilter->SetInput( m_ExtractROIFilter->GetOutput() );


  // Urban Area Detection Filter
  if( m_SubSampleLevel != 1 )
  {
    m_UrbanAreaDetectionFilter->SetInput(m_ShrinkFilter->GetOutput());
  }
  else
  {
    m_UrbanAreaDetectionFilter->SetInput(m_ShrinkFilter->GetInput());
  }

  // Polygons Manager
  m_PolygonsManager->SetReaderImage(m_Reader->GetOutput());
  m_PolygonsManager->SetInputImage(m_ExpandFilter->GetOutput());

  // Expand
  m_ExpandFilter->SetInput(m_UrbanAreaDetectionFilter->GetOutput());

}

void
UrbanAreaExtractionModel
::FindUrbanAreas()
{

  m_HasOutputImage = true;
  this->InitPipeline();

  m_UrbanAreaDetectionFilter->SetGreenIndex(m_GreenIndex);
  m_UrbanAreaDetectionFilter->SetRedIndex(m_RedIndex);
  m_UrbanAreaDetectionFilter->SetNIRIndex(m_NIRIndex);

  m_UrbanAreaDetectionFilter->SetThresholdValue(m_ThresholdValue);
  m_UrbanAreaDetectionFilter->SetThresholdDensity(m_ThresholdDensity);
  m_UrbanAreaDetectionFilter->SetSobelLowerThreshold(m_SobelLowerThreshold);
  m_UrbanAreaDetectionFilter->SetSobelUpperThreshold(m_SobelUpperThreshold);

  // Watcher
  FltkFilterWatcher qlwatcher(m_UrbanAreaDetectionFilter, 0, 0, 250, 20,"Searching urban areas...");
  m_UrbanAreaDetectionFilter->Update();

  m_ShrinkFilter->SetShrinkFactor(m_SubSampleLevel);
  m_ExpandFilter->SetExpandFactors(m_SubSampleLevel);


  // Launch execution
  m_PolygonsManager->GeneratePolygons(m_MinRegionSize, m_SubSampleLevel);
  this->DisplayNewLayer();

  for(unsigned int i=0; i<m_PolygonsManager->GetVectorDataList()->Size(); i++)
  {
      m_FocusingModel->SetColor( m_PolygonsManager->GetColoredPolyList()->GetNthElement(i)->GetColor() );
      m_FocusingModel->AddVectorData(m_PolygonsManager->GetVectorDataList()->GetNthElement(i));

      if(m_FocusingModel->GetMasterVectorDataGlList()->GetNthElement( 0 )->GetVisible() == false)
       {
         m_FocusingModel->GetMasterVectorDataGlList()->GetNthElement( i )->SetVisible(false);
         m_FocusingModel->GetWorkingVectorDataGlList()->GetNthElement( i )->SetVisible(false);
         m_FocusingModel->UpdateMasterVectorData(i);
         m_FocusingModel->UpdateWorkingVectorData(i);
       }
  }

}

void
UrbanAreaExtractionModel
::DisplayNewLayer()
{

  // Resize the extracted area
//   if( ( m_SubSampleLevel != 1 ) &&
//       ( m_PolygonsManager->GetRGBImage() != NULL ) )
//   {
//     if( m_PolygonsManager->GetRGBImage()->GetLargestPossibleRegion().GetSize() != m_FocusingModel->GetExtractRegion().GetSize() )
//     {
//       SizeType lSize = m_PolygonsManager->GetRGBImage()->GetLargestPossibleRegion().GetSize();
//       IndexType lStart = m_SavedIndex;
//       IndexType lStop;
//       lStop[0] = lStart[0] + lSize[0];
//       lStop[1] = lStart[1] + lSize[1];
//
//       m_FocusingModel->SetExtractRegionByIndex(lStart, lStop);
//       m_FocusingModel->Update();
//     }
//   }

  /** To avoid adding an old computed layer on a new extract */
  if ( ( m_PolygonsManager->GetRGBImage() != NULL ) &&
       ( m_SubSampleLevel == 1 ) &&
       (( m_Reader->GetOutput()->GetLargestPossibleRegion().GetSize() == m_PolygonsManager->GetRGBImage()->GetLargestPossibleRegion().GetSize() ) ||
        ( m_PolygonsManager->GetRGBImage()->GetLargestPossibleRegion().GetSize() == m_FocusingModel->GetExtractRegion().GetSize()))
       )
  {

    VectorLayerGeneratorType::Pointer lGenerator = VectorLayerGeneratorType::New();
    lGenerator->SetImage(m_PolygonsManager->GetRGBImage());
    lGenerator->GenerateLayer();


    typedef RenderingFunctionType::ParametersType ParametersType;
    unsigned int nbComponents = 3; //To be displayed
    ParametersType parameters(2*nbComponents);
    for(unsigned int i = 0; i<nbComponents; ++i)
    {
      parameters[i]=0;
      ++i;
      parameters[i]=255;
    }
    lGenerator->GetLayer()->GetRenderingFunction()->SetParameters(parameters);
//     VectorPixelType min(3), max(3);
//     min.Fill(0);
//     max.Fill(255);


//     lGenerator->GetLayer()->GetRenderingFunction()->SetMinimum(min);
//     lGenerator->GetLayer()->GetRenderingFunction()->SetMaximum(max);


    //FIXME define a generic method SetParameter with a map for rendering function for example
    // to avoid these ugly cast...
    typedef Function::StandardRenderingFunction<VectorLayerType::PixelType, VectorLayerType::OutputPixelType> StandardRenderingFunctionUcharType;
    static_cast<StandardRenderingFunctionUcharType *>(lGenerator->GetLayer()->GetRenderingFunction())->SetAutoMinMax( false );
//     lGenerator->GetLayer()->SetAutoMinMax( false );


    /** Display the image result */
    unsigned int size = m_FocusingModel->GetWorkingModel()->GetNumberOfLayers();
    if(size>1)
    {
      for(unsigned int j=1; j<size; j++)
      {
        m_FocusingModel->GetWorkingModel()->DeleteLayer(j);
      }
    }

    typedef Function::BlendingFunction<RGBPixelType>      BlendingFunctionType;
    typedef BlendingFunctionType::Pointer                 BlendingFunctionPointerType;
    typedef Function::UniformAlphaBlendingFunction<RGBPixelType>  UniformAlphaBlendingFunctionType;
    typedef UniformAlphaBlendingFunctionType::Pointer             UniformAlphaBlendingFunctionPointerType;

    BlendingFunctionPointerType blendfunc = lGenerator->GetLayer()->GetBlendingFunction();

    dynamic_cast<UniformAlphaBlendingFunctionType *>(static_cast<BlendingFunctionType *>((blendfunc)))->SetAlpha(m_Alpha);

    m_FocusingModel->GetWorkingModel()->AddLayer(lGenerator->GetLayer());
    m_FocusingModel->GetWorkingModel()->Update();
  }
}

void
UrbanAreaExtractionModel
::SetAlpha(double alpha)
{
  m_Alpha = alpha;
   if ( m_FocusingModel->GetWorkingModel()->GetHasExtract() )
   {
    this->DisplayNewLayer();
   }
}

UrbanAreaExtractionModel::PolygonListPointerType
UrbanAreaExtractionModel::GetPolygonsList()
{
  if(m_PolygonsManager.IsNotNull())
  {
    return m_PolygonsManager->GetPolygonsList();
  }
  return PolygonListType::New();
}

UrbanAreaExtractionModel::ColoredPolyListPointerType
UrbanAreaExtractionModel::GetColoredPolyList()
{
  if(m_PolygonsManager.IsNotNull())
  {
    return m_PolygonsManager->GetColoredPolyList();
  }
  return ColoredPolyListType::New();
}


void
UrbanAreaExtractionModel
:: SaveImage(const char * filename)
{
  typedef StreamingImageFileWriter<VectorImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetInput(m_PolygonsManager->GetRGBImage());
  writer->SetFileName(filename);

  // Watcher
  FltkFilterWatcher qlwatcher(writer, 0, 0, 250, 20,"Saving image...");
  writer->Update();
}

void
UrbanAreaExtractionModel
::SaveImageOnFull(const char * filename)
{
  m_WhichImage = FULL_IMAGE;
  this->FindUrbanAreas();
  this->SaveImage(filename);
}


void
UrbanAreaExtractionModel
::InitVector()
{
  m_FullVector = VectorDataType::New();
  DataNodeType::Pointer document     = DataNodeType::New();
  DataNodeType::Pointer root         = m_FullVector->GetDataTree()->GetRoot()->Get();
  document->SetNodeType(otb::DOCUMENT);
  document->SetNodeId("DOCUMENT");
  m_FullVector->GetDataTree()->Add(document, root);
  m_Folder = DataNodeType::New();
  m_Folder->SetNodeType(otb::FOLDER);
  m_FullVector->GetDataTree()->Add(m_Folder, document);
}

void
UrbanAreaExtractionModel
::BuildVector()
{
  // concatenate the nodes to the FullVector
  for(unsigned int i=0; i<m_PolygonsManager->GetVectorDataList()->Size(); i++)
  {
    VectorDataType::Pointer vectDat = m_PolygonsManager->GetVectorDataList()->GetNthElement(i);
    itk::PreOrderTreeIterator<DataTreeType> it(vectDat->GetDataTree(), vectDat->GetDataTree()->GetRoot());
    it.GoToBegin();
    while( !it.IsAtEnd() )
    {
      if( it.Get()->GetNodeType() != otb::ROOT )
      {
        if(it.GetParent()->Get()->GetNodeType() == otb::FOLDER)
        {
          m_FullVector->GetDataTree()->Add(it.Get(), m_Folder);
        }
      }
      ++it;
    }
  }
}

void
UrbanAreaExtractionModel
::SaveVector(const char * filename)
{
  std::string outName = filename;
  int length = outName.size()-(outName.find_last_of('.')+1);

  std::string ext = outName.substr(outName.find_last_of('.')+1, length);

  if (ext.compare("shp")==0 && ext.compare("kml")==0 )
  {
    itkExceptionMacro(<<ext<<" : Not recognized file vector format. Must be shp or kml");
  }

  VectorDataWriterType::Pointer writer = VectorDataWriterType::New();
  writer->SetInput( m_FullVector );
  writer->SetFileName(outName);

  // Watcher
  FltkFilterWatcher qlwatcher(writer, 0, 0, 250, 20,"Saving vector data...");
  writer->Update();
}

void
UrbanAreaExtractionModel
::SaveVectorOnExtract(const char * filename)
{
  this->InitVector();
  this->BuildVector();
  this->SaveVector(filename);
}

void
UrbanAreaExtractionModel
::SaveVectorOnFull(const char * filename)
{

  m_WhichImage = FULL_IMAGE;

  /** Divide the input image into pieces */
  RegionType lLargestRegion = m_Reader->GetOutput()->GetLargestPossibleRegion();

  // Setting up local streaming capabilities
  SplitterPointerType lSplitter = SplitterType::New();
  unsigned int lNumberOfStreamDivisions = StreamingTraitsType::CalculateNumberOfStreamDivisions(m_Reader->GetOutput(),
                                          lLargestRegion,
                                          lSplitter,
                                          otb::SET_AUTOMATIC_NUMBER_OF_STREAM_DIVISIONS,
                                          0, 0, 0);

  // local streaming variables
  unsigned int piece = 0;
  for(piece=0; piece<lNumberOfStreamDivisions; piece++)
  {
    m_StreamingRegion = lSplitter->GetSplit(piece, lNumberOfStreamDivisions, lLargestRegion);
    this->FindUrbanAreas();
    this->BuildVector();
  }
  this->SaveVector(filename);
}

}

