/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbUrbanAreaPolygonsManager_h
#define __otbUrbanAreaPolygonsManager_h


#include "itkProcessObject.h"
#include "otbColoredPoly.h"
#include "otbObjectList.h"
#include "itkFixedArray.h"
#include "otbImage.h"

#include "otbChangeLabelImageFilter.h"
#include "itkConnectedComponentImageFilter.h"
#include "itkRelabelComponentImageFilter.h"
#include "otbPersistentVectorizationImageFilter.h"

#include "otbVectorData.h"
#include "otbVectorDataFileWriter.h"
#include "otbVectorDataProjectionFilter.h"

#include "otbUrbanAreaExtractionModel.h"

namespace otb
{

/** \class UrbanAreaPolygonsManager
 *
 *
 *
 */
template < class TInputImage >
class ITK_EXPORT UrbanAreaPolygonsManager
      : public itk::ProcessObject
{

public:

  /** Standard class typedefs */
  typedef UrbanAreaPolygonsManager<TInputImage>          Self;
  typedef itk::ProcessObject                             Superclass;
  typedef itk::SmartPointer<Self>                        Pointer;
  typedef itk::SmartPointer<const Self>                  ConstPointer;

  /** Standard type macro */
  itkTypeMacro(UrbanAreaPolygonsManager, itk::ProcessObject);

  itkNewMacro(Self);

  /** Images typedefs*/
  typedef unsigned int                                   LabeledPixelType;
  typedef Image<LabeledPixelType, 2>                     LabeledImageType;
  typedef typename LabeledImageType::Pointer             LabeledImagePointerType;

  typedef unsigned char                                  VectorPixelType;
  typedef VectorImage<VectorPixelType, 2>                 VectorImageType;
  typedef typename VectorImageType::Pointer              VectorImagePointerType;
  

  /** Polygons typedef */
  typedef Polygon<double>                                PolygonType;
  typedef typename PolygonType::Pointer                  PolygonPointerType;
  typedef PolygonType::ContinuousIndexType               PolygonIndexType;
  typedef ObjectList<PolygonType>                        PolygonListType;
  typedef typename PolygonListType::Pointer              PolygonListPointerType;

  /** Filters typedef */
  typedef itk::ConnectedComponentImageFilter<LabeledImageType, LabeledImageType> ConnectedFilterType;
  typedef typename ConnectedFilterType::Pointer                                 ConnectedFilterPointerType;
  typedef itk::RelabelComponentImageFilter<LabeledImageType, LabeledImageType>   RelabelFilterType;
  typedef typename RelabelFilterType::Pointer                                   RelabelFilterPointerType;
  typedef PersistentVectorizationImageFilter<LabeledImageType, PolygonType>      PersistentVectorizationFilterType;
  typedef typename PersistentVectorizationFilterType::Pointer                   PersistentVectorizationFilterPointerType;
  typedef ChangeLabelImageFilter<LabeledImageType, VectorImageType>              ChangeLabelImageFilterType;
  typedef typename ChangeLabelImageFilterType::Pointer                          ChangeLabelImageFilterPointerType;


  /** Vector Data */
  typedef VectorData<double, 2>                             VectorDataType;
  typedef typename VectorDataType::Pointer                 VectorDataPointerType;
  typedef VectorDataType::DataNodeType                     DataNodeType;
  typedef VectorDataType::DataTreeType                     DataTreeType;
  typedef DataNodeType::PolygonType                        DataNodePolygonType;
  typedef PolygonType::VertexType                          DataNodeVertexType;
  typedef VectorDataFileWriter<VectorDataType>             VectorDataWriterType;
  typedef VectorDataWriterType::Pointer                    VectorDataWriterPointerType;
  typedef ObjectList<VectorDataType>                       VectorDataListType;
  typedef typename VectorDataListType::Pointer             VectorDataListPointerType;

  typedef TInputImage                                      ReaderImageType;
  typedef typename ReaderImageType::Pointer                ReaderImagePointerType;
  typedef typename ReaderImageType::PointType              PointType;
  typedef typename ReaderImageType::SpacingType            SpacingType;

  /** Colors, Lists and Map */
  typedef itk::FixedArray<double, 4>                        ColorType;
  typedef ColoredPoly<LabeledPixelType, ColorType>          ColoredPolyType;
  typedef typename ColoredPolyType::Pointer                ColoredPolyPointerType;
  typedef ObjectList<ColoredPolyType>                      ColoredPolyListType;
  typedef typename ColoredPolyListType::Pointer            ColoredPolyListPointerType;


  /** Getters and setters */
  itkSetObjectMacro(InputImage, LabeledImageType);
  itkGetObjectMacro(InputImage, LabeledImageType);
  itkGetObjectMacro(LabeledImage, LabeledImageType);
  itkGetObjectMacro(RGBImage, VectorImageType);
  itkGetObjectMacro(PolygonsList, PolygonListType);
  itkGetObjectMacro(ColoredPolyList, ColoredPolyListType);
  itkSetObjectMacro(ReaderImage, ReaderImageType);
  itkGetObjectMacro(VectorDataList, VectorDataListType);

  itkSetMacro(CurrentPolygonIndex, unsigned int);

  void InitManager();
  void GeneratePolygons(unsigned int pMinRegionSize, unsigned int pSubSampleLevel);
  void GenerateColorImage();
  void GenerateVectorData();
  void ClearAllPolygons();
  void DeletePolygon(unsigned int index);
  void DeleteColoredPoly(unsigned int index);

protected:
  /** Constructor */
  UrbanAreaPolygonsManager();
  /** Destructor */
  ~UrbanAreaPolygonsManager(){};


private:
  UrbanAreaPolygonsManager(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented


  // Data Lists
  /** Store extract output polygon */
  PolygonListPointerType m_PolygonsList;
  /** Store the color list */
  ColoredPolyListPointerType m_ColoredPolyList;
  /** Store the list of polygon translated in VD*/
  VectorDataListPointerType m_VectorDataList;

  // Filters
  /** Vectorization filter type */
  PersistentVectorizationFilterPointerType m_PersistentVectorizationFilter;
  /** Connected filter */
  ConnectedFilterPointerType m_ConnectedFilter;
  /** Relabel filter */
  RelabelFilterPointerType m_RelabelFilter;
  /** Reader */
  ReaderImagePointerType m_ReaderImage;
  /** Change Labvel Filter */
  ChangeLabelImageFilterPointerType m_ChangeLabelImageFilter;


  // Images
  /** Input Image */
  LabeledImagePointerType m_InputImage;
  LabeledImagePointerType m_LabeledImage;
  VectorImagePointerType m_RGBImage;

  unsigned int m_CurrentPolygonIndex;

};

} // end namespace otb

#ifndef  OTB_MANUAL_INSTANTIATION
#include "otbUrbanAreaPolygonsManager.txx"
#endif

#endif

