/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/
#ifndef __otbbUrbanAreaPolygonsManager_txx
#define __otbbUrbanAreaPolygonsManager_txx

#include "otbUrbanAreaPolygonsManager.h"
#include "itkPreOrderTreeIterator.h"


namespace otb
{


template < class TInputImage >
UrbanAreaPolygonsManager<TInputImage>
::UrbanAreaPolygonsManager()
{

  // Data Lists
  m_PolygonsList = PolygonListType::New();
  m_ColoredPolyList = ColoredPolyListType::New();
  m_VectorDataList = VectorDataListType::New();

  this->InitManager();
}

template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::InitManager()
{

  // Images
  m_InputImage = LabeledImageType::New();
  m_LabeledImage = LabeledImageType::New();
  m_RGBImage = NULL;

  // Filters
  m_ConnectedFilter = ConnectedFilterType::New();
  m_RelabelFilter = RelabelFilterType::New();
  m_ChangeLabelImageFilter = ChangeLabelImageFilterType::New();
  m_PersistentVectorizationFilter = PersistentVectorizationFilterType::New();
}

/**
 * Take a binary input image and convert it into colored and labeled image.
 * Then compute the vector data using labeled image
 */
template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::GeneratePolygons(unsigned int pMinRegionSize, unsigned int pSubSampleLevel)
{
  if(m_InputImage.IsNotNull())
  {

    m_ConnectedFilter = ConnectedFilterType::New();
    m_RelabelFilter = RelabelFilterType::New();
    m_PersistentVectorizationFilter = PersistentVectorizationFilterType::New();

    // Spacing / Pixel surface
    SpacingType lSpacing = m_ReaderImage->GetSpacing();
    double lPixelSurface = vcl_abs( (lSpacing[0]*lSpacing[1]) * static_cast<double>(pSubSampleLevel*pSubSampleLevel) );
    unsigned int lNbPixels = static_cast<unsigned int>(pMinRegionSize / lPixelSurface);

    // Pipeline
    m_ConnectedFilter->SetInput(m_InputImage);
    m_RelabelFilter->SetInput(m_ConnectedFilter->GetOutput());
    m_RelabelFilter->SetMinimumObjectSize(lNbPixels);

    // Turn the picture into vectors
    m_PersistentVectorizationFilter->SetInput(m_RelabelFilter->GetOutput());
    m_PersistentVectorizationFilter->Reset();
    m_LabeledImage = m_RelabelFilter->GetOutput();
    m_PersistentVectorizationFilter->Update();

    // We do not clear the lists in order to add the new results
    PolygonListPointerType lTempPolygonsList = PolygonListType::New();
    lTempPolygonsList = m_PersistentVectorizationFilter->GetPathList();

    LabeledPixelType lId = 0;
    ColorType lColor;

    // Add each new element to the polygons list.
    for (PolygonListType::Iterator it = lTempPolygonsList->Begin(); it!=lTempPolygonsList->End(); ++it)
    {
      // We reject the void-considered region
      if(it.Get()->GetArea() > 1e-6 )
      {
        m_PolygonsList->PushBack(it.Get());

        lId ++;
        lColor[0]=rand()/(RAND_MAX+1.0);
        lColor[1]=rand()/(RAND_MAX+1.0);
        lColor[2]=rand()/(RAND_MAX+1.0);
        lColor[3]=1.0;

        ColoredPolyPointerType lColored = ColoredPolyType::New();
        lColored->SetId(lId);
        lColored->SetColor(lColor);

        m_ColoredPolyList->PushBack(lColored);
      }
    }

    // Make the colored areas
    this->GenerateColorImage();
    // Make the associated vector data
    this->GenerateVectorData();

  }
}

/**
 * Uses the labeled image to transform it into colored image
 */
template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::GenerateColorImage()
{
  if(m_LabeledImage.IsNotNull())
  {

    ChangeLabelImageFilterPointerType m_ChangeLabelImageFilter = ChangeLabelImageFilterType::New();

    m_ChangeLabelImageFilter->SetInput(m_LabeledImage);
    m_ChangeLabelImageFilter->SetNumberOfComponentsPerPixel(3);

    typename VectorImageType::PixelType lRGBPixel;
    lRGBPixel.SetSize(3);
    // For the black
    lRGBPixel.Fill(0);
    m_ChangeLabelImageFilter->SetChange(0, lRGBPixel);

    LabeledPixelType lLabeledPixel;

    for( ColoredPolyListType::Iterator it = m_ColoredPolyList->Begin(); it!=m_ColoredPolyList->End(); ++it)
    {
      lLabeledPixel =it.Get()->GetId();
      ColorType lColor =it.Get()->GetColor();

      lRGBPixel[0]= static_cast<unsigned char>(255*lColor[0]);
      lRGBPixel[1]= static_cast<unsigned char>(255*lColor[1]);
      lRGBPixel[2]= static_cast<unsigned char>(255*lColor[2]);

      m_ChangeLabelImageFilter->SetChange(lLabeledPixel, lRGBPixel);
    }
    m_ChangeLabelImageFilter->Update();
    m_RGBImage = m_ChangeLabelImageFilter->GetOutput();
  }
}

/**
 * Uses the polygons list to make the vector data
 */
template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::GenerateVectorData()
{

  // Build one vector data which contains all the polygons
  VectorDataType::Pointer  vectorData  = VectorDataType::New();

  DataNodeType::Pointer document = DataNodeType::New();
  DataNodeType::Pointer root     = vectorData->GetDataTree()->GetRoot()->Get();
  document->SetNodeType(otb::DOCUMENT);
  document->SetNodeId("DOCUMENT");
  vectorData->GetDataTree()->Add(document, root);

  DataNodeType::Pointer folder = DataNodeType::New();
  folder->SetNodeType(otb::FOLDER);
  vectorData->GetDataTree()->Add(folder, document);

  for (unsigned int i=0; i<m_PolygonsList->Size(); i++)
  {
    //PolygonType::Pointer p = PolygonType::New();
    DataNodeType::Pointer poly   = DataNodeType::New();
    poly->SetNodeType(otb::FEATURE_POLYGON);
    itk::OStringStream oss;
    oss.str("");
    oss<<i+1;
    poly->SetNodeId(oss.str().c_str());
    poly->SetPolygonExteriorRing( m_PolygonsList->GetNthElement(i) );
    vectorData->GetDataTree()->Add(poly, folder);
  }
  
  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType> ProjectionFilterType;
  ProjectionFilterType::Pointer vectorDataProjection = ProjectionFilterType::New();
  vectorDataProjection->SetInput(vectorData);

  PointType lNewOrigin;
  // polygons are recorded with a 0.5 shift...
  lNewOrigin[0] = m_InputImage->GetOrigin()[0]+0.5;
  lNewOrigin[1] = m_InputImage->GetOrigin()[1]+0.5;

  vectorDataProjection->SetInputOrigin(lNewOrigin);
  vectorDataProjection->SetInputSpacing(m_ReaderImage->GetSpacing());

  std::string projectionRef;
  itk::ExposeMetaData<std::string>(m_ReaderImage->GetMetaDataDictionary(),
                                   MetaDataKey::ProjectionRefKey, projectionRef );
  vectorDataProjection->SetInputProjectionRef(projectionRef);
  vectorDataProjection->SetInputKeywordList(m_ReaderImage->GetImageKeywordlist());
  vectorDataProjection->Update();

  // Add the new polygon to the vector data list
  itk::PreOrderTreeIterator<DataTreeType> it(vectorDataProjection->GetOutput()->GetDataTree(), vectorDataProjection->GetOutput()->GetDataTree()->GetRoot());
  it.GoToBegin();
  unsigned int count = m_VectorDataList->Size();
  unsigned int countCur = 1;
  
  while( !it.IsAtEnd() )
    {
      // start at the end of the last recorded VD
      if( countCur > count )
       {
         if( it.Get()->GetNodeType() != otb::ROOT )
           {
             if(it.GetParent()->Get()->GetNodeType() == otb::FOLDER)
              {
                VectorDataType::Pointer  vectorDataTmp  = VectorDataType::New();
                DataNodeType::Pointer documentTmp = DataNodeType::New();
                DataNodeType::Pointer rootTmp     = vectorDataTmp->GetDataTree()->GetRoot()->Get();
                documentTmp->SetNodeType(otb::DOCUMENT);
                documentTmp->SetNodeId("DOCUMENT");
                vectorDataTmp->GetDataTree()->Add(documentTmp, rootTmp);
                DataNodeType::Pointer folderTmp = DataNodeType::New();
                folderTmp->SetNodeType(otb::FOLDER);
                vectorDataTmp->GetDataTree()->Add(folderTmp, documentTmp);
                vectorDataTmp->GetDataTree()->Add(it.Get(), folderTmp);
                
                m_VectorDataList->PushBack(vectorDataTmp);
              }
           }
       }
      ++it;
      ++countCur;
    }

}

template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::ClearAllPolygons()
{
  m_PolygonsList = PolygonListType::New();
  m_ColoredPolyList = ColoredPolyListType::New();
  m_VectorDataList = VectorDataListType::New();
  m_RGBImage = NULL;
}

template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::DeletePolygon(unsigned int index)
{
  // Suppress one color into the colored list
  this->DeleteColoredPoly(index);
  // Generate the Output image and the selected area will appear in black
  this->GenerateColorImage();
  m_VectorDataList->Erase(index);
  m_PolygonsList->Erase(index);
}


template <class TInputImage>
void
UrbanAreaPolygonsManager<TInputImage>
::DeleteColoredPoly(unsigned int index)
{
  // Turn the color into black
  ColorType lColor;
  lColor.Fill(0);
  m_ColoredPolyList->GetNthElement(index)->SetColor(lColor);

  // Then get the selected polygon to paste it at the end of the list
  m_ColoredPolyList->PushBack(m_ColoredPolyList->GetNthElement(index));

  // And finally erase the selected element
  m_ColoredPolyList->Erase(index);
}


}
#endif


