/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
See OTBCopyright.txt for details.


    This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include "otbUrbanAreaExtractionViewGUI.h"
#include <FL/Fl_File_Chooser.H>
#include <FL/Fl_Color_Chooser.H>

#include "otbMacro.h"
#include "otbMsgReporter.h"

namespace otb
{

UrbanAreaExtractionViewGUI::UrbanAreaExtractionViewGUI()
{
  m_UrbanAreaExtractionModel = UrbanAreaExtractionModel::GetInstance();
  m_UrbanAreaExtractionModel->RegisterListener(this);
  m_FocusingView = FocusingViewType::New();

  m_FocusingView->SetFocusingModel( m_UrbanAreaExtractionModel->GetFocusingModel() );
  m_LastPath = "";
  m_MasterViewSelected = true;
  // Init with blue
  m_ClassColor.Fill(0);
  m_ClassColor[2]=1.;
  m_ClassColor[3]=1;

  this->Build();
  bDisplay->value(0);
  wMainWindow->show();
}

void
UrbanAreaExtractionViewGUI
::InitWidgets()
{
  assert(m_UrbanAreaExtractionController.IsNotNull() && "The controller is not created");

  // MasterScroll
  gMasterScroll->add(m_FocusingView->GetMasterScrollWidget());
  gMasterScroll->resizable(m_FocusingView->GetMasterScrollWidget());
  gMasterScroll->box(FL_NO_BOX);
  m_FocusingView->GetMasterScrollWidget()->show();
  m_FocusingView->GetMasterScrollWidget()->resize( gMasterScroll->x(), gMasterScroll->y(), gMasterScroll->w(), gMasterScroll->h() );

  // WorkingScroll
  gWorkingScroll->add(m_FocusingView->GetWorkingScrollWidget());
  gWorkingScroll->resizable(m_FocusingView->GetWorkingScrollWidget());
  gWorkingScroll->box(FL_NO_BOX);
  m_FocusingView->GetWorkingScrollWidget()->show();
  m_FocusingView->GetWorkingScrollWidget()->resize( gWorkingScroll->x(), gWorkingScroll->y(), gWorkingScroll->w(), gWorkingScroll->h() );

  // WorkingFull
  gWorkingFull->add(m_FocusingView->GetWorkingFullWidget());
  gWorkingFull->resizable(m_FocusingView->GetWorkingFullWidget());
  gWorkingFull->box(FL_NO_BOX);
  m_FocusingView->GetWorkingFullWidget()->hide();
  m_FocusingView->GetWorkingFullWidget()->resize( gWorkingFull->x(), gWorkingFull->y(), gWorkingFull->w(), gWorkingFull->h() );

  gMasterScroll->show();
  gWorkingScroll->show();
  gWorkingFull->hide();
}

void UrbanAreaExtractionViewGUI::InitConfigIndices()
{
  if (m_UrbanAreaExtractionModel->GetInputImage() != NULL )
  {
    unsigned int nbComponent = m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
    // Initialization of the widget
    if (nbComponent > 3)
    {
      iRedChannelChoice->value("3");
      iGreenChannelChoice->value("2");
      iNIRChannelChoice->value("4");
      return;
    }
  }
  iRedChannelChoice->value("1");
  iGreenChannelChoice->value("2");
  iNIRChannelChoice->value("3");
}

void
UrbanAreaExtractionViewGUI
::Show()
{
  wMainWindow->show();
}

void UrbanAreaExtractionViewGUI::SetUrbanAreaExtractionController(UrbanAreaExtractionControllerInterface * pController)
{
  m_UrbanAreaExtractionController = pController;
  m_FocusingView->SetFocusingController(pController->GetFocusingController());
}

void UrbanAreaExtractionViewGUI::SetUrbanAreaExtractionModel(UrbanAreaExtractionModel * pModel)
{
  m_UrbanAreaExtractionModel = pModel;
}


void UrbanAreaExtractionViewGUI::Notify()
{
  if( m_UrbanAreaExtractionModel->GetHasNewInput()==true )
    {
      gMasterScroll->remove(m_FocusingView->GetMasterScrollWidget());
      gWorkingScroll->remove(m_FocusingView->GetWorkingScrollWidget());
      gWorkingFull->remove(m_FocusingView->GetWorkingFullWidget());
      this->InitWidgets();
      this->InitConfigIndices();
      this->IndicesOkCallback();
      this->UpdateInformation();
      this->UpdateViewerSetup();
    }

  // Notify Display button
  bDisplay->value(m_UrbanAreaExtractionModel->GetVectorDisplayed());

  // Clear the displayed list
  dROIList->clear();

  // Update polygons List
  PolygonListPointerType lPolygonList = m_UrbanAreaExtractionModel->GetPolygonsList();

  itk::OStringStream oss;
  unsigned int ROICounter = 0;

  for (PolygonListType::Iterator it = lPolygonList->Begin(); it!=lPolygonList->End(); ++it)
  {
      ++ROICounter;
      oss.str("");
      oss<<"ROI "<<ROICounter<<", "<<it.Get()->GetArea()<<" pixels";
      dROIList->add(oss.str().c_str());
  }

  if( ROICounter > 0 )
  {
    bFocus->activate();
    bEraseLastROI->activate();
    bClearROIs->activate();
    bChangeROIColor->activate();
  }

  if(m_UrbanAreaExtractionModel->GetState() == IMAGE_OPENED)
  {
    bRun->activate();
    bSaveImageResult->activate();
    bSaveVectorResult->activate();
    bSaveVectorOnFull->activate();
  }
}

void UrbanAreaExtractionViewGUI::OpenImageCallback()
{
  const char * cfname = fl_file_chooser("Pick an image file", "*.*", m_LastPath.c_str());
  Fl::check();
  wMainWindow->redraw();
  if (cfname == NULL || strlen(cfname)<1)
  {
    return;
  }

  m_LastPath = itksys::SystemTools::GetFilenamePath(cfname);
  m_UrbanAreaExtractionController->OpenInputImage(cfname);
}

void UrbanAreaExtractionViewGUI::OpenVectorCallback()
{
  const char * cfname = fl_file_chooser("Open Vector Data...", "*.{shp, kml}", m_LastPath.c_str());
  Fl::check();
  wMainWindow->redraw();
  if (cfname == NULL || strlen(cfname)<1)
  {
    return;
  }
  m_UrbanAreaExtractionController->OpenInputVector(cfname);
  bDisplay->activate();
}

void UrbanAreaExtractionViewGUI::SaveImageOnExtractCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output image file", "*",".");

  if (filename == NULL)
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_UrbanAreaExtractionController->SetVectorDisplayed(true);
  m_UrbanAreaExtractionController->SaveImageOnExtract(filename);
}

void UrbanAreaExtractionViewGUI::SaveVectorOnExtractCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output vector file", "*.shp\t*.kml",".");

  if (filename == NULL)
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_UrbanAreaExtractionController->SetVectorDisplayed(true);
  m_UrbanAreaExtractionController->SaveVectorOnExtract(filename);
}

void UrbanAreaExtractionViewGUI::SaveImageOnFullCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output image file", "*",".");

  if (filename == NULL)
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_UrbanAreaExtractionController->SetVectorDisplayed(true);
  m_UrbanAreaExtractionController->SaveImageOnFull(filename);
  bSaveImageOnExtract->activate();
  bSaveVectorOnExtract->activate();
  bDisplay->activate();
}

void UrbanAreaExtractionViewGUI::SaveVectorOnFullCallback()
{
  const char * filename = NULL;
  filename = fl_file_chooser("Select an output vector file", "*.shp\t*.kml",".");

  if (filename == NULL)
  {
    otbMsgDebugMacro(<<"Empty file name!");
    return;
  }
  m_UrbanAreaExtractionController->SetVectorDisplayed(true);
  m_UrbanAreaExtractionController->SaveVectorOnFull(filename);
  bSaveImageOnExtract->activate();
  bSaveVectorOnExtract->activate();
  bDisplay->activate();
}

void UrbanAreaExtractionViewGUI::QuitCallback()
{
  wMainWindow->hide();
  wIndicesConfig->hide();
  MsgReporter::GetInstance()->Hide();
}

void UrbanAreaExtractionViewGUI::MinDetailCallback()
{
  m_UrbanAreaExtractionController->SetMinDetail(static_cast<unsigned int>(vMinDetail->value()));
}

void UrbanAreaExtractionViewGUI::SetSubSampleLevelCallback()
{
  m_UrbanAreaExtractionController->SetSubSampleLevel(static_cast<unsigned int>(vMaxDetail->value()));
  if (vMaxDetail->value() == 1)
    vAlpha->activate();
  else
    vAlpha->deactivate();
}

void UrbanAreaExtractionViewGUI::ThresholdCallback()
{
  m_UrbanAreaExtractionController->SetThreshold(static_cast<double>(vThreshold->value()));
}

void UrbanAreaExtractionViewGUI::ThresholdDensityCallback()
{
  if(vThresholdDensity->value()<0)
    vThresholdDensity->value(0);
  m_UrbanAreaExtractionController->SetThresholdDensity(static_cast<double>(vThresholdDensity->value()));
}

void UrbanAreaExtractionViewGUI::SobelLowerThresholdCallback()
{
  m_UrbanAreaExtractionController->SetSobelLowerThreshold(static_cast<double>(vSobelLowerThreshold->value()));
}

void UrbanAreaExtractionViewGUI::SobelUpperThresholdCallback()
{
  m_UrbanAreaExtractionController->SetSobelUpperThreshold(static_cast<double>(vSobelUpperThreshold->value()));
}

void UrbanAreaExtractionViewGUI::PolygonSelectedCallback()
{
  if ( (dROIList->value()>0 ) &&
       (!m_Running) )
  {
    ColoredPolyListPointerType lColoredPolyList = m_UrbanAreaExtractionModel->GetColoredPolyList();
    ColoredPolyListType::Iterator it = lColoredPolyList->Begin();

    // Look for the good color in the list
    it = it+dROIList->value()-1;
    ColorType lColor =it.Get()->GetColor();

    fl_color(static_cast<unsigned char>(255*lColor[0]),
            static_cast<unsigned char>(255*lColor[1]),
            static_cast<unsigned char>(255*lColor[2]));
    dROIList->selection_color(fl_color());
    dROIList->redraw();

    dROIColor->color(fl_color());
    dROIColor->redraw();
  }
}

void UrbanAreaExtractionViewGUI::DeleteCallback()
{
  int index = dROIList->value()-1;

  // No region selected, return
  if (index < 0)
  {
    return;
  }
  m_UrbanAreaExtractionController->DeletePolygon(index);
  this->Notify();
}


void UrbanAreaExtractionViewGUI::ClearROIsCallback()
{
  m_UrbanAreaExtractionController->ClearAllPolygons();
  dROIList->clear();
}

void UrbanAreaExtractionViewGUI::FocusCallback()
{
  int index = dROIList->value()-1;
  // No region selected, return
  if (index < 0)
  {
    return;
  }

  // This should maybe be moved into the controller, but much more easier to implement here
  UrbanAreaExtractionModel::RegionType boundingRegion = m_UrbanAreaExtractionModel->GetPolygonsList()->GetNthElement(index)->GetBoundingRegion().GetImageRegion();
  UrbanAreaExtractionModel::IndexType upperLeft, downerRight;
  upperLeft[0] = boundingRegion.GetIndex()[0] + m_UrbanAreaExtractionModel->GetSavedIndex()[0];
  upperLeft[1] = boundingRegion.GetIndex()[1] + m_UrbanAreaExtractionModel->GetSavedIndex()[1];
  downerRight[0] = upperLeft[0] + boundingRegion.GetSize()[0];
  downerRight[1] = upperLeft[1] + boundingRegion.GetSize()[1];

  m_UrbanAreaExtractionModel->GetFocusingModel()->SetExtractRegionByIndex(upperLeft, downerRight);
  m_UrbanAreaExtractionModel->GetFocusingModel()->Update();
}

void UrbanAreaExtractionViewGUI::ChangeROIColorCallback()
{

  double r = (double)m_ClassColor[0];
  double g = (double)m_ClassColor[1];
  double b = (double)m_ClassColor[2];
  int ok = fl_color_chooser("Changed class color:", r, g, b);

  if (ok)
  {
    m_ClassColor[0]=(float)r;
    m_ClassColor[1]=(float)g;
    m_ClassColor[2]=(float)b;

    fl_color(static_cast<unsigned char>(255*m_ClassColor[0]),
        static_cast<unsigned char>(255*m_ClassColor[1]),
        static_cast<unsigned char>(255*m_ClassColor[2]));

    dROIColor->color(fl_color());
    dROIList->selection_color(fl_color());
    dROIList->redraw();
  }

  if (dROIList->value()>0)
  {
    ColoredPolyListPointerType lColoredPolyList = m_UrbanAreaExtractionModel->GetColoredPolyList();
    ColoredPolyListType::Iterator it = lColoredPolyList->Begin();
    // Look for the good color to change in the coloredPoly list
    unsigned int idNb = dROIList->value()-1;
    it = it+idNb;
    it.Get()->SetColor(m_ClassColor);

    if( idNb > m_UrbanAreaExtractionModel->GetLoadedVDIndex() )
      idNb--;
    m_UrbanAreaExtractionModel->GetFocusingModel()->ChangeVectorDataColor( idNb, m_ClassColor );
    m_UrbanAreaExtractionModel->GetPolygonsManager()->GenerateColorImage();
    m_UrbanAreaExtractionModel->DisplayNewLayer();
    m_UrbanAreaExtractionModel->NotifyAll();
  }
}

void UrbanAreaExtractionViewGUI::SwitchScrollCallback()
{

  if( m_MasterViewSelected == true )
    {
      m_FocusingView->GetMasterScrollWidget()->hide();
      gMasterScroll->hide();
      m_FocusingView->GetWorkingFullWidget()->show();
      gWorkingFull->show();
    }
  else
    {
      m_FocusingView->GetWorkingFullWidget()->hide();
      gWorkingFull->hide();
      m_FocusingView->GetMasterScrollWidget()->show();
      gMasterScroll->show();
    }
  m_MasterViewSelected = !m_MasterViewSelected;
}

void UrbanAreaExtractionViewGUI::RunCallback()
{
  if (!m_Running)
  {
    m_Running = true;
    m_UrbanAreaExtractionController->SetVectorDisplayed(true);
    m_UrbanAreaExtractionController->Run();
    bSaveImageOnExtract->activate();
    bSaveVectorOnExtract->activate();
    bDisplay->activate();
    m_Running = false;
  }
}

void UrbanAreaExtractionViewGUI::DisplayCallback()
{
  bool lVDDisplayed = !m_UrbanAreaExtractionModel->GetVectorDisplayed();
  m_UrbanAreaExtractionController->SetVectorDisplayed(lVDDisplayed);
  m_FocusingView->SetGLComponentsStatus(lVDDisplayed);
}

void UrbanAreaExtractionViewGUI::UpdateInformation()
{
  std::string imName = m_UrbanAreaExtractionModel->GetImageFilename();
  itk::OStringStream oss;
  oss.str("");
  oss<<"Urban Area Extraction Application";
  oss<<" : "<<imName.substr(imName.find_last_of("/")+1, imName.size());
  oss<<" ("<<m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
  if(m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel() != 1)
    oss<<" bands , ";
  else
    oss<<" band , ";

  oss<<m_UrbanAreaExtractionModel->GetInputImage()->GetLargestPossibleRegion().GetSize()<<")";
  wMainWindow->label(oss.str().c_str());
}

void UrbanAreaExtractionViewGUI::AlphaCallback()
{
  m_UrbanAreaExtractionController->SetAlpha(vAlpha->value());
}

void UrbanAreaExtractionViewGUI::ConfigIndicesCallback()
{

  unsigned int nbComponent = m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
  itk::OStringStream oss;
  oss.str("");

  //Clear all the choices
  iRedChannelChoice->clear();
  iGreenChannelChoice->clear();
  iNIRChannelChoice->clear();

  for (unsigned int i = 0; i<nbComponent; ++i)
  {
    oss.str("");
    oss<<i+1;
    iRedChannelChoice->add(oss.str().c_str());
    iGreenChannelChoice->add(oss.str().c_str());
    iNIRChannelChoice->add(oss.str().c_str());
  }

  wIndicesConfig->show();
  wIndicesConfig->redraw();
}

void UrbanAreaExtractionViewGUI::IndicesOkCallback()
{
  m_UrbanAreaExtractionController->SetIndices(::atoi(iGreenChannelChoice->value()), ::atoi(iRedChannelChoice->value()), ::atoi(iNIRChannelChoice->value()));
  wIndicesConfig->hide();
}

void UrbanAreaExtractionViewGUI::IndicesCancelCallback()
{
  wIndicesConfig->hide();
}

 /**
  * Update the viewer setup window
  */
void
UrbanAreaExtractionViewGUI::UpdateViewerSetup()
 {
   if (m_UrbanAreaExtractionModel->GetState() == NO_IMAGE)
   {
     // no image selected, return
     return;
   }

   unsigned int lNbComponent = m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
   itk::OStringStream oss;
   oss.str("");

   //Clear all the choices
   iGrayscaleChannelChoice->clear();
   iRChannelChoice->clear();
   iGChannelChoice->clear();
   iBChannelChoice->clear();

   for (unsigned int i = 0; i<lNbComponent; ++i)
   {
     oss.str("");
     oss<<i+1;
     iGrayscaleChannelChoice->add(oss.str().c_str());
     iRChannelChoice->add(oss.str().c_str());
     iGChannelChoice->add(oss.str().c_str());
     iBChannelChoice->add(oss.str().c_str());
   }
   wViewerSetupWindow->redraw();
   iGrayscaleChannelChoice->redraw();

   switch(lNbComponent){
   case 1 :
   case 2:
     this->GrayscaleSetCallback();
     break;
   default :
    this->RGBSetCallback();
     break;

   }
 }

 /**
  * RGBSet();
  */
 void
 UrbanAreaExtractionViewGUI
 ::RGBSetCallback()
 {
   rViewerSetupColorMode->set();
   rViewerSetupGrayscaleMode->clear();

   iGrayscaleChannelChoice->deactivate();
   iRChannelChoice->activate();
   iGChannelChoice->activate();
   iBChannelChoice->activate();

   if (m_UrbanAreaExtractionModel->GetState() == NO_IMAGE)
   {
     // no image selected, return
     return;
   }

   unsigned int lNbComponent = m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
   if( m_UrbanAreaExtractionModel->GetMasterModel()->GetLayer(0) != NULL )
   {
      StandardRenderingFunctionType::PixelRepresentationFunctionType::ChannelListType channels;
      channels = m_UrbanAreaExtractionModel->GetRenderingFunction()->GetChannelList();
      unsigned int i=0;
      while (channels.size() < 3)
      {
       channels.push_back(i);
       ++i;
      }

      iRChannelChoice->value(std::min(channels[0], lNbComponent-1));
      iGChannelChoice->value(std::min(channels[1], lNbComponent-1));
      iBChannelChoice->value(std::min(channels[2], lNbComponent-1));
   }
 }


/**
  * GrayscaleSet();
  */
 void
 UrbanAreaExtractionViewGUI::GrayscaleSetCallback()
 {
   rViewerSetupGrayscaleMode->set();
   rViewerSetupColorMode->clear();

   iRChannelChoice->deactivate();
   iGChannelChoice->deactivate();
   iBChannelChoice->deactivate();
   iGrayscaleChannelChoice->activate();

   if (m_UrbanAreaExtractionModel->GetState() == NO_IMAGE)
   {
     // no image selected, return
     return;
   }

   unsigned int lNbComponent = m_UrbanAreaExtractionModel->GetInputImage()->GetNumberOfComponentsPerPixel();
   if( m_UrbanAreaExtractionModel->GetMasterModel()->GetLayer(0) != NULL )
   {

      StandardRenderingFunctionType::PixelRepresentationFunctionType::ChannelListType channels;
      channels = m_UrbanAreaExtractionModel->GetRenderingFunction()->GetChannelList();
      if (channels.size() < 1)
      {
      channels.push_back(0);
      }
      iGrayscaleChannelChoice->value(std::min(channels[0], lNbComponent-1));
   }
 }


 /**
  * ViewerSetup();
  */
 void
 UrbanAreaExtractionViewGUI::ViewerSetupCallback()
 {
   wViewerSetupWindow->show();
 }

 /**
  * ViewerSetupOk();
  */
 void
 UrbanAreaExtractionViewGUI::ViewerSetupOkCallback()
 {

   if (m_UrbanAreaExtractionModel->GetState() != IMAGE_OPENED)
   {
     // no image selected, return
     wViewerSetupWindow->hide();
     return;
   }

   if (rViewerSetupColorMode->value())
     {
       m_UrbanAreaExtractionController->UpdateRGBChannelOrder(atoi(iRChannelChoice->value())-1,
                                                              atoi(iGChannelChoice->value())-1,
                                                              atoi(iBChannelChoice->value())-1);

     }
   else if (rViewerSetupGrayscaleMode->value())
     {
       m_UrbanAreaExtractionController->UpdateGrayscaleChannel(atoi(iGrayscaleChannelChoice->value())-1);
     }
   wViewerSetupWindow->show();
 }

 void
 UrbanAreaExtractionViewGUI::ViewerSetupCancelCallback()
 {
   wViewerSetupWindow->hide();
 }

}
