/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include <stdlib.h>
#include <iostream>
#include "itkExceptionObject.h"
#include "otbCommandLineArgumentParser.h"
#include "otbImageFileReader.h"
#include "otbImageViewer.h"
#include "otbMacro.h"

#include "otbI18n.h"

int main( int argc, char * argv[] )
{
  otbI18nMacro();
  try
  {
    Fl::gl_visual(FL_ALPHA);

    // Parse command line parameters
    typedef otb::CommandLineArgumentParser ParserType;
    ParserType::Pointer parser = ParserType::New();
    parser->AddInputImage();
    parser->AddOption("--RedChannelIndex","For multi-spectral images, red channel index. Default is 1. (If there are less than three channels, the viewer will display the red channel component only.)","-ri", 1, false);
    parser->AddOption("--GreenChannelIndex","For multi-spectral images, green channel index. Default is 2.","-gi", 1, false);
    parser->AddOption("--BlueChannelIndex","For multi-spectral images, blue channel index. Default is 3.","-bi", 1, false);
    parser->AddOption("--NormalizationFactor","Coefficient of sigma used for normalization. Default is 3.","-nf", 1, false);
    parser->AddOption("--QuicklookRatio","Adjust quicklook quality. Default is 2. Lower values will make the generation faster, but the quality poorer.","-qr", 1, false);
    parser->AddOption("--ComplexModulus","View the complex modulus of the image. Red channel index indicates the real part and green channel index the imaginary part.","-cm", 0, false);
    parser->AddOption("--ComplexPhase","View the complex phase of the image. Red channel index indicates the real part and green channel index the imaginary part.","-cp", 0, false);
    parser->AddOption("--Overlay","Add overlay to display","-ov", 1, false);
    typedef otb::CommandLineArgumentParseResult ParserResultType;
    ParserResultType::Pointer  parseResult = ParserResultType::New();
    try
    {
      parser->ParseCommandLine(argc, argv, parseResult);
    }
    catch ( itk::ExceptionObject & err )
    {
      std::string descriptionException = err.GetDescription();
      if (descriptionException.find("ParseCommandLine(): Help Parser") != std::string::npos)
      {
        return EXIT_SUCCESS;
      }
      if (descriptionException.find("ParseCommandLine(): Version Parser") != std::string::npos)
      {
        return EXIT_SUCCESS;
      }
      return EXIT_FAILURE;
    }

    typedef double PixelType;
    typedef otb::ImageViewer<PixelType>  ImageViewerType;
    typedef ImageViewerType::ImageType ImageType;
    typedef otb::ImageFileReader<ImageType> ReaderType;

    // instantiation
    ImageViewerType::Pointer viewer = ImageViewerType::New();

    // check for input images
    ReaderType::Pointer reader = ReaderType::New();
    reader->SetFileName(parseResult->GetInputImage().c_str());
    reader->UpdateOutputInformation();
    viewer->SetImage(reader->GetOutput());

    if (parseResult->IsOptionPresent("--RedChannelIndex"))
      viewer->SetRedChannelIndex(parseResult->GetParameterInt("--RedChannelIndex")-1);
    if (parseResult->IsOptionPresent("--GreenChannelIndex"))
      viewer->SetGreenChannelIndex(parseResult->GetParameterInt("--GreenChannelIndex")-1);
    if (parseResult->IsOptionPresent("--BlueChannelIndex"))
      viewer->SetBlueChannelIndex(parseResult->GetParameterInt("--BlueChannelIndex")-1);
    if (parseResult->IsOptionPresent("--NormalizationFactor"))
      viewer->SetNormalizationFactor(parseResult->GetParameterDouble("--NormalizationFactor"));
    if (parseResult->IsOptionPresent("--QuicklookRatio"))
      viewer->SetQuicklookRatioCoef(parseResult->GetParameterDouble("--QuicklookRatio"));

    viewer->SetLabel(parseResult->GetInputImage().c_str());

    typedef otb::ImageFileReader<ImageViewerType::OverlayImageType> OverlayReaderType;
    OverlayReaderType::Pointer overlayReader = OverlayReaderType::New();

    if (parseResult->IsOptionPresent("--Overlay"))
    {
      viewer->SetUseImageOverlay(true);
      std::string overlayfname = parseResult->GetParameterString("--Overlay", 0);
      overlayReader->SetFileName(overlayfname.c_str());
      overlayReader->UpdateOutputInformation();
      viewer->SetImageOverlay(overlayReader->GetOutput());
    }

    // build the app
    viewer->Build();

    if (parseResult->IsOptionPresent("--ComplexModulus"))
      viewer->SetViewModel(ImageViewerType::ScrollWidgetType::COMPLEX_MODULUS);
    if (parseResult->IsOptionPresent("--ComplexPhase"))
      viewer->SetViewModel(ImageViewerType::ScrollWidgetType::COMPLEX_PHASE);

    viewer->Show();

    // if not testing the app
    if (!parseResult->IsOptionPresent("--OTBTesting"))
    {
      otbGenericMsgDebugMacro(<<"Running GUI ...");
      return Fl::run();
    }
  }
  catch ( itk::ExceptionObject & err )
  {
    std::cout << "Following otbException catch :" << std::endl;
    std::cout << err << std::endl;
    return EXIT_FAILURE;
  }
  catch ( std::bad_alloc & err )
  {
    std::cout << "Exception bad_alloc : "<<(char*)err.what()<< std::endl;
    return EXIT_FAILURE;
  }
  catch ( ... )
  {
    std::cout << "Unknown Exception found !" << std::endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}


