/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbEstimateFeaturesStatistics.h"

#include <iostream>
#include "otbCommandLineArgumentParser.h"

#include "itkListSample.h"
#include "otbVectorImage.h"
#include "otbImage.h"
#include "otbImageList.h"
#include "otbImageFileReader.h"
#include "itkMeanCalculator.h"
#include "itkCovarianceCalculator.h"
#include "otbConcatenateSampleListFilter.h"
#include "otbVectorImageToImageListFilter.h"
#include "itkFixedArray.h"
#include "otbVectorImageToImageListFilter.h"

#include "otbStandardMetaImageFunctionBuilder.h"

#include "otbVectorData.h"
#include "otbVectorDataProjectionFilter.h"

#include "otbLabeledSampleLocalizationGenerator.h"
#include "otbDescriptorsListSampleGenerator.h"
#include "otbConcatenateSampleListFilter.h"
#include "otbImageToEnvelopeVectorDataFilter.h"

#include "itkMersenneTwisterRandomVariateGenerator.h"

#include "otbStatisticsXMLFileWriter.h"

namespace otb
{

int EstimateFeaturesStatistics::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("EstimateFeaturesStatistics");
  descriptor->SetDescription("Estimate mean/variance for each feature of the feature set.");
  descriptor->AddOptionNParams("InputImages", "A list of images from which to estimate features mean and variance",
                               "in", true, ApplicationDescriptor::InputImage);
  descriptor->AddOption("OutputFile", "A xml file containing mean and variance for each feature",
                        "out", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Radius", "Radius of the window for features (default is 10)",
                        "r", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("Density","Density of points in ppp (points per pixels) for statistics estimation (default is 0.0005)",
                        "d", 1, false, ApplicationDescriptor::Real);
  descriptor->AddOption("Seed","Set random process, for testing purpose only",
                        "s", 1, false, ApplicationDescriptor::Integer);
  return EXIT_SUCCESS;
}


int EstimateFeaturesStatistics::Execute(otb::ApplicationOptionsResult* parseResult)
{
  // Images
  typedef unsigned short                                  PixelType;
  typedef otb::VectorImage<PixelType, 2>                   VectorImageType;
  typedef otb::Image<PixelType, 2>                         ImageType;
  typedef otb::ImageFileReader<VectorImageType>           ReaderType;
  typedef otb::ImageList<ImageType>                       ImageListType;
  typedef otb::VectorImageToImageListFilter
    <VectorImageType, ImageListType>                      VI2ILFilterType;

  // Samples
  typedef double                                          ValueType;
  typedef itk::VariableLengthVector<ValueType>            MeasurementType;
  typedef itk::Statistics::ListSample<MeasurementType>    ListSampleType;
  typedef otb::Statistics::ConcatenateSampleListFilter
    <ListSampleType>                                      ConcatelateListSampleFilterType;
  typedef itk::Statistics::MeanCalculator<ListSampleType> MeanCalculatorType;
  typedef itk::Statistics::CovarianceCalculator
    <ListSampleType>                                      CovarianceCalculatorType;
  typedef itk::FixedArray<unsigned short, 1>               LabelType;
  typedef itk::Statistics::ListSample<LabelType>          LabeledListSampleType;

  typedef otb::VectorData<>                               VectorDataType;
  typedef otb::ImageToEnvelopeVectorDataFilter
    <VectorImageType, VectorDataType>                      EnvelopeFilterType;
  typedef otb::VectorDataProjectionFilter
    <VectorDataType, VectorDataType>                       VectorDataProjectionFilterType;
  typedef otb::LabeledSampleLocalizationGenerator
    <VectorDataType>                                      LocalizationGeneratorType;
  typedef otb::DescriptorsListSampleGenerator
    <ImageType, VectorDataType, ListSampleType,
    LabeledListSampleType, ValueType>                      SampleGeneratorType;
  
  // Functions
  typedef otb::StandardMetaImageFunctionBuilder<ValueType, double>
                                                          MetaImageFunctionBuilderType;
  typedef MetaImageFunctionBuilderType::MetaImageFunctionType
                                                          MetaImageFunctionType;
  typedef MetaImageFunctionBuilderType::DataObjectContainerType
                                                          DataObjectContainerType;
  
  // Build global ListSample
  ConcatelateListSampleFilterType::Pointer concatenate = ConcatelateListSampleFilterType::New();

  typedef itk::Statistics::MersenneTwisterRandomVariateGenerator RandomGeneratorType;

  // TODO: add textures here
  if(parseResult->IsOptionPresent("Seed"))
      {
      RandomGeneratorType::Pointer randomGenerator = RandomGeneratorType::New();
      randomGenerator->SetSeed(parseResult->GetParameterUInt("Seed"));
      }
  
  unsigned int radius = 10;
  if(parseResult->IsOptionPresent("Radius"))
    {
    radius = parseResult->GetParameterUInt("Radius");
    }

  double density = 0.0005;
  if(parseResult->IsOptionPresent("Density"))
    {
    density = parseResult->GetParameterDouble("Density");
    }

  unsigned int nbInputImages = parseResult->GetNumberOfParameters("InputImages");

  std::cout<<"Found "<<nbInputImages<<" images for statistics estimation"<<std::endl;

  for(unsigned int imageId = 0; imageId < nbInputImages; ++imageId)
    {
    std::string filename = parseResult->GetParameterString("InputImages", imageId);

    std::cout<<"Processing image "<<imageId<<": "<<filename<<std::endl;

    ReaderType::Pointer reader = ReaderType::New();
    reader->SetFileName(filename);
    reader->GenerateOutputInformation();

    // First, generate input image envelope
    EnvelopeFilterType::Pointer envelope = EnvelopeFilterType::New();
    envelope->SetInput(reader->GetOutput());

    VectorDataProjectionFilterType::Pointer vectorDataProjectionFilter = VectorDataProjectionFilterType::New();
    vectorDataProjectionFilter->SetInput(envelope->GetOutput());
    vectorDataProjectionFilter->SetOutputKeywordList(reader->GetOutput()->GetImageKeywordlist());
    vectorDataProjectionFilter->SetOutputProjectionRef(reader->GetOutput()->GetProjectionRef());

    // Then, generate random localization inside the envelope
    LocalizationGeneratorType::Pointer localizationGenerator = LocalizationGeneratorType::New();
    localizationGenerator->PushBackInput(vectorDataProjectionFilter->GetOutput());
    localizationGenerator->SetRandomLocalizationDensity(density);

    // Next, generate the samples
    SampleGeneratorType::Pointer samplesGenerator = SampleGeneratorType::New();
    samplesGenerator->SetNeighborhoodRadius(radius);
    samplesGenerator->SetSamplesLocations(localizationGenerator->GetOutput());
   
    // Splitting image into channels
    VI2ILFilterType::Pointer vi2ilFilter = VI2ILFilterType::New();
    vi2ilFilter->SetInput(reader->GetOutput());
    vi2ilFilter->GenerateOutputInformation();

    MetaImageFunctionBuilderType::Pointer builder = MetaImageFunctionBuilderType::New();
    builder->SetNeighborhoodRadius(radius);

    samplesGenerator->SetDescriptorsFunction(builder->GetMetaImageFunction());

    ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();
    samplesGenerator->SetInputImage(ilIt.Get());
    while(ilIt != vi2ilFilter->GetOutput()->End())
      {
      builder->AddImage(ilIt.Get().GetPointer());
      ++ilIt;
      }
    DataObjectContainerType imageContainer = builder->GetDataObjectContainer();

    samplesGenerator->Update();

    concatenate->AddInput(samplesGenerator->GetListSample());
    }
  
  concatenate->Update();

  std::cout<<"Mean and variance estimated from "<<concatenate->GetOutputSampleList()->Size()<<" samples."<<std::endl;

  CovarianceCalculatorType::Pointer covarianceCalculator = CovarianceCalculatorType::New();

  // Compute Covariance
  covarianceCalculator->SetInputSample(concatenate->GetOutputSampleList());
  covarianceCalculator->Update();
  const CovarianceCalculatorType::OutputType * cm = covarianceCalculator->GetOutput();
  
  // Build a Measurement Vector of mean
  MeasurementType mean;
  mean.SetSize(cm->Rows());
  for(unsigned int i = 0; i < (unsigned int)concatenate->GetOutputSampleList()->GetMeasurementVectorSize(); ++i)
    {
    mean.SetElement(i, (*covarianceCalculator->GetMean())[i]);
    }
  
  // Build a MeasurementVector of variance
  MeasurementType  stddev;
  stddev.SetSize(cm->Rows());
  for(unsigned int i = 0; i < cm->Rows(); ++i)
    {
    stddev.SetElement(i, vcl_sqrt((*cm)(i, i)));
    }
  
  // Write the Statistics via the statistic writer
  typedef otb::StatisticsXMLFileWriter<MeasurementType>    StatisticsWriter;
  StatisticsWriter::Pointer writer = StatisticsWriter::New();
  writer->SetFileName(parseResult->GetParameterString("OutputFile").c_str());
  writer->AddInput("mean", mean);
  writer->AddInput("stddev", stddev);
  writer->Update();
  
  return EXIT_SUCCESS;
}

}
