/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbEvaluateHOGObjectDetectorResponse.h"


//Image
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageList.h"
#include "otbImageFileReader.h"
#include "otbVectorImageToImageListFilter.h"
#include "itkExtractImageFilter.h"
#include "otbVectorImageToIntensityImageFilter.h"

// ListSample
#include "itkListSample.h"
#include "itkVariableLengthVector.h"
#include "itkFixedArray.h"

// Functions to use
#include "otbHistogramOfOrientedGradientCovariantImageFunction.h"
#include "otbLocalHistogramImageFunction.h"
#include "itkGradientImageFilter.h"
#include "otbImageFunctionAdaptor.h"
#include "otbMetaImageFunction.h"

// Localisation estimator filter
#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataFileWriter.h"

// Descriptor generator
#include "otbObjectDetectionClassifier.h"

#include "otbStandardFilterWatcher.h"

#include "otbImageFileWriter.h"

namespace otb
{

int EvaluateHOGObjectDetectorResponse::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("EvaluateHOGObjectDetectorResponse");

  descriptor->SetDescription("Allows to evaluate the detector response around some user-defined location.");

  descriptor->AddInputImage();
  descriptor->AddOption("SVMModel","SVM model trained to perform the detection","svm", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Radius","Radius of the window for features (default is 10)","r", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("WindowRadius","Radius of the window of analysis (default is equal to radius)","wr", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("Position","Position where to do the test in the image (index)","p", 2, true, ApplicationDescriptor::Integer);
  descriptor->AddOption("OutputLabelMap","Detection map on the premises of the point","lm", 1, true, ApplicationDescriptor::OutputImage);
  descriptor->AddOption("OutputHyperplaneDistance","Image the the hyperplane distance on the premises of the point","hd", 1, true, ApplicationDescriptor::OutputImage);
  descriptor->AddOption("OutputImage","Image extract","ie", 1, true, ApplicationDescriptor::OutputImage);

  return EXIT_SUCCESS;
}

int EvaluateHOGObjectDetectorResponse::Execute(otb::ApplicationOptionsResult* parseResult)
{
  // Input Image
  typedef double                                          ValueType;
  typedef double                                          PixelType;
  typedef double                                          CoordRepType;
  typedef double                                          FunctionPrecisionType;
  typedef otb::VectorImage<PixelType, 2>                   VectorImageType;
  typedef otb::Image<PixelType, 2>                         ImageType;
  typedef otb::ImageFileReader<VectorImageType>           ReaderType;
  typedef otb::ImageList<ImageType>                       ImageListType;
  typedef itk::ExtractImageFilter
    <VectorImageType, VectorImageType>                   ExtractFilterType;
  typedef otb::VectorImageToIntensityImageFilter
    <VectorImageType, ImageType>                           IntensityFilterType;
  typedef itk::GradientImageFilter<ImageType>             GradientFilterType;
  typedef GradientFilterType::OutputImageType             GradientImageType;
  typedef otb::VectorImageToImageListFilter
    <VectorImageType, ImageListType>                      VI2ILFilterType;

  // VectorData
  typedef otb::VectorData<>                               VectorDataType;
  typedef otb::VectorDataFileReader<VectorDataType>       VectorDataReaderType;
  typedef otb::VectorDataProjectionFilter
    <VectorDataType, VectorDataType>                     VectorDataProjectionFilterType;
  typedef otb::VectorDataFileWriter
    <VectorDataType>                                    VectorDataWriterType;

  // Functions
  typedef otb::HistogramOfOrientedGradientCovariantImageFunction
    <GradientImageType>                                   HOGFunctionType;
  typedef otb::ImageFunctionAdaptor<HOGFunctionType>      AdapatedHOGFunctionType;
  typedef otb::LocalHistogramImageFunction
    <ImageType>                                           LocalHistogramFunctionType;
  typedef otb::ImageFunctionAdaptor<LocalHistogramFunctionType>
    AdaptedLocalHistogramFunctionType;
  typedef otb::MetaImageFunction<>                        MetaImageFunctionType;

  // SampleList manipulation
  typedef unsigned int                                    LabelType;

  typedef otb::ObjectDetectionClassifier
    < ImageType,
    VectorDataType,
    LabelType,
    FunctionPrecisionType,
    CoordRepType > ObjectDetectionClassifierType;

  typedef ObjectDetectionClassifierType::SVMModelType SVMModelType;
  typedef ObjectDetectionClassifierType::SVMModelPointerType SVMModelPointerType;
  
  // Read the image
  ReaderType::Pointer    reader  = ReaderType::New();
  reader->SetFileName(parseResult->GetParameterString("InputImage"));
  reader->UpdateOutputInformation();

  SVMModelPointerType svmModel = SVMModelType::New();
  svmModel->LoadModel(parseResult->GetParameterString("SVMModel"));

  unsigned int neighborhood = 10;
  if(parseResult->IsOptionPresent("Radius"))
    {
    neighborhood = parseResult->GetParameterUInt("Radius");
    }
  std::cout<<"Radius: "<<neighborhood<<std::endl;


  unsigned int windowRadius = neighborhood;
  if(parseResult->IsOptionPresent("WindowRadius"))
    {
    windowRadius = parseResult->GetParameterUInt("WindowRadius");
    }

  ImageType::IndexType position;
  position[0] = parseResult->GetParameterInt("Position", 0);
  position[1] = parseResult->GetParameterInt("Position", 1);

  // Extract the appropriate image portion
  ExtractFilterType::Pointer extractFilter = ExtractFilterType::New();

  VectorImageType::RegionType region;
  VectorImageType::IndexType  startIndex;
  VectorImageType::SizeType   size;
  startIndex[0] = position[0]-windowRadius;
  startIndex[1] = position[1]-windowRadius;
  size.Fill(2*windowRadius+1);
  region.SetSize(size);
  region.SetIndex(startIndex);

  VectorImageType::RegionType regionForReading = region;
  regionForReading.PadByRadius(neighborhood);

  startIndex.Fill(0);
  region.SetIndex(startIndex);

  extractFilter->SetInput(reader->GetOutput());
  extractFilter->SetExtractionRegion(regionForReading);

  // compose the imagefunction
  // Splitting image into channels
  VI2ILFilterType::Pointer vi2ilFilter = VI2ILFilterType::New();
  vi2ilFilter->SetInput(extractFilter->GetOutput());
  vi2ilFilter->Update();

  IntensityFilterType::Pointer intensityFilter = IntensityFilterType::New();
  intensityFilter->SetInput(extractFilter->GetOutput());

  GradientFilterType::Pointer gradientFilter = GradientFilterType::New();
  gradientFilter->SetInput(intensityFilter->GetOutput());
  gradientFilter->Update();

  AdapatedHOGFunctionType::Pointer hog = AdapatedHOGFunctionType::New();
  hog->SetInputImage(gradientFilter->GetOutput());
  hog->GetInternalImageFunction()->SetNeighborhoodRadius(neighborhood);

  MetaImageFunctionType::Pointer metaFunction = MetaImageFunctionType::New();
  metaFunction->AddFunction(hog);

  for(ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();
      ilIt != vi2ilFilter->GetOutput()->End(); ++ilIt)
    {
    AdaptedLocalHistogramFunctionType::Pointer localHistogramFunction = AdaptedLocalHistogramFunctionType::New();
    localHistogramFunction->SetInputImage(ilIt.Get());
    localHistogramFunction->GetInternalImageFunction()->SetNeighborhoodRadius(neighborhood);
    localHistogramFunction->GetInternalImageFunction()->SetHistogramMin(0);
    localHistogramFunction->GetInternalImageFunction()->SetHistogramMax(1000);
    localHistogramFunction->GetInternalImageFunction()->SetNumberOfHistogramBins(20);
    metaFunction->AddFunction(localHistogramFunction);
    }
  std::cout<<"Beep"<<std::endl;
  ImageType::PointType point;
  extractFilter->GetOutput()->TransformIndexToPhysicalPoint(position, point);

  MetaImageFunctionType::OutputType descriptor = metaFunction->Evaluate(point);

  SVMModelType::MeasurementType svmMeasurement(descriptor.Size());

  std::cout<<"Position: "<<position<<" ( "<< point << " )"<<std::endl;
  std::cout<<"Normalized measurement: ";


  for(unsigned int i = 0; i<descriptor.Size(); ++i)
    {
    svmMeasurement[i] = descriptor[i];
    std::cout<<svmMeasurement[i]<<"\t";
    }
  std::cout<<std::endl;


  std::cout<<"Label: "<<svmModel->EvaluateLabel(svmMeasurement)<<std::endl;
  std::cout<<"Hyperplane distance: "<<svmModel->EvaluateHyperplanesDistances(svmMeasurement)[0]<<std::endl;

  ImageType::Pointer labels = ImageType::New();
  labels->SetRegions(region);
  labels->Allocate();
  labels->FillBuffer(2);

  typedef otb::Image<float, 2> FloatImageType;
  FloatImageType::Pointer distances = FloatImageType::New();
  distances->SetRegions(region);
  distances->Allocate();
  distances->FillBuffer(10.);

  for(unsigned int i = 0; i < 2*windowRadius+1; ++i)
    {
    for(unsigned int j = 0; j < 2*windowRadius+1; ++j)
      {
      ImageType::IndexType index;
      index[0] = i;
      index[1] = j;
      ImageType::IndexType currentPosition = position;
      currentPosition[0]+=(int)i-(int)windowRadius;
      currentPosition[1]+=(int)j-(int)windowRadius;
      extractFilter->GetOutput()->TransformIndexToPhysicalPoint(currentPosition, point);

      descriptor = metaFunction->Evaluate(point);

      for(unsigned int id = 0; id<descriptor.Size(); ++id)
        {
        svmMeasurement[id] = descriptor[id];
        }
      labels->SetPixel(index, svmModel->EvaluateLabel(svmMeasurement) );
      distances->SetPixel(index, (svmModel->EvaluateHyperplanesDistances(svmMeasurement))[0]);
      }
    }

  typedef otb::ImageFileWriter<ImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName(parseResult->GetParameterString("OutputLabelMap"));
  writer->SetInput(labels);
  writer->Update();

  typedef otb::ImageFileWriter<FloatImageType> FloatWriterType;
  FloatWriterType::Pointer fwriter = FloatWriterType::New();
  fwriter->SetFileName(parseResult->GetParameterString("OutputHyperplaneDistance"));
  fwriter->SetInput(distances);
  fwriter->Update();

  typedef otb::ImageFileWriter<VectorImageType> VecWriterType;
  VecWriterType::Pointer vecWriter = VecWriterType::New();
  vecWriter->SetInput(extractFilter->GetOutput());
  vecWriter->SetFileName(parseResult->GetParameterString("OutputImage") );
  vecWriter->Update();

  return EXIT_SUCCESS;
}

}
