/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbHOGObjectDetector.h"

#include <iostream>
#include "otbCommandLineArgumentParser.h"

//Image
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageList.h"
#include "otbImageFileReader.h"
#include "otbImageFileWriter.h"
#include "otbVectorImageToImageListFilter.h"

// ListSample
#include "itkListSample.h"
#include "itkVariableLengthVector.h"
#include "itkFixedArray.h"

// Functions to use
//#include "otbStandardMetaImageFunctionBuilder.h"
// Functions to use
#include "otbHistogramOfOrientedGradientCovariantImageFunction.h"
#include "otbLocalHistogramImageFunction.h"
#include "itkGradientImageFilter.h"
#include "otbImageFunctionAdaptor.h"
#include "otbVectorImageToIntensityImageFilter.h"
#include "otbMetaImageFunction.h"

// Localisation estimator filter
#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataFileWriter.h"

// Statistic XML Reader
#include "otbStatisticsXMLFileReader.h"

// Descriptor generator
#include "otbObjectDetectionClassifier.h"

#include "itkMersenneTwisterRandomVariateGenerator.h"

#include "otbStandardFilterWatcher.h"


namespace otb
{

int HOGObjectDetector::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("HOGObjectDetector");
  descriptor->SetDescription("Perform object detection according to a SVM model.");
  descriptor->AddInputImage();
  descriptor->AddOption("SVMModel", "SVM model trained to perform the detection",
                        "svm", 1, true, ApplicationDescriptor::FileName);
//   descriptor->AddOption("FeaturesStatistics", "XML file containing mean and variance of each feature for normalization.",
//                         "fs", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Output", "Output vector data containing location of detected objects",
                        "out", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Radius", "Radius of the window for features (default is 10)",
                        "r", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("GridStep", "Step of the detection grid (default is 5)",
                        "gs", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("Seed","Set random process, for testing purpose only",
                        "s", 1, false, ApplicationDescriptor::Integer);

  descriptor->AddOption("UseRadiometryHistogram","Add the local histograms of each channel",
                        "rad", 0, false, ApplicationDescriptor::Boolean);

  return EXIT_SUCCESS;
}

// Input Image
typedef double                                          ValueType;
typedef unsigned short                                  PixelType;
typedef double                                          CoordRepType;
typedef double                                          FunctionPrecisionType;
typedef otb::VectorImage<PixelType, 2>                   VectorImageType;
typedef otb::Image<PixelType, 2>                         ImageType;
typedef otb::ImageFileReader<VectorImageType>           ReaderType;
typedef otb::ImageList<ImageType>                       ImageListType;

typedef otb::VectorImageToImageListFilter
  <VectorImageType, ImageListType>                      VI2ILFilterType;

// VectorData
typedef otb::VectorData<>                               VectorDataType;
typedef otb::VectorDataFileReader<VectorDataType>       VectorDataReaderType;
typedef otb::VectorDataProjectionFilter
    <VectorDataType, VectorDataType>                     VectorDataProjectionFilterType;
typedef otb::VectorDataFileWriter
    <VectorDataType>                                    VectorDataWriterType;
typedef itk::GradientImageFilter<ImageType>             GradientFilterType;
typedef GradientFilterType::OutputImageType             GradientImageType;
typedef otb::VectorImageToIntensityImageFilter
    <VectorImageType, ImageType>                           IntensityFilterType;
// Functions
typedef otb::HistogramOfOrientedGradientCovariantImageFunction
  <GradientImageType>                                   HOGFunctionType;
typedef otb::ImageFunctionAdaptor<HOGFunctionType>      AdaptedHOGFunctionType;
typedef otb::LocalHistogramImageFunction
    <ImageType>                                           LocalHistogramFunctionType;
  typedef otb::ImageFunctionAdaptor<LocalHistogramFunctionType>
    AdaptedLocalHistogramFunctionType;
  typedef otb::MetaImageFunction<>                        MetaImageFunctionType;

typedef otb::ImageFileWriter<GradientImageType>           WriterType;

// SampleList manipulation
typedef unsigned int                                    LabelType;

typedef otb::ObjectDetectionClassifier
           < GradientImageType,
             VectorDataType,
             LabelType,
             FunctionPrecisionType,
             CoordRepType > ObjectDetectionClassifierType;

typedef ObjectDetectionClassifierType::SVMModelType SVMModelType;
typedef ObjectDetectionClassifierType::SVMModelPointerType SVMModelPointerType;

typedef itk::Statistics::MersenneTwisterRandomVariateGenerator RandomGeneratorType;

int HOGObjectDetector::Execute(otb::ApplicationOptionsResult* parseResult)
{
  if(parseResult->IsOptionPresent("Seed"))
    {
    RandomGeneratorType::Pointer randomGenerator = RandomGeneratorType::New();
    randomGenerator->SetSeed(parseResult->GetParameterUInt("Seed"));
    }

  // Read the image
  ReaderType::Pointer    reader  = ReaderType::New();
  reader->SetFileName(parseResult->GetParameterString("InputImage", 0).c_str());
  reader->UpdateOutputInformation();
  
  IntensityFilterType::Pointer intensityFilter = IntensityFilterType::New();
  intensityFilter->SetInput(reader->GetOutput());

  GradientFilterType::Pointer gradientFilter = GradientFilterType::New();
  gradientFilter->SetInput(intensityFilter->GetOutput());

  VI2ILFilterType::Pointer vi2ilFilter = VI2ILFilterType::New();
  vi2ilFilter->SetInput(reader->GetOutput());
  vi2ilFilter->GetOutput()->UpdateOutputInformation();

  SVMModelPointerType svmModel = SVMModelType::New();
  svmModel->LoadModel(parseResult->GetParameterString("SVMModel"));

  unsigned int neighborhood = 10;
  if(parseResult->IsOptionPresent("Radius"))
    {
    neighborhood = parseResult->GetParameterUInt("Radius");
    }

  unsigned int gridStep = 5;
  if(parseResult->IsOptionPresent("GridStep"))
    {
    gridStep = parseResult->GetParameterUInt("GridStep");
    }


  AdaptedHOGFunctionType::Pointer hog = AdaptedHOGFunctionType::New();
  hog->SetInputImage(gradientFilter->GetOutput());
  hog->GetInternalImageFunction()->SetNeighborhoodRadius(neighborhood);
  
  MetaImageFunctionType::Pointer metaFunction = MetaImageFunctionType::New();
  metaFunction->AddFunction(hog);

  if(parseResult->IsOptionPresent("UseRadiometryHistogram"))
    {

    for(ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();
        ilIt != vi2ilFilter->GetOutput()->End(); ++ilIt)
      {
      AdaptedLocalHistogramFunctionType::Pointer localHistogramFunction = AdaptedLocalHistogramFunctionType::New();
      localHistogramFunction->SetInputImage(ilIt.Get());
      localHistogramFunction->GetInternalImageFunction()->SetNeighborhoodRadius(neighborhood);
      localHistogramFunction->GetInternalImageFunction()->SetHistogramMin(0);
      localHistogramFunction->GetInternalImageFunction()->SetHistogramMax(2047);
      localHistogramFunction->GetInternalImageFunction()->SetNumberOfHistogramBins(32);
      metaFunction->AddFunction(localHistogramFunction);
      }
    }
    

  ObjectDetectionClassifierType::Pointer classifier = ObjectDetectionClassifierType::New();
  classifier->SetInputImage(gradientFilter->GetOutput());
  classifier->SetNeighborhoodRadius(neighborhood);
  classifier->SetSVMModel(svmModel);
  classifier->SetDescriptorsFunction(metaFunction);
  classifier->SetNoClassLabel(0);
  classifier->SetClassKey("Class");
  classifier->SetGridStep(gridStep);
  
   if(parseResult->IsOptionPresent("UseRadiometryHistogram"))
    {
    for(ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();
        ilIt != vi2ilFilter->GetOutput()->End();
        ++ilIt)
      {
      classifier->AddInput(ilIt.Get());
      }
    }

  AdaptedHOGFunctionType::OutputType vshift(18*5+4*32);
  vshift.Fill(0.);
  AdaptedHOGFunctionType::OutputType vscale(18*5+4*32);
  vscale.Fill(1.);
  classifier->SetShifts(vshift);
  classifier->SetScales(vscale);

  otb::StandardFilterWatcher watcher(classifier,"Performing object detection ...");

  classifier->Update();

  VectorDataProjectionFilterType::Pointer proj = VectorDataProjectionFilterType::New();
  proj->SetInput(classifier->GetOutputVectorData());
  proj->SetInputKeywordList(reader->GetOutput()->GetImageKeywordlist());
  proj->SetInputProjectionRef(reader->GetOutput()->GetProjectionRef());

  VectorDataWriterType::Pointer writer = VectorDataWriterType::New();
  writer->SetFileName(parseResult->GetParameterString("Output"));
  writer->SetInput(proj->GetOutput());
  writer->Update();

return EXIT_SUCCESS;
}

}
