/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbMeanShiftModesDetection.h"

#include "otbCommandLineArgumentParser.h"

#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbVectorDataFileWriter.h"
#include "itkListSample.h"
#include "itkSampleMeanShiftClusteringFilter.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbImage.h"
#include "otbImageFileReader.h"
#include <map>


namespace otb
{

int MeanShiftModesDetection::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("MeanShift modes detection");
  descriptor->SetDescription("Perform modes detection by mean-shift within a regular grid detection map.");
  descriptor->AddOption("Output", "Output vector data containing location of detected modes",
                        "out", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Input", "Input vector data to detect modes in",
                        "in", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("ReferenceImage", "Reference image to reproject input vector data so that mode detection happens in pixel coordinates","ref", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Bandwidth", "Bandwith of mode detection, in pixel coordinates (default in 10)","b", 1, false, ApplicationDescriptor::Real);
  descriptor->AddOption("MinClusterSize", "Minimum size of the retained clusters (default is 5)","ms", 1, false, ApplicationDescriptor::Integer);
  return EXIT_SUCCESS;
}


typedef otb::VectorData<>                         VectorDataType;
typedef VectorDataType::DataNodeType              DataNodeType;
typedef VectorDataType::DataTreeType              DataTreeType;
typedef otb::VectorDataFileReader<VectorDataType> ReaderType;
typedef otb::VectorDataFileWriter<VectorDataType> WriterType;
typedef VectorDataType::PointType                 PointType;
typedef itk::Statistics::ListSample<PointType>    ListSampleType;
typedef itk::Statistics::SampleMeanShiftClusteringFilter<ListSampleType> MeanShiftFilterType;

typedef otb::Image<unsigned short, 2>              ImageType;
typedef otb::ImageFileReader<ImageType>           ImageReaderType;

typedef otb::VectorDataProjectionFilter
    <VectorDataType, VectorDataType>               VectorDataProjectionFilterType;


int MeanShiftModesDetection::Execute(otb::ApplicationOptionsResult* parseResult)
{
  double bandwidth = 10.;
  unsigned int minClusterSize = 5;

  if(parseResult->IsOptionPresent("Bandwidth"))
    {
    bandwidth = parseResult->GetParameterDouble("Bandwidth");
    }
  if(parseResult->IsOptionPresent("MinClusterSize"))
    {
    minClusterSize = parseResult->GetParameterUInt("MinClusterSize");
    }
  
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName(parseResult->GetParameterString("Input", 0).c_str());
  reader->Update();

  ImageReaderType::Pointer ireader = ImageReaderType::New();
  ireader->SetFileName(parseResult->GetParameterString("ReferenceImage", 0).c_str());
  ireader->GenerateOutputInformation();
  
  VectorDataProjectionFilterType::Pointer vectorDataProjectionFilter = VectorDataProjectionFilterType::New();
  vectorDataProjectionFilter->SetInput(reader->GetOutput());
  vectorDataProjectionFilter->SetOutputKeywordList(ireader->GetOutput()->GetImageKeywordlist());
  vectorDataProjectionFilter->SetOutputProjectionRef(ireader->GetOutput()->GetProjectionRef());
  vectorDataProjectionFilter->Update();


  ListSampleType::Pointer listSample = ListSampleType::New();


  typedef VectorDataType::DataTreeType            DataTreeType;
  typedef itk::PreOrderTreeIterator<DataTreeType> TreeIteratorType;
  TreeIteratorType it(vectorDataProjectionFilter->GetOutput()->GetDataTree());

  for (it.GoToBegin(); !it.IsAtEnd(); ++it)
    {
    if(it.Get()->GetNodeType() == otb::FEATURE_POINT)
      {
      listSample->PushBack(it.Get()->GetPoint());
      }
    }

  MeanShiftFilterType::Pointer msFilter = MeanShiftFilterType::New();
  msFilter->SetInputSample(listSample);
  msFilter->SetThreshold(bandwidth);
  msFilter->SetMinimumClusterSize(0);
  msFilter->Update();
  
  MeanShiftFilterType::ClusterLabelsType clusters = msFilter->GetOutput();
  
  std::map<unsigned int, PointType> pointMap;
  std::map<unsigned int, unsigned int> countMap;

  for(unsigned int i = 0; i < listSample->Size(); ++i)
    {
    if(clusters[i]!=0)
      {
      if(pointMap.find(clusters[i])==pointMap.end())
        {
        pointMap[clusters[i]] = listSample->GetMeasurementVector(i);
        countMap[clusters[i]]=1;
        }
      else
        {
        pointMap[clusters[i]][0] += listSample->GetMeasurementVector(i)[0];
        pointMap[clusters[i]][1] += listSample->GetMeasurementVector(i)[1];
        countMap[clusters[i]]+=1;
        }
      }
    }

  VectorDataType::Pointer output = VectorDataType::New();
  
  DataNodeType::Pointer document = DataNodeType::New();
  document->SetNodeType(otb::DOCUMENT);
  document->SetNodeId("Detected objects");
  DataNodeType::Pointer folder = DataNodeType::New();
  folder->SetNodeType(otb::FOLDER);
  
  DataTreeType::Pointer tree = output->GetDataTree();
  DataNodeType::Pointer root = tree->GetRoot()->Get();

  tree->Add(document, root);
  tree->Add(folder, document);


  std::map<unsigned int, PointType>::const_iterator mIt;
  for(mIt=pointMap.begin(); mIt != pointMap.end(); ++mIt)
    {
    if(countMap[mIt->first]>minClusterSize)
      {
      PointType finalP = mIt->second;
      finalP[0]/=countMap[mIt->first];
      finalP[1]/=countMap[mIt->first];
      std::cout<<"Mode found: "<<finalP<<" label=  "<<mIt->first <<std::endl;

      DataNodeType::Pointer newPoint = DataNodeType::New();
      newPoint->SetPoint(finalP);
      
      tree->Add(newPoint, folder);
      }
    }

  
  VectorDataProjectionFilterType::Pointer vectorDataProjectionFilter2 = VectorDataProjectionFilterType::New();
  vectorDataProjectionFilter2->SetInput(output);
  vectorDataProjectionFilter2->SetInputKeywordList(ireader->GetOutput()->GetImageKeywordlist());
  vectorDataProjectionFilter2->SetInputProjectionRef(ireader->GetOutput()->GetProjectionRef());
  vectorDataProjectionFilter2->Update();

  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName(parseResult->GetParameterString("Output", 0).c_str());
  writer->SetInput(vectorDataProjectionFilter2->GetOutput());
  writer->Update();

  return EXIT_SUCCESS;
}
}
