/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbObjectDetector.h"

#include <iostream>
#include "otbCommandLineArgumentParser.h"

//Image
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageList.h"
#include "otbImageFileReader.h"
#include "otbVectorImageToImageListFilter.h"

// ListSample
#include "itkListSample.h"
#include "itkVariableLengthVector.h"
#include "itkFixedArray.h"

// Functions to use
#include "otbStandardMetaImageFunctionBuilder.h"

// Localisation estimator filter
#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataFileWriter.h"

// Statistic XML Reader
#include "otbStatisticsXMLFileReader.h"

// Descriptor generator
#include "otbObjectDetectionClassifier.h"

#include "itkMersenneTwisterRandomVariateGenerator.h"

#include "otbStandardFilterWatcher.h"


namespace otb
{

int ObjectDetector::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("ObjectDetector");
  descriptor->SetDescription("Perform object detection according to a SVM model.");
  descriptor->AddInputImage();
  descriptor->AddOption("SVMModel", "SVM model trained to perform the detection",
                        "svm", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("FeaturesStatistics", "XML file containing mean and variance of each feature for normalization.",
                        "fs", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Output", "Output vector data containing location of detected objects",
                        "out", 1, true, ApplicationDescriptor::FileName);
  descriptor->AddOption("Radius", "Radius of the window for features (default is 10)",
                        "r", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("Seed","Set random process, for testing purpose only",
                        "s", 1, false, ApplicationDescriptor::Integer);
    descriptor->AddOption("GridStep", "Step of the detection grid (default is 5)",
                        "gs", 1, false, ApplicationDescriptor::Integer);

  return EXIT_SUCCESS;
}

// Input Image
typedef double                                          ValueType;
typedef unsigned short                                  PixelType;
typedef double                                          CoordRepType;
typedef double                                          FunctionPrecisionType;
typedef otb::VectorImage<PixelType, 2>                   VectorImageType;
typedef otb::Image<PixelType, 2>                         ImageType;
typedef otb::ImageFileReader<VectorImageType>           ReaderType;
typedef otb::ImageList<ImageType>                       ImageListType;

typedef otb::VectorImageToImageListFilter
  <VectorImageType, ImageListType>                      VI2ILFilterType;

// VectorData
typedef otb::VectorData<>                               VectorDataType;
typedef otb::VectorDataFileReader<VectorDataType>       VectorDataReaderType;
typedef otb::VectorDataProjectionFilter
    <VectorDataType, VectorDataType>                     VectorDataProjectionFilterType;
typedef otb::VectorDataFileWriter
    <VectorDataType>                                    VectorDataWriterType;

// Functions
typedef otb::StandardMetaImageFunctionBuilder<FunctionPrecisionType, CoordRepType>
                                                        MetaImageFunctionBuilderType;
typedef MetaImageFunctionBuilderType::MetaImageFunctionType
                                                        MetaImageFunctionType;
typedef MetaImageFunctionBuilderType::DataObjectContainerType
                                                        DataObjectContainerType;

// SampleList manipulation
typedef unsigned int                                    LabelType;

typedef otb::ObjectDetectionClassifier
           < ImageType,
             VectorDataType,
             LabelType,
             FunctionPrecisionType,
             CoordRepType > ObjectDetectionClassifierType;

typedef ObjectDetectionClassifierType::SVMModelType SVMModelType;
typedef ObjectDetectionClassifierType::SVMModelPointerType SVMModelPointerType;

typedef otb::StatisticsXMLFileReader<MetaImageFunctionType::OutputType> StatisticsXMLFileReaderType;

typedef itk::Statistics::MersenneTwisterRandomVariateGenerator RandomGeneratorType;

int ObjectDetector::Execute(otb::ApplicationOptionsResult* parseResult)
{
  if(parseResult->IsOptionPresent("Seed"))
    {
    RandomGeneratorType::Pointer randomGenerator = RandomGeneratorType::New();
    randomGenerator->SetSeed(parseResult->GetParameterUInt("Seed"));
    }
  
  unsigned int gridStep = 5;
  if(parseResult->IsOptionPresent("GridStep"))
    {
    gridStep = parseResult->GetParameterUInt("GridStep");
    }

  // Read the image
  ReaderType::Pointer    reader  = ReaderType::New();
  reader->SetFileName(parseResult->GetParameterString("InputImage", 0).c_str());
  reader->UpdateOutputInformation();

  StatisticsXMLFileReaderType::Pointer statisticsReader = StatisticsXMLFileReaderType::New();
  statisticsReader->SetFileName(parseResult->GetParameterString("FeaturesStatistics"));

  SVMModelPointerType svmModel = SVMModelType::New();
  svmModel->LoadModel(parseResult->GetParameterString("SVMModel"));

  // compose the imagefunction
  // Splitting image into channels
  VI2ILFilterType::Pointer vi2ilFilter = VI2ILFilterType::New();
  vi2ilFilter->SetInput(reader->GetOutput());
  vi2ilFilter->UpdateOutputInformation();
  ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();

  unsigned int neighborhood = 10;
  if(parseResult->IsOptionPresent("Radius"))
    {
    neighborhood = parseResult->GetParameterUInt("Radius");
    }

  MetaImageFunctionBuilderType::Pointer builder = MetaImageFunctionBuilderType::New();
  
  builder->SetNeighborhoodRadius(neighborhood);
  while(ilIt != vi2ilFilter->GetOutput()->End())
    {
    builder->AddImage(ilIt.Get().GetPointer()); //reader->GetOutput());
    
    ++ilIt;
    }
  
  MetaImageFunctionType::Pointer metaFunction   = builder->GetMetaImageFunction();
  DataObjectContainerType        imageContainer = builder->GetDataObjectContainer();
  
  ObjectDetectionClassifierType::Pointer classifier = ObjectDetectionClassifierType::New();
  classifier->SetInputImage(vi2ilFilter->GetOutput()->Begin().Get());
  classifier->SetNeighborhoodRadius(neighborhood);
  classifier->SetSVMModel(svmModel);
  classifier->SetDescriptorsFunction(metaFunction);
  classifier->SetNoClassLabel(0);
  classifier->SetClassKey("Class");
  classifier->SetGridStep(gridStep);
  classifier->SetShifts(statisticsReader->GetStatisticVectorByName("mean"));
  classifier->SetScales(statisticsReader->GetStatisticVectorByName("stddev"));

  otb::StandardFilterWatcher watcher(classifier,"Performing object detection ...");

  classifier->Update();

  VectorDataProjectionFilterType::Pointer proj = VectorDataProjectionFilterType::New();
  proj->SetInput(classifier->GetOutputVectorData());
  proj->SetInputKeywordList(reader->GetOutput()->GetImageKeywordlist());
  proj->SetInputProjectionRef(reader->GetOutput()->GetProjectionRef());

  VectorDataWriterType::Pointer writer = VectorDataWriterType::New();
  writer->SetFileName(parseResult->GetParameterString("Output"));
  writer->SetInput(proj->GetOutput());
  writer->Update();

       return EXIT_SUCCESS;
}

}
