/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbTrainDeepSVMObjectDetector.h"

#include <iostream>
#include "otbCommandLineArgumentParser.h"

//Image
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageList.h"
#include "otbImageFileReader.h"
#include "otbVectorImageToImageListFilter.h"

// ListSample
#include "itkListSample.h"
#include "itkVariableLengthVector.h"
#include "itkFixedArray.h"

// Functions to use
#include "otbStandardMetaImageFunctionBuilder.h"

// Localisation estimator filter
#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbLabeledSampleLocalizationGenerator.h"
#include "otbVectorDataProjectionFilter.h"
#include "otbVectorDataFileWriter.h"

// Descriptor generator
#include "otbDescriptorsListSampleGenerator.h"

// Normalize the samples
#include "otbShiftScaleSampleListFilter.h"

// Balancing ListSample
#include "otbListSampleToBalancedListSampleFilter.h"

// List sample concatenation
#include "otbConcatenateSampleListFilter.h"

// SVM estimator
#include "otbSVMSampleListModelEstimator.h"

// Statistic XML Reader
#include "otbStatisticsXMLFileReader.h"

// Validation
#include "otbSVMClassifier.h"
#include "otbConfusionMatrixCalculator.h"

#include "itkTimeProbe.h"
#include "otbStandardFilterWatcher.h"


namespace otb
{

int TrainDeepSVMObjectDetector::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("TrainObjectDetector");
  descriptor->SetDescription("Perform SVM training per group of features, followed by performance evaluation of each classifier. Perform SVM training on the vector of hyperplanes distances of each classifier, and performances estimation.");
  descriptor->AddOptionNParams("InputImages", "A list of images",
                               "in", true, ApplicationDescriptor::InputImage);
  descriptor->AddOptionNParams("VectorDataSamples", "Vector Data of sample used to train the estimator",
                               "vd", true, ApplicationDescriptor::FileName);
  descriptor->AddOptionNParams("WorkingAreas", "Vector Data containing polygons to delineate working areas",
                               "wa", true, ApplicationDescriptor::FileName);
  //descriptor->AddOptionNParams("TrainingSet", "List of images / ground-truth (im1 gt1 im2 gt2 ... imN gtN)",
  //                             "ts", true, ApplicationDescriptor::FileName);
  descriptor->AddOption("FeaturesStatistics", "XML file containing mean and variance of each feature for normalization.",
                        "fs", 1, false, ApplicationDescriptor::FileName);

  descriptor->AddOption("Radius", "Radius of the window for features (default is 10)",
                        "r", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("Margin", "Margin for SVM learning.",
                        "m", 1, false, ApplicationDescriptor::Real);
  descriptor->AddOption("Balancing", "Balance and grow the training set.",
                        "b", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("NegativeDensity","Density of negative points in ppp (points per pixels, default is 0.0005)",
                        "nd", 1, false, ApplicationDescriptor::Real);
  descriptor->AddOption("PositiveRate","Number of random positive examples generated for each ground truth point (default is 50)",
                        "pr", 1, false, ApplicationDescriptor::Integer);
  descriptor->AddOption("InhibitionRadius","Inhibition radius around positive point to generate negative examples (default is 10)",
                        "ir", 1, false, ApplicationDescriptor::Integer);


  return EXIT_SUCCESS;
}

int TrainDeepSVMObjectDetector::Execute(otb::ApplicationOptionsResult* parseResult)
{

  // Input Image
  typedef double                                          ValueType;
  typedef unsigned short                                  PixelType;
  typedef otb::VectorImage<PixelType, 2>                   VectorImageType;
  typedef otb::Image<PixelType, 2>                         ImageType;
  typedef otb::ImageFileReader<VectorImageType>           ReaderType;
  typedef otb::ImageList<ImageType>                       ImageListType;
  typedef otb::VectorImageToImageListFilter
      <VectorImageType, ImageListType>                      VI2ILFilterType;

  // Training vectordata
  typedef otb::VectorData<>                               VectorDataType;
  typedef otb::VectorDataFileReader<VectorDataType>       VectorDataReaderType;
  typedef otb::VectorDataProjectionFilter
      <VectorDataType, VectorDataType>                     VectorDataProjectionFilterType;
  typedef otb::VectorDataFileWriter
      <VectorDataType>                                    VectorDataWriterType;

  // Functions
  typedef otb::StandardMetaImageFunctionBuilder<ValueType, double>
  MetaImageFunctionBuilderType;
  typedef MetaImageFunctionBuilderType::MetaImageFunctionType
      MetaImageFunctionType;
  typedef MetaImageFunctionBuilderType::DataObjectContainerType
      DataObjectContainerType;

  // SampleList manipulation
  typedef unsigned int                                    LabelType;
  typedef itk::VariableLengthVector<ValueType>            MeasurementType;
  typedef itk::Statistics::ListSample<MeasurementType>    ListSampleType;
  typedef itk::FixedArray<LabelType, 1>                    LabelSampleType;
  typedef itk::Statistics::ListSample<LabelSampleType>    LabelListSampleType;
  typedef otb::Statistics::ConcatenateSampleListFilter
      <ListSampleType>                                    ConcatenateListSampleFilterType;
  typedef otb::Statistics::ConcatenateSampleListFilter
      <LabelListSampleType>                               ConcatenateLabelListSampleFilterType;

  // Filter to process listSample
  typedef double                                          CoordRepType;
  typedef double                                          FunctionPrecisionType;
  typedef otb::LabeledSampleLocalizationGenerator<VectorDataType>  GeneratorType;
  typedef otb::DescriptorsListSampleGenerator<
      ImageType,
      VectorDataType,
      ListSampleType,
      LabelListSampleType,
      FunctionPrecisionType,
      CoordRepType >                                        DescriptorsListSampleGeneratorType;

  typedef otb::Statistics::ListSampleToBalancedListSampleFilter<
      ListSampleType,
      LabelListSampleType>                                  BalancingListSampleFilterType;

  typedef otb::Statistics::ShiftScaleSampleListFilter<
      ListSampleType, ListSampleType>                        ShiftScaleFilterType;

  // SVM Estimator
  typedef otb::Functor::VariableLengthVectorToMeasurementVectorFunctor<
      MeasurementType>                                      MeasurementVectorFunctorType;
  typedef otb::SVMSampleListModelEstimator<
      ListSampleType,
      LabelListSampleType,
      MeasurementVectorFunctorType>                         SVMEstimatorType;
  typedef otb::SVMClassifier<ListSampleType, LabelType>   ClassifierType;
  typedef otb::ConfusionMatrixCalculator<LabelListSampleType,
      LabelListSampleType>                            ConfusionMatrixCalculatorType;
  typedef ClassifierType::OutputType ClassifierOutputType;


  // Statistic XML file Reader
  typedef otb::StatisticsXMLFileReader<MeasurementType>   StatisticsReader;

  ConcatenateLabelListSampleFilterType::Pointer concatenateTrainingLabels = ConcatenateLabelListSampleFilterType::New();
  ConcatenateListSampleFilterType::Pointer concatenateTrainingSamples = ConcatenateListSampleFilterType::New();
  ConcatenateLabelListSampleFilterType::Pointer concatenateValidationLabels = ConcatenateLabelListSampleFilterType::New();
  ConcatenateListSampleFilterType::Pointer concatenateValidationSamples = ConcatenateListSampleFilterType::New();

  typedef itk::Statistics::MersenneTwisterRandomVariateGenerator RandomGeneratorType;

  // Random Generator
  RandomGeneratorType::Pointer randomGenerator = RandomGeneratorType::New();
  if(parseResult->IsOptionPresent("Seed"))
    {
    randomGenerator->SetSeed(parseResult->GetParameterUInt("Seed"));
    }
  
  double       negativeDensity  = 0.0005;
  unsigned int positiveRate     = 50;
  unsigned int inhibitionRadius = 10;

  if(parseResult->IsOptionPresent("NegativeDensity"))
    {
    negativeDensity = parseResult->GetParameterDouble("NegativeDensity");
    }
  if(parseResult->IsOptionPresent("PositiveRate"))
    {
    positiveRate = parseResult->GetParameterUInt("PositiveRate");
    }
  if(parseResult->IsOptionPresent("InhibitionRadius"))
    {
    inhibitionRadius = parseResult->GetParameterUInt("InhibitionRadius");
    }

  for(int imgIndex = 0; imgIndex<parseResult->GetNumberOfParameters("InputImages"); ++imgIndex)
    {
    // Read the image
    ReaderType::Pointer    reader  = ReaderType::New();
    reader->SetFileName(parseResult->GetParameterString("InputImages", imgIndex));
    reader->UpdateOutputInformation();

    std::cout<<"Processing image ("<<imgIndex<<"): "<<reader->GetFileName()<<std::endl;

    // read the vectordata
    VectorDataReaderType::Pointer vdreader = VectorDataReaderType::New();
    vdreader->SetFileName(parseResult->GetParameterString("VectorDataSamples", imgIndex));

    VectorDataProjectionFilterType::Pointer proj1 = VectorDataProjectionFilterType::New();
    proj1->SetInput(vdreader->GetOutput());
    proj1->SetOutputKeywordList(reader->GetOutput()->GetImageKeywordlist());
    proj1->SetOutputProjectionRef(reader->GetOutput()->GetProjectionRef());

    VectorDataReaderType::Pointer workingAreaReader = VectorDataReaderType::New();
    workingAreaReader->SetFileName(parseResult->GetParameterString("WorkingAreas", imgIndex));

    VectorDataProjectionFilterType::Pointer proj2 = VectorDataProjectionFilterType::New();
    proj2->SetInput(workingAreaReader->GetOutput());
    proj2->SetOutputKeywordList(reader->GetOutput()->GetImageKeywordlist());
    proj2->SetOutputProjectionRef(reader->GetOutput()->GetProjectionRef());

    // compose the imagefunction
    // Splitting image into channels
    VI2ILFilterType::Pointer vi2ilFilter = VI2ILFilterType::New();
    vi2ilFilter->SetInput(reader->GetOutput());
    vi2ilFilter->UpdateOutputInformation();

    unsigned int radius = 10;
    if(parseResult->IsOptionPresent("Radius"))
      {
      radius = parseResult->GetParameterUInt("Radius");
      }
    MetaImageFunctionBuilderType::Pointer builder = MetaImageFunctionBuilderType::New();
    builder->SetNeighborhoodRadius(radius);

    ImageListType::Iterator ilIt = vi2ilFilter->GetOutput()->Begin();

    while(ilIt != vi2ilFilter->GetOutput()->End())
      {
      builder->AddImage(ilIt.Get().GetPointer()); //reader->GetOutput());

      ++ilIt;
      }
    DataObjectContainerType imageContainer = builder->GetDataObjectContainer();

    // Localization filter for training
    GeneratorType::Pointer   trainingGenerator   = GeneratorType::New();
    trainingGenerator->PushBackInput(proj1->GetOutput());
    trainingGenerator->PushBackInput(proj2->GetOutput());
    trainingGenerator->SetInhibitionRadius(inhibitionRadius);
    trainingGenerator->SetRandomLocalizationDensity(negativeDensity);
    trainingGenerator->SetNumberOfPositiveSamplesPerPoint(positiveRate);
    if(!parseResult->IsOptionPresent("Seed"))
    {
    trainingGenerator->SetSeed(time(0));
    }

    // Localization filter for validation
    GeneratorType::Pointer   validationGenerator   = GeneratorType::New();
    validationGenerator->PushBackInput(proj1->GetOutput());
    validationGenerator->PushBackInput(proj2->GetOutput());
    validationGenerator->SetInhibitionRadius(inhibitionRadius);
    validationGenerator->SetRandomLocalizationDensity(negativeDensity);
    validationGenerator->SetNumberOfPositiveSamplesPerPoint(positiveRate);
    if(!parseResult->IsOptionPresent("Seed"))
      {
      validationGenerator->SetSeed(time(0)+1);
      }

    // Generate descriptors for training
    DescriptorsListSampleGeneratorType::Pointer trainingDescriptorsGenerator = DescriptorsListSampleGeneratorType::New();
    trainingDescriptorsGenerator->SetInputImage(vi2ilFilter->GetOutput()->GetNthElement(0));
    trainingDescriptorsGenerator->SetSamplesLocations(trainingGenerator->GetOutput());
    trainingDescriptorsGenerator->SetDescriptorsFunction(builder->GetMetaImageFunction());
    trainingDescriptorsGenerator->SetNeighborhoodRadius(radius);
    // --> without calling Update it doesn't seem to work proprely

    // Generate descriptors for validation
    DescriptorsListSampleGeneratorType::Pointer validationDescriptorsGenerator = DescriptorsListSampleGeneratorType::New();
    validationDescriptorsGenerator->SetInputImage(vi2ilFilter->GetOutput()->GetNthElement(0));
    validationDescriptorsGenerator->SetSamplesLocations(validationGenerator->GetOutput());
    validationDescriptorsGenerator->SetDescriptorsFunction(builder->GetMetaImageFunction());
    validationDescriptorsGenerator->SetNeighborhoodRadius(radius);

    itk::TimeProbe chrono;
    chrono.Start();
    trainingDescriptorsGenerator->Update();
    validationDescriptorsGenerator->Update();
    chrono.Stop();

    std::cout<<trainingDescriptorsGenerator->GetListSample()->Size()
        +validationDescriptorsGenerator->GetListSample()->Size()<<" samples generated in "<<chrono.GetMeanTime()<<" seconds."<<std::endl;

    concatenateTrainingSamples->AddInput(trainingDescriptorsGenerator->GetListSampleObject());
    concatenateTrainingLabels->AddInput(trainingDescriptorsGenerator->GetLabelListSampleObject());
    concatenateValidationSamples->AddInput(validationDescriptorsGenerator->GetListSampleObject());
    concatenateValidationLabels->AddInput(validationDescriptorsGenerator->GetLabelListSampleObject());

    }

  concatenateTrainingSamples->Update();
  concatenateTrainingLabels->Update();
  concatenateValidationSamples->Update();
  concatenateValidationLabels->Update();


  // Normalize the samples
  // Read the mean and variance form the XML file
  StatisticsReader::Pointer  statisticsReader = StatisticsReader::New();
  statisticsReader->SetFileName(parseResult->GetParameterString("FeaturesStatistics").c_str());
  MeasurementType  meanMeasurentVector     = statisticsReader->GetStatisticVectorByName("mean");
  MeasurementType  varianceMeasurentVector = statisticsReader->GetStatisticVectorByName("stddev");

  // Shift scale the samples
  ShiftScaleFilterType::Pointer trainingShiftScaleFilter = ShiftScaleFilterType::New();
  trainingShiftScaleFilter->SetInput(concatenateTrainingSamples->GetOutput());
  trainingShiftScaleFilter->SetShifts(meanMeasurentVector);
  trainingShiftScaleFilter->SetScales(varianceMeasurentVector);
  trainingShiftScaleFilter->Update();

  ShiftScaleFilterType::Pointer validationShiftScaleFilter = ShiftScaleFilterType::New();
  validationShiftScaleFilter->SetInput(concatenateValidationSamples->GetOutput());
  validationShiftScaleFilter->SetShifts(meanMeasurentVector);
  validationShiftScaleFilter->SetScales(varianceMeasurentVector);
  validationShiftScaleFilter->Update();

  std::cout<<std::endl;

  ListSampleType::Pointer listSample;
  LabelListSampleType::Pointer labelListSample;

  if(parseResult->IsOptionPresent("Balancing"))
    {
    // Balance the list sample.
    std::cout<<"Number of training samples before balancing: "<<concatenateTrainingSamples->GetOutputSampleList()->Size()<<std::endl;
    BalancingListSampleFilterType::Pointer balancingFilter = BalancingListSampleFilterType::New();
    balancingFilter->SetInput(trainingShiftScaleFilter->GetOutput()/*GetOutputSampleList()*/);
    balancingFilter->SetInputLabel(concatenateTrainingLabels->GetOutput()/*GetOutputSampleList()*/);
    balancingFilter->SetBalancingFactor(parseResult->GetParameterUInt("Balancing"));
    balancingFilter->Update();
    listSample = balancingFilter->GetOutputSampleList();
    labelListSample = balancingFilter->GetOutputLabelSampleList();
    std::cout<<"Number of samples after balancing: "<<balancingFilter->GetOutputSampleList()->Size()<<std::endl;
    }
  else
    {
    listSample = trainingShiftScaleFilter->GetOutputSampleList();
    labelListSample = concatenateTrainingLabels->GetOutputSampleList();
    std::cout<<"Number of training samples: "<<concatenateTrainingSamples->GetOutputSampleList()->Size()<<std::endl;
    }

  // spit the data set into training/validation set
  ListSampleType::Pointer trainingListSample = listSample;
  ListSampleType::Pointer validationListSample = validationShiftScaleFilter->GetOutputSampleList();
  LabelListSampleType::Pointer trainingLabeledListSample = labelListSample;
  LabelListSampleType::Pointer validationLabeledListSample = concatenateValidationLabels->GetOutputSampleList();

  ListSampleType::Pointer trainingHDSampleList = ListSampleType::New();
  ListSampleType::Pointer validationHDSampleList = ListSampleType::New();


  // Initialize hyperplane distances list samples
  MeasurementType hdVector(12);
  hdVector.Fill(0);

  ListSampleType::Iterator tmpIt = trainingListSample->Begin();
  while(tmpIt!=trainingListSample->End())
    {
    trainingHDSampleList->PushBack(hdVector);
    ++tmpIt;
    }

  tmpIt = validationListSample->Begin();
  while(tmpIt!=validationListSample->End())
    {
    validationHDSampleList->PushBack(hdVector);
    ++tmpIt;
    }

  std::cout<<"Size of training set: "<<trainingListSample->Size()<<std::endl;
  std::cout<<"Size of validation set: "<<validationListSample->Size()<<std::endl;

  typedef std::pair<unsigned int, unsigned int> IndexPairType;
  typedef std::map<std::string, IndexPairType> IndexMapType;

  unsigned int nbfm = 16;
  unsigned int nbstats = 4;
  unsigned int nbflusser = 11;
  unsigned int total = nbfm + nbstats + nbflusser;

  // Per features SVM
  IndexMapType featureMap;
  featureMap["[blue  statistics_mvsk]"]=std::make_pair(0, nbstats-1);
  featureMap["[blue  flusser_moments]"]=std::make_pair(nbstats, nbstats+nbflusser-1);
  featureMap["[blue  fourrier_mellin]"]=std::make_pair(nbstats+nbflusser, total-1);
  featureMap["[green statistics_mvsk]"]=std::make_pair(total, total+nbstats-1);
  featureMap["[green flusser_moments]"]=std::make_pair(total+nbstats, total+nbstats+nbflusser-1);
  featureMap["[green fourrier_mellin]"]=std::make_pair(total+nbstats+nbflusser, 2*total-1);
  featureMap["[red   statistics_mvsk]"]=std::make_pair(2*total, 2*total+nbstats-1);
  featureMap["[red   flusser_moments]"]=std::make_pair(2*total+nbstats, 2*total+nbstats+nbflusser-1);
  featureMap["[red   fourrier_mellin]"]=std::make_pair(2*total+nbstats+nbflusser, 3*total-1);
  featureMap["[nir   statistics_mvsk]"]=std::make_pair(3*total, 3*total+nbstats-1);
  featureMap["[nir   flusser_moments]"]=std::make_pair(3*total+nbstats, 3*total+nbstats+nbflusser-1);
  featureMap["[nir   fourrier_mellin]"]=std::make_pair(3*total+nbstats+nbflusser, 4*total-1);

  std::cout<<"Performances per group of features: "<<std::endl;

  unsigned int findex = 0;

  // For each group of feature
  for(IndexMapType::const_iterator mIt = featureMap.begin(); mIt!=featureMap.end(); ++mIt, ++findex)
    {
    ListSampleType::Pointer currentTrainingListSample = ListSampleType::New();
    ListSampleType::Pointer currentValidationListSample = ListSampleType::New();

    unsigned int currentMeasurementSize = 1 + mIt->second.second - mIt->second.first;

    //    std::cout<<mIt->first<<" - measurement size = "<<currentMeasurementSize<<std::endl;

    // Extracting training samples
    ListSampleType::Iterator sampleIt = trainingListSample->Begin();
    while(sampleIt != trainingListSample->End())
      {
      MeasurementType currentMeasurement(currentMeasurementSize);

      unsigned int idx = 0;
      for(unsigned int i = mIt->second.first; i<=mIt->second.second; ++i)
        {
        currentMeasurement[idx]=sampleIt.GetMeasurementVector()[i];
        ++idx;
        }
      currentTrainingListSample->PushBack(currentMeasurement);
      ++sampleIt;
      }

    // Extracting validation samples
    sampleIt = validationListSample->Begin();
    while(sampleIt != validationListSample->End())
      {
      MeasurementType currentMeasurement(currentMeasurementSize);

      unsigned int idx = 0;
      for(unsigned int i = mIt->second.first; i<=mIt->second.second; ++i)
        {
        currentMeasurement[idx]=sampleIt.GetMeasurementVector()[i];
        ++idx;
        }
      currentValidationListSample->PushBack(currentMeasurement);
      ++sampleIt;
      }


    // Estimate the svm model
    SVMEstimatorType::Pointer svmestimator = SVMEstimatorType::New();
    svmestimator->SetInputSampleList(currentTrainingListSample);
    svmestimator->SetTrainingSampleList(trainingLabeledListSample);

    //svmestimator->SetKernelType(RBF);
    //  svmestimator->SetParametersOptimization(true);

    if(parseResult->IsOptionPresent("Margin"))
      {
      svmestimator->SetC(parseResult->GetParameterDouble("Margin"));
      }

    svmestimator->Update();

    // Performances estimation
    ClassifierType::Pointer validationClassifier = ClassifierType::New();
    validationClassifier->SetSample(currentValidationListSample);
    validationClassifier->SetNumberOfClasses(svmestimator->GetModel()->GetNumberOfClasses());
    validationClassifier->SetModel(svmestimator->GetModel());
    validationClassifier->Update();

    // Build validation hd sample list
    ClassifierType::HyperplanesDistancesListSampleType::Pointer hyperplaneDistances = validationClassifier->GetHyperplanesDistancesOutput();
    ClassifierType::HyperplanesDistancesListSampleType::Iterator hdIt = hyperplaneDistances->Begin();
    ListSampleType::Iterator sIt = validationHDSampleList->Begin();
    ListSampleType::Pointer tmp = ListSampleType::New();

    while(hdIt!= hyperplaneDistances->End()
        && sIt != validationHDSampleList->End())
      {
      MeasurementType sMes = sIt.GetMeasurementVector();
      sMes[findex] = hdIt.GetMeasurementVector()[0];
      tmp->PushBack(sMes);
      ++hdIt;
      ++sIt;
      }
    validationHDSampleList = tmp;

    // Estimate performances
    ClassifierOutputType::ConstIterator it = validationClassifier->GetOutput()->Begin();
    ClassifierOutputType::ConstIterator itEnd = validationClassifier->GetOutput()->End();

    LabelListSampleType::Pointer classifierListLabel = LabelListSampleType::New();

    while (it != itEnd)
      {
      // Due to a bug in SVMClassifier, outlier in one-class SVM are labeled with unsigned int max
      classifierListLabel->PushBack(it.GetClassLabel()==itk::NumericTraits<unsigned int>::max() ? 2 : it.GetClassLabel());
      ++it;
      }

    ConfusionMatrixCalculatorType::Pointer confMatCalc =  ConfusionMatrixCalculatorType::New();

    confMatCalc->SetReferenceLabels(validationLabeledListSample);
    confMatCalc->SetProducedLabels(classifierListLabel);

    confMatCalc->Update();

    double tp = confMatCalc->GetConfusionMatrix()(1, 1);
    double fp = confMatCalc->GetConfusionMatrix()(1, 0);
    double fn = confMatCalc->GetConfusionMatrix()(0, 1);

    double precision = tp/(tp+fp);
    double recall    = tp/(tp+fn);

    std::cout<<mIt->first<<" - precision= "<<precision<<", recall= "<<recall<<", fscore= "<<2*precision*recall/(precision+recall)<<std::endl;

    // Fill the hyperplane distance descriptor
    validationClassifier = ClassifierType::New();
    validationClassifier->SetSample(currentTrainingListSample);
    validationClassifier->SetNumberOfClasses(svmestimator->GetModel()->GetNumberOfClasses());
    validationClassifier->SetModel(svmestimator->GetModel());
    validationClassifier->Update();

    hyperplaneDistances = validationClassifier->GetHyperplanesDistancesOutput();
    ClassifierType::HyperplanesDistancesListSampleType::Iterator hd2It = hyperplaneDistances->Begin();
    ListSampleType::Iterator s2It = trainingHDSampleList->Begin();

    tmp = ListSampleType::New();

    while(hd2It!= hyperplaneDistances->End()
        && s2It != trainingHDSampleList->End())
      {
      MeasurementType sMes = s2It.GetMeasurementVector();
      sMes[findex] = hd2It.GetMeasurementVector()[0];
      tmp->PushBack(sMes);
      ++hd2It;
      ++s2It;
      }

    trainingHDSampleList = tmp;

    }

  SVMEstimatorType::Pointer svmestimator = SVMEstimatorType::New();
  svmestimator->SetInputSampleList(trainingHDSampleList);
  svmestimator->SetTrainingSampleList(trainingLabeledListSample);

  //svmestimator->SetKernelType(RBF);

  if(parseResult->IsOptionPresent("Margin"))
    {
    svmestimator->SetC(parseResult->GetParameterDouble("Margin"));
    }

  svmestimator->Update();

  // Performances estimation
  ClassifierType::Pointer validationClassifier = ClassifierType::New();
  validationClassifier->SetSample(validationHDSampleList);
  validationClassifier->SetNumberOfClasses(svmestimator->GetModel()->GetNumberOfClasses());
  validationClassifier->SetModel(svmestimator->GetModel());
  validationClassifier->Update();

  // Estimate performances
  ClassifierOutputType::ConstIterator it = validationClassifier->GetOutput()->Begin();
  ClassifierOutputType::ConstIterator itEnd = validationClassifier->GetOutput()->End();

  LabelListSampleType::Pointer classifierListLabel = LabelListSampleType::New();

  while (it != itEnd)
    {
    // Due to a bug in SVMClassifier, outlier in one-class SVM are labeled with unsigned int max
    classifierListLabel->PushBack(it.GetClassLabel()==itk::NumericTraits<unsigned int>::max() ? 2 : it.GetClassLabel());
    ++it;
    }

  ConfusionMatrixCalculatorType::Pointer confMatCalc =  ConfusionMatrixCalculatorType::New();

  confMatCalc->SetReferenceLabels(validationLabeledListSample);
  confMatCalc->SetProducedLabels(classifierListLabel);

  confMatCalc->Update();

  double tp = confMatCalc->GetConfusionMatrix()(1, 1);
  double fp = confMatCalc->GetConfusionMatrix()(1, 0);
  double fn = confMatCalc->GetConfusionMatrix()(0, 1);

  double precision = tp/(tp+fp);
  double recall    = tp/(tp+fn);

  std::cout<<"Hyperplane distances classifier performances :"<<std::endl;
  std::cout<<"precision = "<<precision<<", recall= "<<recall<<", fscore= "<<2*precision*recall/(precision+recall)<<std::endl;

  return EXIT_SUCCESS;
}

}
