/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE,  See the above copyright notices for more information.

=========================================================================*/

#include <iostream>

#include "otbCommandLineArgumentParser.h"
#include "otbImageFileReader.h"
#include "otbMapProjections.h"
#include "otbGenericRSTransform.h"
#include "otbImage.h"

template<typename TMapProjection>
int generic_main(int argc, char* argv[],
                 TMapProjection* mapProjection,
                 otb::CommandLineArgumentParseResult* parseResult)
{
  try
  {
    typedef otb::Image<unsigned int, 2>     ImageType;
    typedef otb::ImageFileReader<ImageType>  ReaderType;
    typedef otb::GenericRSTransform<> RSTransformType;
    typedef typename RSTransformType::InputPointType PointType;

    //Allocate pointer
    ReaderType::Pointer reader=ReaderType::New();

    typename RSTransformType::Pointer rsTransform = RSTransformType::New();

    // Set parameters ...
    reader->SetFileName(parseResult->GetInputImage().c_str());
    reader->GenerateOutputInformation();

    typename ImageType::SizeType lsize = reader->GetOutput()->GetLargestPossibleRegion().GetSize();

    // Generate GCPs from physical sensor model
    // First, set-up the RS transform
    rsTransform->SetInputKeywordList(reader->GetOutput()->GetImageKeywordlist());
    rsTransform->SetInputProjectionRef(reader->GetOutput()->GetProjectionRef());
    if(parseResult->IsOptionPresent("--DEMDirectory"))
      {
      otb::DEMHandler::Instance()->OpenDEMDirectory(parseResult->GetParameterString("--DEMDirectory", 0));
      }
    rsTransform->SetOutputProjectionRef(mapProjection->GetWkt());
    std::cout<<"Map projection: "<<mapProjection->GetWkt()<<std::endl;
    rsTransform->InstanciateTransform();

    // Creating optimal output size, index, origin, spacing
    PointType ul, ur, ll, lr, ul_carto, ur_carto, ll_carto, lr_carto, ul_final, ur_final, ll_final, lr_final, ul_back, ur_back, ll_back, lr_back;

    // Create raw geometry corners
    ul.Fill(0);
    ur.Fill(0);
    ur[0] = lsize[0]-1;
    ll.Fill(0);
    ll[1]=lsize[1]-1;
    lr.Fill(0);
    lr[0] = lsize[0]-1;
    lr[1] = lsize[1]-1;

    // Transform corners into ortho geometry
    ul_carto = rsTransform->TransformPoint(ul);
    ur_carto = rsTransform->TransformPoint(ur);
    ll_carto = rsTransform->TransformPoint(ll);
    lr_carto = rsTransform->TransformPoint(lr);

    // Display
    std::cout<<"Image corners (raw -> carto): "<<std::endl;
    std::cout<<"Upper left : "<<ul<<"\t -> "<<ul_carto<<std::endl;
    std::cout<<"Upper right: "<<ur<<"\t -> "<<ur_carto<<std::endl;
    std::cout<<"Lower left : "<<ll<<"\t -> "<<ll_carto<<std::endl;
    std::cout<<"Lower right: "<<lr<<"\t -> "<<lr_carto<<std::endl;
    std::cout<<std::endl;

    // Give center
    std::cout<<"Image center (carto): "<<std::endl;
    std::cout<<(ul_carto[0]+ur_carto[0])/2<<"\t"<<(ul_carto[1]+ll_carto[1])/2<<std::endl;

    // Compute carto bounding box
    ul_final[0] = std::min(std::min(ul_carto[0], ur_carto[0]), std::min(ll_carto[0], lr_carto[0]));
    ul_final[1] = std::min(std::min(ul_carto[1], ur_carto[1]), std::min(ll_carto[1], lr_carto[1]));
    lr_final[0] = std::max(std::max(ul_carto[0], ur_carto[0]), std::max(ll_carto[0], lr_carto[0]));
    lr_final[1] = std::max(std::max(ul_carto[1], ur_carto[1]), std::max(ll_carto[1], lr_carto[1]));
    ur_final[0] = lr_final[0];
    ur_final[1] = ul_final[1];
    ll_final[0] = ul_final[0];
    ll_final[1] = lr_final[1];

    std::cout<<"Bounding box (carto): "<<std::endl;
    std::cout<<"Upper left : "<<ul_final<<std::endl;
    std::cout<<"Upper right: "<<ur_final<<std::endl;
    std::cout<<"Lower left : "<<ll_final<<std::endl;
    std::cout<<"Lower right: "<<lr_final<<std::endl;
    std::cout<<std::endl;

    ImageType::SpacingType spacing;
    // If spacing option is there
    if(parseResult->IsOptionPresent("--Spacing"))
    {
      spacing[0]=parseResult->GetParameterDouble("--Spacing");
      spacing[1]=-parseResult->GetParameterDouble("--Spacing");
    }
    // Else, compute an optimal spacing by hand
    else
    {
    typename RSTransformType::Pointer inverse = RSTransformType::New();
    rsTransform->GetInverse(inverse);

    ul_back = inverse->TransformPoint(ul_final);
    ur_back = inverse->TransformPoint(ur_final);
    ll_back = inverse->TransformPoint(ll_final);
    lr_back = inverse->TransformPoint(lr_final);

    // Spacing is the length of the line in carto divided by the length of back-projected line
    double sX = (ur_carto[0]-ul_carto[0])/vcl_sqrt(
          (ul_back[0] - ur_back[0])*(ul_back[0] - ur_back[0])
         +(ul_back[1] - ur_back[1])*(ul_back[1] - ur_back[1]));
    double sY = (ll_carto[1]-ul_carto[1])/vcl_sqrt(
              (ll_back[0] - ul_back[0])*(ll_back[0] - ul_back[0])
             +(ll_back[1] - ul_back[1])*(ll_back[1] - ul_back[1]));

    double s = std::max(sX, sY);

    spacing[0] = s;
    spacing[1] = -s;
    }

    std::cout<<"Advised spacing: "<<spacing<<std::endl;

    ImageType::SizeType size;
    size[0]= static_cast<unsigned int>((lr_final[0]-ul_final[0])/vcl_abs(spacing[0]));
    size[1]= static_cast<unsigned int>((lr_final[1]-ul_final[1])/vcl_abs(spacing[1]));

    std::cout<<"Advised size:    "<<size<<std::endl;

    ImageType::PointType origin;
    origin[0]=ul_final[0];
    origin[1]=lr_final[1];


    std::cout<<"Advised origin:  "<<origin<<std::endl;

  }
  catch ( itk::ExceptionObject & err )
  {
    std::cout << "Exception itk::ExceptionObject raised !" << std::endl;
    std::cout << err << std::endl;
    return EXIT_FAILURE;
  }
  catch ( std::bad_alloc & err )
  {
    std::cout << "Exception bad_alloc : "<<(char*)err.what()<< std::endl;
    return EXIT_FAILURE;
  }
  catch ( ... )
  {
    std::cout << "Unknown exception raised !" << std::endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;

}//End main()


int main(int argc, char* argv[])
{
  try
  {
    // Parse command line parameters
    typedef otb::CommandLineArgumentParser ParserType;
    ParserType::Pointer parser = ParserType::New();

    parser->SetProgramDescription("Display useful information about image projection (advised origin, spacing and size).");
    parser->AddInputImage();
    parser->AddOption("--Spacing","Requested spacing","-spacing", 1, false);
    parser->AddOption("--DEMDirectory","Directory were to find the DEM tiles","-dem", 1, false);
    parser->AddOptionNParams("--MapProjectionType","Type (UTM/LAMBERT/LAMBERT2/LAMBERT93/SINUS/ECKERT4/TRANSMERCATOR/MOLLWEID) and parameters of map projection used","-mapProj");

    typedef otb::CommandLineArgumentParseResult ParserResultType;
    ParserResultType::Pointer  parseResult = ParserResultType::New();

    try
    {
      parser->ParseCommandLine(argc, argv, parseResult);
    }
    catch ( itk::ExceptionObject & err )
    {
      std::string descriptionException = err.GetDescription();
      if (descriptionException.find("ParseCommandLine(): Help Parser") != std::string::npos)
      {
        std::cout << "WARNING : output file pixels are converted in 'unsigned char'" << std::endl;
        return EXIT_SUCCESS;
      }
      if (descriptionException.find("ParseCommandLine(): Version Parser") != std::string::npos)
      {
        return EXIT_SUCCESS;
      }
      return EXIT_FAILURE;
    }

    // Code

    std::string typeMap = parseResult->GetParameterString("--MapProjectionType", 0);
//        std::cout << "typeMap " << typeMap << std::endl;
    int nbParams = parseResult->GetNumberOfParameters("--MapProjectionType");
//        std::cout << "nbParams " << nbParams << std::endl;
    nbParams--;

    if ((typeMap == "UTM")&&(nbParams==2))
    {
      int numZone = parseResult->GetParameterUInt("--MapProjectionType", 1);
      char hemisphere = parseResult->GetParameterChar("--MapProjectionType", 2);

      typedef otb::UtmInverseProjection UtmProjectionType;
      UtmProjectionType::Pointer utmProjection = UtmProjectionType::New();

      utmProjection->SetZone(numZone);
      utmProjection->SetHemisphere(hemisphere);

      return generic_main<UtmProjectionType>(argc, argv, utmProjection, parseResult);
    }
    else
    {
      std::vector<double> parameters;

      for (int i=1; i<nbParams+1; i++)
      {
        parameters.push_back(parseResult->GetParameterDouble("--MapProjectionType", i));
      }

      if ((typeMap == "LAMBERT")&&(nbParams==4))
      {
        typedef otb::LambertConformalConicInverseProjection LambertProjectionType;
        LambertProjectionType::Pointer lambertProjection = LambertProjectionType::New();

        lambertProjection->SetParameters(parameters[0], parameters[1], parameters[2], parameters[3]);

        return generic_main<LambertProjectionType>(argc, argv, lambertProjection, parseResult);
      }
      else if ((typeMap == "LAMBERT2")&&(nbParams==0))
      {
        typedef otb::Lambert2EtenduInverseProjection Lambert2ProjectionType;
        Lambert2ProjectionType::Pointer lambert2Projection = Lambert2ProjectionType::New();

        return generic_main<Lambert2ProjectionType>(argc, argv, lambert2Projection, parseResult);
      }
      else if ((typeMap == "LAMBERT93")&&(nbParams==0))
      {
             typedef otb::Lambert93InverseProjection Lambert93ProjectionType;
             Lambert93ProjectionType::Pointer lambert93Projection = Lambert93ProjectionType::New();

             return generic_main<Lambert93ProjectionType>(argc, argv, lambert93Projection, parseResult);
      }
      else if ((typeMap == "SINUS")&&(nbParams==2))
      {
        typedef otb::SinusoidalInverseProjection SinusoidalProjectionType;
        SinusoidalProjectionType::Pointer sinusoidalProjection = SinusoidalProjectionType::New();

        sinusoidalProjection->SetParameters(parameters[0], parameters[1]);

        return generic_main<SinusoidalProjectionType>(argc, argv, sinusoidalProjection, parseResult);
      }
      else if ((typeMap == "ECKERT4")&&(nbParams==2))
      {
        typedef otb::Eckert4InverseProjection Eckert4ProjectionType;
        Eckert4ProjectionType::Pointer eckert4Projection = Eckert4ProjectionType::New();

        eckert4Projection->SetParameters(parameters[0], parameters[1]);

        return generic_main<Eckert4ProjectionType>(argc, argv, eckert4Projection, parseResult);
      }
      else if ((typeMap == "TRANSMERCATOR")&&(nbParams==3))
      {
        typedef otb::TransMercatorInverseProjection TransMercatorProjectionType;
        TransMercatorProjectionType::Pointer transMercatorProjection = TransMercatorProjectionType::New();

        transMercatorProjection->SetParameters(parameters[0], parameters[1], parameters[2]);

        return generic_main<TransMercatorProjectionType>(argc, argv, transMercatorProjection, parseResult);
      }
      else if ((typeMap == "MOLLWEID")&&(nbParams==2))
      {
        typedef otb::MollweidInverseProjection MollweidProjectionType;
        MollweidProjectionType::Pointer mollweidProjection = MollweidProjectionType::New();

        mollweidProjection->SetParameters(parameters[0], parameters[1]);

        return generic_main<MollweidProjectionType>(argc, argv, mollweidProjection, parseResult);
      }
      else
      {
        itkGenericExceptionMacro(<< "TypeMap not recognized, choose one with (parameters) : UTM(2), LAMBERT(4), LAMBERT2(0), SINUS(2), ECKERT4(2), TRANSMERCATOR(3), MOLLWEID(2)");
      }

    }


  }
  catch ( itk::ExceptionObject & err )
  {
    std::cout << "Exception itk::ExceptionObject raised !" << std::endl;
    std::cout << err << std::endl;
    return EXIT_FAILURE;
  }
  catch ( std::bad_alloc & err )
  {
    std::cout << "Exception bad_alloc : "<<(char*)err.what()<< std::endl;
    return EXIT_FAILURE;
  }
  catch ( ... )
  {
    std::cout << "Unknown exception raised !" << std::endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

