/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbImageSimulator.h"

#include "otbCommandLineArgumentParser.h"
#include "otbVectorImage.h"
#include "otbImage.h"
#include "otbImageFileWriter.h"
#include "otbVectorDataFileReader.h"

#include "otbVectorDataToLabelMapWithAttributesFilter.h"
#include "otbSpatialisationFilter.h"
#include "otbProspectModel.h"
#include "otbSailModel.h"
#include "otbImageSimulationMethod.h"
#include "otbAttributesMapLabelObject.h"
#include "otbImageFileReader.h"


namespace otb
{

int ImageSimulator::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("ImageSimulator");

  descriptor->SetDescription("Image Simulation from shape file.");
  descriptor->AddInputImage();
  descriptor->AddOutputImage();

  descriptor->AddOption("--SatRSRFileName","Satellite relative spectral response file","rs", 1, true, ApplicationDescriptor::String);
  descriptor->AddOption("--NumberOfBands","Number of spectral bands","nb", 1, true, ApplicationDescriptor::Integer);
  descriptor->AddOption("--RootPath","Root Path","rp", 1, true, ApplicationDescriptor::String);
  descriptor->AddOption("--MTFRadius","Radius of the prolate function for the MTF","rd", 1, true, ApplicationDescriptor::Integer);

  return EXIT_SUCCESS;
}

int ImageSimulator::Execute(otb::ApplicationOptionsResult* parseResult)
{


  typedef unsigned short LabelType;
  const unsigned int Dimension = 2;
  typedef otb::Image<LabelType, Dimension>                                           LabelImageType;
  typedef otb::VectorImage<double, Dimension>                                        OutputImageType;
  typedef otb::ImageFileWriter<OutputImageType>                                      ImageWriterType;
  typedef otb::ImageFileWriter<LabelImageType>                                       LabelImageWriterType;
  typedef otb::ImageFileReader<LabelImageType>                                       ImageReaderType;
  typedef otb::VectorData<double, Dimension>                                         VectorDataType;
  typedef otb::AttributesMapLabelObject<LabelType, Dimension, std::string>           LabelObjectType;
  typedef itk::LabelMap<LabelObjectType>                                             LabelMapType;
  typedef otb::VectorDataFileReader<VectorDataType>                                  VectorDataFileReaderType;
   
   
  typedef otb::SpatialisationFilter<LabelMapType>                                    SpatialisationFilterType;
  typedef otb::ProspectModel                                                         SimulationStep1Type;
  typedef otb::SailModel                                                             SimulationStep2Type;
  typedef otb::ProlateInterpolateImageFunction<LabelImageType>                       FTMType;
  typedef otb::ImageSimulationMethod<VectorDataType, SpatialisationFilterType,
    SimulationStep1Type, SimulationStep2Type, FTMType , OutputImageType>               ImageSimulationMethodType;
   
  std::string infilename = parseResult->GetInputImage();
  std::string satRSRFilename = parseResult->GetParameterString("--SatRSRFileName", 0);
  unsigned int nbBands = parseResult->GetParameterUInt("--NumberOfBands");
  std::string rootPath = parseResult->GetParameterString("--RootPath", 0);
  unsigned int radius = parseResult->GetParameterUInt("--MTFRadius");
  std::string outfilename = parseResult->GetOutputImage();
   


   
  /** Instantiation of pointer objects*/
  VectorDataFileReaderType::Pointer reader = VectorDataFileReaderType::New();
  ImageWriterType::Pointer writer = ImageWriterType::New();
  LabelImageWriterType::Pointer labelWriter = LabelImageWriterType::New();
  ImageSimulationMethodType::Pointer imageSimulation = ImageSimulationMethodType::New();
  SpatialisationFilterType::Pointer spatialisationFilter = SpatialisationFilterType::New();

  reader->SetFileName(infilename);
  reader->Update();
 
  SpatialisationFilterType::SizeType objectSize;
  objectSize[0]=300;
  objectSize[1]=300;
   
  SpatialisationFilterType::SizeType nbOjects;
  nbOjects[0]=3;
  nbOjects[1]=3;
   
  std::vector<std::string> pathVector(9);
  pathVector[0]="JHU/becknic/rocks/sedimentary/powder/0_75/txt/greywa1f.txt";
  pathVector[1]="";
  pathVector[2]="JHU/becknic/manmade/txt/0092uuu.txt";
  pathVector[3]="JHU/becknic/vegetation/txt/conifers.txt";
  pathVector[4]="JHU/becknic/manmade/txt/0834uuu.txt";
  pathVector[5]="JHU/becknic/vegetation/txt/grass.txt";
  pathVector[6]="JHU/becknic/water/txt/coarse.txt";
  pathVector[7]="JHU/becknic/rocks/igneous/solid/txt/andesi1s.txt";
  pathVector[8]="JHU/becknic/soils/txt/0015c.txt";
   
  std::vector<std::string> areaVector(9);
  areaVector[0]="sedimentaryRock";
  areaVector[1]="prosail";
  areaVector[2]="manmade";
  areaVector[3]="conifers";
  areaVector[4]="manmade";
  areaVector[5]="grass";
  areaVector[6]="water";
  areaVector[7]="igneousRocks";
  areaVector[8]="soils";
   
  std::vector<LabelType> labels(9);
  labels[0]=1;
  labels[1]=2;
  labels[2]=1;
  labels[3]=2;
  labels[4]=3;
  labels[5]=2;
  labels[6]=4;
  labels[7]=5;
  labels[8]=3;

  spatialisationFilter->SetObjectSize(objectSize);
  spatialisationFilter->SetNumberOfObjects(nbOjects);
  spatialisationFilter->SetPathVector(pathVector);
  spatialisationFilter->SetAreaVector(areaVector);
  spatialisationFilter->SetLabels(labels);
   
  imageSimulation->SetInputVectorData(reader->GetOutput());
  imageSimulation->SetSpatialisation(spatialisationFilter);
  imageSimulation->SetNumberOfComponentsPerPixel(nbBands);
  imageSimulation->SetSatRSRFilename(satRSRFilename);
  imageSimulation->SetPathRoot(rootPath);
  imageSimulation->SetRadius(radius);
//    imageSimulation->SetMean();
//    imageSimulation->SetVariance();
  imageSimulation->UpdateData();

   
  
  //Write the result to an image file
  writer->SetFileName(outfilename);
  writer->SetInput(imageSimulation->GetOutputReflectanceImage());
  writer->Update();

  return EXIT_SUCCESS;
}
}
