/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "otbLAIAndPROSAILToSensorResponse.h"

#include "otbCommandLineArgumentParser.h"
#include "otbVectorImage.h"
#include "otbImage.h"
#include "otbImageFileWriter.h"
#include "otbVectorDataFileReader.h"

#include "otbVectorDataToLabelMapWithAttributesFilter.h"
#include "otbSpatialisationFilter.h"
#include "otbProspectModel.h"
#include "otbSailModel.h"
#include "otbImageSimulationMethod.h"
#include "otbAttributesMapLabelObject.h"
#include "otbImageFileReader.h"

#include "otbCommandLineArgumentParser.h"
#include "otbLeafParameters.h"
#include "otbSailModel.h"
#include "otbProspectModel.h"
#include "otbSatelliteRSR.h"
#include "otbReduceSpectralResponse.h"
#include "otbImage.h"
#include "otbVectorImage.h"
#include "otbImageFileReader.h"
#include "otbImageFileWriter.h"
#include "otbSpectralResponse.h"
#include "itkTernaryFunctorImageFilter.h"
#include "otbVegetationIndicesFunctor.h"
#include "otbRAndNIRIndexImageFilter.h"
#include "otbStandardWriterWatcher.h"
#include "itkImageRegionConstIterator.h"
#include "otbMultiChannelExtractROI.h"
#include "itkOrImageFilter.h"


namespace otb
{

template < class TImage >
class ITK_EXPORT ImageUniqueValuesCalculator : public itk::Object
{
public:
  typedef ImageUniqueValuesCalculator<TImage>   Self;
  typedef itk::Object                           Superclass;
  typedef itk::SmartPointer<Self>               Pointer;
  typedef itk::SmartPointer<const Self>         ConstPointer;

  itkNewMacro(Self);

  itkTypeMacro(ImageUniqueValuesCalculator, itk::Object);

  itkStaticConstMacro(ImageDimension, unsigned int,
                      TImage::ImageDimension);

  typedef typename TImage::PixelType PixelType;

  typedef TImage ImageType;

  typedef std::vector<PixelType>           ArrayType;

  typedef typename ImageType::Pointer      ImagePointer;
  typedef typename ImageType::ConstPointer ImageConstPointer;

  virtual void SetImage( const ImageType * image )
    {
    if ( m_Image != image )
      {
      m_Image = image;
      this->Modified();
      }
    }


  ArrayType GetUniqueValues() const
  {
    typedef typename ImageType::IndexType IndexType;

    if( !m_Image )
      {
       itkExceptionMacro(<<"GetUniqueValues(): Null input image pointer.");
      }

    itk::ImageRegionConstIterator< ImageType > it( m_Image,
                                              m_Image->GetRequestedRegion() );

    ArrayType uniqueValues;

    uniqueValues.push_back(it.Get());
    ++it;
    while( !it.IsAtEnd() )
      {
      if( std::find(uniqueValues.begin(), uniqueValues.end(), it.Get()) == uniqueValues.end())
        {
        uniqueValues.push_back(it.Get());
        }
      ++it;
      }

    return uniqueValues;
  }


protected:
  ImageUniqueValuesCalculator()
    {
    m_Image = NULL;
    }
  virtual ~ImageUniqueValuesCalculator()
  {
  }
  void PrintSelf(std::ostream& os, itk::Indent indent) const
  {
    Superclass::PrintSelf(os, indent);
    os << indent << "Image: " << m_Image.GetPointer() << std::endl;
  }

private:
  ImageUniqueValuesCalculator(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

  ImageConstPointer         m_Image;

};  // class ImageUniqueValuesCalculator

namespace Functor
{

template <class TLAI, class TLabel, class TMask, class TOutput, class TLabelSpectra, class TLabelParameter, class TAcquistionParameter, class TSatRSR>
class ProsailSimulatorFunctor
{
public:
  /** Standard class typedefs */
  typedef TLAI LAIPixelType;
  typedef TLabel LabelPixelType;
  typedef TMask MaskPixelType;
  typedef TOutput OutputPixelType;
  typedef TLabelSpectra LabelSpectraType;
  typedef TLabelParameter LabelParameterType;
  typedef TAcquistionParameter AcquistionParameterType;
  typedef TSatRSR SatRSRType;
  typedef typename SatRSRType::Pointer SatRSRPointerType;
  typedef typename otb::ProspectModel ProspectType;
  typedef typename otb::SailModel SailType;
  static const unsigned int SimNbBands = 2000;

  typedef double PrecisionType;
  typedef std::pair<PrecisionType, PrecisionType> PairType;
  typedef typename std::vector<PairType> VectorPairType;
  typedef otb::SpectralResponse<PrecisionType, PrecisionType> ResponseType;
  typedef ResponseType::Pointer ResponsePointerType;
  typedef otb::ReduceSpectralResponse<ResponseType, SatRSRType> ReduceResponseType;
  typedef typename ReduceResponseType::Pointer ReduceResponseTypePointerType;

  /** Constructor */
  ProsailSimulatorFunctor()
  {
    m_SatRSR = SatRSRType::New();
    m_InvertedMask = false;
    m_CompleteSpectralResponse = false;
  }

  /** Destructor */
  ~ProsailSimulatorFunctor()
  {
  }
;

  void SetInvertedMask(bool ivm)
  {
    m_InvertedMask = ivm;
  }

  void SetCompleteSpectralResponse(bool ivm)
  {
    m_CompleteSpectralResponse = ivm;
  }

  const unsigned int GetNumberOfOutputBands() const
  {
    if (!m_CompleteSpectralResponse)
      return m_SatRSR->GetNbBands();
    else return SimNbBands;

  }

  /** Implementation of the () operator*/
  inline OutputPixelType operator ()(const LAIPixelType &lai, const LabelPixelType& label, const MaskPixelType& mask)
  {

    OutputPixelType pix;
    unsigned int pixSize;

    if (!m_CompleteSpectralResponse)
      pixSize = m_SatRSR->GetNbBands();
    else pixSize = SimNbBands;

    // std::cout<<"pixSize "<<pixSize<<std::endl;

    pix.SetSize(pixSize);
    if ((!mask && !m_InvertedMask) || (mask && m_InvertedMask))
      {
      for (unsigned int i = 0; i < pixSize; i++)
        pix[i] = static_cast<typename OutputPixelType::ValueType> (0);
      return pix;
      }

    VectorPairType hxSpectrum;

    for (unsigned int i = 0; i < SimNbBands; i++)
      {
      PairType resp;
      resp.first = static_cast<PrecisionType> ((400.0 + i) / 1000);
      resp.second = static_cast<PrecisionType> (0.0);
      hxSpectrum.push_back(resp);
      }

    // either the spectrum has to be simulated by Prospect+Sail
    if (m_LabelParameters.find(label) != m_LabelParameters.end())
      {
      ProspectType::Pointer prospect = ProspectType::New();
      prospect->SetInput(m_LabelParameters[label]);

      SailType::Pointer sail = SailType::New();

      sail->SetLAI(lai);
      sail->SetAngl(m_AcquisitionParameters[std::string("Angl")]);
      sail->SetPSoil(m_AcquisitionParameters[std::string("PSoil")]);
      sail->SetSkyl(m_AcquisitionParameters[std::string("Skyl")]);
      sail->SetHSpot(m_AcquisitionParameters[std::string("HSpot")]);
      sail->SetTTS(m_AcquisitionParameters[std::string("TTS")]);
      sail->SetTTO(m_AcquisitionParameters[std::string("TTO")]);
      sail->SetPSI(m_AcquisitionParameters[std::string("PSI")]);
      sail->SetReflectance(prospect->GetReflectance());
      sail->SetTransmittance(prospect->GetTransmittance());
      sail->Update();

      for (unsigned int i = 0; i < SimNbBands; i++)
        {
        hxSpectrum[i].second = static_cast<typename OutputPixelType::ValueType> (sail->GetHemisphericalReflectance()->GetResponse()[i].second);

        }

      }
    // or the spectra has been set from outside the functor (ex. bare soil, etc.)
    else
      if (m_LabelSpectra.find(label) != m_LabelSpectra.end())
        {
        for (unsigned int i = 0; i < SimNbBands; i++)
          hxSpectrum[i].second = static_cast<typename OutputPixelType::ValueType> (m_LabelSpectra[label][i]);

        }
      // or the class does not exist
      else
        {
        for (unsigned int i = 0; i < SimNbBands; i++)
          hxSpectrum[i].second = static_cast<typename OutputPixelType::ValueType> (0);
        }

    if (!m_CompleteSpectralResponse)
      {
      ResponseType::Pointer aResponse = ResponseType::New();
      aResponse->SetResponse(hxSpectrum);
      ReduceResponseTypePointerType reduceResponse = ReduceResponseType::New();
      reduceResponse->SetInputSatRSR(m_SatRSR);
      reduceResponse->SetInputSpectralResponse(aResponse);
      reduceResponse->CalculateResponse();
      VectorPairType reducedResponse = reduceResponse->GetReduceResponse()->GetResponse();

      for (unsigned int i = 0; i < m_SatRSR->GetNbBands(); i++)
        pix[i] = static_cast<typename OutputPixelType::ValueType> (reducedResponse[i].second);
      }
    else
      {
      for (unsigned int i = 0; i < SimNbBands; i++)
        {
        pix[i] = static_cast<typename OutputPixelType::ValueType> (hxSpectrum[i].second);
        }
      }

    return pix;

  }

  bool operator !=(const ProsailSimulatorFunctor& other) const
  {
    return (m_AcquisitionParameters != other.m_AcquisitionParameters || m_LabelParameters != other.m_LabelParameters);
  }

  bool operator ==(const ProsailSimulatorFunctor& other) const
  {
    return (m_AcquisitionParameters == other.m_AcquisitionParameters && m_LabelParameters == other.m_LabelParameters);
  }

  /** The spectra associated with labels are supposed to have the same
   * sampling as a Prosail simulation, that is, 2000 bands starting at
   * 400nm up to 2400 nm*/
  inline void SetLabelSpectra(const TLabelSpectra& lSpectra)
  {
    m_LabelSpectra = lSpectra;
  }

  inline void SetLabelParameters(const TLabelParameter& lParameters)
  {
    m_LabelParameters = lParameters;
  }

  inline void SetAcquisitionParameters(const TAcquistionParameter& aParameters)
  {
    m_AcquisitionParameters = aParameters;
  }

  inline void SetRSR(const SatRSRPointerType rsr)
  {
    m_SatRSR = rsr;
  }

  inline SatRSRPointerType GetRSR() const
  {
    return m_SatRSR;
  }

protected:

  /** Spectra associated to labels*/
  LabelSpectraType m_LabelSpectra;
  /** Prospect+sail parameters to labels*/
  LabelParameterType m_LabelParameters;
  /** Prospect+sail acquisition parameters*/
  AcquistionParameterType m_AcquisitionParameters;
  /** Satellite Relative spectral response*/
  SatRSRPointerType m_SatRSR;

  /** Simulate pixels with mask != 0 ==> m_InvertedMask = false; */
  bool m_InvertedMask;

  /** do no perform reduce spectral response **/
  bool m_CompleteSpectralResponse;

};


} // namespace Functor

template <class TInputImage1, class TInputImage2, class TInputImage3, class TOutputImage, class TFunctor>
class ITK_EXPORT TernaryFunctorImageFilterWithNBands :
  public itk::TernaryFunctorImageFilter< TInputImage1, TInputImage2, TInputImage3, TOutputImage, TFunctor >
{
public:
  typedef TernaryFunctorImageFilterWithNBands Self;
  typedef itk::TernaryFunctorImageFilter< TInputImage1, TInputImage2, TInputImage3, TOutputImage, TFunctor > Superclass;
  typedef itk::SmartPointer<Self>       Pointer;
  typedef itk::SmartPointer<const Self> ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);

  /** Macro defining the type*/
  itkTypeMacro(TernaryFunctorImageFilterWithNBands, SuperClass);

  /** Accessors for the number of bands*/
  itkSetMacro(NumberOfOutputBands, unsigned int);
  itkGetConstMacro(NumberOfOutputBands, unsigned int);

protected:
  TernaryFunctorImageFilterWithNBands() {}
  virtual ~TernaryFunctorImageFilterWithNBands() {}

  void GenerateOutputInformation()
  {
    Superclass::GenerateOutputInformation();
    this->GetOutput()->SetNumberOfComponentsPerPixel( m_NumberOfOutputBands );
  }
private:
  TernaryFunctorImageFilterWithNBands(const Self &); //purposely not implemented
  void operator =(const Self&); //purposely not implemented

  unsigned int m_NumberOfOutputBands;


};


int LAIAndPROSAILToSensorResponse::Describe(ApplicationDescriptor* descriptor)
{
  descriptor->SetName("LAIAndPROSAILToSensorResponse");

  descriptor->SetDescription(
                             "Use LAI, label, Mask, (CloudMask) images, PROSAIL parameters and Sensor RSR to generate Spectral Image adapted to the sensor type");
  descriptor->AddInputImage();
  descriptor->AddOutputImage();

  descriptor->AddOption("--LAIImage", "LAI image", "laii", 1, true, ApplicationDescriptor::String);
  descriptor->AddOption("--MaskImage", "Mask for valid pixels", "mi", 1, true, ApplicationDescriptor::String);
  descriptor->AddOption("--CloudMask", "Mask for cloudy pixels", "cmi", 1, false, ApplicationDescriptor::String);
  descriptor->AddOption("--LabelImage", "Image of labels", "li", 1, true, ApplicationDescriptor::String);
  descriptor->AddOption("--LabelParameterFile", "File containing the parameters for the labels", "lp", 1, true,
                        ApplicationDescriptor::String);
  descriptor->AddOption("--AcquisitionParameterFile", "File containing the parameters for the acquisition", "ap", 1,
                        true, ApplicationDescriptor::String);
  descriptor->AddOption("--RSRFile", "File name for the relative spectral response to be used", "sr", 1, false,
                        ApplicationDescriptor::String);
  descriptor->AddOption("--SROutImage", "File name for the complete spectral response output image", "srout", 1, false,
                          ApplicationDescriptor::String);
  descriptor->AddOption("--NumberOfBands", "Number of bands of the simulated image", "nb", 1, false,
                        ApplicationDescriptor::Integer);
  descriptor->AddOption("--MinWavelength", "User Customized Spectral Response minimum WaveLength (in micrometer)", "lmin", 1, false,
                          ApplicationDescriptor::Real);
  descriptor->AddOption("--MaxWavelength", "User Customized Spectral Response maximum WaveLength (in micrometer)", "lmax", 1, false,
                          ApplicationDescriptor::Real);
  descriptor->AddOption("--Sampling", "User Customized Spectral Response bandwidth", "samp", 1, false,
                          ApplicationDescriptor::Real);
  descriptor->AddOption("--NormalizationCoef", "Satellite Normalisation coefficient", "norm", 1, false,
                            ApplicationDescriptor::Real);
  descriptor->AddOption("--RootPath", "Root Path", "rp", 1, false, ApplicationDescriptor::String);

  return EXIT_SUCCESS;
}

int LAIAndPROSAILToSensorResponse::Execute(otb::ApplicationOptionsResult* parseResult)
{

  // Read the label parameter file. It is assumed to have the form
  // label 1 Cab Car CBrown Cw Cm N   --> for vegetation classes
  // label 0 /path/to/spectra/file    --> for other classes

  typedef double LAIPixelType;
  typedef unsigned short LabelType;
  typedef unsigned short MaskPixelType;
  typedef float OutputPixelType;

  // Image typedef
  typedef otb::Image<LAIPixelType, 2> LAIImageType;
  typedef otb::Image<LabelType, 2> LabelImageType;
  typedef otb::Image<MaskPixelType, 2> MaskImageType;
  typedef otb::VectorImage<OutputPixelType, 2> SimulatedImageType;

  // reader typedef
  typedef otb::ImageFileReader<LAIImageType> LAIReaderType;
  typedef otb::ImageFileReader<LabelImageType> LabelReaderType;
  typedef otb::ImageFileReader<MaskImageType> MaskReaderType;

  // leaf parameter typedef
  typedef otb::LeafParameters LeafParametersType;
  typedef LeafParametersType::Pointer LeafParametersPointerType;

  // label parameters typedef
  typedef std::map<LabelType, LeafParametersPointerType> LabelParameterMapType;

  // spectra  typedef
  typedef double PrecisionType;
  typedef double ValuePrecisionType;
  typedef std::vector<PrecisionType> SpectraType;
  typedef std::map<LabelType, SpectraType> SpectraParameterType;

  typedef std::map<std::string, double> AcquistionParsType;

  typedef otb::SatelliteRSR<PrecisionType, PrecisionType> SatRSRType;

  typedef otb::Functor::ProsailSimulatorFunctor<LAIPixelType, LabelType, MaskPixelType, SimulatedImageType::PixelType,
      SpectraParameterType, LabelParameterMapType, AcquistionParsType, SatRSRType> SimuFunctorType;
  typedef otb::TernaryFunctorImageFilterWithNBands<LAIImageType, LabelImageType, MaskImageType, SimulatedImageType,
      SimuFunctorType> SimulatorType;

  SpectraParameterType spectraParameters;

  std::string outfname = parseResult->GetOutputImage();

  std::string cmifname; //cloud mask
  if (parseResult->IsOptionPresent("--CloudMask"))
    {
    cmifname = parseResult->GetParameterString("--CloudMask", 0);
    }

  std::string lpfname = parseResult->GetParameterString("--LabelParameterFile", 0);
  std::string apfname = parseResult->GetParameterString("--AcquisitionParameterFile", 0);

  std::string laiifname = parseResult->GetParameterString("--LAIImage", 0);
  std::string lifname = parseResult->GetParameterString("--LabelImage", 0);
  std::string mifname = parseResult->GetParameterString("--MaskImage", 0);
  std::string rsfname;

  PrecisionType lmin;
  PrecisionType lmax;
  PrecisionType smp;
  ValuePrecisionType norm = 1.0;

  if (parseResult->IsOptionPresent("--RSRFile"))
    {
    rsfname = parseResult->GetParameterString("--RSRFile", 0);
    if (rsfname.empty())
      {
      std::cerr << "problem in satellite filename loading " << std::endl;
      return EXIT_FAILURE;
      }
    }
  else
    if (parseResult->IsOptionPresent("--MinWavelength") && parseResult->IsOptionPresent("--MaxWavelength"))
      {
      lmin = static_cast<PrecisionType> (parseResult->GetParameterDouble("--MinWavelength"));
      lmax = static_cast<PrecisionType> (parseResult->GetParameterDouble("--MaxWavelength"));

      smp = 0.0; //by default sampling is set to wavelengthprecision const defined in SatelliteRSR class.
      if (parseResult->IsOptionPresent("--Sampling")) smp = static_cast<PrecisionType> (parseResult->GetParameterDouble(
                                                                                                                        "--Sampling"));

      if ((0 == lmin * lmax))
        {
        std::cerr << "problem in customized parameter loading" << std::endl;
        }
      if (lmin > lmax)
        {
        PrecisionType temp = lmax;
        lmax = lmin;
        lmin = temp;
        }

      }
    else
      {
      std::cerr << "at least a satellite SR (rs) filename or user defined SR (lmin lmax and smp) must be given "
          << std::endl;
      return EXIT_FAILURE;
      }

  unsigned int nbBands = 1;
  if (parseResult->IsOptionPresent("--NumberOfBands")) nbBands = parseResult->GetParameterUInt("--NumberOfBands");

  if (parseResult->IsOptionPresent("--NormalizationCoef")) norm = parseResult->GetParameterDouble("--NormalizationCoef");

  // Read the acquisition parameter file which is like
  // Angl val
  // PSoil val
  // Skyl val
  // HSpot val
  // TTS val
  // TTO val
  // PSI val

  AcquistionParsType acquistionPars;
  acquistionPars[std::string("Angl")] = 0.0;
  acquistionPars[std::string("PSoil")] = 0.0;
  acquistionPars[std::string("Skyl")] = 0.0;
  acquistionPars[std::string("HSpot")] = 0.0;
  acquistionPars[std::string("TTS")] = 0.0;
  acquistionPars[std::string("TTO")] = 0.0;
  acquistionPars[std::string("PSI")] = 0.0;

  std::ifstream acquistionParsFile;

  try
    {
    acquistionParsFile.open(apfname.c_str());
    }
  catch (...)
    {
    std::cerr << "Could not open file " << apfname << std::endl;
    return EXIT_FAILURE;
    }

  unsigned int acPar = 0;
  while (acquistionParsFile.good())
    {
    std::string line;
    std::getline(acquistionParsFile, line);
    std::stringstream ss(line);
    std::string parName;
    ss >> parName;
    double parValue;
    ss >> parValue;
    acquistionPars[parName] = parValue;

    }

  acquistionParsFile.close();

  LabelParameterMapType labelParameters;

  std::string rootPath = ""; // path to object param
  if (parseResult->IsOptionPresent("--RootPath"))
    {
    rootPath = parseResult->GetParameterString("--RootPath", 0);
    }

  std::ifstream labelParsFile;

  try
    {
    labelParsFile.open(lpfname.c_str(), std::ifstream::in);
    }
  catch (...)
    {
    std::cerr << "Could not open file " << lpfname << std::endl;
    return EXIT_FAILURE;
    }

  while (labelParsFile.good())
    {
    char fileLine[256];
    labelParsFile.getline(fileLine, 256);

    if (fileLine[0] != '#')
      {
      std::stringstream ss(fileLine);
      unsigned short label;
      ss >> label;
      unsigned short paramsOrSpectra;
      ss >> paramsOrSpectra;
      if (paramsOrSpectra == 1)
        {
        double Cab;
        ss >> Cab;
        double Car;
        ss >> Car;
        double CBrown;
        ss >> CBrown;
        double Cw;
        ss >> Cw;
        double Cm;
        ss >> Cm;
        double N;
        ss >> N;

        LeafParametersType::Pointer leafParams = LeafParametersType::New();

        leafParams->SetCab(Cab);
        leafParams->SetCar(Car);
        leafParams->SetCBrown(CBrown);
        leafParams->SetCw(Cw);
        leafParams->SetCm(Cm);
        leafParams->SetN(N);

        labelParameters[label] = leafParams;
        }
      else
        {
        std::string spectraFilename = rootPath;
        ss >> spectraFilename;
        spectraFilename = rootPath + spectraFilename;
        typedef otb::SpectralResponse<PrecisionType, PrecisionType> ResponseType;
        ResponseType::Pointer resp = ResponseType::New();

        // Coefficient 100 since Aster database is given in % reflectance
        resp->Load(spectraFilename, 100.0);

        SpectraType spec;
        for (unsigned int i = 0; i < SimuFunctorType::SimNbBands; i++)
          //Prosail starts at 400 and lambda in Aster DB is in micrometers
          spec.push_back(static_cast<PrecisionType> ((*resp)((i + 400.0) / 1000.0)));

        spectraParameters[label] = spec;

        }

      }
    }

  labelParsFile.close();

  LAIReaderType::Pointer laiReader = LAIReaderType::New();
  laiReader->SetFileName(laiifname);
  laiReader->Update();
  LAIImageType::Pointer laiImage = laiReader->GetOutput();

  // Check that all the labels of the GT are present in the labelParameters

  LabelReaderType::Pointer labelReader = LabelReaderType::New();
  labelReader->SetFileName(lifname);
  labelReader->Update();

  LabelImageType::Pointer labelImage = labelReader->GetOutput();

  typedef otb::ImageUniqueValuesCalculator<LabelImageType> UniqueCalculatorType;

  UniqueCalculatorType::Pointer uniqueCalculator = UniqueCalculatorType::New();

  uniqueCalculator->SetImage(labelImage);

  UniqueCalculatorType::ArrayType uniqueVals = uniqueCalculator->GetUniqueValues();
  std::cout << "GT Labels are " << std::endl;
  UniqueCalculatorType::ArrayType::const_iterator uvIt = uniqueVals.begin();

  while (uvIt != uniqueVals.end())
    {
    std::cout << (*uvIt) << ", ";
    ++uvIt;
    }

  std::cout << std::endl;

  uvIt = uniqueVals.begin();

  while (uvIt != uniqueVals.end())
    {
    if (labelParameters.find(static_cast<LabelType> (*uvIt)) == labelParameters.end() &&
        spectraParameters.find(static_cast<LabelType> (*uvIt)) == spectraParameters.end() &&
        static_cast<LabelType> (*uvIt) != 0)
      {
      std::cout << "GT label " << (*uvIt) << " not found in " << lpfname << std::endl;
      return EXIT_FAILURE;
      }
    ++uvIt;
    }

  // Read the input images and the image of labels

  MaskReaderType::Pointer miReader = MaskReaderType::New();
  miReader->SetFileName(mifname);
  miReader->UpdateOutputInformation();
  MaskImageType::Pointer maskImage = miReader->GetOutput();

  if (parseResult->IsOptionPresent("--CloudMask"))
    {

    MaskReaderType::Pointer cmiReader = MaskReaderType::New();
    cmiReader->SetFileName(cmifname);
    cmiReader->UpdateOutputInformation();

    typedef itk::OrImageFilter<MaskImageType, MaskImageType, MaskImageType> OrType;
    OrType::Pointer orfilter = OrType::New();

    orfilter->SetInput1(miReader->GetOutput());
    orfilter->SetInput2(cmiReader->GetOutput());

    orfilter->Update();
    maskImage = orfilter->GetOutput();

    }

  if (laiImage->GetLargestPossibleRegion().GetSize()[0] != labelImage->GetLargestPossibleRegion().GetSize()[0] ||
      laiImage->GetLargestPossibleRegion().GetSize()[1] != labelImage->GetLargestPossibleRegion().GetSize()[1] ||
      laiImage->GetLargestPossibleRegion().GetSize()[0] != maskImage->GetLargestPossibleRegion().GetSize()[0] ||
      laiImage->GetLargestPossibleRegion().GetSize()[1] != maskImage->GetLargestPossibleRegion().GetSize()[1])
    {
    std::cerr << "Image of labels, mask and LAI image must have the same size" << std::endl;
    return EXIT_FAILURE;
    }

  SatRSRType::Pointer satRSR = SatRSRType::New();
  satRSR->SetNbBands(nbBands);
  satRSR->SetSortBands(false);
  if (parseResult->IsOptionPresent("--RSRFile"))
    {
    satRSR->Load(rsfname, norm);
    }
  else
    {
    satRSR->Load(lmin, lmax, smp, norm);
    }

  for (unsigned int i = 0; i < nbBands; ++i)
    std::cout << i << " " << (satRSR->GetRSR())[i]->GetInterval().first << " "
        << (satRSR->GetRSR())[i]->GetInterval().second << std::endl;

  SimuFunctorType simuFunctor;

  simuFunctor.SetLabelParameters(labelParameters);
  simuFunctor.SetLabelSpectra(spectraParameters);
  simuFunctor.SetAcquisitionParameters(acquistionPars);
  simuFunctor.SetRSR(satRSR);
  simuFunctor.SetInvertedMask(true);

  SimulatorType::Pointer simulator = SimulatorType::New();

  simulator->SetInput1(laiImage);
  simulator->SetInput2(labelImage);
  simulator->SetInput3(maskImage);
  simulator->SetFunctor(simuFunctor);
  simulator->SetNumberOfOutputBands(nbBands);

  // Write output image to disk

  typedef otb::ImageFileWriter<SimulatedImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();

  writer->SetFileName(outfname);
  writer->SetInput(simulator->GetOutput());
  // writer->SetBufferNumberOfLinesDivisions(nbStreams);

  otb::StandardWriterWatcher watcher(writer, simulator, "Simulation");

  writer->Update();

  /** complete spectral response is needed **/
  if (parseResult->IsOptionPresent("--SROutImage"))
    {
    SimuFunctorType SRsimuFunctor;

    SRsimuFunctor.SetLabelParameters(labelParameters);
    SRsimuFunctor.SetLabelSpectra(spectraParameters);
    SRsimuFunctor.SetAcquisitionParameters(acquistionPars);
    SRsimuFunctor.SetRSR(satRSR);
    SRsimuFunctor.SetInvertedMask(true);
    SRsimuFunctor.SetCompleteSpectralResponse(true);

    SimulatorType::Pointer SRsimulator = SimulatorType::New();

    SRsimulator->SetInput1(laiImage);
    SRsimulator->SetInput2(labelImage);
    SRsimulator->SetInput3(maskImage);
    SRsimulator->SetFunctor(SRsimuFunctor);

    SRsimulator->SetNumberOfOutputBands(SRsimuFunctor.GetNumberOfOutputBands());

    SRsimulator->Update();
    typedef otb::ImageFileWriter<SimulatedImageType> WriterType;
    WriterType::Pointer srwriter = WriterType::New();
    std::string sroutfname = parseResult->GetParameterString("--SROutImage", 0);
    srwriter->SetFileName(sroutfname);
    srwriter->SetInput(SRsimulator->GetOutput());

    otb::StandardWriterWatcher watcher(srwriter, SRsimulator, "Simulation_2");

    srwriter->Update();

    }

  return EXIT_SUCCESS;

}
}
