/*=========================================================================

  Program:   ORFEO Toolbox
  Language:  C++
  Date:      $Date$
  Version:   $Revision$


  Copyright (c) Centre National d'Etudes Spatiales. All rights reserved.
  See OTBCopyright.txt for details.


  This software is distributed WITHOUT ANY WARRANTY; without even
  the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the above copyright notices for more information.

=========================================================================*/


#include "otbImage.h"
#include "otbImageFileReader.h"
#include "otbImageFileWriter.h"

#include "itkCannyEdgeDetectionImageFilter.h"
#include "otbMeanShiftVectorImageFilter.h"
#include "otbLineSegmentDetector.h"
#include "otbDrawLineSpatialObjectListFilter.h"
#include "otbImageFileReader.h"
#include "otbImageFileWriter.h"

#include "otbVectorData.h"
#include "otbVectorDataFileReader.h"
#include "otbVectorDataFileWriter.h"
#include "otbVectorDataToValidSegmentsFilter.h"
#include "otbVectorDataProjectionFilter.h"

#include "otbStandardImageViewer.h"


int  otbImageToDBRegistrationProcess( int argc, char* argv[] )
{
  
  const char  *     infname       = argv[1];
  const char  *     outfname      = argv[2];
  const char  *     vfname        = argv[3];
  unsigned int verbose = atoi(argv[4]);
  

  /** Variables for the canny detector*/
  const double    upperThreshold   = atof(argv[4]);
  const double    lowerThreshold   = atof(argv[5]);
  const double    threshold   = atof(argv[6]);
    
  const   unsigned int                                    Dimension = 2;
  typedef float                                           PixelType;
  typedef unsigned short                                  LabeledPixelType;
  
  typedef otb::Image<PixelType, Dimension>                                         ImageType;
  typedef otb::Image<LabeledPixelType, Dimension>                                  LabeledImageType;
  typedef otb::ImageFileReader<ImageType>                                         ReaderType;
  typedef otb::ImageFileWriter<ImageType>                                         WriterType;
  typedef itk::CannyEdgeDetectionImageFilter<ImageType , ImageType>               CannyDetectorType;
  typedef otb::LineSegmentDetector<ImageType , PixelType>                         lsdFilterType;
  typedef lsdFilterType::LineSpatialObjectListType                                LineListType;
  typedef otb::DrawLineSpatialObjectListFilter< ImageType , ImageType>             DrawLineListType;
  
  typedef otb::VectorData<double, 2>                                               VectorDataType;
  typedef VectorDataType::DataNodeType                                            DataNodeType;
  typedef DataNodeType::LineType                                                  LineType;
  typedef LineType::VertexType                                                    VertexType;
  typedef otb::VectorDataFileReader<VectorDataType>                               VReaderType;
  typedef otb::VectorDataFileWriter<VectorDataType>                               VWriterType;
  typedef otb::VectorDataProjectionFilter<VectorDataType, VectorDataType>          VectorDataFilterType;

  typedef otb::VectorDataToValidSegmentsFilter<VectorDataType>                    VectorDataToValidSegmentsType;

  typedef otb::StandardImageViewer<ImageType>                                     ViewerType;

  //
  ReaderType::Pointer            reader         = ReaderType::New();
  WriterType::Pointer            writer         = WriterType::New();
  CannyDetectorType::Pointer     canny          = CannyDetectorType::New();
  lsdFilterType::Pointer         lsd            = lsdFilterType::New();
  DrawLineListType::Pointer      drawLineFilter = DrawLineListType::New();
  VReaderType::Pointer           vreader        = VReaderType::New();
  VectorDataToValidSegmentsType::Pointer  vhough  = VectorDataToValidSegmentsType::New();
  
  //
  reader->SetFileName(infname);
  reader->Update();
  
  canny->SetInput(reader->GetOutput());
  canny->SetUpperThreshold(static_cast<ImageType::PixelType>(upperThreshold)); /** 30*/
  canny->SetLowerThreshold(static_cast<ImageType::PixelType>(lowerThreshold)); /** 10*/
  canny->SetVariance(1.); //1.
  canny->SetMaximumError(0.01); ///0.01f

  lsd->SetInput(canny->GetOutput());
  drawLineFilter->SetInput(reader->GetOutput());
  drawLineFilter->SetInputLineSpatialObjectList(lsd->GetOutput());
    
  writer->SetInput(drawLineFilter->GetOutput());
  writer->SetFileName(outfname);
  writer->Update();


  //Create a vectorData
  
  VectorDataType::Pointer       data       = VectorDataType::New();
  DataNodeType::Pointer         document   = DataNodeType::New();
  DataNodeType::Pointer         folder    = DataNodeType::New();
  DataNodeType::Pointer         root = data->GetDataTree()->GetRoot()->Get();
  
  document->SetNodeType(otb::DOCUMENT);
  folder->SetNodeType(otb::FOLDER);
  data->GetDataTree()->Add(document, root);
  data->GetDataTree()->Add(folder, document);
  data->SetProjectionRef(reader->GetOutput()->GetProjectionRef());
  
  LineListType::const_iterator it = lsd->GetOutput()->begin();
  LineListType::const_iterator itEnd = lsd->GetOutput()->end();
  while(it != itEnd)
    {
      VertexType p1, p2;
      LineType::Pointer l = LineType::New();
      LineListType::LineType::PointListType & pointsList = (*it)->GetPoints();
      LineListType::LineType::PointListType::const_iterator itPoints = pointsList.begin();
      p1[0] =(*itPoints).GetPosition()[0];
      p1[1] =(*itPoints).GetPosition()[1];
      itPoints++;
      p2[0] =(*itPoints).GetPosition()[0];
      p2[1] =(*itPoints).GetPosition()[1];
      l->AddVertex(p1);
      l->AddVertex(p2);
      DataNodeType::Pointer node = DataNodeType::New();
      node->SetNodeType(otb::FEATURE_LINE);
      node->SetLine(l);
      data->GetDataTree()->Add(node, folder);
      ++it;
    }
  
  //Projection of the polygons
  VectorDataFilterType::Pointer vectorDataProjection = VectorDataFilterType::New();

  vectorDataProjection->SetInputOrigin(reader->GetOutput()->GetOrigin());
  vectorDataProjection->SetInputSpacing(reader->GetOutput()->GetSpacing());
  vectorDataProjection->SetInput(data);

  //...
  vhough->SetInput(vectorDataProjection->GetOutput());
  vhough->SetFrequencyThreshold(threshold);
  vhough->SetOrigin(reader->GetOutput()->GetOrigin());
  

  //Write the shapefile
  VWriterType::Pointer vwriter = VWriterType::New();
  vwriter->SetFileName(vfname);
  vwriter->SetInput(vhough->GetOutput()/*vectorDataProjection->GetOutput()*/);
  vwriter->Update();

  
  //Instanciate the viewer
  ViewerType::Pointer             viewer =       ViewerType::New();
  viewer->SetImage(reader->GetOutput());
  viewer->SetVectorData(vhough->GetOutput());
  viewer->SetLabel("Registation Process ");
  viewer->Update();

  if(verbose)
    Fl::check();
  else
    Fl::run();
  
  


  return EXIT_SUCCESS;
  

}
